@@ -84,6 +84,11 @@ struct SVAR_context
    Svar_state state_D;
 };
 
+//-----------------------------------------------------------------------------
+const char * prog_name()
+{
+   return "AP210";
+}
 //-----------------------------------------------------------------------------
 int
 set_ACK(Coupled_var & var, int exco)
@@ -148,8 +153,8 @@ seek_variable(FILE * file, int code, int rn)
 
               const int nb = header.get_nb();
               const int rest = nb - sizeof(CDR_header);
-              uint8_t buffer[nb];
-              const int rlen = fread(buffer, 1, rest, file);
+              DynArray(uint8_t, buffer, nb);
+              const int rlen = fread(&buffer[0], 1, rest, file);
               if (rlen != rest)   return -44;   // delimiter not found
             }
 
@@ -188,13 +193,13 @@ read_variable(FILE * file, int code, Coupled_var & var_D,
 
         const int nb = header.get_nb();
         const int rest = nb - sizeof(CDR_header);
-        uint8_t buffer[nb];
-        memcpy(buffer, &header, sizeof(CDR_header));
-        const int rlen = fread(buffer + sizeof(CDR_header), 1, rest, file);
+        DynArray(uint8_t, buffer, nb);
+        memcpy(&buffer[0], &header, sizeof(CDR_header));
+        const int rlen = fread(&buffer[sizeof(CDR_header)], 1, rest, file);
         if (rlen != rest)   return -44;   // delimiter not found
 
         delete var_D.data;
-        var_D.data = new CDR_string(buffer, nb);
+        var_D.data = new CDR_string(&buffer[0], nb);
 
         return 0;
       }
@@ -257,15 +262,15 @@ const CDR_string & cdr = *ctx.var_C.data;
 
    if (cdr.check())
       {
-        CERR << "Bad CDR record:" << endl;
-        cdr.debug(CERR, LOC);
+        get_CERR() << "Bad CDR record:" << endl;
+        cdr.debug(get_CERR(), LOC);
         set_ACK(ctx.var_C, -47);
         return;
       }
 
-   if (cdr.get_rank() > 1)   // not a skalar or vector
+   if (cdr.get_rank() > 1)   // not a scalar or vector
       {
-        CERR << "Bad CDR rank (" << cdr.get_rank() << endl;
+        get_CERR() << "Bad CDR rank (" << cdr.get_rank() << endl;
         set_ACK(ctx.var_C, -32);   // C rank error
         return;
       }
@@ -314,25 +319,25 @@ const int nelm = cdr.header().get_nelm();
    switch(op)
       {
         case 0:  // read fixed-len record
-                 CERR << " INVALID SUBCOMMAND at " << LOC << endl;
+                 get_CERR() << " INVALID SUBCOMMAND at " << LOC << endl;
                  set_ACK(ctx.var_C, -47);    // -47 := INVALID SUBCOMMAND
                  return;
                  break;
 
         case 1:  // write fixed-len record
-                 CERR << " INVALID SUBCOMMAND at " << LOC << endl;
+                 get_CERR() << " INVALID SUBCOMMAND at " << LOC << endl;
                  set_ACK(ctx.var_C, -47);    // -47 := INVALID SUBCOMMAND
                  return;
                  break;
 
         case 2:  // read direct
-                 CERR << " INVALID SUBCOMMAND at " << LOC << endl;
+                 get_CERR() << " INVALID SUBCOMMAND at " << LOC << endl;
                  set_ACK(ctx.var_C, -47);    // -47 := INVALID SUBCOMMAND
                  return;
                  break;
 
         case 3:  // write direct
-                 CERR << " INVALID SUBCOMMAND at " << LOC << endl;
+                 get_CERR() << " INVALID SUBCOMMAND at " << LOC << endl;
                  set_ACK(ctx.var_C, -47);    // -47 := INVALID SUBCOMMAND
                  return;
                  break;
@@ -340,7 +345,7 @@ const int nelm = cdr.header().get_nelm();
         case 4:  // read variable (APL object or line)
                  if (ctx.write)
                     {
-                      CERR << " INVALID SUBCOMMAND at " << LOC << endl;
+                      get_CERR() << " INVALID SUBCOMMAND at " << LOC << endl;
                       set_ACK(ctx.var_C, -47);    // -47 INVALID SUBCOMMAND
                       return;
                     }
@@ -357,7 +362,7 @@ const int nelm = cdr.header().get_nelm();
         case 5:  // write variable (APL object or line)
                  if (!ctx.write)
                     {
-                      CERR << " INVALID SUBCOMMAND at " << LOC << endl;
+                      get_CERR() << " INVALID SUBCOMMAND at " << LOC << endl;
                       set_ACK(ctx.var_C, -47);    // -47 INVALID SUBCOMMAND
                       return;
                     }
@@ -371,13 +376,13 @@ const int nelm = cdr.header().get_nelm();
                  break;
 
         case 6:  // read variable (APL object or line) without CR/LF
-                 CERR << " INVALID SUBCOMMAND at " << LOC << endl;
+                 get_CERR() << " INVALID SUBCOMMAND at " << LOC << endl;
                  set_ACK(ctx.var_C, -47);    // -47 := INVALID SUBCOMMAND
                  return;
                        break;
 
         default: set_ACK(ctx.var_C, -47);    // -47 := INVALID SUBCOMMAND
-                 CERR << " INVALID SUBCOMMAND at " << LOC << endl;
+                 get_CERR() << " INVALID SUBCOMMAND at " << LOC << endl;
                  return;
       }
 }
@@ -585,49 +590,18 @@ is_valid_varname(const uint32_t * varname)
    return false;
 }
 //-----------------------------------------------------------------------------
-/// return true iff ctl/dat is CTL/DAT, or Cxxx/Dxxx (or vice versa)
-static bool
-compare_other(const uint32_t * ctl, const uint32_t * dat)
-{
-   if (*ctl == 'D' && *dat == 'C')   // maybe vice versa: change roles
-      return compare_other(dat, ctl);
-
-   if (*ctl++ != 'C')   return false;   // not CTL or Cxxx
-   if (*dat++ != 'D')   return false;   // not DAT or Dxxx
-
-   // check CTL vs. DAT
-   //
-   if (ctl[0] == 'T' && ctl[1] == 'L' && ctl[2] == 0)
-      {
-        if (dat[0] != 'A')   return false;
-        if (dat[1] != 'T')   return false;
-        if (dat[2] == 0)     return true;
-
-        // continue since CTL might match DTL as well
-      }
-
-   for (; ; ++ctl, ++dat)
-       {
-         if (*ctl != *dat)   return false;
-         if (*ctl == 0)      return true;
-       }
-
-   /* not reached */
-   return false;
-}
-//-----------------------------------------------------------------------------
 bool
 initialize(Coupled_var & var)
 {
    coupled_vars.push_back(var);
-const offered_SVAR * svar =  Svar_DB::find_var(var.key);
-   if (!svar)   return true;
+const uint32_t * varname = Svar_DB::get_varname(var.key);
+   if (!varname)   return true;
 
-   if (*svar->varname == 'C')
+   if (*varname == 'C')
       {
-        Svar_DB::set_control(svar->key, USE_BY_1);
+        Svar_DB::set_control(var.key, USE_BY_1);
       }
-   else if (*svar->varname == 'D')
+   else if (*varname == 'D')
       {
       }
    else                             return true;  // error
@@ -653,10 +627,10 @@ const CDR_header * header = (const CDR_header *)data.c_str();
 
 const int rank = header->rank;
 
-offered_SVAR * svar =  Svar_DB::find_var(var.key);
-   if (svar == 0)   { error_loc = LOC;   return E_VALUE_ERROR; }
+const uint32_t * varname = Svar_DB::get_varname(var.key);
+   if (!varname)   return E_VALUE_ERROR;
 
-   if (*svar->varname == 'D')   // e.g. DAT←value
+   if (*varname == 'D')   // e.g. DAT←value
       {
         delete var.data;
         var.data = new CDR_string((const uint8_t *)data.c_str(), data.size());
@@ -668,7 +642,7 @@ offered_SVAR * svar =  Svar_DB::find_var(var.key);
    if (rank > 1)
       { error_loc = LOC;   return E_RANK_ERROR; }
 
-SV_key key_D = Svar_DB::pairing_key(var.key, compare_other);
+SV_key key_D = Svar_DB::find_pairing_key(var.key);
    if (key_D == 0)   { error_loc = LOC;   return E_VALUE_ERROR; }
 Coupled_var * var_D = 0;
    for (int c = 0; c < coupled_vars.size(); ++c)
@@ -681,9 +655,9 @@ Coupled_var * var_D = 0;
        }
    if (var_D == 0)
       {
-        CERR << "key_D = " << (key_D & 0xFFFF) << endl;
+        get_CERR() << "key_D = " << (key_D & 0xFFFF) << endl;
         for (int c = 0; c < coupled_vars.size(); ++c)
-            CERR << "key = " << (coupled_vars[c].key & 0xFFFF) << endl;
+            get_CERR() << "key = " << (coupled_vars[c].key & 0xFFFF) << endl;
 
         error_loc = LOC;   return E_VALUE_ERROR;
       }
@@ -701,23 +675,23 @@ SVAR_context * ctx = var.context;
         // a sub-command was triggered by a write to C, so C is always written.
         // in addition D may have been read  or written.
         //
-        svar->set_state(false, LOC);
+        Svar_DB::set_state(var.key, false, LOC);
 
         if (ctx->state_D != SVS_NOT_SHARED)   // var D read or written
            {
-             offered_SVAR * svar_D =  Svar_DB::find_var(key_D);
-             if (svar_D == 0)   { error_loc = LOC;   return E_VALUE_ERROR; }
-             svar_D->set_state(ctx->state_D == SVS_IDLE, LOC);
+             if (!Svar_DB::get_varname(key_D)) { error_loc = LOC;
+                                                 return E_VALUE_ERROR; }
+             Svar_DB::set_state(key_D, ctx->state_D == SVS_IDLE, LOC);
            }
       }
    else
       {
         const char * cmd = (const char *)(header + 1) + 4*rank;
         handle_cmd(cmd, data.size(), var, *var_D);
-offered_SVAR * svar_D =  Svar_DB::find_var(key_D);
-   if (svar_D == 0)   { error_loc = LOC;   return E_VALUE_ERROR; }
-   svar->set_state(false, LOC);
-   svar_D->set_state(false, LOC);
+        if (!Svar_DB::get_varname(key_D)) { error_loc = LOC;
+                                            return E_VALUE_ERROR; }
+   Svar_DB::set_state(var.key, false, LOC);
+   Svar_DB::set_state(key_D, false, LOC);
       }
 
    error_loc = "no_error";   return E_NO_ERROR;
