@@ -27,6 +27,7 @@
 #include <sys/types.h>
 
 #include "Archive.hh"
+#include "buildtag.hh"   // for ARCHIVE_SVN
 #include "Common.hh"
 #include "CharCell.hh"
 #include "ComplexCell.hh"
@@ -44,14 +45,15 @@
 #include "Token.hh"
 #include "UCS_string.hh"
 #include "UserFunction.hh"
-#include "Value.hh"
+#include "Value.icc"
 #include "Workspace.hh"
 
 using namespace std;
 
-#define NEED(x, cm)   if (space < (x))                                   \
-                         { if (cm)   out << UNI_PAD_U0;   out << "\n";   \
-                           space = do_indent(); char_mode = false; } out
+/// if there is less than x chars left on the current line then leave
+/// char mode and start a new line indented.
+#define NEED(x)   if (space < (x))   { leave_char_mode();   out << "\n";   \
+                                       space = do_indent(); } out
 
 //-----------------------------------------------------------------------------
 bool
@@ -78,7 +80,7 @@ XML_Saving_Archive::do_indent()
 const int spaces = indent * INDENT_LEN;
    loop(s, spaces)   out << " ";
 
-   return 78 - spaces;
+   return 72 - spaces;
 }
 //-----------------------------------------------------------------------------
 int
@@ -108,33 +110,27 @@ XML_Saving_Archive::find_vid(const Value & val)
 }
 //-----------------------------------------------------------------------------
 void
-XML_Saving_Archive::emit_unicode(Unicode uni, int & space, bool & char_mode)
+XML_Saving_Archive::emit_unicode(Unicode uni, int & space)
 {
-   NEED(2, char_mode) << "";
    if (uni == UNI_ASCII_LF)
       {
-        NEED(2, char_mode) << UNI_PAD_U1 << "A";
-        char_mode = false;
-        NEED(space + 4, char_mode);   // force new line
+        leave_char_mode();
+        out << UNI_PAD_U1 << "A" << "\n";
+        space = do_indent();
       }
    else if (!xml_allowed(uni))
       {
+        space -= leave_char_mode();
         char cc[40];
         snprintf(cc, sizeof(cc), "%X", uni);
-        NEED(1 + strlen(cc), char_mode) << UNI_PAD_U1 << decr(space, cc);
+        NEED(1 + strlen(cc)) << UNI_PAD_U1 << decr(space, cc);
         space--;   // PAD_U1
-        char_mode = false;
-      }
-
-   else if (!char_mode)
-      {
-        NEED(2, char_mode) << UNI_PAD_U2 << uni;
-        space -= 2;   // PAD_U2 + uni
-        char_mode = true;
       }
    else 
       {
-        NEED(1, char_mode) << uni;
+        NEED(1) << "";
+        space -= enter_char_mode();
+        out << uni;
         space--;   // uni
       }
 }
@@ -143,17 +139,13 @@ void
 XML_Saving_Archive::save_UCS(const UCS_string & ucs)
 {
 int space = do_indent();
-char cc[40];
-bool char_mode = false;
    out << decr(space, "<UCS uni=\"");
+   Assert(char_mode == false);
    ++indent;
-   loop(u, ucs.size())
-      {
-        const Unicode uni = ucs[u];
-        emit_unicode(ucs[u], space, char_mode);
-      }
-
-   NEED(3, false) << "\"/>" << endl;
+   loop(u, ucs.size())   emit_unicode(ucs[u], space);
+   leave_char_mode();
+   out << "\"/>" << endl;
+   space -= 2;
    --indent;
 }
 //-----------------------------------------------------------------------------
@@ -180,7 +172,8 @@ unsigned int parent_vid = -1;
 
    loop (r, v.get_rank())
       {
-        snprintf(cc, sizeof(cc), " sh-%llu=\"%llu\"", r, v.get_shape_item(r));
+        snprintf(cc, sizeof(cc), " sh-%llu=\"%llu\"", (unsigned long long)r,
+                 (unsigned long long)(v.get_shape_item(r)));
         out << cc;
       }
 
@@ -192,7 +185,6 @@ XML_Saving_Archive &
 XML_Saving_Archive::save_ravel(const Value & v)
 {
 int space = do_indent();
-bool char_mode = false;
 
 char cc[80];
    snprintf(cc, sizeof(cc), "<Ravel vid=\"%d\" cells=\"", vid);
@@ -201,82 +193,84 @@ char cc[80];
    ++indent;
 const ShapeItem len = v.nz_element_count();
 const Cell * C = &v.get_ravel(0);
-   loop(l, len)
-      {
-        if (char_mode && C->get_cell_type() != CT_CHAR)
-           {
-             out << UNI_PAD_U0;
-             char_mode = false;
-             --space;
-           }
+   loop(l, len)   emit_cell(*C++, space);
 
-        emit_cell(*C++, space, char_mode);
-      }
-
-   NEED(3, false) << "\"/>" << endl;
+   space -= leave_char_mode();
+   out<< "\"/>" << endl;
+   space -= 2;
    --indent;
 
    return *this;
 }
 //-----------------------------------------------------------------------------
 void
-XML_Saving_Archive::emit_cell(const Cell & cell, int & space, bool & char_mode)
+XML_Saving_Archive::emit_cell(const Cell & cell, int & space)
 {
 char cc[80];
    switch(cell.get_cell_type())
       {
-            case CT_CHAR:   // UNI_PAD_U0, UNI_PAD_U1, and UNI_PAD_U2
-                 emit_unicode(cell.get_char_value(), space, char_mode);
+            case CT_CHAR:   // uses UNI_PAD_U0, UNI_PAD_U1, and UNI_PAD_U2
+                 emit_unicode(cell.get_char_value(), space);
                  break;
 
-            case CT_INT:
-                 snprintf(cc, sizeof(cc), "%lld", cell.get_int_value());
-                 NEED(1 + strlen(cc), false) << UNI_PAD_U3 << decr(space, cc);
-                 space--;   // PAD_U3
+            case CT_INT:   // uses UNI_PAD_U3
+                 space -= leave_char_mode();
+                 snprintf(cc, sizeof(cc), "%lld",
+                          (long long)cell.get_int_value());
+                 NEED(1 + strlen(cc)) << UNI_PAD_U3 << decr(--space, cc);
                  break;
 
-            case CT_FLOAT:
+            case CT_FLOAT:   // uses UNI_PAD_U4
+                 space -= leave_char_mode();
                  snprintf(cc, sizeof(cc), "%g", cell.get_real_value());
-                 NEED(1 + strlen(cc), false) << UNI_PAD_U4 << decr(space, cc);
-                 space--;   // PAD_U4
+                 NEED(1 + strlen(cc)) << UNI_PAD_U4 << decr(--space, cc);
                  break;
 
-            case CT_COMPLEX:
+            case CT_COMPLEX:   // uses UNI_PAD_U5
+                 space -= leave_char_mode();
                  snprintf(cc, sizeof(cc), "%gJ%g",
                           cell.get_real_value(), cell.get_imag_value());
-                 NEED(1 + strlen(cc), false) << UNI_PAD_U5 << decr(space, cc);
-                 space--;   // PAD_U5
+                 NEED(1 + strlen(cc)) << UNI_PAD_U5 << decr(--space, cc);
                  break;
 
-            case CT_POINTER:
+            case CT_POINTER:   // uses UNI_PAD_U6
+                 space -= leave_char_mode();
                  {
                    const int vid = find_vid(*cell.get_pointer_value());
                    snprintf(cc, sizeof(cc), "%d", vid);
-                   NEED(1 + strlen(cc), false) << UNI_PAD_U6 << decr(space, cc);
-                   space--;   // PAD_U6
+                   NEED(1 + strlen(cc)) << UNI_PAD_U6 << decr(--space, cc);
                  }
                  break;
 
-            case CT_CELLREF:
+            case CT_CELLREF:   // uses UNI_PAD_U7
+                 space -= leave_char_mode();
                  {
                    Cell * cp = cell.get_lval_value();
                    const int vid = find_owner(cp);
                    const Value & val = values[vid]._val;
                    const ShapeItem offset = val.get_offset(cp);
-                   snprintf(cc, sizeof(cc), "%d[%lld]", vid, offset);
-                   NEED(1 + strlen(cc), false) << UNI_PAD_U7 << decr(space, cc);
-                   space--;   // PAD_U6
+                   snprintf(cc, sizeof(cc), "%d[%lld]", vid, (long long)offset);
+                   NEED(1 + strlen(cc)) << UNI_PAD_U7 << decr(--space, cc);
                  }
                  break;
+
+            default: Assert(0);
       }
 }
 //-----------------------------------------------------------------------------
 void
 XML_Saving_Archive::save_function(const Function & fun)
 {
+const int * eprops = fun.get_exec_properties();
+const APL_time_us creation_time = fun.get_creation_time();
    do_indent();
-   out << "<Function";
+   out << "<Function creation-time=\"" << creation_time
+       << "\" exec-properties=\""
+       << eprops[0] << "," << eprops[1] << ","
+       << eprops[2] << "," << eprops[3] << "\"";
+
    if (fun.is_native())   out << " native=\"1\"";
+
    out << ">" << endl;
    ++indent;
 
@@ -315,8 +309,8 @@ XML_Saving_Archive::save_symtab(const SymbolTable & symtab)
    // collect all symbols in this symbol table.
    //
 int symbol_count = symtab.symbols_allocated();
-Symbol * all_symbols[symbol_count];
-   symtab.get_all_symbols(all_symbols, symbol_count);
+DynArray(Symbol *, all_symbols, symbol_count);
+   symtab.get_all_symbols(&all_symbols[0], symbol_count);
 
    // remove erased symbols
    //
@@ -368,30 +362,8 @@ const Executable & exec = *si.get_executable();
    do_indent();
    out << "<SI-entry level=\"" << si.get_level()
        << "\" pc=\"" << si.get_PC()
-       << "\" line=\"" << exec.get_line(si.get_PC()) << "\"" << flush;
-
-   if (si.eval_arg_F)   // function pending
-      {
-        if (!!si.eval_arg_A)   // function has a left arg
-           {
-             const unsigned int vid_A = find_vid(*si.eval_arg_A.get());
-             out << " vid_arg_A=\"" << vid_A << "\"";
-           }
-
-        {
-          char cc[80];
-          snprintf(cc, sizeof(cc), " Id_arg_F=\"%X\"", si.eval_arg_F->get_Id());
-          out << cc;
-        }
-
-        if (!!si.eval_arg_B)   // function has a right arg
-           {
-             const int vid_B = find_vid(*si.eval_arg_B.get());
-             out << " vid_arg_B=\"" << vid_B << "\"";
-           }
-      }
-
-   out <<">" << endl << flush;
+       << "\" line=\"" << exec.get_line(si.get_PC()) << "\""
+       <<">" << endl << flush;
 
    ++indent;
    do_indent();
@@ -467,7 +439,7 @@ char cc[80];
        << "\" tag=\"" << cc << "\"";
    emit_token_val(tloc.tok);
 
-   out << ">" << endl;
+   out << "/>" << endl;
 }
 //-----------------------------------------------------------------------------
 void
@@ -475,34 +447,33 @@ XML_Saving_Archive::emit_token_val(const Token & tok)
 {
    switch(tok.get_ValueType())
       {
-        case TV_NONE:  out << "/";
+        case TV_NONE:  break;
+
+        case TV_CHAR:  out << " char=\"" << int(tok.get_char_val()) << "\"";
                        break;
 
-        case TV_CHAR:  out << " char=\"" << int(tok.get_char_val()) << "\"/";
+        case TV_INT:   out << " int=\"" << tok.get_int_val() << "\"";
                        break;
 
-        case TV_INT:   out << " int=\"" << tok.get_int_val() << "\"/";
-                       break;
-
-        case TV_FLT:   out << " float=\"" << tok.get_flt_val() << "\"/";
+        case TV_FLT:   out << " float=\"" << tok.get_flt_val() << "\"";
                        break;
 
         case TV_CPX:   out << " real=\"" << tok.get_cpx_real()
-                           << "\" imag=\"" << tok.get_cpx_imag() << "\"/";
+                           << "\" imag=\"" << tok.get_cpx_imag() << "\"";
                        break;
 
         case TV_SYM:   {
                          Symbol * sym = tok.get_sym_ptr();
                          const UCS_string name = sym->get_name();
-                         out << " sym=\"" << name << "\"/";
+                         out << " sym=\"" << name << "\"";
                        }
                        break;
 
-        case TV_LIN:   out << " line=\"" << tok.get_fun_line() << "\"/";
+        case TV_LIN:   out << " line=\"" << tok.get_fun_line() << "\"";
                        break;
 
         case TV_VAL:   { const int vid = find_vid(*tok.get_apl_val());
-                         out << " vid=\"" << vid << "\"/";
+                         out << " vid=\"" << vid << "\"";
                        }
                        break;
 
@@ -514,8 +485,8 @@ XML_Saving_Archive::emit_token_val(const Token & tok)
                                if (i)   out << ",";
                                const Value & val = *idx.values[i].get();
                                if (&val)   out << "vid_" << find_vid(val);
-                               else       out << "-";
-                                out << "\"/>";
+                               else        out << "-";
+                                out << "\"";
                              }
                        }
                        break;
@@ -532,7 +503,7 @@ XML_Saving_Archive::emit_token_val(const Token & tok)
                               const int sym_depth = sym->get_ufun_depth(ufun);
                               out << " ufun-name=\"" << fname
                                   << "\" symbol-level=\"" << sym_depth
-                                  << "\"/>" << endl;
+                                  << "\"" << endl;
                             }
                          else        // primitive or quad function
                             {
@@ -541,7 +512,7 @@ XML_Saving_Archive::emit_token_val(const Token & tok)
                               out << " fun-id=\"" << cc;
                             }
                        }
-                       out << "\"/";
+                       out << "\"";
                        break;
 
         default:       FIXME;
@@ -581,6 +552,7 @@ XML_Saving_Archive::save_vstack_item(const ValueStackItem & vsi)
                  << "\"/>" << endl;
              break;
 
+        default: Assert(0);
       }
 }
 //-----------------------------------------------------------------------------
@@ -594,6 +566,8 @@ tm * t;
      t = gmtime(&now.tv_sec);
    }
 
+const int offset = Workspace::get_v_Quad_TZ().get_offset();   // timezone offset
+
    out <<
 "<?xml version='1.0' encoding='UTF-8' standalone='yes'?>\n"
 "\n"
@@ -657,17 +631,16 @@ tm * t;
 "            <!ATTLIST SI-entry level     CDATA #REQUIRED>\n"
 "            <!ATTLIST SI-entry pc        CDATA #REQUIRED>\n"
 "            <!ATTLIST SI-entry line      CDATA #REQUIRED>\n"
-"            <!ATTLIST SI-entry vid_arg_A CDATA #IMPLIED>\n"
-"            <!ATTLIST SI-entry Id_arg_F  CDATA #IMPLIED>\n"
-"            <!ATTLIST SI-entry vid_arg_B CDATA #IMPLIED>\n"
 "\n"
 "                <!ELEMENT Statements (UCS)>\n"
 "\n"
 "                <!ELEMENT Execute (UCS)>\n"
 "\n"
 "                <!ELEMENT UserFunction (#PCDATA)>\n"
-"                <!ATTLIST UserFunction ufun-name    CDATA #REQUIRED>\n"
-"                <!ATTLIST UserFunction symbol-level CDATA #REQUIRED>\n"
+"                <!ATTLIST UserFunction ufun-name       CDATA #REQUIRED>\n"
+"                <!ATTLIST UserFunction symbol-level    CDATA #REQUIRED>\n"
+"                <!ATTLIST UserFunction creation-time   CDATA #IMPLIED>\n"
+"                <!ATTLIST UserFunction exec-properties CDATA #IMPLIED>\n"
 "\n"
 "                <!ELEMENT Parser (Token*)>\n"
 "                <!ATTLIST Parser assign-pending CDATA #REQUIRED>\n"
@@ -703,7 +676,8 @@ tm * t;
 "           hour=\""     <<  t->tm_hour
      << "\" minute=\""   <<  t->tm_min
      << "\" second=\""   <<  t->tm_sec
-     << "\" timezone=\"" << Workspace::get_v_Quad_TZ().get_offset()
+     << "\" timezone=\"" << offset << "\"" << endl <<
+"           saving_SVN=\"" << ARCHIVE_SVN
      << "\">\n" << endl;
 
    ++indent;
@@ -719,7 +693,6 @@ tm * t;
        {
          const Value & val(*(const Value *)obj);
 
-         if (val.is_forever())   continue;   // e.g. ⍞
          if (val.is_marked())    continue;   // stale
 
          val.unmark();
@@ -811,15 +784,24 @@ CERR << "LVAL CELL in " << p << " at " LOC << endl;
    return *this;
 }
 //=============================================================================
-XML_Loading_Archive::XML_Loading_Archive(const char * _filename)
-   : line(1),
+XML_Loading_Archive::XML_Loading_Archive(const char * _filename, int & dump_fd)
+   : fd(-1),
+     map_start(0),
+     map_length(0),
+     file_start(0),
+     line_start(0),
+     line_no(1),
+     current_char(UNI_ASCII_SPACE),
      data(0),
-     end(0),
+     file_end(0),
      copying(false),
      protection(false),
      reading_vids(false),
+     have_allowed_objects(false),
      filename(_filename)
 {
+   Log(LOG_archive)   CERR << "Loading file " << filename << endl;
+
    fd = open(filename, O_RDONLY);
    if (fd == -1)   return;
 
@@ -832,26 +814,55 @@ struct stat st;
         return;
       }
 
-void * map = mmap(0, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
-   if (map == MAP_FAILED)
+   map_length = st.st_size;
+   map_start = mmap(0, map_length, PROT_READ, MAP_SHARED, fd, 0);
+   if (map_start == MAP_FAILED)
       {
         CERR << "mmap() failed: " << strerror(errno) << endl;
         close(fd);
         fd = -1;
+        return;
       }
 
    // success
    //
-   file_start = (const UTF8 *)map;
-   end = file_start + st.st_size;
+   file_start = (const UTF8 *)map_start;
+   file_end = file_start + map_length;
+
    reset();
+
+   if (!strncmp((const char *)file_start, "#!", 2) ||    // a )DUMP file
+       !strncmp((const char *)file_start, "⍝!", 4))      // a library
+      {
+        // the file was either written with )DUMP or is a library.
+        // Return the open file descriptor (the destructor will unmap())
+        //
+        dump_fd = fd;
+        fd = -1;   // file will be closed via dump_fd
+        return;
+      }
+
+   if (strncmp((const char *)file_start, "<?xml", 5))   // not an xml file
+      {
+        CERR << "file " << filename << " does not " << endl
+             << "have the format of a GNU APL .xml or .apl file" << endl;
+        close(fd);
+        fd = -1;
+        return;
+      }
+}
+//-----------------------------------------------------------------------------
+XML_Loading_Archive::~XML_Loading_Archive()
+{
+   if (map_start)   munmap((char *)map_start, map_length);
+   if (fd != -1)   close(fd);
 }
 //-----------------------------------------------------------------------------
 void
 XML_Loading_Archive::reset()
 {
    line_start = data = file_start;
-   line = 1;
+   line_no = 1;
    next_tag(LOC);
 }
 //-----------------------------------------------------------------------------
@@ -881,7 +892,7 @@ XML_Loading_Archive::read_vids()
 void
 XML_Loading_Archive::where(ostream & out)
 {
-   out << "line=" << line << "+" << (data - line_start) << " '";
+   out << "line=" << line_no << "+" << (data - line_start) << " '";
 
    loop(j, 40)   { if (data[j] == 0x0A)   break;   out << data[j]; }
    out << "'" << endl;
@@ -890,7 +901,7 @@ XML_Loading_Archive::where(ostream & out)
 void
 XML_Loading_Archive::where_att(ostream & out)
 {
-   out << "line=" << line << "+" << (attributes - line_start) << " '";
+   out << "line=" << line_no << "+" << (attributes - line_start) << " '";
 
    loop(j, 40)
       {
@@ -905,12 +916,12 @@ XML_Loading_Archive::where_att(ostream & out)
 bool
 XML_Loading_Archive::get_uni()
 {
-   if (data > end)   return true;   // EOF
+   if (data > file_end)   return true;   // EOF
 
 int len = 0;
    current_char = UTF8_string::toUni(data, len);
    data += len;
-   if (current_char == 0x0A)   { ++line;   line_start = data; }
+   if (current_char == 0x0A)   { ++line_no;   line_start = data; }
    return false;
 }
 //-----------------------------------------------------------------------------
@@ -928,7 +939,7 @@ XML_Loading_Archive::expect_tag(const char * prefix, const char * loc) const
         CERR << "   Got tag ";
         print_tag(CERR);
         CERR << " when expecting tag " << prefix
-             << " at " << loc << "  line " << line << endl;
+             << " at " << loc << "  line " << line_no << endl;
         DOMAIN_ERROR;
       }
 }
@@ -1032,7 +1043,7 @@ again:
 /*
    CERR << "See tag ";
    for (const UTF8 * t = tag_name; t < attributes; ++t)   CERR << (char)*t;
-   CERR << " at " << loc << " line " << line << endl;
+   CERR << " at " << loc << " line " << line_no << endl;
 */
 
    return false;
@@ -1043,14 +1054,49 @@ XML_Loading_Archive::read_Workspace()
 {
    expect_tag("Workspace", LOC);
 
+const int offset   = find_int_attr("timezone", false, 10);
 const UTF8 * wsid  = find_attr("wsid",         false);
-const int year     = find_int_attr("year",     false, 10);
-const int month    = find_int_attr("month",    false, 10);
-const int day      = find_int_attr("day",      false, 10);
-const int hour     = find_int_attr("hour",     false, 10);
-const int minute   = find_int_attr("minute",   false, 10);
-const int second   = find_int_attr("second",   false, 10);
-const int tzone    = find_int_attr("timezone", false, 10);
+
+int year  = find_int_attr("year",     false, 10);
+int mon   = find_int_attr("month",    false, 10);
+int day   = find_int_attr("day",      false, 10);
+int hour  = find_int_attr("hour",     false, 10);
+int min   = find_int_attr("minute",   false, 10);
+int sec   = find_int_attr("second",   false, 10);
+
+   sec  += offset          % 60;
+   min  += (offset /   60) % 60;
+   hour += (offset / 3600) % 60;
+   if (sec  >= 60)   { sec  -= 60;   ++min;  }
+   if (min  >= 60)   { min  -= 60;   ++hour; }
+   if (hour >= 24)   { hour -= 24;   ++day;  }
+
+bool next_month = false;
+   switch(day)
+      {
+        case 32: next_month = true;
+                 break;
+
+        case 31: if (mon == 4 || mon == 6 || mon == 9 || mon == 11)
+                    next_month = true;
+                 break;
+
+        case 30: if (mon == 2)   next_month = true;
+                 break;
+
+        case 29: if (mon != 2)         break;               // not februry
+                 if (year & 3)         next_month = true;   // not leap yer
+                 // the above fails if someone loads a workspace that
+                 // was saved around midnight on 2/28/2100. Bad luck!
+                 break;
+
+        default: break;
+      }
+
+   if (next_month)   { day = 1; ++mon;  }
+   if (mon > 12)     { mon = 1; ++year; }
+
+   Log(LOG_archive)   CERR << "read_Workspace() " << endl;
 
 const char * tag_order[] = { "Value",          "Ravel",
                              "SymbolTable",    "Symbol",
@@ -1087,10 +1133,18 @@ const char ** tag_pos = tag_order;
    //
    Value::erase_stale(LOC);
 
-const char * tz_sign = (tzone < 0) ? "" : "+";
-   COUT << "SAVED " << year << "-" << month << "-" << day
-        << "  " << hour << ":" << minute << ":" << second
-        << " (GMT" << tz_sign << tzone/3600 << ")" << endl;
+   if (reading_vids)   return;
+
+const char * tz_sign = (offset < 0) ? "" : "+";
+   COUT << "SAVED "
+        << setfill('0') << year        << "-"
+        << setw(2)      << mon         << "-"
+        << setw(2)      << day         << " "
+        << setw(2)      << hour        << ":"
+        << setw(2)      << min         << ":"
+        << setw(2)      << sec         << " (GMT"
+        << tz_sign      << offset/3600 << ")"
+        << setfill(' ') <<  endl;
 
    if (!copying)
       {
@@ -1099,6 +1153,13 @@ const char * tz_sign = (tzone < 0) ? "" : "+";
 
         Workspace::set_WS_name(UCS_string(UTF8_string(wsid, end - wsid)));
       }
+
+   if (have_allowed_objects && allowed_objects.size())
+      {
+        CERR << "NOT COPIED:";
+        loop(a, allowed_objects.size())   CERR << " " << allowed_objects[a];
+        CERR << endl;
+      }
 }
 //-----------------------------------------------------------------------------
 void
@@ -1106,11 +1167,12 @@ XML_Loading_Archive::read_Value()
 {
    expect_tag("Value", LOC);
 
-const int  flg = find_int_attr("flg", false, 16);
 const int  vid = find_int_attr("vid", false, 10);
 const int  parent = find_int_attr("parent", true, 10);
 const int  rk  = find_int_attr("rk",  false, 10);
 
+   Log(LOG_archive)   CERR << "  read_Value() vid=" << vid << endl;
+
    if (reading_vids)
       {
          parents.push_back(parent);
@@ -1178,6 +1240,7 @@ const Unicode type = UTF8_string::toUni(first, len);
    switch (type)
       {
         case UNI_PAD_U0: // end of UNI_PAD_U2
+        case '\n':       // end of UNI_PAD_U2 (fix old bug)
              first += len;
              break;
 
@@ -1251,7 +1314,7 @@ const Unicode type = UTF8_string::toUni(first, len);
              }
              break;
 
-        default: Q1(type) Q1(line) DOMAIN_ERROR;
+        default: Q1(type) Q1(line_no) DOMAIN_ERROR;
       }
 }
 //-----------------------------------------------------------------------------
@@ -1264,7 +1327,7 @@ XML_Loading_Archive::read_chars(UCS_string & ucs, const UTF8 * & utf)
        {
          if (*utf == '"')   break;   // end of attribute value
 
-          if (char_mode && *utf < 0x80)
+          if (char_mode && *utf < 0x80 && *utf != '\n')
              {
                ucs.append(Unicode(*utf++));
                continue;
@@ -1298,6 +1361,18 @@ XML_Loading_Archive::read_chars(UCS_string & ucs, const UTF8 * & utf)
               continue;
             }
 
+         if (type == '\n')   // end of char_mode (fix old bug)
+            {
+              // due to an old bug, the trailing UNI_PAD_U0 may be missing
+              // at the end of the line. We therefore reset char_mode at the
+              // end of the line so that workspaces save with that fault can
+              // be read.
+              //
+              utf += len;   // skip UNI_PAD_U0
+              char_mode = false;
+              continue;
+            }
+
          break;
        }
 
@@ -1312,6 +1387,8 @@ XML_Loading_Archive::read_Ravel()
 const int vid = find_int_attr("vid", false, 10);
 const UTF8 * cells = find_attr("cells", false);
 
+   Log(LOG_archive)   CERR << "    read_Ravel() vid=" << vid << endl;
+
    Assert(vid < values.size());
 Value_P val = values[vid];
 
@@ -1330,7 +1407,11 @@ Cell * end = C + count;
         while (*cells <= ' ')   ++cells;   // skip trailing whitespace
       }
 
-   Assert(C == end);   // all cells read
+   if (C != end)   // unless all cells read
+      {
+        CERR << "vid=" << vid << endl;
+        FIXME;
+      }
 
    {
      int len = 0;
@@ -1355,15 +1436,22 @@ XML_Loading_Archive::read_Variable(int d, Symbol & symbol)
 const unsigned int vid = find_int_attr("vid", false, 10);
    Assert(vid < values.size());
 
-   if (d != 0)   symbol.push();
+   Log(LOG_archive)   CERR << "      read_Variable() vid=" << vid
+                           << " name=" << symbol.get_name() << endl;
 
    // some system variables are  saved for troubleshooting purposes, but
    // should not be loaded...
    //
-   if (symbol.is_readonly())   return;
-   if (symbol.get_name() == id_name(ID_Quad_NLT))    return;
+   if (symbol.is_readonly())                         return;
+   if (symbol.get_name().starts_iwith("⎕NLT"))       return;
+   if (symbol.get_name() == id_name(ID_Quad_SVE))    return;
    if (symbol.get_name() == id_name(ID_Quad_SYL))    return;
 
+   if (!values[vid])   return;   // value filtered out
+
+   if (d != 0)   symbol.push();
+
+
    try
       {
         symbol.assign(values[vid], LOC);
@@ -1379,7 +1467,7 @@ void
 XML_Loading_Archive::read_Label(int d, Symbol & symbol)
 {
 const int value = find_int_attr("value", false, 10);
-   if (d == 0)   symbol.pop(false);
+   if (d == 0)   symbol.pop();
    symbol.push_label(Function_Line(value));
 }
 //-----------------------------------------------------------------------------
@@ -1387,6 +1475,16 @@ void
 XML_Loading_Archive::read_Function(int d, Symbol & symbol)
 {
 const int native = find_int_attr("native", true, 10);
+const APL_time_us creation_time = find_int_attr("creation-time", true, 10);
+int eprops[4] = { 0, 0, 0, 0 };
+const UTF8 * ep = find_attr("exec-properties", true);
+   if (ep)
+      {
+        sscanf((const char *)ep, "%d,%d,%d,%d",
+               eprops, eprops + 1, eprops + 2, eprops+ 3);
+      }
+
+   Log(LOG_archive)   CERR << "      read_Function() native=" << native << endl;
 
    next_tag(LOC);
    expect_tag("UCS", LOC);
@@ -1402,31 +1500,40 @@ UCS_string text;
         NativeFunction * nfun = NativeFunction::fix(text, symbol.get_name());
         if (nfun)   // fix succeeded
            {
-             if (d == 0)   symbol.pop(false);
+             nfun->set_creation_time(creation_time);
+             if (d == 0)   symbol.pop();
              symbol.push_function(nfun);
            }
         else        // fix failed
            {
              CERR << "   *** loading of native function " << text
-                  << " failed" << endl << endl;;
-             if (d == 0)   symbol.pop(false);
+                  << " failed" << endl << endl;
+             if (d == 0)   symbol.pop();
              symbol.push();
            }
       }
    else
       {
         int err = 0;
-        UserFunction * fun = UserFunction::fix(text, err, false, LOC, filename);
+        UCS_string creator(filename);
+        creator.append(UNI_ASCII_COLON);
+        creator.append_number(line_no);
+        UTF8_string creator_utf8(creator);
 
-        if (d == 0)   symbol.pop(false);
-        if (fun)
+        UserFunction * ufun = UserFunction::fix(text, err, false,
+                                               LOC, creator_utf8);
+
+        if (d == 0)   symbol.pop();
+        if (ufun)
            {
-             symbol.push_function(fun);
+             ufun->set_creation_time(creation_time);
+             symbol.push_function(ufun);
+             ufun->set_exec_properties(eprops);
            }
         else
            {
-             CERR << "fix(" << symbol.get_name() << ") failed at line "
-                  << err << " (" << Workspace::more_error() << ")" << endl;
+             CERR << "    ⎕FX " << symbol.get_name() << " failed: "
+                  << Workspace::more_error() << endl;
              symbol.push();
            }
       }
@@ -1435,10 +1542,14 @@ UCS_string text;
 void
 XML_Loading_Archive::read_Shared_Variable(int d, Symbol & symbol)
 {
-const SV_key key = find_int_attr("key", false, 10);
+// const SV_key key = find_int_attr("key", false, 10);
    if (d != 0)   symbol.push();
 
-   symbol.share_var(key);
+   CERR << "WARNING: workspace was )SAVEd with a shared variable "
+        << symbol.get_name() << endl
+        << " (shared variables are not restored by )LOAD or )COPY)" << endl;
+
+   // symbol.share_var(key);
 }
 //-----------------------------------------------------------------------------
 void
@@ -1446,6 +1557,8 @@ XML_Loading_Archive::read_SymbolTable()
 {
 const int size = find_int_attr("size", false, 10);
 
+   Log(LOG_archive)   CERR << "  read_SymbolTable() vid=" << endl;
+
    loop(s, size)
       {
         next_tag(LOC);
@@ -1470,6 +1583,17 @@ UCS_string name_ucs;
      name_ucs = UCS_string(name_utf);
    }
 
+   Log(LOG_archive)   CERR << "    read_Symbol() name=" << name_ucs << endl;
+
+   // ⎕NLT was removed, but could lurk around in old workspaces.
+   //
+   if (name_ucs == UCS_string(UTF8_string("⎕NLT")))
+      {
+        Log(LOG_archive)   CERR << "        skipped at " << LOC << endl;
+        skip_to_tag("/Symbol");
+        return;
+      }
+
 const int depth = find_int_attr("stack-size", false, 10);
 
    // lookup symbol, trying ⎕xx first
@@ -1478,10 +1602,12 @@ Symbol * symbol = Workspace::lookup_existing_symbol(name_ucs);
 
    // we do NOT copy if:
    //
-   //  )PCOPY and the symbol exists, or 
-   //  there is an object list and this symbol is not contained in the list
+   // 1. )PCOPY and the symbol exists, or 
+   // 2.  there is an object list and this symbol is not contained in the list
    //
-const bool no_copy = (symbol && protection) || ! is_allowed(name_ucs);
+const bool is_protected = symbol && protection;
+const bool is_selected = name_ucs.contained_in(allowed_objects);
+bool no_copy = is_protected || (have_allowed_objects && !is_selected);
 
    if (reading_vids)
       {
@@ -1490,8 +1616,7 @@ const bool no_copy = (symbol && protection) || ! is_allowed(name_ucs);
         //
         if (no_copy || (depth == 0))
            {
-             //
-             COUT << "NOT COPIED: " << name_ucs << endl;
+             Log(LOG_archive)   CERR << "        skipped at " << LOC << endl;
              skip_to_tag("/Symbol");
              return;
            }
@@ -1508,6 +1633,28 @@ const bool no_copy = (symbol && protection) || ! is_allowed(name_ucs);
         return;
       }
 
+   // in a )COPY without dedicated objects only
+   // ⎕CT, ⎕FC, ⎕IO, ⎕LX, ⎕PP, ⎕PR, and ⎕RL shall be copied
+   // shall be copied.
+   //
+   if (!have_allowed_objects       &&   // no dedicated object list
+        copying                    &&   // )COPY
+        Avec::is_quad(name_ucs[0]) &&   // ⎕xx
+        name_ucs.size() == 3       &&
+        ! (
+            ( name_ucs[1] == 'C' && name_ucs[2] == 'T' ) ||
+            ( name_ucs[1] == 'F' && name_ucs[2] == 'C' ) ||
+            ( name_ucs[1] == 'I' && name_ucs[2] == 'O' ) ||
+            ( name_ucs[1] == 'L' && name_ucs[2] == 'X' ) ||
+            ( name_ucs[1] == 'P' && name_ucs[2] == 'P' ) ||
+            ( name_ucs[1] == 'P' && name_ucs[2] == 'R' ) ||
+            ( name_ucs[1] == 'R' && name_ucs[2] == 'L' )
+          ))
+      {
+        Log(LOG_archive)   CERR << name_ucs << " not copied at " << LOC << endl;
+        no_copy = true;
+      }
+
    if (copying)
       {
         if (no_copy || (depth == 0))
@@ -1517,7 +1664,22 @@ const bool no_copy = (symbol && protection) || ! is_allowed(name_ucs);
            }
       }
 
-   if (symbol == 0)   symbol = Workspace::lookup_symbol(name_ucs);
+   // remove this symbol from allowed_objects so that we can print NOT COPIED
+   // at the end for all objects that are still in the list.
+   //
+   loop(a, allowed_objects.size())
+      {
+        if (allowed_objects[a] == name_ucs)
+             {
+               allowed_objects.erase(allowed_objects.begin() + a);
+               break;
+           }
+      }
+
+   if (symbol == 0)
+      {
+        symbol = Workspace::lookup_symbol(name_ucs);
+      }
    Assert(symbol);
 
    loop(d, depth)
@@ -1551,6 +1713,8 @@ XML_Loading_Archive::read_StateIndicator()
           return;
       }
 
+   Log(LOG_archive)   CERR << "read_StateIndicator()" << endl;
+
 const int levels = find_int_attr("levels", false, 10);
 
    loop(l, levels)
@@ -1571,14 +1735,11 @@ void
 XML_Loading_Archive::read_SI_entry(int lev)
 {
 const int level     = find_int_attr("level",     false, 10);
-const int pc        = find_int_attr("pc",        false, 10);
-const int line      = find_int_attr("line",      false, 10);
-const int vid_arg_A = find_int_attr("vid_arg_A", true, 10);
-const int Id_arg_F  = find_int_attr("Id_arg_F",  true, 16);
-const int vid_arg_B = find_int_attr("vid_arg_B", true, 10);
+
+   Log(LOG_archive)   CERR << "read_SI_entry() level=" << level << endl;
 
    next_tag(LOC);
-const Executable * exec = 0;
+Executable * exec = 0;
    if      (is_tag("Execute"))        exec = read_Execute();
    else if (is_tag("Statements"))     exec = read_Statement();
    else if (is_tag("UserFunction"))   exec = read_UserFunction();
@@ -1590,17 +1751,10 @@ const Executable * exec = 0;
    Workspace::push_SI(exec, LOC);
 StateIndicator * si = Workspace::SI_top();
    Assert(si);
-   if (Id_arg_F != -1)
-      {
-        si->eval_arg_F = get_system_function(Id(Id_arg_F));
-        if (vid_arg_A != -1)  si->eval_arg_A = values[vid_arg_A];
-        if (vid_arg_B != -1)  si->eval_arg_B = values[vid_arg_B];
-      }
-
    read_Parsers(*si);
 }
 //-----------------------------------------------------------------------------
-const Executable *
+Executable *
 XML_Loading_Archive::read_Execute()
 {
    next_tag(LOC);
@@ -1611,11 +1765,11 @@ UCS_string text;
    next_tag(LOC);
    expect_tag("/Execute", LOC);
 
-const Executable * exec = new ExecuteList(text, LOC);
+Executable * exec = new ExecuteList(text, LOC);
    return exec;
 }
 //-----------------------------------------------------------------------------
-const Executable *
+Executable *
 XML_Loading_Archive::read_Statement()
 {
    next_tag(LOC);
@@ -1627,11 +1781,11 @@ UCS_string text;
    next_tag(LOC);
    expect_tag("/Statements", LOC);
 
-const Executable * exec = new StatementList(text, LOC);
+Executable * exec = new StatementList(text, LOC);
    return exec;
 }
 //-----------------------------------------------------------------------------
-const Executable *
+Executable *
 XML_Loading_Archive::read_UserFunction()
 {
 const int level = find_int_attr("symbol-level", false, 10);
@@ -1647,9 +1801,9 @@ Symbol * symbol = Workspace::lookup_symbol(name_ucs);
    Assert(level < symbol->value_stack_size());
 ValueStackItem & vsi = (*symbol)[level];
    Assert(vsi.name_class == NC_FUNCTION);
-const Function * fun = vsi.sym_val.function;
+Function * fun = vsi.sym_val.function;
    Assert(fun);
-const UserFunction * ufun = fun->get_ufun1();
+UserFunction * ufun = fun->get_ufun1();
    Assert(fun == ufun);
 
    return ufun;
@@ -1756,7 +1910,6 @@ const TokenTag tag = TokenTag(find_int_attr("tag", false, 16));
 
         case TV_INDEX: 
              {
-               const unsigned int rank = find_int_attr("rank", false, 10);
                char * vids = (char *)find_attr("index", false);
                IndexExpr & idx = *new IndexExpr(ASS_none, LOC);
                while (*vids != '"')
@@ -1772,7 +1925,6 @@ const TokenTag tag = TokenTag(find_int_attr("tag", false, 16));
                          Assert1(*vids == 'i');   ++vids;
                          Assert1(*vids == 'd');   ++vids;
                          Assert1(*vids == '_');   ++vids;
-                         char * end;
                          const unsigned int vid = strtol(vids, &vids, 10);
                          Assert(vid < values.size());
                          idx.add(values[vid]);
