@@ -29,7 +29,7 @@
 #include "Parser.hh"
 #include "PrintOperator.hh"
 #include "UTF8_string.hh"
-#include "Value.hh"
+#include "Value.icc"
 
 using namespace std;
 
@@ -244,7 +244,7 @@ Avec::map_alternative_char(Unicode alt_av)
    // map characters that look similar to characters used in GNU APL
    // to the GNU APL character.
    //
-   switch(alt_av)
+   switch((int)alt_av)
       {
         case 0x005E: return AV_AND;              //  map ^ to ∧
         case 0x007C: return AV_DIVIDES;          //  map | to ∣
@@ -261,6 +261,7 @@ Avec::map_alternative_char(Unicode alt_av)
         case 0x22BC: return AV_NAND;             //  map ⊽ to ⍲
         case 0x22BD: return AV_NOR;              //  map ⊼ to ⍱
         case 0x22C4: return AV_DIAMOND;          //  map ⋄ to ◊
+        case 0x2377: return AV_EPSILON_UBAR;     //  map ⍷ to ⋸
         case 0x25AF: return AV_Quad_Quad;        //  map ▯ to ⎕
         case 0x25E6: return AV_RING_OPERATOR;    //  map ◦ to ∘
         case 0x26AA: return AV_CIRCLE;           //  map ⚪ to ○
@@ -269,6 +270,7 @@ Avec::map_alternative_char(Unicode alt_av)
         case 0x2B25: return AV_DIAMOND;          //  map ⬥ to ◊
         case 0x2B26: return AV_DIAMOND;          //  map ⬦ to ◊
         case 0x2B27: return AV_DIAMOND;          //  map ⬧ to ◊
+        default:     break;
       }
 
    return Invalid_CHT;
@@ -283,6 +285,8 @@ Avec::is_known_char(Unicode av)
 #define char_df1(_n, u, _t, _f, _p)  case u: 
 #include "Avec.def"
           return true;
+
+        default: break;
       }
 
    return false;   // not found
@@ -296,7 +300,7 @@ Avec::need_UCS(Unicode uni)
 
 const CHT_Index idx = find_char(uni);
    if (idx == Invalid_CHT)   return true;           // char not in GNU APL's ⎕AV
-   if (Command::unicode_to_cp(uni) == 0xB0)   return true;   // not in IBM's ⎕AV
+   if (unicode_to_cp(uni) == 0xB0)   return true;   // not in IBM's ⎕AV
 
    return false;
 }
@@ -391,3 +395,175 @@ Avec::superscript(uint32_t i)
    return Unicode(0x207A);
 }
 //-----------------------------------------------------------------------------
+/* the IBM APL2 character set shown in lrm figure 68 on page 470
+
+   The table is indexed with an 8-bit position in IBM's ⎕AV and returns
+   the Unicode for that position
+ */
+static const uint32_t ibm_av[] =
+{
+  0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007, 
+  0x0008, 0x0009, 0x000A, 0x000B, 0x000C, 0x000D, 0x000E, 0x000F, 
+  0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016, 0x0017, 
+  0x0018, 0x0019, 0x001A, 0x001B, 0x001C, 0x001D, 0x001E, 0x001F, 
+  0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027, 
+  0x0028, 0x0029, 0x002A, 0x002B, 0x002C, 0x002D, 0x002E, 0x002F, 
+  0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037, 
+  0x0038, 0x0039, 0x003A, 0x003B, 0x003C, 0x003D, 0x003E, 0x003F, 
+  0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047, 
+  0x0048, 0x0049, 0x004A, 0x004B, 0x004C, 0x004D, 0x004E, 0x004F, 
+  0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057, 
+  0x0058, 0x0059, 0x005A, 0x005B, 0x005C, 0x005D, 0x005E, 0x005F, 
+  0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067, 
+  0x0068, 0x0069, 0x006A, 0x006B, 0x006C, 0x006D, 0x006E, 0x006F, 
+  0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077, 
+  0x0078, 0x0079, 0x007A, 0x007B, 0x007C, 0x007D, 0x007E, 0x007F, 
+  0x00C7, 0x00FC, 0x00E9, 0x00E2, 0x00E4, 0x00E0, 0x00E5, 0x00E7,
+  0x00EA, 0x00EB, 0x00E8, 0x00EF, 0x00EE, 0x00EC, 0x00C4, 0x00C5,
+  0x2395, 0x235E, 0x2339, 0x00F4, 0x00F6, 0x00F2, 0x00FB, 0x00F9,
+  0x22A4, 0x00D6, 0x00DC, 0x00F8, 0x00A3, 0x22A5, 0x2376, 0x2336,
+  0x00E1, 0x00ED, 0x00F3, 0x00FA, 0x00F1, 0x00D1, 0x00AA, 0x00BA,
+  0x00BF, 0x2308, 0x00AC, 0x00BD, 0x222A, 0x00A1, 0x2355, 0x234E,
+  0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x235F, 0x2206, 0x2207,
+  0x2192, 0x2563, 0x2551, 0x2557, 0x255D, 0x2190, 0x230A, 0x2510,
+  0x2514, 0x2534, 0x252C, 0x251C, 0x2500, 0x253C, 0x2191, 0x2193,
+  0x255A, 0x2554, 0x2569, 0x2566, 0x2560, 0x2550, 0x256C, 0x2261,
+  0x2378, 0x22F8, 0x2235, 0x2337, 0x2342, 0x233B, 0x22A2, 0x22A3,
+  0x25CA, 0x2518, 0x250C, 0x2588, 0x2584, 0x00A6, 0x00CC, 0x2580,
+  0x237A, 0x2379, 0x2282, 0x2283, 0x235D, 0x2372, 0x2374, 0x2371,
+  0x233D, 0x2296, 0x25CB, 0x2228, 0x2373, 0x2349, 0x2208, 0x2229,
+  0x233F, 0x2340, 0x2265, 0x2264, 0x2260, 0x00D7, 0x00F7, 0x2359,
+  0x2218, 0x2375, 0x236B, 0x234B, 0x2352, 0x00AF, 0x00A8, 0x00A0
+};
+
+const Unicode *
+Avec::IBM_quad_AV()
+{
+   return (const Unicode *)ibm_av;
+}
+//-----------------------------------------------------------------------------
+Avec::Unicode_to_IBM_codepoint Avec::inverse_ibm_av[128] =
+{
+  { 0x00A0, 255 }, { 0x00A1, 173 }, { 0x00A3, 156 }, { 0x00A6, 221 },
+  { 0x00A8, 254 }, { 0x00AA, 166 }, { 0x00AC, 170 }, { 0x00AF, 253 },
+  { 0x00BA, 167 }, { 0x00BD, 171 }, { 0x00BF, 168 }, { 0x00C4, 142 },
+  { 0x00C5, 143 }, { 0x00C7, 128 }, { 0x00CC, 222 }, { 0x00D1, 165 },
+  { 0x00D6, 153 }, { 0x00D7, 245 }, { 0x00DC, 154 }, { 0x00E0, 133 },
+  { 0x00E1, 160 }, { 0x00E2, 131 }, { 0x00E4, 132 }, { 0x00E5, 134 },
+  { 0x00E7, 135 }, { 0x00E8, 138 }, { 0x00E9, 130 }, { 0x00EA, 136 },
+  { 0x00EB, 137 }, { 0x00EC, 141 }, { 0x00ED, 161 }, { 0x00EE, 140 },
+  { 0x00EF, 139 }, { 0x00F1, 164 }, { 0x00F2, 149 }, { 0x00F3, 162 },
+  { 0x00F4, 147 }, { 0x00F6, 148 }, { 0x00F7, 246 }, { 0x00F8, 155 },
+  { 0x00F9, 151 }, { 0x00FA, 163 }, { 0x00FB, 150 }, { 0x00FC, 129 },
+  { 0x2190, 189 }, { 0x2191, 198 }, { 0x2192, 184 }, { 0x2193, 199 },
+  { 0x2206, 182 }, { 0x2207, 183 }, { 0x2208, 238 }, { 0x2218, 248 },
+  { 0x2228, 235 }, { 0x2229, 239 }, { 0x222A, 172 }, { 0x2235, 210 },
+  { 0x2260, 244 }, { 0x2261, 207 }, { 0x2264, 243 }, { 0x2265, 242 },
+  { 0x2282, 226 }, { 0x2283, 227 }, { 0x2296, 233 }, { 0x22A2, 214 },
+  { 0x22A3, 215 }, { 0x22A4, 152 }, { 0x22A5, 157 }, { 0x22F8, 209 },
+  { 0x2308, 169 }, { 0x230A, 190 }, { 0x2336, 159 }, { 0x2337, 211 },
+  { 0x2339, 146 }, { 0x233B, 213 }, { 0x233D, 232 }, { 0x233F, 240 },
+  { 0x2340, 241 }, { 0x2342, 212 }, { 0x2349, 237 }, { 0x234B, 251 },
+  { 0x234E, 175 }, { 0x2352, 252 }, { 0x2355, 174 }, { 0x2359, 247 },
+  { 0x235D, 228 }, { 0x235E, 145 }, { 0x235F, 181 }, { 0x236B, 250 },
+  { 0x2371, 231 }, { 0x2372, 229 }, { 0x2373, 236 }, { 0x2374, 230 },
+  { 0x2375, 249 }, { 0x2376, 158 }, { 0x2378, 208 }, { 0x2379, 225 },
+  { 0x237A, 224 }, { 0x2395, 144 }, { 0x2500, 196 }, { 0x2502, 179 },
+  { 0x250C, 218 }, { 0x2510, 191 }, { 0x2514, 192 }, { 0x2518, 217 },
+  { 0x251C, 195 }, { 0x2524, 180 }, { 0x252C, 194 }, { 0x2534, 193 },
+  { 0x253C, 197 }, { 0x2550, 205 }, { 0x2551, 186 }, { 0x2554, 201 },
+  { 0x2557, 187 }, { 0x255A, 200 }, { 0x255D, 188 }, { 0x2560, 204 },
+  { 0x2563, 185 }, { 0x2566, 203 }, { 0x2569, 202 }, { 0x256C, 206 },
+  { 0x2580, 223 }, { 0x2584, 220 }, { 0x2588, 219 }, { 0x2591, 176 },
+  { 0x2592, 177 }, { 0x2593, 178 }, { 0x25CA, 216 }, { 0x25CB, 234 }
+};
+
+// the inverse mapping table of ibm_av above. We provide only the upper
+// (non-ASCII) half to spead up bsearch() in the map.
+
+#if 0
+void
+Avec::print_inverse_IBM_quad_AV()
+{
+   // a helper function that sorts ibm_av by Unicode and prints it on CERR
+   //
+   // To use it change #if 0 to #if ` bove, recompile, start apl
+   // and enter command )OUT QQQ
+   ..
+   loop(c, 128)
+      {
+        inverse_ibm_av[c].uni = -1;
+        inverse_ibm_av[c].cp  = -1;
+      }
+
+int current_max = 0;
+const uint32_t * upper = ibm_av + 128;
+Unicode_to_IBM_codepoint * map = inverse_ibm_av;
+   loop(c, 128)
+      {
+        // find next Unicode after current_max
+        //
+        int next_idx = -1;
+        loop(n, 128)
+           {
+             const int uni = upper[n];
+             if (uni <= current_max)   continue;   // already done
+             if (next_idx == -1)               next_idx = n;
+             else if (uni < upper[next_idx])   next_idx = n;
+           }
+        current_max = map->uni = upper[next_idx];
+        map->cp  = next_idx;
+        ++map;
+      }
+
+   loop(row, 32)
+      {
+        CERR << " ";
+        loop(col, 4)
+           {
+             const int pos = col + 4*row;
+             CERR << " { " << HEX4(inverse_ibm_av[pos].uni) << ", "
+                  << 128 + inverse_ibm_av[pos].cp << " }";
+             if (pos < 127)   CERR << ",";
+           }
+        CERR << endl;
+      }
+
+   exit(1);
+}
+#endif
+//-----------------------------------------------------------------------------
+/// compare the unicodes of two chars in the IBM ⎕AV
+int
+Avec::compare_uni(const void * key, const void * entry)
+{
+   return *(const Unicode *)key -
+          ((const Unicode_to_IBM_codepoint *)entry)->uni;
+}
+//-----------------------------------------------------------------------------
+unsigned char
+Avec::unicode_to_cp(Unicode uni)
+{
+   if (uni <= 0x80)                return uni;
+   if (uni == UNI_STAR_OPERATOR)   return '*';   // ⋆ → *
+   if (uni == UNI_AND)             return '^';   // ∧ → ^
+   if (uni == UNI_TILDE_OPERATOR)  return 126;   // ∼ → ~
+
+   // search in uni_to_cp_map table
+   //
+const void * where = bsearch(&uni, inverse_ibm_av, 128,
+                             sizeof(Unicode_to_IBM_codepoint), compare_uni);
+
+   if (where == 0)
+      {
+        // the workspace being )OUT'ed can contain characters that are not
+        // in IBM's APL character set. We replace such characters by 0xB0
+        //
+        return 0xB0;
+      }
+
+   Assert(where);
+   return ((const Unicode_to_IBM_codepoint *)where)->cp;
+}
+//-----------------------------------------------------------------------------
+
