@@ -33,20 +33,19 @@ Bif_OPER1_EACH::eval_ALB(Value_P A, Token & _LO, Value_P B)
 Function * LO = _LO.get_function();
    Assert1(LO);
 
-
    if (A->is_empty() || B->is_empty())
       {
         if (!LO->has_result())   return Token(TOK_VOID);
 
-        Value_P Fill_A = Bif_F12_TAKE::fun.eval_B(A).get_apl_val();
-        Value_P Fill_B = Bif_F12_TAKE::fun.eval_B(B).get_apl_val();
+        Value_P Fill_A = Bif_F12_TAKE::first(A);
+        Value_P Fill_B = Bif_F12_TAKE::first(B);
         Shape shape_Z;
 
         if (A->is_empty())          shape_Z = A->get_shape();
-        else if (!A->is_skalar())   DOMAIN_ERROR;
+        else if (!A->is_scalar())   DOMAIN_ERROR;
 
         if (B->is_empty())          shape_Z = B->get_shape();
-        else if (!B->is_skalar())   DOMAIN_ERROR;
+        else if (!B->is_scalar())   DOMAIN_ERROR;
 
         Value_P Z1 = LO->eval_fill_AB(Fill_A, Fill_B).get_apl_val();
 
@@ -64,24 +63,24 @@ EACH_ALB & _arg = arg.u.u_EACH_ALB;
    _arg.dB = 1;
 
 
-   if (A->nz_element_count() == 1)
-      {
-        if (B->is_skalar_or_len1_vector())   return LO->eval_AB(A, B);
+   if (A->is_scalar() && B->is_scalar())   return LO->eval_AB(A, B);
 
+   if (A->is_scalar())
+      {
         _arg.count = B->element_count();
 
         _arg.dA = 0;
         if (LO->has_result())
            arg.Z = Value_P(new Value(B->get_shape(), LOC));
       }
-   else if (B->nz_element_count() == 1)
+   else if (B->is_scalar())
       {
         _arg.dB = 0;
         _arg.count = A->element_count();
         if (LO->has_result())
            arg.Z = Value_P(new Value(A->get_shape(), LOC));
       }
-   else if (A->same_shape(B))
+   else if (A->same_shape(*B))
       {
         _arg.count = B->element_count();
         if (LO->has_result())
@@ -89,7 +88,7 @@ EACH_ALB & _arg = arg.u.u_EACH_ALB;
       }
    else
       {
-        if (A->same_rank(B))   LENGTH_ERROR;
+        if (A->same_rank(*B))   LENGTH_ERROR;
         RANK_ERROR;
       }
 
@@ -109,7 +108,6 @@ EACH_ALB & _arg = arg.u.u_EACH_ALB;
 
    if (_arg.how == 1)   goto how_1;
 
-how_0:
    Assert1(_arg.how == 0);
    _arg.z = 0;
 
@@ -133,7 +131,7 @@ loop_z:
           Value_P vZ = result.get_apl_val();
 
           if (!_arg.sub)   arg.Z->next_ravel()->init_from_value(vZ, LOC);
-          else if (vZ->is_simple_skalar())
+          else if (vZ->is_simple_scalar())
                            arg.Z->next_ravel()->init(vZ->get_ravel(0));
           else             new (arg.Z->next_ravel())   PointerCell(vZ);
 
@@ -184,7 +182,7 @@ EACH_ALB & _arg = arg.u.u_EACH_ALB;
        Value_P vZ = token.get_apl_val();
 
        if (!_arg.sub)   arg.Z->next_ravel()->init_from_value(vZ, LOC);
-       else if (vZ->is_simple_skalar())
+       else if (vZ->is_simple_scalar())
                         arg.Z->next_ravel()->init(vZ->get_ravel(0));
        else             new (arg.Z->next_ravel())   PointerCell(vZ);
       }
@@ -209,6 +207,19 @@ Bif_OPER1_EACH::eval_LB(Token & _LO, Value_P B)
 Function * LO = _LO.get_function();
    Assert1(LO);
 
+   if (B->is_scalar())   return LO->eval_B(B);
+
+   if (B->is_empty())
+      {
+        if (!LO->has_result())   return Token(TOK_VOID);
+
+        Value_P Fill_B = Bif_F12_TAKE::first(B);
+        Token tZ = LO->eval_fill_B(Fill_B);
+        Value_P Z = tZ.get_apl_val();
+        Z->set_shape(B->get_shape());
+        return Token(TOK_APL_VALUE1, Z);
+      }
+
 EOC_arg arg(B);
 EACH_LB & _arg = arg.u.u_EACH_LB;
 
@@ -229,7 +240,6 @@ EACH_LB & _arg = arg.u.u_EACH_LB;
 
    if (_arg.how == 1)   goto how_1;
 
-how_0:
    Assert1(_arg.how == 0);
    _arg.z = 0;
 
@@ -260,7 +270,7 @@ loop_z:
              Value_P vZ = result.get_apl_val();
 
              if (!_arg.sub)   arg.Z->next_ravel()->init_from_value(vZ, LOC);
-             else if (vZ->is_simple_skalar())
+             else if (vZ->is_simple_scalar())
                               arg.Z->next_ravel()->init(vZ->get_ravel(0));
              else             new (arg.Z->next_ravel())   PointerCell(vZ);
 
@@ -312,7 +322,7 @@ loop_z:
              Value_P vZ = result.get_apl_val();
 
              if (!_arg.sub)   arg.Z->next_ravel()->init_from_value(vZ, LOC);
-             else if (vZ->is_simple_skalar())
+             else if (vZ->is_simple_scalar())
                               arg.Z->next_ravel()->init(vZ->get_ravel(0));
              else             new (arg.Z->next_ravel())   PointerCell(vZ);
 
@@ -353,7 +363,7 @@ EACH_LB & _arg = arg.u.u_EACH_LB;
         Value_P vZ = token.get_apl_val();
 
         if (!_arg.sub)   arg.Z->next_ravel()->init_from_value(vZ, LOC);
-        else if (vZ->is_simple_skalar())
+        else if (vZ->is_simple_scalar())
                          arg.Z->next_ravel()->init(vZ->get_ravel(0));
         else             new (arg.Z->next_ravel())   PointerCell(vZ);
       }
