@@ -24,15 +24,15 @@
 #include "FloatCell.hh"
 #include "IntCell.hh"
 #include "PointerCell.hh"
-#include "Value.hh"
+#include "Value.icc"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
 void
-CDR::to_CDR(CDR_string & result, Value_P value)
+CDR::to_CDR(CDR_string & result, const Value & value)
 {
-const int type = value->get_CDR_type();
-const int len = value->total_size_brutto(type);
+const CDR_type type = value.get_CDR_type();
+const int len = value.total_size_brutto(type);
    result.reserve(len + 1);
    result.clear();
 
@@ -40,7 +40,7 @@ const int len = value->total_size_brutto(type);
 }
 //-----------------------------------------------------------------------------
 void
-CDR::fill(CDR_string & result, int type, int len, Value_P val)
+CDR::fill(CDR_string & result, int type, int len, const Value & val)
 {
    Assert((len & 0x0F) == 0);
 
@@ -74,7 +74,7 @@ const uint32_t ptr = 0x00002020;
 
    // nelm
    //
-const uint32_t nelm = val->element_count();
+const uint32_t nelm = val.element_count();
    result.append(Unicode(nelm >> 24 & 0xFF));
    result.append(Unicode(nelm >> 16 & 0xFF));
    result.append(Unicode(nelm >>  8 & 0xFF));
@@ -83,15 +83,15 @@ const uint32_t nelm = val->element_count();
    // type, rank, 0, 0
    //
    result.append(Unicode(type));
-   result.append(Unicode(val->get_rank()));
+   result.append(Unicode(val.get_rank()));
    result.append(Unicode_0);
    result.append(Unicode_0);
 
    // shape
    //
-   loop(r, val->get_rank())
+   loop(r, val.get_rank())
       {
-        const uint32_t sh = val->get_shape_item(r);
+        const uint32_t sh = val.get_shape_item(r);
         result.append(Unicode(sh >> 24 & 0xFF));
         result.append(Unicode(sh >> 16 & 0xFF));
         result.append(Unicode(sh >>  8 & 0xFF));
@@ -107,7 +107,7 @@ const uint32_t nelm = val->element_count();
         loop(e, nelm)
            {
              const int bit = e%8;
-             const APL_Integer i = val->get_ravel(e).get_int_value();
+             const APL_Integer i = val.get_ravel(e).get_int_value();
              Assert(i == 0 || i == 1);
              if (i)   accu |= 0x80 >> bit;
              if (bit == 7)
@@ -127,7 +127,7 @@ const uint32_t nelm = val->element_count();
       {
         loop(e, nelm)
            {
-             uint64_t i = val->get_ravel(e).get_int_value();
+             uint64_t i = val.get_ravel(e).get_int_value();
              result.append(Unicode(i & 0xFF));   i >>= 8;
              result.append(Unicode(i & 0xFF));   i >>= 8;
              result.append(Unicode(i & 0xFF));   i >>= 8;
@@ -139,7 +139,7 @@ const uint32_t nelm = val->element_count();
       {
         loop(e, nelm)
            {
-             const double v = val->get_ravel(e).get_real_value();
+             const double v = val.get_ravel(e).get_real_value();
              uint64_t i = *(uint64_t *)&v;
              result.append(Unicode(i & 0xFF));   i >>= 8;
              result.append(Unicode(i & 0xFF));   i >>= 8;
@@ -155,7 +155,7 @@ const uint32_t nelm = val->element_count();
       {
         loop(e, nelm)
            {
-             double v = val->get_ravel(e).get_real_value();
+             double v = val.get_ravel(e).get_real_value();
              uint64_t i = *(uint64_t *)&v;
              result.append(Unicode(i & 0xFF));   i >>= 8;
              result.append(Unicode(i & 0xFF));   i >>= 8;
@@ -166,7 +166,7 @@ const uint32_t nelm = val->element_count();
              result.append(Unicode(i & 0xFF));   i >>= 8;
              result.append(Unicode(i & 0xFF));   i >>= 8;
 
-             v = val->get_ravel(e).get_imag_value();
+             v = val.get_ravel(e).get_imag_value();
              i = *(uint64_t *)&v;
              result.append(Unicode(i & 0xFF));   i >>= 8;
              result.append(Unicode(i & 0xFF));   i >>= 8;
@@ -182,7 +182,7 @@ const uint32_t nelm = val->element_count();
       {
         loop(e, nelm)
            {
-             const Unicode uni = val->get_ravel(e).get_char_value();
+             const Unicode uni = val.get_ravel(e).get_char_value();
              Assert(uni >= 0);
              Assert(uni < 256);
              result.append(uni);
@@ -192,7 +192,7 @@ const uint32_t nelm = val->element_count();
       {
         loop(e, nelm)
            {
-             uint32_t i = val->get_ravel(e).get_char_value();
+             uint32_t i = val.get_ravel(e).get_char_value();
              result.append(Unicode(i & 0xFF));   i >>= 8;
              result.append(Unicode(i & 0xFF));   i >>= 8;
              result.append(Unicode(i & 0xFF));   i >>= 8;
@@ -228,7 +228,7 @@ const uint32_t nelm = val->element_count();
         //   +------------------+
         //
       
-        uint32_t offset = 16 + 4*val->get_rank() + 4*nelm;
+        uint32_t offset = 16 + 4*val.get_rank() + 4*nelm;
         while (offset & 0x0F)   ++offset;
         loop(e, nelm)
            {
@@ -237,7 +237,7 @@ const uint32_t nelm = val->element_count();
              result.append(Unicode(offset >> 16 & 0xFF));
              result.append(Unicode(offset >> 24 & 0xFF));
 
-             const Cell & cell = val->get_ravel(e);
+             const Cell & cell = val.get_ravel(e);
              if (cell.is_character_cell() || cell.is_numeric())
                 {
                   // a non-pointer sub value: 16 byte header,
@@ -247,9 +247,9 @@ const uint32_t nelm = val->element_count();
                 }
              else if (cell.is_pointer_cell())
                 {
-                  Value_P sub_val = cell.get_pointer_value();
-                  const int sub_type = sub_val->get_CDR_type();
-                  offset += sub_val->total_size_brutto(sub_type);
+                  const Value & sub_val = *cell.get_pointer_value();
+                  const CDR_type sub_type = sub_val.get_CDR_type();
+                  offset += sub_val.total_size_brutto(sub_type);
                 }
               else
                 DOMAIN_ERROR;
@@ -264,23 +264,23 @@ const uint32_t nelm = val->element_count();
         //
         loop(e, nelm)
            {
-             const Cell & cell = val->get_ravel(e);
+             const Cell & cell = val.get_ravel(e);
              if (cell.is_character_cell() || cell.is_numeric())
                 {
                   Value_P sub_val(new Value(LOC));
                   sub_val->get_ravel(0).init(cell);
 
-                  const int sub_type = sub_val->get_CDR_type();
+                  const CDR_type sub_type = sub_val->get_CDR_type();
                   const int sub_len = sub_val->total_size_brutto(sub_type);
-                  fill(result, sub_type, sub_len, sub_val);
+                  fill(result, sub_type, sub_len, *sub_val);
                 }
               else
                 {
                   Value_P sub_val = cell.get_pointer_value();
 
-                  const int sub_type = sub_val->get_CDR_type();
+                  const CDR_type sub_type = sub_val->get_CDR_type();
                   const int sub_len = sub_val->total_size_brutto(sub_type);
-                  fill(result, sub_type, sub_len, sub_val);
+                  fill(result, sub_type, sub_len, *sub_val);
                 }
            }
 
@@ -316,7 +316,7 @@ struct tf3_header
 const uint8_t * data = cdr.get_items();
 
 const uint32_t nelm = get_4_be(data + 8);
-const uint32_t vtype = data[12];
+const CDR_type vtype = (CDR_type)(data[12]);
 const Rank rank = data[13];
 Shape shape;
    loop(r, rank)
@@ -329,7 +329,7 @@ Value_P ret(new Value(shape, loc));
 
 const uint8_t * ravel = data + 16 + 4*rank;
 
-   if (vtype == 0)        // packed bit vector
+   if (vtype == CDR_BOOL1)        // packed bit vector
       {
         loop(n, nelm)
            {
@@ -338,7 +338,7 @@ const uint8_t * ravel = data + 16 + 4*rank;
              else       new (&ret->get_ravel(n)) IntCell(0);
            }
       }
-   else if (vtype == 1)   // 4 byte ints vector
+   else if (vtype == CDR_INT32)   // 4 byte ints vector
       {
         loop(n, nelm)
            {
@@ -347,7 +347,7 @@ const uint8_t * ravel = data + 16 + 4*rank;
              new (&ret->get_ravel(n)) IntCell(d);
            }
       }
-   else if (vtype == 2)   // 8 byte float vector
+   else if (vtype == CDR_FLT64)   // 8 byte float vector
       {
         loop(n, nelm)
            {
@@ -355,7 +355,7 @@ const uint8_t * ravel = data + 16 + 4*rank;
              new (&ret->get_ravel(n)) FloatCell(v);
            }
       }
-  else if (vtype == 3)   // 16 byte complex vector
+  else if (vtype == CDR_CPLX128)   // 16 byte complex vector
       {
         loop(n, nelm)
            {
@@ -364,7 +364,7 @@ const uint8_t * ravel = data + 16 + 4*rank;
              new (&ret->get_ravel(n)) ComplexCell(vr, vi);
            }
       }
-   else if (vtype == 4)   // 1 byte char vector
+   else if (vtype == CDR_CHAR8)   // 1 byte char vector
       {
         loop(n, nelm)
            {
@@ -373,7 +373,7 @@ const uint8_t * ravel = data + 16 + 4*rank;
              new (&ret->get_ravel(n)) CharCell(Unicode(d));
            }
       }
-   else if (vtype == 5)   // 4 byte UNICODE vector
+   else if (vtype == CDR_CHAR32)   // 4 byte UNICODE vector
       {
         loop(n, nelm)
            {
@@ -381,7 +381,7 @@ const uint8_t * ravel = data + 16 + 4*rank;
              new (&ret->get_ravel(n)) CharCell(Unicode(d));
            }
       }
-   else if (vtype == 7)   // packed vector with 4 bytes offsets.
+   else if (vtype == CDR_NEST32)   // packed vector with 4 bytes offsets.
       {
         loop(n, nelm)
             {
@@ -391,10 +391,10 @@ const uint8_t * ravel = data + 16 + 4*rank;
               const Rank sub_rank = sub_data[13];
               const uint8_t * sub_ravel = sub_data + 16 + 4*sub_rank;
 
-              // if the sub-item is a non-nested skalar then we append it to
+              // if the sub-item is a non-nested scalar then we append it to
               // ret directly; otherwise we append a pointer to it,
               //
-              if (sub_rank == 0)   // skalar
+              if (sub_rank == 0)   // scalar
                  {
                    if (sub_vtype == 0)        // bit
                       {
@@ -439,10 +439,10 @@ const uint8_t * ravel = data + 16 + 4*rank;
                         continue;   // next n
                       }
 
-                   // sub_vtype 6 (progression vector) can't be skalar
+                   // sub_vtype 6 (progression vector) can't be scalar
                  }
 
-              // at this point the sub-item is not a skalar
+              // at this point the sub-item is not a scalar
               //
               if (sub_vtype == 6)        // arithmetic progression vector
                  {
@@ -478,9 +478,12 @@ const uint8_t * ravel = data + 16 + 4*rank;
       }
    else
       {
+        CERR << "Unsupprted CDR type " << vtype << endl;
         Assert(0 && "Bad/unsupported CDR type");
       }
 
+   ret->set_default_Zero();
+   ret->check_value(LOC);
    return ret;
 }
 //-----------------------------------------------------------------------------
