@@ -18,15 +18,50 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-#include "Value.hh"
-#include "Error.hh"
-#include "Workspace.hh"
 #include "CharCell.hh"
-#include "ComplexCell.hh"
-#include "FloatCell.hh"
-#include "IntCell.hh"
-#include "PointerCell.hh"
 
+//-----------------------------------------------------------------------------
+CellType
+CharCell::get_cell_subtype() const
+{
+   if (value.ival < 0)   // negative char (only fits in signed containers)
+      {
+        if (-value.ival <= 0x80)
+           return (CellType)(CT_CHAR | CTS_S8 | CTS_S16 | CTS_S32 | CTS_S64);
+
+        if (-value.ival <= 0x8000)
+           return (CellType)(CT_CHAR | CTS_S16 | CTS_S32 | CTS_S64);
+
+        return (CellType)(CT_CHAR | CTS_S32 | CTS_S64);
+      }
+
+   // positive char
+   //
+   if (value.ival <= 0x7F)
+      return (CellType)(CT_CHAR | CTS_X8  | CTS_S8  | CTS_U8  |
+                                  CTS_X16 | CTS_S16 | CTS_U16 |
+                                  CTS_X32 | CTS_S32 | CTS_U32 |
+                                  CTS_X64 | CTS_S64 | CTS_U64);
+
+   if (value.ival <= 0xFF)
+      return (CellType)(CT_CHAR |                     CTS_U8  |
+                                  CTS_X16 | CTS_S16 | CTS_U16 |
+                                  CTS_X32 | CTS_S32 | CTS_U32 |
+                                  CTS_X64 | CTS_S64 | CTS_U64);
+
+   if (value.ival <= 0x7FFF)
+      return (CellType)(CT_CHAR | CTS_X16 | CTS_S16 | CTS_U16 |
+                                  CTS_X32 | CTS_S32 | CTS_U32 |
+                                  CTS_X64 | CTS_S64 | CTS_U64);
+
+   if (value.ival <= 0xFFFF)
+      return (CellType)(CT_CHAR |                     CTS_U16 |
+                                  CTS_X32 | CTS_S32 | CTS_U32 |
+                                  CTS_X64 | CTS_S64 | CTS_U64);
+
+   return (CellType)(CT_CHAR | CTS_X32 | CTS_S32 | CTS_U32 |
+                               CTS_X64 | CTS_S64 | CTS_U64);
+}
 //-----------------------------------------------------------------------------
 bool
 CharCell::greater(const Cell * other, bool ascending) const
