@@ -25,11 +25,10 @@
 #include <string.h>
 
 #include "ComplexCell.hh"
-#include "Error.hh"
+#include "ErrorCode.hh"
 #include "FloatCell.hh"
 #include "IntCell.hh"
 #include "Output.hh"
-#include "Value.hh"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
@@ -66,40 +65,6 @@ ComplexCell::get_imag_value() const
 }
 //-----------------------------------------------------------------------------
 bool
-ComplexCell::get_near_bool(APL_Float qct)  const   
-{
-   if (value.cpxp->imag() >  qct)   DOMAIN_ERROR;
-   if (value.cpxp->imag() < -qct)   DOMAIN_ERROR;
-
-   if (value.cpxp->real() > qct)   // 1 or invalid
-      {
-        if (value.cpxp->real() < (1.0 - qct))   DOMAIN_ERROR;
-        if (value.cpxp->real() > (1.0 + qct))   DOMAIN_ERROR;
-        return true;
-      }
-   else
-      {
-        if (value.cpxp->real() < -qct)   DOMAIN_ERROR;
-        return false;
-      }
-}
-//-----------------------------------------------------------------------------
-APL_Integer
-ComplexCell::get_near_int(APL_Float qct) const
-{
-// if (value.cpxp->imag() >  qct)   DOMAIN_ERROR;
-// if (value.cpxp->imag() < -qct)   DOMAIN_ERROR;
-
-const double val = value.cpxp->real();
-const double result = round(val);
-const double diff = val - result;
-   if (diff > qct)    DOMAIN_ERROR;
-   if (diff < -qct)   DOMAIN_ERROR;
-
-   return APL_Integer(result + 0.3);
-}
-//-----------------------------------------------------------------------------
-bool
 ComplexCell::is_near_int(APL_Float qct) const
 {
    return Cell::is_near_int(value.cpxp->real(), qct) &&
@@ -172,56 +137,7 @@ bool
 ComplexCell::equal(const Cell & A, APL_Float qct) const
 {
    if (!A.is_numeric())   return false;
-
-   // real part
-   //
-   {
-     const APL_Float valA = A.get_real_value();
-     const APL_Float valB =   get_real_value();
-     const APL_Float posA = valA < 0 ? -valA : valA;
-     const APL_Float posB = valB < 0 ? -valB : valB;
-     const APL_Float maxAB = posA > posB ? posA : posB;
-     const APL_Float maxCT = qct * maxAB;
-
-     if (posB < (posA - maxCT))   return false;
-     if (posB > (posA + maxCT))   return false;
-   }
-
-   // imag part
-   //
-   {
-     const APL_Float valA = A.get_imag_value();
-     const APL_Float valB =   get_imag_value();
-     const APL_Float posA = valA < 0 ? -valA : valA;
-     const APL_Float posB = valB < 0 ? -valB : valB;
-     const APL_Float maxAB = posA > posB ? posA : posB;
-     const APL_Float maxCT = qct * maxAB;
-
-     if (posB < (posA - maxCT))   return false;
-     if (posB > (posA + maxCT))   return false;
-   }
-
-   return true;
-}
-//-----------------------------------------------------------------------------
-bool
-ComplexCell::greater(const Cell * other, bool ascending) const
-{
-   switch(other->get_cell_type())
-      {
-        case CT_CHAR:    return ascending;
-        case CT_INT:
-        case CT_FLOAT:
-        case CT_COMPLEX: break;
-        case CT_POINTER: return !ascending;
-        case CT_CELLREF: DOMAIN_ERROR;
-        defaulkt:        Assert(0 && "Bad celltype");
-      }
-
-const Comp_result comp = compare(*other);
-   if (comp == COMP_EQ)   return this > other;
-   if (comp == COMP_GT)   return ascending;
-   else                   return !ascending;
+   return tolerantly_equal(A.get_complex_value(), get_complex_value(), qct);
 }
 //-----------------------------------------------------------------------------
 void
@@ -231,58 +147,16 @@ ComplexCell::release(const char * loc)
    new (this) Cell;
 }
 //-----------------------------------------------------------------------------
-Comp_result
-ComplexCell::compare(const Cell & other) const
-{
-   // comparison of complex numbers gives DOMAN ERROR if one of the cells
-   // is not near-real.
-   //
-const APL_Float qct = Workspace::get_CT();
-
-   if (get_imag_value() < -qct)   DOMAIN_ERROR;   // not near-real
-   if (get_imag_value() >  qct)   DOMAIN_ERROR;   // not near-real
-
-const APL_Float rval = get_real_value();
-   if (other.is_integer_cell())   // integer
-      {
-        if (equal(other, qct))   return COMP_EQ;
-        return (rval < other.get_int_value()) ? COMP_LT : COMP_GT;
-      }
-
-   if (other.is_float_cell())
-      {
-        if (equal(other, qct))   return COMP_EQ;
-        return (rval < other.get_real_value()) ? COMP_LT : COMP_GT;
-      }
-
-   if (other.is_complex_cell())   // complex
-      {
-        if (other.get_imag_value() < -qct)   DOMAIN_ERROR;   // not near-real
-        if (other.get_imag_value() >  qct)   DOMAIN_ERROR;   // not near-real
-
-        if (equal(other, qct))   return COMP_EQ;
-        return (rval < other.get_real_value()) ? COMP_LT : COMP_GT;
-        return COMP_EQ;
-      }
-
-   if (other.is_character_cell())
-      {
-        return COMP_GT;
-      }
-
-   DOMAIN_ERROR;
-}
-//-----------------------------------------------------------------------------
 // monadic build-in functions...
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 ComplexCell::bif_factorial(Cell * Z) const
 {
    if (is_near_real(Workspace::get_CT()))
       {
         const FloatCell fc(get_real_value());
         fc.bif_factorial(Z);
-        return;
+        return E_NO_ERROR;
       }
 
 const APL_Float zr = get_real_value() + 1.0;
@@ -290,58 +164,75 @@ const APL_Float zi = get_imag_value();
 const APL_Complex z(zr, zi);
 
    new (Z) ComplexCell(gamma(get_real_value(), get_imag_value()));
+   if (errno)   return E_DOMAIN_ERROR;
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_conjugate(Cell * Z) const
+ErrorCode
+ComplexCell::bif_conjugate(Cell * Z) const
 {
    new (Z) ComplexCell(conj(*value.cpxp));
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_negative(Cell * Z) const
+ErrorCode
+ComplexCell::bif_negative(Cell * Z) const
 {
    new (Z) ComplexCell(- *value.cpxp);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_direction(Cell * Z) const
+ErrorCode
+ComplexCell::bif_direction(Cell * Z) const
 {
 const APL_Complex val = *value.cpxp; 
 const APL_Float mag = abs(val);
    if (mag == 0.0)   new (Z) IntCell(0);
    else              new (Z) ComplexCell(val/mag);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_magnitude(Cell * Z) const
+ErrorCode
+ComplexCell::bif_magnitude(Cell * Z) const
 {
    new (Z) FloatCell(abs(*value.cpxp));
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_reciprocal(Cell * Z) const
+ErrorCode
+ComplexCell::bif_reciprocal(Cell * Z) const
 {
 const APL_Float qct = Workspace::get_CT();
-   if (is_near_zero(qct))  DOMAIN_ERROR;
+   if (is_near_zero(qct))  return E_DOMAIN_ERROR;
 
    if (is_near_real(qct))   new (Z) FloatCell(1.0/value.cpxp->real());
    else                     new (Z) ComplexCell(APL_Complex(1.0)/ *value.cpxp);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_roll(Cell * Z) const
+ErrorCode
+ComplexCell::bif_roll(Cell * Z) const
 {
 const APL_Integer qio = Workspace::get_IO();
-   if (!is_near_int(Workspace::get_CT()))   DOMAIN_ERROR;
+   if (!is_near_int(Workspace::get_CT()))   return E_DOMAIN_ERROR;
 
 const APL_Integer set_size = get_checked_near_int();
-   if (set_size <= 0)   DOMAIN_ERROR;
+   if (set_size <= 0)   return E_DOMAIN_ERROR;
 
-const APL_Integer rnd = Workspace::get_RL();
+const uint64_t rnd = Workspace::get_RL();
    new (Z) IntCell(qio + (rnd % set_size));
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_pi_times(Cell * Z) const
+ErrorCode
+ComplexCell::bif_pi_times(Cell * Z) const
 {
    new (Z) ComplexCell(M_PI * *value.cpxp);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_ceiling(Cell * Z) const
+ErrorCode
+ComplexCell::bif_ceiling(Cell * Z) const
 {
 const double r = value.cpxp->real();
 const double i = value.cpxp->imag();
@@ -354,9 +245,11 @@ const double di = i - floor(i);
       new (Z) ComplexCell(floor(r), ceil(i));
    else                 // round real up and imag down
       new (Z) ComplexCell(ceil(r), floor(i));
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_floor(Cell * Z) const
+ErrorCode
+ComplexCell::bif_floor(Cell * Z) const
 {
 const double r = value.cpxp->real();
 const double i = value.cpxp->imag();
@@ -369,50 +262,64 @@ const double di = i - floor(i);
       new (Z) ComplexCell(floor(r), ceil (i));
    else                // round real up and imag down
       new (Z) ComplexCell(ceil (r), floor(i));
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_exponential(Cell * Z) const
+ErrorCode
+ComplexCell::bif_exponential(Cell * Z) const
 {
    new (Z) ComplexCell(exp(*value.cpxp));
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_nat_log(Cell * Z) const
+ErrorCode
+ComplexCell::bif_nat_log(Cell * Z) const
 {
    new (Z) ComplexCell(log(*value.cpxp));
    Z->demote_complex_to_real(Workspace::get_CT());
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 // dyadic build-in functions...
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_add(Cell * Z, const Cell * A) const
+ErrorCode
+ComplexCell::bif_add(Cell * Z, const Cell * A) const
 {
    new (Z) ComplexCell(A->get_complex_value() + get_complex_value());
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_subtract(Cell * Z, const Cell * A) const
+ErrorCode
+ComplexCell::bif_subtract(Cell * Z, const Cell * A) const
 {
    new (Z) ComplexCell(A->get_complex_value() - get_complex_value());
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_multiply(Cell * Z, const Cell * A) const
+ErrorCode
+ComplexCell::bif_multiply(Cell * Z, const Cell * A) const
 {
    new (Z) ComplexCell(A->get_complex_value() * get_complex_value());
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_divide(Cell * Z, const Cell * A) const
+ErrorCode
+ComplexCell::bif_divide(Cell * Z, const Cell * A) const
 {
 const APL_Float qct = Workspace::get_CT();
 
    if (is_near_zero(qct))
       {
          if (A->is_near_zero(qct))   new (Z) IntCell(1);
-         else                        DOMAIN_ERROR;
+         else                        return E_DOMAIN_ERROR;
       }
 
    new (Z) ComplexCell(A->get_complex_value() / get_complex_value());
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void ComplexCell::bif_residue(Cell * Z, const Cell * A) const
+ErrorCode
+ComplexCell::bif_residue(Cell * Z, const Cell * A) const
 {
 const APL_Float qct = Workspace::get_CT();
 
@@ -437,11 +344,12 @@ const APL_Float qct = Workspace::get_CT();
       }
    else
       {
-        DOMAIN_ERROR;
+        return E_DOMAIN_ERROR;
       }
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 ComplexCell::bif_maximum(Cell * Z, const Cell * A) const
 {
    // maximum of complex numbers gives DOMAN ERROR if one of the cells
@@ -449,8 +357,8 @@ ComplexCell::bif_maximum(Cell * Z, const Cell * A) const
    //
 const APL_Float qct = Workspace::get_CT();
 
-   if (get_imag_value() < -qct)   DOMAIN_ERROR;   // not near-real
-   if (get_imag_value() >  qct)   DOMAIN_ERROR;   // not near-real
+   if (get_imag_value() < -qct)   return E_DOMAIN_ERROR;   // not near-real
+   if (get_imag_value() >  qct)   return E_DOMAIN_ERROR;   // not near-real
 
 const APL_Float rval = get_real_value();
 
@@ -468,17 +376,18 @@ const APL_Float rval = get_real_value();
       }
    else
       {
-        if (A->get_imag_value() < -qct)   DOMAIN_ERROR;   // not near-real
-        if (A->get_imag_value() >  qct)   DOMAIN_ERROR;   // not near-real
+        if (A->get_imag_value() < -qct)   return E_DOMAIN_ERROR;   // not near-real
+        if (A->get_imag_value() >  qct)   return E_DOMAIN_ERROR;   // not near-real
 
          const APL_Float a = A->get_real_value();
          if (a >= rval)   new (Z) FloatCell(a);
          else             new (Z) FloatCell(rval);
       }
 
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 ComplexCell::bif_minimum(Cell * Z, const Cell * A) const
 {
    // minimum of complex numbers gives DOMAN ERROR if one of the cells
@@ -486,8 +395,8 @@ ComplexCell::bif_minimum(Cell * Z, const Cell * A) const
    //
 const APL_Float qct = Workspace::get_CT();
 
-   if (get_imag_value() < -qct)   DOMAIN_ERROR;   // not near-real
-   if (get_imag_value() >  qct)   DOMAIN_ERROR;   // not near-real
+   if (get_imag_value() < -qct)   return E_DOMAIN_ERROR;   // not near-real
+   if (get_imag_value() >  qct)   return E_DOMAIN_ERROR;   // not near-real
 
 const APL_Float rval = get_real_value();
 
@@ -505,16 +414,17 @@ const APL_Float rval = get_real_value();
       }
    else
       {
-        if (A->get_imag_value() < -qct)   DOMAIN_ERROR;   // not near-real
-        if (A->get_imag_value() >  qct)   DOMAIN_ERROR;   // not near-real
+        if (A->get_imag_value() < -qct)   return E_DOMAIN_ERROR;   // not near-real
+        if (A->get_imag_value() >  qct)   return E_DOMAIN_ERROR;   // not near-real
 
          const APL_Float a = A->get_real_value();
          if (a <= rval)   new (Z) FloatCell(a);
          else             new (Z) FloatCell(rval);
       }
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 ComplexCell::bif_equal(Cell * Z, const Cell * A) const
 {
 const APL_Float qct = Workspace::get_CT();
@@ -540,31 +450,62 @@ const APL_Float qct = Workspace::get_CT();
       {
         new (Z) IntCell(0);
       }
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 ComplexCell::bif_power(Cell * Z, const Cell * A) const
 {
-   if (A->is_complex_cell())
+   // some A to the complex B-th power
+   //
+   if (!A->is_numeric())   return E_DOMAIN_ERROR;
+
+const APL_Float qct = Workspace::get_CT();
+   if (Cell::is_near_int(value.cpxp->real(), qct) &&
+       Cell::is_near_zero(value.cpxp->imag(), qct))
       {
-        APL_Complex z = pow(A->get_complex_value(), get_complex_value());
-        new (Z) ComplexCell(z);
-        Z->demote_complex_to_real(Workspace::get_CT());
+        APL_Integer b = get_near_int(qct);
+        const bool invert_Z = b < 0;
+        if (invert_Z)   b = - b;
+
+        if (b <= 1)   // special cases A⋆1, A⋆0, and A⋆¯1
+           {
+             if (b == 0)   return IntCell::z1(Z);  // A⋆0 is 1
+
+             if (invert_Z)               return A->bif_reciprocal(Z);
+             if (A->is_real_cell())      return A->bif_conjugate(Z);
+             new (Z) ComplexCell(A->get_complex_value());
+             return E_NO_ERROR;
+           }
       }
-   else if (A->is_numeric())   // real to the power of complex
+
+   if (A->is_near_zero(qct))
+      {
+        if (get_real_value() < 0)   return E_DOMAIN_ERROR;
+        return IntCell::z0(Z);
+      }
+
+   if (A->is_real_cell())
       {
         APL_Complex complex_A(A->get_real_value(), 0);
         APL_Complex z = pow(complex_A, get_complex_value());
         new (Z) ComplexCell(z);
         Z->demote_complex_to_real(Workspace::get_CT());
+        return E_NO_ERROR;
       }
-   else
-      {
-        DOMAIN_ERROR;
-      }
+
+   // complex A to the complex B-th power
+   //
+   {
+     APL_Complex z = pow(A->get_complex_value(), get_complex_value());
+     new (Z) ComplexCell(z);
+     Z->demote_complex_to_real(Workspace::get_CT());
+
+     return E_NO_ERROR;
+   }
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 ComplexCell::bif_logarithm(Cell * Z, const Cell * A) const
 {
    if (A->is_real_cell())
@@ -577,11 +518,13 @@ ComplexCell::bif_logarithm(Cell * Z, const Cell * A) const
       }
    else
       {
-        DOMAIN_ERROR;
+        return E_DOMAIN_ERROR;
       }
+
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 ComplexCell::bif_circle_fun(Cell * Z, const Cell * A) const
 {
 const APL_Integer fun = A->get_near_int(Workspace::get_CT());
@@ -589,10 +532,10 @@ const APL_Complex b = *value.cpxp;
 
    switch(fun)
       {
-        case -12: ComplexCell(-b.imag(), b.real()).bif_exponential(Z);   return;
-        case -11: new (Z) ComplexCell(-b.imag(), b.real());              return;
-        case -10: new (Z) ComplexCell(b.real(), -b.imag());              return;
-        case  -9: new (Z) ComplexCell(            b   );                 return;
+        case -12: ComplexCell(-b.imag(), b.real()).bif_exponential(Z);   break;
+        case -11: new (Z) ComplexCell(-b.imag(), b.real());              break;
+        case -10: new (Z) ComplexCell(b.real(), -b.imag());              break;
+        case  -9: new (Z) ComplexCell(            b   );                 break;
         case  -7: // arctanh(z) = 0.5 (ln(1 + z) - ln(1 - z))
                   {
                     const APL_Complex b1      = ONE() + b;
@@ -601,7 +544,7 @@ const APL_Complex b = *value.cpxp;
                     const APL_Complex log_b_1 = log(b_1);
                     const APL_Complex diff    = log_b1 - log_b_1;
                     const APL_Complex half    = 0.5 * diff;
-                    new (Z) ComplexCell(half);                           return;
+                    new (Z) ComplexCell(half);                           break;
                   }
         case  -6: // arccosh(z) = ln(z + sqrt(z + 1) sqrt(z - 1))
                   {
@@ -612,7 +555,7 @@ const APL_Complex b = *value.cpxp;
                     const APL_Complex prod   = root1 * root_1;
                     const APL_Complex sum    = b + prod;
                     const APL_Complex loga   = log(sum);
-                    new (Z) ComplexCell(loga);                           return;
+                    new (Z) ComplexCell(loga);                           break;
                   }
         case  -5: // arcsinh(z) = ln(z + sqrt(z^2 + 1))
                   {
@@ -621,9 +564,9 @@ const APL_Complex b = *value.cpxp;
                     const APL_Complex root = sqrt(b2_1);
                     const APL_Complex sum =  b + root;
                     const APL_Complex loga = log(sum);
-                    new (Z) ComplexCell(loga);                           return;
+                    new (Z) ComplexCell(loga);                           break;
                   }
-        case  -4: new (Z) ComplexCell(sqrt(b*b - 1.0));                  return;
+        case  -4: new (Z) ComplexCell(sqrt(b*b - 1.0));                  break;
         case  -3: // arctan(z) = i/2 (ln(1 - iz) - ln(1 + iz))
                   {
                     const APL_Complex iz = APL_Complex(- b.imag(), b.real());
@@ -633,7 +576,7 @@ const APL_Complex b = *value.cpxp;
                     const APL_Complex log_niz = log(niz);
                     const APL_Complex diff = log_niz - log_piz;
                     const APL_Complex prod = APL_Complex(0, 0.5) * diff;
-                    new (Z) ComplexCell(prod);                           return;
+                    new (Z) ComplexCell(prod);                           break;
                   }
         case  -2: // arccos(z) = -i (ln( z + sqrt(z^2 - 1)))
                   {
@@ -643,7 +586,7 @@ const APL_Complex b = *value.cpxp;
                     const APL_Complex sum = b + root;
                     const APL_Complex loga = log(sum);
                     const APL_Complex prod = MINUS_i() * loga;
-                    new (Z) ComplexCell(prod);                           return;
+                    new (Z) ComplexCell(prod);                           break;
                   }
         case  -1: // arcsin(z) = -i (ln(iz + sqrt(1 - z^2)))
                   {
@@ -654,16 +597,16 @@ const APL_Complex b = *value.cpxp;
                                            + root;
                     const APL_Complex loga = log(sum);
                     const APL_Complex prod = MINUS_i() * loga;
-                    new (Z) ComplexCell(prod);                           return;
+                    new (Z) ComplexCell(prod);                           break;
                   }
-        case   0: new (Z) ComplexCell(sqrt (1.0 - b*b));                 return;
-        case   1: new (Z) ComplexCell( sin       (b));                   return;
-        case   2: new (Z) ComplexCell( cos       (b));                   return;
-        case   3: new (Z) ComplexCell( tan       (b));                   return;
-        case   4: new (Z) ComplexCell(sqrt (1.0 + b*b));                 return;
-        case   5: new (Z) ComplexCell( sinh      (b));                   return;
-        case   6: new (Z) ComplexCell( cosh      (b));                   return;
-        case   7: new (Z) ComplexCell( tanh      (b));                   return;
+        case   0: new (Z) ComplexCell(sqrt (1.0 - b*b));                 break;
+        case   1: new (Z) ComplexCell( sin       (b));                   break;
+        case   2: new (Z) ComplexCell( cos       (b));                   break;
+        case   3: new (Z) ComplexCell( tan       (b));                   break;
+        case   4: new (Z) ComplexCell(sqrt (1.0 + b*b));                 break;
+        case   5: new (Z) ComplexCell( sinh      (b));                   break;
+        case   6: new (Z) ComplexCell( cosh      (b));                   break;
+        case   7: new (Z) ComplexCell( tanh      (b));                   break;
         case  -8:
         case   8: {
                     bool pos_8 = false;
@@ -678,16 +621,18 @@ const APL_Complex b = *value.cpxp;
 
                     if (pos_8)   new (Z) ComplexCell(sq);
                     else         new (Z) ComplexCell(-sq);
-                  }                                                      return;
+                  }                                                      break;
 
-        case   9: new (Z) FloatCell(     b.real());                      return;
+        case   9: new (Z) FloatCell(     b.real());                      break;
         case  10: new (Z) FloatCell(sqrt(b.real()*b.real()
-                                       + b.imag()* b.imag()));           return;
-        case  11: new (Z) FloatCell(b.imag());                           return;
-        case  12: new (Z) FloatCell(phase());                            return;
+                                       + b.imag()* b.imag()));           break;
+        case  11: new (Z) FloatCell(b.imag());                           break;
+        case  12: new (Z) FloatCell(phase());                            break;
+
+        return E_DOMAIN_ERROR;
       }
 
-   DOMAIN_ERROR;
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 APL_Float
@@ -736,7 +681,7 @@ const APL_Complex ret( (sqrt(2*M_PI) / z)
                        * exp(-(z + 5.5))
                      );
 
-   if (errno)   DOMAIN_ERROR;
+   // errno may be set and is checked by the caller
 
    return ret;
 
@@ -748,17 +693,163 @@ const APL_Complex ret( (sqrt(2*M_PI) / z)
 #undef p5
 #undef p6
 }
+//=============================================================================
+// throw/nothrow boundary. Functions above MUST NOT (directly or indirectly)
+// throw while funcions below MAY throw.
+//=============================================================================
+
+#include "Error.hh"   // throws
+
+//-----------------------------------------------------------------------------
+bool
+ComplexCell::get_near_bool(APL_Float qct)  const   
+{
+   if (value.cpxp->imag() >  qct)   DOMAIN_ERROR;
+   if (value.cpxp->imag() < -qct)   DOMAIN_ERROR;
+
+   if (value.cpxp->real() > qct)   // 1 or invalid
+      {
+        if (value.cpxp->real() < (1.0 - qct))   DOMAIN_ERROR;
+        if (value.cpxp->real() > (1.0 + qct))   DOMAIN_ERROR;
+        return true;
+      }
+   else
+      {
+        if (value.cpxp->real() < -qct)   DOMAIN_ERROR;
+        return false;
+      }
+}
+//-----------------------------------------------------------------------------
+APL_Integer
+ComplexCell::get_near_int(APL_Float qct) const
+{
+// if (value.cpxp->imag() >  qct)   DOMAIN_ERROR;
+// if (value.cpxp->imag() < -qct)   DOMAIN_ERROR;
+
+const double val = value.cpxp->real();
+const double result = round(val);
+const double diff = val - result;
+   if (diff > qct)    DOMAIN_ERROR;
+   if (diff < -qct)   DOMAIN_ERROR;
+
+   return APL_Integer(result + 0.3);
+}
+//-----------------------------------------------------------------------------
+bool
+ComplexCell::greater(const Cell * other, bool ascending) const
+{
+   switch(other->get_cell_type())
+      {
+        case CT_CHAR:    return ascending;
+        case CT_INT:
+        case CT_FLOAT:
+        case CT_COMPLEX: break;
+        case CT_POINTER: return !ascending;
+        case CT_CELLREF: DOMAIN_ERROR;
+        default:         Assert(0 && "Bad celltype");
+      }
+
+const Comp_result comp = compare(*other);
+   if (comp == COMP_EQ)   return this > other;
+   if (comp == COMP_GT)   return ascending;
+   else                   return !ascending;
+}
+//-----------------------------------------------------------------------------
+Comp_result
+ComplexCell::compare(const Cell & other) const
+{
+   // comparison of complex numbers gives DOMAN ERROR if one of the cells
+   // is not near-real.
+   //
+const APL_Float qct = Workspace::get_CT();
+
+   if (get_imag_value() < -qct)   DOMAIN_ERROR;   // not near-real
+   if (get_imag_value() >  qct)   DOMAIN_ERROR;   // not near-real
+
+const APL_Float rval = get_real_value();
+   if (other.is_integer_cell())   // integer
+      {
+        if (equal(other, qct))   return COMP_EQ;
+        return (rval < other.get_int_value()) ? COMP_LT : COMP_GT;
+      }
+
+   if (other.is_float_cell())
+      {
+        if (equal(other, qct))   return COMP_EQ;
+        return (rval < other.get_real_value()) ? COMP_LT : COMP_GT;
+      }
+
+   if (other.is_complex_cell())   // complex
+      {
+        if (other.get_imag_value() < -qct)   DOMAIN_ERROR;   // not near-real
+        if (other.get_imag_value() >  qct)   DOMAIN_ERROR;   // not near-real
+
+        if (equal(other, qct))   return COMP_EQ;
+        return (rval < other.get_real_value()) ? COMP_LT : COMP_GT;
+        return COMP_EQ;
+      }
+
+   if (other.is_character_cell())
+      {
+        return COMP_GT;
+      }
+
+   DOMAIN_ERROR;
+}
 //-----------------------------------------------------------------------------
 PrintBuffer
 ComplexCell::character_representation(const PrintContext & pctx) const
 {
+   if (pctx.get_PP() < MAX_Quad_PP)
+      {
+         // 10⋆get_PP()
+         //
+         APL_Float ten_to_PP = 1.0;
+         loop(p, pctx.get_PP())   ten_to_PP *= 10.0;
+      
+         // lrm p. 13: In J notation, the real or imaginary part is not
+         // displayed if it is less than the other by more than ⎕PP orders
+         // of magnitude (unless ⎕PP is at its maximum).
+         //
+         const APL_Float pos_real = value.cpxp->real() < 0
+                                  ? -value.cpxp->real() : value.cpxp->real();
+         const APL_Float pos_imag = value.cpxp->imag() < 0
+                                  ? -value.cpxp->imag() : value.cpxp->imag();
+
+         if (pos_real >= pos_imag)   // pos_real dominates pos_imag
+            {
+              if (pos_real > pos_imag*ten_to_PP)
+                 {
+                   const FloatCell real_cell(value.cpxp->real());
+                   return real_cell.character_representation(pctx);
+                 }
+            }
+         else                        // pos_imag dominates pos_real
+            {
+              if (pos_imag > pos_real*ten_to_PP)
+                 {
+                   const FloatCell imag_cell(value.cpxp->imag());
+                   PrintBuffer ret = imag_cell.character_representation(pctx);
+                   ret.pad_l(UNI_ASCII_J, 1);
+                   ret.pad_l(UNI_ASCII_0, 1);
+                   
+                   ret.get_info().flags |= CT_COMPLEX;
+                   ret.get_info().imag_len = 1 + ret.get_info().real_len;
+                   ret.get_info().int_len = 1;
+                   ret.get_info().fract_len = 0;
+                   ret.get_info().real_len = 1;
+                   return ret;
+                 }
+            }
+      }
+
 bool scaled_real = pctx.get_scaled();   // may be changed by print function
 UCS_string ucs(value.cpxp->real(), scaled_real, pctx);
 
 ColInfo info;
    info.flags |= CT_COMPLEX;
    if (scaled_real)   info.flags |= real_has_E;
-int int_fract = ucs.size();;
+int int_fract = ucs.size();
    info.real_len = ucs.size();
    info.int_len = ucs.size();
    loop(u, ucs.size())
