@@ -24,7 +24,7 @@
 #include "PrintOperator.hh"
 #include "UCS_string.hh"
 #include "UserFunction.hh"
-#include "Value.hh"
+#include "Value.icc"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
@@ -41,7 +41,7 @@ UCS_string line;
         const Unicode uni = ucs[t];
         switch (uni)
            {
-             case UNI_ASCII_CR: break;;
+             case UNI_ASCII_CR:                                         break;
              case UNI_ASCII_LF: text.push_back(line);   line.clear();   break;
              default: line.append(uni);
           }
@@ -66,12 +66,68 @@ Executable::~Executable()
              << " (body size=" <<  body.size() << ")" << endl;
       }
 
-   loop(b, body.size())   body[b].extract_apl_val(LOC);
+   clear_body();
+}
+//-----------------------------------------------------------------------------
+void
+Executable::clear_body()
+{
+   loop(b, body.size())
+      {
+        body[b].extract_apl_val(LOC);
+      }
+
+   clear_lambdas();
+   body.clear();
+}
+//-----------------------------------------------------------------------------
+void
+Executable::clear_lambdas()
+{
+   while (named_lambdas.size())
+      {
+        // remove named lambda, unless it is bound to a symbol
+        const UserFunction * ufun = named_lambdas.back();
+        if (ufun == 0)   continue;
+
+        const Function * fun = ufun;
+
+        const UCS_string & ufun_name = ufun->get_name();
+        const Symbol * ufun_sym = Workspace::lookup_existing_symbol(ufun_name);
+
+        bool used = false;
+        if (ufun_sym)
+           {
+              loop(v, ufun_sym->value_stack_size())
+                 {
+                   const ValueStackItem vs = (*ufun_sym)[v];
+                   if ((vs.name_class == NC_FUNCTION ||
+                        vs.name_class == NC_OPERATOR) &&
+                        vs.sym_val.function == fun)
+                      {
+                        // the named lambda was assigned to ufun_name so we
+                        // should not delete it.
+                        //
+                        used = true;
+                        break;
+                      }
+                 }
+           }
+
+        if (!used)   delete named_lambdas.back();
+        named_lambdas.pop_back();
+      }
+
+   while (unnamed_lambdas.size())
+      {
+        delete unnamed_lambdas.back();
+        unnamed_lambdas.pop_back();
+      }
 }
 //-----------------------------------------------------------------------------
 void
 Executable::parse_body_line(Function_Line line, const UCS_string & ucs_line,
-                            const char * loc)
+                            bool trace, const char * loc)
 {
    Log(LOG_UserFunction__set_line)
       CERR << "[" << line << "]" << ucs_line << endl;
@@ -83,12 +139,12 @@ const Parser parser(get_parse_mode(), loc);
      if (ec != E_NO_ERROR)   throw_parse_error(ec, LOC, LOC);
    }
 
-   parse_body_line(line, in, loc);
+   parse_body_line(line, in, trace, loc);
 } 
 //-----------------------------------------------------------------------------
 void
 Executable::parse_body_line(Function_Line line, const Token_string & in,
-                            const char * loc)
+                            bool trace, const char * loc)
 {
 Source<Token> src(in);
 
@@ -112,77 +168,62 @@ Token_string out;
       {
         // find next diamond. If there is one, copy statement in reverse order,
         // append TOK_ENDL, and continue.
+        
+        // 1. determine statement length
         //
-        {
-          uint32_t stat_len = 0;   // statement length, not counting ◊
-          bool got_diamond = false;
-          loop(t, src.rest())
-              {
+        ShapeItem stat_len = 0;   // statement length, not counting ◊
+        int diamond = 0;
+        loop(t, src.rest())
+            {
               if (src[t].get_tag() == TOK_DIAMOND)
                  {
-                   got_diamond = true;
+                   diamond = 1;
                    break;
                  }
               else
                  {
                    ++stat_len;
                  }
-              }
+            }
 
-          if (got_diamond)
-             {
-               loop(s, stat_len)
-                   {
-                     const Token & tok = src[stat_len - s - 1];
-                     if (tok.get_Class() >= TC_MAX_PERM &&
-                         tok.get_tag() != TOK_L_CURLY &&
-                         tok.get_tag() != TOK_R_CURLY)
-                        {
-                          CERR << "Line " << line << endl
-                               << "Offending token: (tag > TC_MAX_PERM) "
-                               << tok.get_tag() << " " << tok << endl
-                               << "Statement: ";
-                          Token::print_token_list(CERR, in);
-                          SYNTAX_ERROR;
-                        }
-                     out.append(tok);
-                   }
-               if (stat_len)   out.append(Token(TOK_ENDL), LOC);
-               src.skip(stat_len + 1);
-               continue;
-           }
-        }
-
-        // at this point src contains only the last statement.
-        // copy it in reverse order.
+        // 2. check for bad token and append statement to out
         //
-        loop(s, src.rest())
+        loop(s, stat_len)
+            {
+              const Token & tok = src[stat_len - s - 1];
+              if (tok.get_Class() >= TC_MAX_PERM &&
+                  tok.get_tag() != TOK_L_CURLY &&
+                  tok.get_tag() != TOK_R_CURLY)
+                 {
+                   CERR << "Line " << line << endl
+                        << "Offending token: (tag > TC_MAX_PERM) "
+                        << tok.get_tag() << " " << tok << endl
+                        << "Statement: ";
+                   Token::print_token_list(CERR, in);
+                   SYNTAX_ERROR;
+                 }
+              out.append(tok);
+            }
+        src.skip(stat_len + diamond);   // skip statement and maybe ◊
+
+        // 3. maybe add a TOK_END or TOK_ENDL
+        //
+        if (stat_len && (diamond || get_parse_mode() != PM_EXECUTE))
            {
-             const Token & tok = src[src.rest() - s - 1];
-             if (tok.get_Class() >= TC_MAX_PERM &&
-                 tok.get_tag()   != TOK_L_CURLY &&
-                 tok.get_tag()   != TOK_R_CURLY)
-                {
-                  CERR << "Line " << line << endl
-                       << "Offending token: (tag > TC_MAX_PERM) "
-                       << tok.get_tag() << " " << tok << endl
-                       << "Statement: ";
-                  Token::print_token_list(CERR, in);
-                  SYNTAX_ERROR;
-                }
-             out.append(tok);
+             const int64_t tr = trace ? 1 : 0;
+             out.append(Token(TOK_END, tr), LOC);
            }
-        src.skip(src.rest());
-
-        // add an appropriate end token
-        //
-        if (get_parse_mode() != PM_EXECUTE)   out.append(Token(TOK_ENDL), LOC);
       }
 
+   // replace the trailing TOK_END (if any) into TOK_ENDL
+   //
+   if (out.size() && out.last().get_tag() == TOK_END)
+      out.last().ChangeTag(TOK_ENDL);
+
    Log(LOG_UserFunction__set_line)
       {
         CERR << "[non-reverse " << line << "] ";
-        Parser::print_token_list(CERR, out, 0);
+        out.print(CERR);
       } 
 
    loop(t, out.size())   body.append(out[t], LOC);
@@ -202,14 +243,19 @@ StateIndicator & si = *Workspace::SI_top();
    FIXME;
 }
 //-----------------------------------------------------------------------------
-ostream &
-Executable::print(ostream & out) const
+void
+Executable::print_token(ostream & out) const
 {
    out << endl
        <<  "Function body [" << body.size() << " token]:" << endl;
 
-   Parser::print_token_list(out, body);
-   return out;
+   body.print(out);
+}
+//-----------------------------------------------------------------------------
+void
+Executable::print_text(ostream & out) const
+{
+   loop(l, text.size())   out << text[l] << endl;
 }
 //-----------------------------------------------------------------------------
 UCS_string
@@ -230,8 +276,13 @@ const UCS_string & line_txt = text[line];
    // count statement ends before pc
    //
 int stats_before = 0;
-   for (int p = line_start(line); p < pc; ++p)
+   for (Function_PC p = line_start(line); p < pc; ++p)
        {
+         // a TOK_STOP_LINE is immediately followd by a TC_END, but the
+         // TC_END has no ◊ in line_txt. We therefore discount stats_before.
+         //
+         if (body[p].get_tag() == TOK_STOP_LINE)          --stats_before;
+
          if (body[p].get_Class() == TC_END)               ++stats_before;
          else if (body[p].get_tag() == TOK_RETURN_EXEC)   ++stats_before;
        }
@@ -279,7 +330,7 @@ UCS_string & message_2 = error.error_message_2;
 
    if (body[pc_from_to.high].get_Class() == TC_RETURN &&
        pc_from_to.high > pc_from_to.low)
-       pc_from_to.high = Function_PC(pc_from_to.high - 1);;
+       pc_from_to.high = Function_PC(pc_from_to.high - 1);
 
 /*
     The statement is stored in reverse order in the body:
@@ -401,7 +452,7 @@ int count = 0;
         const Token & tok = body[b];
         if (tok.get_ValueType() != TV_VAL)      continue;
 
-        if (tok.get_apl_val().get() == &value)
+        if (tok.get_apl_val()->is_or_contains(value))
            {
              out << prefix << get_name() << "[" << b << "]" << endl;
              ++count;
@@ -414,6 +465,11 @@ int count = 0;
 void
 Executable::setup_lambdas()
 {
+   // setup_lambdas() may be called multiple times for the same executable.
+   // remove old lambdas...
+   //
+   clear_lambdas();
+
    loop(b, body.size())
        {
          // body is in reverse order, so } comes before { (and optional ← NAME)
@@ -422,12 +478,13 @@ Executable::setup_lambdas()
 
          // a named lambda has the form Q←{ ... }
          //
-         // however, eg. Q←{ ... } / ... is not a named lambda by an unnamed
+         // however, eg. Q←{ ... } / ... is not a named lambda but an unnamed
          // lambda for / whose result is assigned to Q. We need to distinguish
          // the two cases and remember if the token after } is the end of
          // the statement or not.
          //
          const bool maybe_named = b == 0 || (body[b-1].get_tag() == TOK_DIAMOND
+                                         || body[b-1].get_tag() == TOK_END
                                          || body[b-1].get_tag() == TOK_ENDL);
 
          body[b++].clear(LOC);   // invalidate }
@@ -443,11 +500,13 @@ Executable::setup_lambdas()
                move_1(t, body[b], LOC);
                body[b].clear(LOC);   // invalidate in main body
 
+               // figure the signature by looking for ⍺, ⍶, ⍵, ⍹, and χ
+               // and compain about ◊ and →
                switch(t.get_tag())
                   {
                     case TOK_L_CURLY:
-                         if (body_nesting == 1)  { goon = false;   continue; }
-                         else               { --body_nesting;      break;    }
+                         if (body_nesting == 1)  { goon = false;     continue; }
+                         else                    { --body_nesting;   break;    }
 
                     case TOK_ALPHA:    signature |= SIG_A;    /* no break */
                     case TOK_OMEGA:    signature |= SIG_B;    break;
@@ -459,6 +518,8 @@ Executable::setup_lambdas()
                     case TOK_DIAMOND:  DEFN_ERROR;
                     case TOK_BRANCH:   DEFN_ERROR;
                     case TOK_ESCAPE:   DEFN_ERROR;
+
+                    default: break;
                   }
 
                rev_lambda_body.append(t, LOC);
@@ -476,7 +537,8 @@ Executable::setup_lambdas()
                 Token ret_lambda(TOK_RETURN_SYMBOL, &Workspace::get_v_LAMBDA());
                 forw_lambda_body.append(ret_lambda, LOC);
 
-                forw_lambda_body.append(Token(TOK_ENDL));
+                const int64_t tr = 0;
+                forw_lambda_body.append(Token(TOK_ENDL, tr));
 
                 Symbol * sym_Z = &Workspace::get_v_LAMBDA();
                 forw_lambda_body.append(Token(TOK_LAMBDA, sym_Z), LOC);
@@ -493,7 +555,6 @@ Executable::setup_lambdas()
                 forw_lambda_body.append(ret_void, LOC);
               }
 
-
          // at this point {} was copied from body to forw_lambda_body and
          // b is at the (now invalidated) { token.
          // check if lambda is named, i.e. Name ← { ... }
@@ -509,9 +570,6 @@ Executable::setup_lambdas()
               lambda_sym = body[b + 2].get_sym_ptr();
               lambda_name = lambda_sym->get_name();
               signature |= SIG_FUN;
-
-               body[b + 1].clear(LOC);
-               body[b + 2].clear(LOC);
             }
          else
             {
@@ -519,61 +577,9 @@ Executable::setup_lambdas()
               lambda_name.append_number(unnamed_lambdas.size() + 1);
             }
 
-         UCS_string lambda_text;
-         if (signature & SIG_Z)   lambda_text.append_utf8("λ←");
-
-         int lambda_skip = unnamed_lambdas.size() + named_lambdas.size();
-         int opening_curly = 0;
-         int closing_curly = 0;
-         int tidx = 0;
-         int tcol = 0;
-
-         for (;;)
-             {
-               // search for the lambda_skip'th { in the body text
-               //
-               Assert(tidx < text.size());
-               const UCS_string & line = text[tidx];
-
-               if (tcol >= line.size())   // end of line: wrap to next line
-                  {
-                    ++tidx;
-                    tcol = 0;
-                    continue;
-                  }
-
-               {
-                 const Unicode uni = line[tcol++];
-                 if      (uni == UNI_ASCII_L_CURLY)   ++opening_curly;
-                 else if (uni == UNI_ASCII_R_CURLY)   ++closing_curly;
-
-                 if (uni != UNI_ASCII_L_CURLY)   continue;   // not {
-               }
-
-               // found { (but maybe from a previous lambda
-               // or from a sub-lambds)
-               //
-               if (lambda_skip)   { --lambda_skip;   continue; }
-
-               if (tcol >= line.size())   SYNTAX_ERROR;
-               for (;;)
-                   {
-                     const Unicode uni = line[tcol++];
-                     if      (uni == UNI_ASCII_L_CURLY)   ++opening_curly;
-                     else if (uni == UNI_ASCII_R_CURLY)   ++closing_curly;
-
-                     // if uni was } then it could be:
-                     //
-                     // 1. a sub-{}      as in { {...} or
-                     // 2. a top-level-} as in { ,,, }
-                     // 
-                     // opening_curly == closing_curly distinguishes 1. and 2.
-                     //
-                     if (opening_curly == closing_curly)   break;
-                     lambda_text.append(uni);
-                   }
-               break;
-             }
+         const UCS_string lambda_text = extract_lambda_text(
+                                (Fun_signature)signature,
+                                unnamed_lambdas.size() + named_lambdas.size());
 
          UserFunction * ufun = new UserFunction(Fun_signature(signature),
                                                 lambda_name, lambda_text,
@@ -582,12 +588,13 @@ Executable::setup_lambdas()
          if (signature & SIG_FUN)   // named lambda
             {
               named_lambdas.push_back(ufun);
+              move_2(body[b], ufun->get_token(), LOC);
 
-              // replace the lambda by its name, but as committed value
-              // so that it doesn't print (similar to ⎕FX).
+              // UserFunction::UserFunction has bound the lambda body to
+              // the lambda name. However, we want that to happen when
+              // V←{ ...} is executed, not now. We therefore unbind it here
               //
-              Value_P val_name(new Value(lambda_name, LOC));
-              move_2(body[b], Token(TOK_APL_VALUE2, val_name), LOC);
+              lambda_sym->set_nc(NC_UNUSED_USER_NAME, 0);
             }
          else                       // unnamed lambda
             {
@@ -598,6 +605,63 @@ Executable::setup_lambdas()
 
    Parser::remove_void_token(body);
 }
+//-----------------------------------------------------------------------------
+UCS_string
+Executable::extract_lambda_text(Fun_signature signature, int skip) const
+{
+UCS_string lambda_text;
+   if (signature & SIG_Z)   lambda_text.append_utf8("λ←");
+
+int level = 0;   // {/} nesting level
+bool copying = false;
+
+int tidx = 0;
+int tcol = 0;
+
+   // skip over the first skip lambdas and copy the next one to lambda_text
+   //
+   for (;;)
+       {
+         if (tidx >= text.size())
+            {
+              Q(copying)
+              Q(tidx)
+              FIXME;
+            }
+
+         const UCS_string & line = text[tidx];
+
+         if (tcol >= line.size())   // end of line: wrap to next line
+            {
+              ++tidx;
+
+              tcol = 0;
+              continue;
+            }
+
+         const Unicode uni = line[tcol++];
+         if (copying)   lambda_text.append(uni);
+
+         if (uni == UNI_ASCII_L_CURLY)
+            {
+              if (level++ == 0)   // top-level {
+                 {
+                   if (skip == 0)   copying = true;
+                 }
+            }
+         else if (uni == UNI_ASCII_R_CURLY)
+            {
+              if (--level == 0)   // top-level }
+                 {
+                   if (copying)   break;   // done
+                   --skip;
+                 }
+            }
+       }
+
+   lambda_text.pop();   // the last }
+   return lambda_text;
+}
 //=============================================================================
 ExecuteList *
 ExecuteList::fix(const UCS_string & data, const char * loc)
@@ -614,7 +678,7 @@ ExecuteList * fun = new ExecuteList(data, loc);
       {
         CERR << "fix pmode=execute list:" << endl << data
              << " addr " << (const void *)fun << endl
-             << "------------------- fix --" << endl;
+             << "------------------- ExecuteList::fix() --" << endl;
       }
 
    {
@@ -635,7 +699,7 @@ ExecuteList * fun = new ExecuteList(data, loc);
 
    try
       {
-        fun->parse_body_line(Function_Line_0, data, loc);
+        fun->parse_body_line(Function_Line_0, data, false, loc);
       }
    catch (Error err)
       {
@@ -645,6 +709,8 @@ ExecuteList * fun = new ExecuteList(data, loc);
         return 0;
       }
 
+   fun->setup_lambdas();
+
    Log(LOG_UserFunction__fix)
       {
         CERR << "fun->body.size() is " << fun->body.size() << endl;
@@ -666,7 +732,7 @@ StatementList * fun = new StatementList(data, loc);
       {
         CERR << "fix pmode=statement list:" << endl << data << endl
              << " addr " << (const void *)fun << endl
-             << "------------------- fix --" << endl;
+             << "------------------- StatementList::fix() --" << endl;
       }
 
    {
@@ -674,7 +740,7 @@ StatementList * fun = new StatementList(data, loc);
      if (err)   err->parser_loc = 0;
    }
 
-   fun->parse_body_line(Function_Line_0, data, loc);
+   fun->parse_body_line(Function_Line_0, data, false, loc);
    fun->setup_lambdas();
 
    Log(LOG_UserFunction__fix)
