@@ -27,7 +27,7 @@
 #include "FloatCell.hh"
 #include "IntCell.hh"
 #include "UTF8_string.hh"
-#include "Value.hh"
+#include "Value.icc"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
@@ -36,29 +36,7 @@ FloatCell::equal(const Cell & A, APL_Float qct) const
 {
    if (!A.is_numeric())       return false;
    if (A.is_complex_cell())   return A.equal(*this, qct);
-   return equal_within_quad_CT(A.get_real_value(), get_real_value(), qct);
-}
-//-----------------------------------------------------------------------------
-bool
-FloatCell::equal_within_quad_CT(APL_Float A, APL_Float B, APL_Float qct)
-{
-const APL_Float posA = A < 0 ? -A : A;
-const APL_Float posB = B < 0 ? -B : B;
-const APL_Float maxAB = posA > posB ? posA : posB;
-
-   // if the signs of A and B differ then they are unequal (standard page 100)
-   // we treat exact 0.0 as having both signs
-   //
-   if (A < 0.0 && B > 0.0)   return false;
-   if (A > 0.0 && B < 0.0)   return false;
-
-   // for large numbers we allow a larger ⎕CT
-   //
-   if (maxAB > 1.0)   qct *= maxAB;
-
-   if (posB < (posA - qct))   return false;
-   if (posB > (posA + qct))   return false;
-   return true;
+   return tolerantly_equal(A.get_real_value(), get_real_value(), qct);
 }
 //-----------------------------------------------------------------------------
 bool
@@ -86,7 +64,7 @@ const APL_Float this_val  = get_real_value();
         case CT_COMPLEX: break;
         case CT_POINTER: return !ascending;
         case CT_CELLREF: DOMAIN_ERROR;
-        defaulkt:        Assert(0 && "Bad celltype");
+        default:         Assert(0 && "Bad celltype");
       }
 
 const Comp_result comp = compare(*other);
@@ -153,7 +131,8 @@ FloatCell::compare(const Cell & other) const
 //-----------------------------------------------------------------------------
 // monadic built-in functions...
 //-----------------------------------------------------------------------------
-void FloatCell::bif_factorial(Cell * Z) const
+ErrorCode
+FloatCell::bif_factorial(Cell * Z) const
 {
 const APL_Float qct = Workspace::get_CT();
 
@@ -162,72 +141,85 @@ const APL_Float qct = Workspace::get_CT();
         const APL_Integer iv = get_near_int(qct);
         IntCell icell(iv);
         icell.bif_factorial(Z);
-        return;
+        return E_NO_ERROR;
       }
 
    // max N! that fits into double is about 170
    //
-   if (value.fval > 170)   DOMAIN_ERROR;
+   if (value.fval > 170)   return E_DOMAIN_ERROR;
 
 const double arg = value.fval + 1.0;
    new (Z) FloatCell(tgamma(arg));
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void FloatCell::bif_conjugate(Cell * Z) const
+ErrorCode
+FloatCell::bif_conjugate(Cell * Z) const
 {
    if (is_near_int(Workspace::get_CT()))
       new (Z) IntCell(get_checked_near_int());
    else
       new (Z) FloatCell(value.fval);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void FloatCell::bif_negative(Cell * Z) const
+ErrorCode
+FloatCell::bif_negative(Cell * Z) const
 {
    new (Z) FloatCell(- value.fval);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 FloatCell::bif_direction(Cell * Z) const
 {
    // bif_direction does NOT use ⎕CT
    //
-const APL_Float qct = Workspace::get_CT();
    if      (value.fval == 0)   new (Z) IntCell( 0);
    else if (value.fval < 0)    new (Z) IntCell(-1);
    else                        new (Z) IntCell( 1);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void FloatCell::bif_magnitude(Cell * Z) const
+ErrorCode
+FloatCell::bif_magnitude(Cell * Z) const
 {
    if (value.fval >= 0.0)   new (Z) FloatCell( value.fval);
    else                     new (Z) FloatCell(-value.fval);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void FloatCell::bif_reciprocal(Cell * Z) const
+ErrorCode
+FloatCell::bif_reciprocal(Cell * Z) const
 {
-   if (is_near_zero(Workspace::get_CT()))  DOMAIN_ERROR;
+   if (is_near_zero(Workspace::get_CT()))  return E_DOMAIN_ERROR;
 
    new (Z) FloatCell(1.0/value.fval);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void FloatCell::bif_roll(Cell * Z) const
+ErrorCode
+FloatCell::bif_roll(Cell * Z) const
 {
 const APL_Integer qio = Workspace::get_IO();
-   if (!is_near_int(Workspace::get_CT()))   DOMAIN_ERROR;
+   if (!is_near_int(Workspace::get_CT()))   return E_DOMAIN_ERROR;
 
 const APL_Integer set_size = get_checked_near_int();
-   if (set_size <= 0)   DOMAIN_ERROR;
+   if (set_size <= 0)   return E_DOMAIN_ERROR;
 
-const APL_Integer rnd = Workspace::get_RL();
+const uint64_t rnd = Workspace::get_RL();
    new (Z) IntCell(qio + (rnd % set_size));
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void FloatCell::bif_pi_times(Cell * Z) const
+ErrorCode
+FloatCell::bif_pi_times(Cell * Z) const
 {
    new (Z) FloatCell(M_PI * value.fval);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 FloatCell::bif_ceiling(Cell * Z) const
 {
    // we want values slightly above an int to be rounded down rather than up
@@ -237,9 +229,11 @@ const double d_ret = ceil(value.fval - Workspace::get_CT());
 
    if (Cell::is_near_int(d_ret, qct))   new (Z) IntCell(llrint(d_ret));
    else                                 new (Z) FloatCell(d_ret);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void FloatCell::bif_floor(Cell * Z) const
+ErrorCode
+FloatCell::bif_floor(Cell * Z) const
 {
    // we want values slightly below an int to be rounded up rather than down
    //
@@ -248,14 +242,20 @@ const double d_ret = floor(value.fval + qct);
 
    if (Cell::is_near_int(d_ret, qct))   new (Z) IntCell(llrint(d_ret)); 
    else                                 new (Z) FloatCell(d_ret);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void FloatCell::bif_exponential(Cell * Z) const
+ErrorCode
+FloatCell::bif_exponential(Cell * Z) const
 {
+   // e to the B-th power
+   //
    new (Z) FloatCell(exp(value.fval));
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void FloatCell::bif_nat_log(Cell * Z) const
+ErrorCode
+FloatCell::bif_nat_log(Cell * Z) const
 {
    if (value.fval > 0)
       {
@@ -269,170 +269,208 @@ void FloatCell::bif_nat_log(Cell * Z) const
       }
    else
       {
-        DOMAIN_ERROR;
+        return E_DOMAIN_ERROR;
       }
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 // dyadic built-in functions...
+//
+// where possible a function with non-real A is delegated to the corresponding
+// member function of A. For numeric cells that is the ComplexCell function
+// and otherwise the default function (that returns E_DOMAIN_ERROR.
+//
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 FloatCell::bif_add(Cell * Z, const Cell * A) const
 {
-   if (A->is_complex_cell())
-      {
-        A->bif_add(Z, this);
-      }
-   else if (A->is_real_cell())
+   if (A->is_real_cell())
       {
         new (Z) FloatCell(A->get_real_value() + get_real_value());
         Z->demote_float_to_int(Workspace::get_CT());
+        return E_NO_ERROR;
       }
-   else
-      {
-        DOMAIN_ERROR;
-      }
+
+   // delegate to A
+   //
+   return A->bif_add(Z, this);
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 FloatCell::bif_subtract(Cell * Z, const Cell * A) const
 {
-const APL_Float qct = Workspace::get_CT();
-
-   if (A->is_complex_cell())
-      {
-        new (Z) ComplexCell(A->get_real_value() - get_real_value(),
-                            A->get_imag_value());
-        Z->demote_complex_to_real(qct);
-        Z->demote_float_to_int(qct);
-      }
-   else if (A->is_real_cell())
+   if (A->is_real_cell())
       {
         new (Z) FloatCell(A->get_real_value() - get_real_value());
-        Z->demote_float_to_int(qct);
-      }
-   else
-      {
-        DOMAIN_ERROR;
+        Z->demote_float_to_int(Workspace::get_CT());
+        return E_NO_ERROR;
       }
+
+   // delegate to A
+   //
+   return A->bif_subtract(Z, this);
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 FloatCell::bif_multiply(Cell * Z, const Cell * A) const
 {
-const APL_Float qct = Workspace::get_CT();
-      
-   if (A->is_complex_cell())
-      {                     
-        A->bif_multiply(Z, this);
-      } 
-   else if (A->is_real_cell())
+   if (A->is_real_cell())
       { 
         new (Z) FloatCell(A->get_real_value() * get_real_value());
-        Z->demote_float_to_int(qct);
-      } 
-   else
-      {
-        DOMAIN_ERROR;
+        Z->demote_float_to_int(Workspace::get_CT());
+        return E_NO_ERROR;
       } 
+
+   // delegate to A
+   //
+   return A->bif_multiply(Z, this);
 }     
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 FloatCell::bif_divide(Cell * Z, const Cell * A) const
 {
 const APL_Float qct = Workspace::get_CT();
 
    if (this->is_near_zero(qct))   // divide by 0
       {
-         if (!A->is_near_zero(qct))   DOMAIN_ERROR;
+         if (!A->is_near_zero(qct))   return E_DOMAIN_ERROR;
          new (Z) IntCell(1);   // 0/0 is 1 in APL
-         return;
+         return E_NO_ERROR;
       }
 
-const APL_Float real = A->get_real_value() / get_real_value();
 
-   if (A->is_complex_cell())
-      {
-        const APL_Float imag = A->get_imag_value() / get_real_value();
-        if (!Cell::is_near_zero(imag, qct))     new (Z) ComplexCell(real, imag);
-        else if (Cell::is_near_int(real, qct))  new (Z) IntCell(real, qct);
-        else                                    new (Z) FloatCell(real);
-      }
-   else if (A->is_real_cell())
+   if (A->is_real_cell())
       {
+        const APL_Float real = A->get_real_value() / get_real_value();
         if (Cell::is_near_int(real, qct))   new (Z) IntCell(real, qct);
         else                                new (Z) FloatCell(real);
+         return E_NO_ERROR;
       }
-   else
-      {
-        DOMAIN_ERROR;
-      }
+
+   // delegate to A
+   //
+   return A->bif_divide(Z, this);
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 FloatCell::bif_power(Cell * Z, const Cell * A) const
 {
+   // some A to the real B-th power
+   //
+   if (!A->is_numeric())   return E_DOMAIN_ERROR;
+
+const APL_Float qct = Workspace::get_CT();
+   if (is_near_int(qct))
+      {
+        APL_Integer b = get_near_int(qct);
+        const bool invert_Z = b < 0;
+        if (invert_Z)   b = - b;
+
+        if (b <= 1)   // special cases A⋆1, A⋆0, and A⋆¯1
+           {
+             if (b == 0)   return IntCell::z1(Z);  // A⋆0 is 1
+
+             if (invert_Z)               return A->bif_reciprocal(Z);
+             if (A->is_real_cell())      return A->bif_conjugate(Z);
+             new (Z) ComplexCell(A->get_complex_value());
+             return E_NO_ERROR;
+           }
+      }
+
+   if (A->is_near_zero(qct))
+      {
+        if (get_real_value() < 0)   return E_DOMAIN_ERROR;
+        return IntCell::z0(Z);
+      }
+          
    if (A->is_real_cell())
       {
         if (A->get_real_value() < 0)
            {
-             const APL_Complex z = pow(A->get_complex_value(),get_real_value());
-             new (Z) ComplexCell(z);
-             Z->demote_complex_to_real(Workspace::get_CT());
+             if (is_near_int(qct))
+                {
+                  const APL_Integer b = get_near_int(qct);
+                  const APL_Float z = pow(A->get_real_value(), (APL_Float)b);
+                  new (Z) FloatCell(z);
+                }
+             else
+                {
+                  const APL_Complex z = pow(A->get_complex_value(),
+                                            get_complex_value());
+                  new (Z) ComplexCell(z);
+                  Z->demote_complex_to_real(Workspace::get_CT());
+                }
            }
         else
            {
-             new (Z) FloatCell(pow(A->get_real_value(), get_real_value()));
+             new (Z) FloatCell(pow(A->get_real_value(),
+                               get_real_value()));
            }
+        return E_NO_ERROR;
       }
-   else if (A->is_complex_cell())
-      {
-        APL_Complex z = pow(A->get_complex_value(), get_real_value());
 
-        new (Z) ComplexCell(z);
-        Z->demote_complex_to_real(Workspace::get_CT());
-      }
-   else
-      {
-        DOMAIN_ERROR;
-      }
+   // complex A to the B-th power
+   //
+   {
+     APL_Complex z = pow(A->get_complex_value(), get_real_value());
+
+     new (Z) ComplexCell(z);
+     Z->demote_complex_to_real(Workspace::get_CT());
+     return E_NO_ERROR;
+   }
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 FloatCell::bif_residue(Cell * Z, const Cell * A) const
 {
-const APL_Float qct = Workspace::get_CT();
-
-   // if A is zero , return B
-   //
-   if (A->is_near_zero(0))
+   if (A->is_complex_cell())
       {
-        Z->init(*this);
-        return;
+        ComplexCell B(get_real_value(), 0);
+        return B.bif_residue(Z, A);
       }
 
-   // if ⎕CT != 0 and B÷A is close to an integer within ⎕CT then return zero.
-   // Otherwise return B mod A
-   //
-const APL_Float remainder = fmod(get_real_value(), A->get_real_value()); 
-   if (qct == 0.0)   // ⎕CT is 0
-      {
-        new (Z) FloatCell(remainder);
-      }
-   else if (Cell::is_near_zero(remainder, qct))
-      {
-        new (Z) IntCell(0);
-      }
-   else if (Cell::is_near_zero(A->get_real_value() - remainder, qct))
-      {
-        new (Z) IntCell(0);
-      }
-   else
+   if (A->is_numeric())
       {
+        const APL_Float qct = Workspace::get_CT();
+
+        // if A is zero , return B
+        //
+        if (A->is_near_zero(0))
+           {
+             Z->init(*this);
+             return E_NO_ERROR;
+           }
+
+        // if ⎕CT != 0 and B÷A is close to an integer within ⎕CT then return 0.
+        // Otherwise return B mod A
+        //
+        const APL_Float remainder = fmod(get_real_value(), A->get_real_value()); 
+        if (qct == 0.0)   // ⎕CT is 0
+           {
+             new (Z) FloatCell(remainder);
+             return E_NO_ERROR;
+           }
+
+        if (Cell::is_near_zero(remainder, qct))
+           {
+             new (Z) IntCell(0);
+             return E_NO_ERROR;
+           }
+
+        if (Cell::is_near_zero(A->get_real_value() - remainder, qct))
+           {
+             new (Z) IntCell(0);
+             return E_NO_ERROR;
+           }
+
         new (Z) FloatCell(remainder);
+        return E_NO_ERROR;
       }
+
+   return E_DOMAIN_ERROR;
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 FloatCell::bif_maximum(Cell * Z, const Cell * A) const
 {
    if (A->is_integer_cell())
@@ -440,20 +478,23 @@ FloatCell::bif_maximum(Cell * Z, const Cell * A) const
          const APL_Integer a = A->get_int_value();
          if (APL_Float(a) >= value.fval)   new (Z) IntCell(a);
          else                              new (Z) FloatCell(value.fval);
+         return E_NO_ERROR;
       }
-   else if (A->is_float_cell())
+
+   if (A->is_float_cell())
       {
          const APL_Float a = A->get_real_value();
          if (a >= value.fval)   new (Z) FloatCell(a);
          else                   new (Z) FloatCell(value.fval);
+         return E_NO_ERROR;
       }
-   else
-      {
-        A->bif_maximum(Z, this);
-      }
+
+   // delegate to A
+   //
+   return A->bif_maximum(Z, this);
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 FloatCell::bif_minimum(Cell * Z, const Cell * A) const
 {
    if (A->is_integer_cell())
@@ -461,19 +502,25 @@ FloatCell::bif_minimum(Cell * Z, const Cell * A) const
          const APL_Integer a = A->get_int_value();
          if (APL_Float(a) <= value.fval)   new (Z) IntCell(a);
          else                              new (Z) FloatCell(value.fval);
+         return E_NO_ERROR;
       }
-   else if (A->is_float_cell())
+
+   if (A->is_float_cell())
       {
          const APL_Float a = A->get_real_value();
          if (a <= value.fval)   new (Z) FloatCell(a);
          else                   new (Z) FloatCell(value.fval);
+         return E_NO_ERROR;
       }
-   else
-      {
-        A->bif_minimum(Z, this);
-      }
+
+   // delegate to A
+   //
+   return A->bif_minimum(Z, this);
 }
-//-----------------------------------------------------------------------------
+//=============================================================================
+// throw/nothrow boundary. Functions above MUST NOT (directly or indirectly)
+// throw while funcions below MAY throw.
+//=============================================================================
 PrintBuffer
 FloatCell::character_representation(const PrintContext & pctx) const
 {
@@ -486,7 +533,7 @@ ColInfo info;
 
    // assume integer.
    //
-int int_fract = ucs.size();;
+int int_fract = ucs.size();
    info.real_len = ucs.size();
    info.int_len = ucs.size();
    loop(u, ucs.size())
@@ -570,6 +617,7 @@ FloatCell::map_FC(UCS_string & ucs)
              case UNI_ASCII_FULLSTOP: ucs[u] = Workspace::get_FC(0);   break;
              case UNI_ASCII_COMMA:    ucs[u] = Workspace::get_FC(1);   break;
              case UNI_OVERBAR:        ucs[u] = Workspace::get_FC(5);   break;
+             default:                 break;
            }
       }
 }
