@@ -31,7 +31,7 @@
 #include "Parser.hh"
 #include "PrintOperator.hh"
 #include "Symbol.hh"
-#include "Value.hh"
+#include "Value.icc"
 
 //-----------------------------------------------------------------------------
 void
@@ -46,13 +46,16 @@ Function::get_attributes(int mode, Cell * dest) const
                 return;
 
         case 2: // creation time (7⍴0 for system functions)
-                new (dest + 0) IntCell(0);
-                new (dest + 1) IntCell(0);
-                new (dest + 2) IntCell(0);
-                new (dest + 3) IntCell(0);
-                new (dest + 4) IntCell(0);
-                new (dest + 5) IntCell(0);
-                new (dest + 6) IntCell(0);
+                {
+                  const YMDhmsu created(get_creation_time());
+                  new (dest + 0) IntCell(created.year);
+                  new (dest + 1) IntCell(created.month);
+                  new (dest + 2) IntCell(created.day);
+                  new (dest + 3) IntCell(created.hour);
+                  new (dest + 4) IntCell(created.minute);
+                  new (dest + 5) IntCell(created.second);
+                  new (dest + 6) IntCell(created.micro/1000);
+                }
                 return;
 
         case 3: // execution properties
@@ -82,7 +85,7 @@ Function::eval_()
 Token
 Function::eval_B(Value_P B)
 {
-   CERR << get_name() << "::" << __FUNCTION__
+   Log(LOG_verbose_error)   CERR << get_name() << "::" << __FUNCTION__
         << "() called (overloaded variant not yet implemented?)" << endl;
    VALENCE_ERROR;
 }
@@ -90,7 +93,7 @@ Function::eval_B(Value_P B)
 Token
 Function::eval_AB(Value_P A, Value_P B)
 {
-   CERR << get_name() << "::" << __FUNCTION__
+   Log(LOG_verbose_error)   CERR << get_name() << "::" << __FUNCTION__
         << "() called (overloaded variant not yet implemented?)" << endl;
    VALENCE_ERROR;
 }
@@ -98,7 +101,7 @@ Function::eval_AB(Value_P A, Value_P B)
 Token
 Function::eval_LB(Token & LO, Value_P B)
 {
-   CERR << get_name() << "::" << __FUNCTION__
+   Log(LOG_verbose_error)   CERR << get_name() << "::" << __FUNCTION__
         << "() called (overloaded variant not yet implemented?)" << endl;
    VALENCE_ERROR;
 }
@@ -106,7 +109,7 @@ Function::eval_LB(Token & LO, Value_P B)
 Token
 Function::eval_ALB(Value_P A, Token & LO, Value_P B)
 {
-   CERR << get_name() << "::" << __FUNCTION__
+   Log(LOG_verbose_error)   CERR << get_name() << "::" << __FUNCTION__
         << "() called (overloaded variant not yet implemented?)" << endl;
    VALENCE_ERROR;
 }
@@ -114,7 +117,7 @@ Function::eval_ALB(Value_P A, Token & LO, Value_P B)
 Token
 Function::eval_LRB(Token & LO, Token & RO, Value_P B)
 {
-   CERR << get_name() << "::" << __FUNCTION__
+   Log(LOG_verbose_error)   CERR << get_name() << "::" << __FUNCTION__
         << "() called (overloaded variant not yet implemented?)" << endl;
    VALENCE_ERROR;
 }
@@ -122,7 +125,7 @@ Function::eval_LRB(Token & LO, Token & RO, Value_P B)
 Token
 Function::eval_ALRB(Value_P A, Token & LO, Token & RO, Value_P B)
 {
-   CERR << get_name() << "::" << __FUNCTION__
+   Log(LOG_verbose_error)   CERR << get_name() << "::" << __FUNCTION__
         << "() called (overloaded variant not yet implemented?)" << endl;
    VALENCE_ERROR;
 }
