@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2014  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,18 +18,24 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
+#include <string.h>
+#include <termios.h>
 #include <unistd.h>
 
+#define Function APL_Function   // make realine happy
 #include "Command.hh"
 #include "Common.hh"   // for HAVE_LIBREADLINE etc.
 #include "Input.hh"
+#include "InputFile.hh"
+#include "IO_Files.hh"
 #include "main.hh"
 #include "Output.hh"
 #include "PrintOperator.hh"
+#include <signal.h>
 #include "SystemVariable.hh"
-#include "TestFiles.hh"
 #include "UTF8_string.hh"
 #include "Workspace.hh"
+#undef Function
 
 int Input::readline_history_len = 500;
 UTF8_string Input::readline_history_path(".apl.history");
@@ -47,31 +53,54 @@ void (*end_input)() = 0;
 
 bool Input::use_readline = true;
 
-namespace readline_lib
-{
-#include <readline/readline.h>
-#include <readline/history.h>
-};   // namespace readline_lib
+#  if defined(HAVE_READLINE_READLINE_H)
+#    include <readline/readline.h>
+#  elif defined(HAVE_READLINE_H)
+#    include <readline.h>
+#  else /* !defined(HAVE_READLINE_H) */
+extern char *readline ();
+#  endif /* !defined(HAVE_READLINE_H) */
+char *cmdline = NULL;
+
+#ifdef HAVE_READLINE_HISTORY
+#  if defined(HAVE_READLINE_HISTORY_H)
+#    include <readline/history.h>
+#  elif defined(HAVE_HISTORY_H)
+#    include <history.h>
+#  else /* !defined(HAVE_HISTORY_H) */
+extern void add_history ();
+extern int write_history ();
+extern int read_history ();
+#  endif /* defined(HAVE_READLINE_HISTORY_H) */
+/* no history */
+#endif /* HAVE_READLINE_HISTORY */
 
 #else // ! HAVE_LIBREADLINE
 
 bool Input::use_readline = false;
 
-namespace readline_lib
-{
-static void add_history(const char * line) {}
-static void rl_initialize() {}
-static void stifle_history(int) {}
-static void read_history(const char *) {}
-static void rl_stuff_char(char) {}
-static char * readline(const char *) {}
-const char *  rl_readline_name = "GnuAPL";
+static void   add_history(const char *)    { }
+static void   read_history(const char *)   { }
 
-};   // namespace readline_lib
+static void   rl_initialize()              { }
+static void   stifle_history(int)          { }
+static void   rl_stuff_char(char)          { }
+static void   rl_list_funmap_names()       { }
 
-#endif //  HAVE_LIBREADLINE
+static char * readline(const char *)       { return 0; }
+static int    rl_bind_key(int, int)        { return 0; }
+static int    rl_crlf()                    { return 0; }
+static int    rl_delete_text(int, int)     { return 0; }
 
-// a function to be used if readline_lib::readline() does not exist or
+
+static const char * rl_readline_name = "GnuAPL";
+static int          rl_catch_signals = 0;
+static int          rl_done = 0;
+void              (*rl_startup_hook)() = 0;
+
+#endif //  don't HAVE_LIBREADLINE
+
+// a function to be used if readline() does not exist or
 // is not wanted
 static char *
 no_readline(const UCS_string * prompt)
@@ -96,36 +125,35 @@ int len = strlen(buffer);
    if (len && buffer[len - 1] == '\r')   buffer[--len] = 0;
    return buffer;
 }
-
-
-const char * Input::input_file_name = 0;
-FILE * Input::input_file_FILE = 0;
-
-bool Input::stdin_from_file = false;
-
 //-----------------------------------------------------------------------------
 void
-Input::init(bool read_history)
+Input::init(bool do_read_history)
 {
+#if HAVE_LIBREADLINE
    if (use_readline)
       {
-        readline_lib::rl_readline_name = "GnuAPL";
-        readline_lib::rl_initialize();
-        if (read_history)
+        rl_readline_name = strdup("GnuAPL");
+        rl_initialize();
+        if (do_read_history)
            {
-             readline_lib::stifle_history(readline_history_len);
-             readline_lib::read_history(readline_history_path.c_str());
+             stifle_history(readline_history_len);
+             read_history(readline_history_path.c_str());
            }
 
-//      readline_lib::rl_function_dumper(1);
+#ifdef ADVANCED_readline
+        rl_startup_hook = (rl_hook_func_t *)(Input::init_readline_control_C);
+#endif
+
+//      rl_function_dumper(1);
       }
+#endif
 }
 //-----------------------------------------------------------------------------
 int
 Input::readline_version()
 {
 #if HAVE_LIBREADLINE
-   return readline_lib::rl_readline_version;
+   return rl_readline_version;
 #else
    return 0;
 #endif
@@ -135,58 +163,24 @@ UCS_string
 Input::get_line()
 {
    Quad_QUOTE::done(true, LOC);
+   InputFile::increment_current_line_no();
 
 const char * input_type = "?";
-const UTF8 * buf = TestFiles::get_testcase_line();
+const UTF8 * buf = IO_Files::get_file_line();
 
    if (buf)   // we got a line from a test file.
       {
         input_type = "-T";
         CIN << Workspace::get_prompt() << buf << endl;
       }
-   else if (input_file_FILE)   // a -f file is open
-      {
-        input_type = "-f";
-        buf = get_f_line(input_file_FILE);
-        if (buf)   CIN << Workspace::get_prompt() << buf << endl;
-      }
-   else if (input_file_name)   // -f present, but not opened yet
-      {
-        // A special case is -f - where we open stdin
-        //
-        stdin_from_file = !strcmp(input_file_name, "-");
-        if (stdin_from_file && scriptname)
-           {
-             input_file_FILE = fopen(scriptname, "r");
-           }
-        else if (stdin_from_file)
-           {
-            input_file_FILE = stdin;
-           }
-        else
-           {
-             input_file_FILE = fopen(input_file_name, "r");
-           }
-
-        if (input_file_FILE)
-           {
-             input_type = "-f";
-             buf = get_f_line(input_file_FILE);
-           }
-        else
-           {
-             CERR << "cannot open file " << input_file_name
-                  << " (from -f) at " LOC << endl;
-             input_file_name = 0;
-           }
-      }
 
 const APL_time_us from = now();
 int control_D_count = 0;
    while (buf == 0)
       {
         input_type = "U";
-        buf = get_user_line(&Workspace::get_prompt(), false);
+        Output::set_color_mode(Output::COLM_INPUT);
+        buf = get_user_line(&Workspace::get_prompt());
 
         if (buf == 0)   // ^D or end of file
            {
@@ -238,13 +232,7 @@ Input::get_f_line(FILE * file)
 static char buffer[MAX_INPUT_LEN];
 
 const char * s = fgets(buffer, sizeof(buffer) - 1, file);
-   if (s == 0)   // end of file
-      {
-        input_file_name = 0;
-        input_file_FILE = 0;
-        stdin_from_file = false;
-        return 0;
-      }
+   if (s == 0)   return 0;// end of file
 
    buffer[sizeof(buffer) - 1] = 0;   // make strlen happy.
 size_t len = strlen(buffer);
@@ -258,17 +246,15 @@ void
 Input::exit_readline()
 {
 #if HAVE_LIBREADLINE
-   readline_lib::write_history(readline_history_path.c_str());
+   write_history(readline_history_path.c_str());
 #endif
 }
 //-----------------------------------------------------------------------------
 const unsigned char *
-Input::get_user_line(const UCS_string * prompt, bool from_nabla)
+Input::get_user_line(const UCS_string * prompt)
 {
    if (start_input)   (*start_input)();
 
-   if (!from_nabla)   Output::set_color_mode(Output::COLM_INPUT);
-
 const APL_time_us from = now();
 char * line;
    if (!use_readline)
@@ -278,11 +264,11 @@ char * line;
    else if (prompt)
       {
         UTF8_string prompt_utf(*prompt);
-        line = readline_lib::readline(prompt_utf.c_str());
+        line = call_readline(prompt_utf.c_str());
       }
    else
       {
-        line = readline_lib::readline(0);
+        line = call_readline(0);
       }
 
    if (end_input)   (*end_input)();
@@ -295,7 +281,7 @@ UTF8 * l = (UTF8 *)line;
       {
         while (*l && *l <= ' ')   ++l;   // skip leading whitespace
 
-        if (use_readline && *l)   readline_lib::add_history(line);
+        if (use_readline && *l)   add_history(line);
       }
 
    return l;
@@ -304,84 +290,21 @@ UTF8 * l = (UTF8 *)line;
 const char *
 Input::get_user_line_nabla(const UCS_string * prompt)
 {
-const UTF8 * line = TestFiles::get_testcase_line();
+   InputFile::increment_current_line_no();
+
+const UTF8 * line = IO_Files::get_file_line();
    if (line)   return (const char *)line;
-
-   if (input_file_FILE)   // a -f file is open
-      {
-        line = get_f_line(input_file_FILE);
-        if (line)   return (const char *)line;
-      }
-
-   if (input_file_name)   // -f present, but not opened yet
-      {
-        // we have a -f option with a filename to rerad from
-        // we open the file and read from it as long as possible.
-        // get_f_line() will set input_file_name and input_file_FILE to 0 when
-        // it reaches EOF.
-        //
-        // A special case is -f - where we open stdin
-        //
-        stdin_from_file = !strcmp(input_file_name, "-");
-        if (stdin_from_file)
-           {
-             input_file_FILE = stdin;
-           }
-        else
-           {
-             input_file_FILE = fopen(input_file_name, "r");
-           }
-        if (input_file_FILE)
-           {
-             line = get_f_line(input_file_FILE);
-             if (line)   return (const char *)line;
-           }
-        else
-           {
-             CERR << "cannot open file " << input_file_name
-                  << "(from -f) at " LOC << endl;
-             input_file_name = 0;
-           }
-      }
-
-   // all reads from files failed.
-   //
-   return (const char *)get_user_line(prompt, true);
-
-
-const APL_time_us from = now();
-   if (!use_readline)
-      {
-        line = (UTF8 *)no_readline(prompt);
-      }
-   else
-     {
-       if (prompt)
-          {
-            UTF8_string prompt_utf(*prompt);
-            loop(p, prompt_utf.size())
-            readline_lib::rl_stuff_char(prompt_utf[p] & 0x00FF);
-          }
-
-       line = (UTF8 *)readline_lib::readline(0);
-     }
-
-   while (*line && *line <= ' ')   ++line;
-
-const APL_time_us to = now();
-   Workspace::add_wait(to - from);
-
-   return (const char *)line;
+   return (const char *)get_user_line(prompt);
 }
 //-----------------------------------------------------------------------------
 UCS_string
 Input::get_quad_cr_line(UCS_string ucs_prompt)
 {
-   if (TestFiles::is_validating())
+   if (InputFile::is_validating())
       {
         Quad_QUOTE::done(true, LOC);
 
-        const UTF8 * line = TestFiles::get_testcase_line();
+        const UTF8 * line = IO_Files::get_file_line();
         if (line)
            {
              // for each leading backspace in line: remove last prompt char
@@ -409,13 +332,125 @@ const char * line;
    else
       {
         loop(s, utf_prompt.size())
-            readline_lib::rl_stuff_char(utf_prompt[s] & 0x00FF);
+            rl_stuff_char(utf_prompt[s] & 0x00FF);
 
         cout << '\r';
-        line = readline_lib::readline(0);
+        line = call_readline(0);
       }
 
+   if (line == 0)   return UCS_string();
+
 const UTF8_string uline(line);
    return UCS_string(uline);
 }
 //-----------------------------------------------------------------------------
+const UTF8 *
+Input::read_file_line()
+{
+static char buf[2000];
+int b = 0;
+   for (;b < sizeof(buf) - 1; ++b)
+       {
+         int cc = fgetc(InputFile::current_file()->file);
+         if (cc == EOF)   // end of file
+            {
+              if (b == 0)   return 0;
+              break;
+            }
+
+         if (cc == '\n' || cc == 2)   // end of line or ^B
+            {
+              break;
+            }
+
+         if (cc == '\r')   continue;   // ignore carrige returns
+
+          buf[b] = cc;
+       }
+
+   buf[b] = 0;
+
+   Log(LOG_test_execution)
+      CERR << "read_file_line() -> " << buf << endl;
+
+   return (UTF8 *)buf;
+}
+//-----------------------------------------------------------------------------
+char *
+Input::call_readline(const char * prompt)
+{
+#ifdef ADVANCED_readline
+
+   // deinstall ^C handler...
+   //
+static struct sigaction control_C_action;
+   memset(&control_C_action, 0, sizeof(struct sigaction));
+   control_C_action.sa_handler = SIG_IGN;
+   sigaction(SIGINT,  &control_C_action, 0);
+
+   set_intr_char(0);   // none
+char * ret = readline(prompt);
+   set_intr_char(3);   // ^C
+
+   control_C_action.sa_handler = &control_C;
+   sigaction(SIGINT,  &control_C_action, 0);
+
+#else
+
+char * ret = readline(prompt);
+
+#endif
+
+   return ret;
+}
+//-----------------------------------------------------------------------------
+void
+Input::set_intr_char(int ch)
+{
+struct termios tios;
+   tcgetattr(STDIN_FILENO, &tios);
+   tios.c_cc[VINTR] = ch;
+   tios.c_lflag &= ~ECHOCTL;
+   tcsetattr(STDIN_FILENO, TCSANOW, &tios);
+}
+
+#ifdef ADVANCED_readline
+
+// the functions below are not present in the readline lib shipped with OS-X
+
+//-----------------------------------------------------------------------------
+int
+Input::init_readline_control_C()
+{
+#if HAVE_LIBREADLINE
+// CERR << "init_readline_control_C()" << endl;
+
+   rl_bind_key(CTRL('C'), &readline_control_C);
+
+   // it seems to suffice if this this function is called once (i.e. after
+   // first time where readline() is being called
+   //
+   rl_startup_hook = 0;
+#endif
+
+   return 0;
+}
+//-----------------------------------------------------------------------------
+int
+Input::readline_control_C(int count, int key)
+{
+#if HAVE_LIBREADLINE
+   control_C(SIGINT);
+   rl_crlf();
+   rl_delete_text(0, rl_end);
+   rl_done = 1;
+
+   attention_raised = false;
+   interrupt_raised = false;
+   interrupt_when = 0;
+#endif
+
+   return 0;
+}
+//-----------------------------------------------------------------------------
+#endif
