@@ -20,7 +20,9 @@
 
 
 #include "Avec.hh"
+#include "Command.hh"
 #include "Input.hh"
+#include "InputFile.hh"
 #include "Logging.hh"
 #include "Nabla.hh"
 #include "Output.hh"
@@ -40,62 +42,88 @@ Nabla nabla(cmd);
 }
 //-----------------------------------------------------------------------------
 Nabla::Nabla(const UCS_string & cmd)
-   : fun_symbol(0),
+   : defn_line_no(InputFile::current_line_no()),
+     fun_symbol(0),
      ecmd(ECMD_NOP),
      edit_from(-1),
      edit_to(-1),
-     function_existed(true),
+     function_existed(false),
      do_close(false),
      locked(false),
-     first_command(cmd),
-     current_line(1)
+     current_line(1),
+     first_command(cmd)
 {
+   Workspace::more_error().clear();
 }
 //-----------------------------------------------------------------------------
 void
-Nabla::throw_edit_error(const char * loc)
+Nabla::throw_edit_error(const char * why)
 {
-   COUT << "DEFN ERROR" << endl
+   COUT << "DEFN ERROR+" << endl
         << "      " << first_command << endl
         << "      " << UCS_string(first_command.size() - 1, UNI_ASCII_SPACE)
         << "^" << endl;
 
-   Workspace::more_error() = UCS_string(loc);
+   if (Workspace::more_error().size() == 0)
+      {
+        Workspace::more_error() = UCS_string(why);
+      }
 
-   throw_define_error(fun_name, first_command, loc);
+   throw_define_error(fun_header, first_command, why);
 }
 //-----------------------------------------------------------------------------
 void
 Nabla::edit()
 {
-   if (const char * loc = start())
+const char * error = start();
+   if (error)
       {
-        CERR << "bad editor command '" << first_command
-             << "' : problem '" << loc << "'" << endl;
-        throw_edit_error(loc);
+        Log(LOG_verbose_error)   if (Workspace::more_error().size() == 0)
+           {
+             UERR << "Bad ∇-open '" << first_command
+                  << "' : '" << error << "'" << endl;
+           }
+        throw_edit_error(error);
       }
 
    // editor loop
    //
-   Log(LOG_nabla)   CERR << "Nabla(" << fun_name << ")..." << endl;
+int control_D_count = 0;
+   Log(LOG_nabla)   UERR << "Nabla(" << fun_header << ")..." << endl;
    while (!do_close)
        {
          const UCS_string prompt = current_line.print_prompt();
          const char * line = Input::get_user_line_nabla(&prompt);
+
+         if (line == 0)   // end-of-input (^D) pressed
+            {
+              ++control_D_count;
+              if (control_D_count < 5)
+                 {
+                    COUT << "^D" << endl;
+                    continue;
+                 }
+               COUT << endl << "      *** end of input" << endl;
+               Command::cmd_OFF(2);
+            }
+
          const UTF8_string utf(line);
-         const UCS_string ucs(utf);
+         UCS_string ucs(utf);
          if (const char * loc = parse_oper(ucs, false))
             {
-              CERR << "???" << endl;
+              UERR << "??? " << loc << endl;
               continue;
             }
 
          if (const char * loc = execute_oper())
-            CERR << "execute_oper() failed at " << loc << endl;;
+            {
+              UERR << "∇-command failed: " << loc << endl;
+              Output::set_color_mode(Output::COLM_INPUT);
+            }
        }
 
    Log(LOG_nabla)
-      CERR << "done '" << fun_name << "'" << endl;
+      UERR << "done '" << fun_header << "'" << endl;
 
 UCS_string fun_text;
    loop(l, lines.size())
@@ -105,8 +133,13 @@ UCS_string fun_text;
       }
 
 int error_line = 0;
+UCS_string creator(InputFile::current_filename());
+   creator.append(UNI_ASCII_COLON);
+   creator.append_number(defn_line_no);
+UTF8_string creator_utf8(creator);
+
 UserFunction * ufun = UserFunction::fix(fun_text, error_line, false,
-                                        LOC, "∇ editor");
+                                        LOC, creator_utf8);
    if (ufun == 0)   NABLA_ERROR;
 
    if (locked)
@@ -114,6 +147,21 @@ UserFunction * ufun = UserFunction::fix(fun_text, error_line, false,
          const int exec_properties[4] = { 1, 1, 1, 1 };
          ufun->set_exec_properties(exec_properties);
       }
+
+   // set stop and trace vectors
+   //
+DynArray(Function_Line, stop_vec,  lines.size());
+DynArray(Function_Line, trace_vec, lines.size());
+int stop_count = 0;
+int trace_count = 0;
+   loop(l, lines.size())
+       {
+         if (lines[l].stop_flag)    stop_vec [stop_count++]  = Function_Line(l);
+         if (lines[l].trace_flag)   trace_vec[trace_count++] = Function_Line(l);
+       }
+
+   ufun->set_trace_stop(stop_vec,  stop_count,  true);
+   ufun->set_trace_stop(trace_vec, trace_count, false);
 }
 //-----------------------------------------------------------------------------
 const char *
@@ -128,17 +176,22 @@ Nabla::start()
    //
 UCS_string::iterator c(first_command.begin());
 
+   // skip leading spaces
+   //
+   while (c.more() && c.get() <= UNI_ASCII_SPACE)   c.next();
+
    // skip leading nabla.
    //
-   if (c.next() != UNI_NABLA)   return LOC;
+   if (c.next() != UNI_NABLA)   return "Bad ∇-command (no ∇)";
 
    // skip leading spaces
    //
-   while (c.more() && c.get() == UNI_ASCII_SPACE)   c.next();
+   while (c.more() && c.get() <= UNI_ASCII_SPACE)   c.next();
 
-   // function name or header.
+   // function header.
    //
-   while (c.more() && c.get() != UNI_ASCII_L_BRACK)   fun_name.append(c.next());
+   while (c.more() && c.get() != UNI_ASCII_L_BRACK)
+         fun_header.append(c.next());
 
    // at this point there could be an axis specification [X] that
    // could be confused with an operation like [⎕]. We take the first char
@@ -151,15 +204,20 @@ UCS_string::iterator c(first_command.begin());
               if (c.get(off) <= UNI_ASCII_SPACE)   continue; 
               if (Avec::is_first_symbol_char(c.get(off)))   // axis
                  {
-                   fun_name.append(c.next());   //  copy the [
+                   fun_header.append(c.next());   //  copy the [
                    while (c.more() && c.get() != UNI_ASCII_L_BRACK)
-                         fun_name.append(c.next());
+                         fun_header.append(c.next());
                  }
               break;
             }
       }
 
-   if (fun_name.size() == 0)   return LOC;   // no function name
+UserFunction_header hdr(fun_header);
+   if (hdr.get_error())   return "Bad function header";
+   fun_symbol = hdr.FUN();
+   Assert(fun_symbol);
+
+   if (fun_header.size() == 0)   return "no function name";
 
    // optional operation
    //
@@ -167,39 +225,110 @@ UCS_string::iterator c(first_command.begin());
       {
         UCS_string oper;
         Unicode cc;
-        do oper.append(cc = c.next());   while (cc != UNI_ASCII_R_BRACK);
-        if (const char * loc = parse_oper(oper, true))   return loc;   // error
+        do 
+          {
+            if (!c.more())   return "no ] in ∇-command";
+            oper.append(cc = c.next());
+          } while (cc != UNI_ASCII_R_BRACK);
+
+        const char * loc = parse_oper(oper, true);
+        if (loc)   return loc;   // error
       }
 
-   if (const char * loc = open_function())   return loc;
+   switch(fun_symbol->get_nc())
+      {
+        case NC_UNUSED_USER_NAME:   // open a new function
+             function_existed = false;
+             {
+               // a new function must not have a command
+               //
+               if (ecmd != ECMD_NOP)   return "∇-command in new function";
+
+               const char * open_loc = open_new_function();
+               if (open_loc)   return open_loc;
+             }
+             break;
+
+        case NC_FUNCTION:
+        case NC_OPERATOR:   // open an existing function
+             function_existed = true;
+             if (InputFile::running_script())   // script
+                {
+                  const char * open_loc = open_new_function();
+                  if (open_loc)   return open_loc;
+                  break;   // continue below
+                }
+
+             // interactive
+             {
+               if (hdr.has_vars())
+                  {
+                       // an existing function was opened with a header that
+                       // contains more than the function name.
+                       //
+                       return "attempt to ∇-open existing function with "
+                              "new function header";
+                  }
+
+               const char * open_loc = open_existing_function();
+               if (open_loc)   return open_loc;
+             }
+             break;
+
+        default:
+             return "attempt to ∇-open a variable at " LOC;
+      }
+
+   // at this point the (new or existing) function was successfully opened.
+   // That means that at least the header is present (lines.size() > 0)
+
 
    // immediate close (only show command is allowed here),
    // e.g. ∇fun[⎕]∇
    //
    if (c.get() == UNI_NABLA)
       {
-        if (ecmd != ECMD_SHOW)   return LOC;
+        if (ecmd != ECMD_SHOW)   return "illegal command between ∇ ... ∇";
         if (const char * loc = execute_oper())
-           CERR << "execute_oper() failed at " << loc << endl;
+           UERR << "execute_oper() failed at " << loc << endl;
         do_close = true;
         return 0;
       }
 
    if (const char * loc = execute_oper())
-      CERR << "execute_oper() failed at " << loc << endl;
+      UERR << "execute_oper() failed at " << loc << endl;
 
    return 0;   // no error
 }
 //-----------------------------------------------------------------------------
 const char *
-Nabla::parse_oper(const UCS_string & oper, bool initial)
+Nabla::parse_oper(UCS_string & oper, bool initial)
 {
    Log(LOG_nabla)
-      CERR << "parsing oper '" << oper << "'" << endl;
+      UERR << "parsing oper '" << oper << "'" << endl;
+
+   // skip trailing spaces
+   //
+   while (oper.size() > 0 && oper.last() < ' ')   oper.pop();
+   if (oper.size() > 0 && oper.last() == UNI_NABLA)
+      {
+        do_close = true;
+        oper.pop();
+        while (oper.size() > 0 && oper.last() < ' ')   oper.pop();
+      }
+   else if (oper.size() > 0 && oper.last() == UNI_DEL_TILDE)
+      {
+        do_close = true;
+        locked = true;
+        oper.pop();
+        while (oper.size() > 0 && oper.last() < ' ')   oper.pop();
+      }
 
    current_text.clear();
    ecmd = ECMD_NOP;
 
+   if (oper.size() == 0 && do_close)   return 0;
+
 UCS_string::iterator c(oper.begin());
 Unicode cc = c.next();
    while (cc == ' ')   cc = c.next();
@@ -212,19 +341,6 @@ Unicode cc = c.next();
    // [n]                                           (goto)
    // text                                          (override text)
 
-   if (cc == UNI_NABLA)       // initial ∇
-      {
-        do_close = true;
-        return 0;
-      }
-
-   if (cc == UNI_DEL_TILDE)   // initial ⍫
-      {
-        locked = true;
-        do_close = true;
-        return 0;
-      }
-
    if (cc != UNI_ASCII_L_BRACK)
       {
         ecmd = ECMD_EDIT;
@@ -249,7 +365,8 @@ command_loop:
 
    // optional edit_from
    //
-   if (Avec::is_digit(c.get()))   edit_from = parse_lineno(c);
+   if (Avec::is_digit(c.get()) || c.get() == UNI_ASCII_FULLSTOP)
+      edit_from = parse_lineno(c);
 
    // operation, which is one of:
    //
@@ -264,17 +381,17 @@ command_loop:
         case UNI_ASCII_R_BRACK: ecmd = ECMD_EDIT;                 break;
         case UNI_DELTA:         ecmd = ECMD_DELETE;   c.next();   break;
         case UNI_RIGHT_ARROW:   ecmd = ECMD_ESCAPE;   c.next();   break;
-        case Invalid_Unicode:   return LOC;
+        case Invalid_Unicode:   return "Bad ∇-command";
 
-        default: CERR << "Bad edit op '" << c.get() << "'" << endl;
-                 return LOC;
+        default: UERR << "Bad edit op '" << c.get() << "'" << endl;
+                 return "Bad ∇-command";
       }
 
    // don't allow delete or escape in the first command
    if (initial)
       {
-        if (ecmd == ECMD_DELETE)   return LOC;
-        if (ecmd == ECMD_ESCAPE)   return LOC;
+        if (ecmd == ECMD_DELETE)   return "Bad initial ∇-command ∆";
+        if (ecmd == ECMD_ESCAPE)   return "Bad initial ∇-command →";
       }
 
 again:
@@ -284,14 +401,14 @@ again:
 
    if (c.get() == UNI_ASCII_MINUS)   // range
       {
-        if (got_minus)   return "error: second -  at " LOC;
+        if (got_minus)   return "error: second -  in ∇-range";
         got_minus = true;
         edit_from = edit_to;
         c.next();   // eat the -
         goto again;
       }
 
-   if (c.next() != UNI_ASCII_R_BRACK)   return LOC;
+   if (c.next() != UNI_ASCII_R_BRACK)   return "missing ] in ∇-range";
 
    // at this point we have parsed an editor command, like:
    //
@@ -299,7 +416,7 @@ again:
    // [from ∆ to]
    // [from]
 
-   while (c.get() == UNI_ASCII_SPACE)   c.next();
+   while (c.get() <= UNI_ASCII_SPACE)   c.next();
 
    if (c.get() == UNI_ASCII_L_BRACK)   // another command: ignore previous
       {
@@ -402,130 +519,108 @@ LineLabel ret(0);
 }
 //-----------------------------------------------------------------------------
 const char *
-Nabla::open_function()
-{
-   // open an existing function according to fun_name, or create a new function.
-
-   // there are two cases when a new function is created:
-   //
-   // 1. the header line contains at least one non-symbol char, or
-   // 2. the header consists of only symbol chars and a function with
-   //    that name exists.
-   //
-   // Examples:
-   //
-   // 1.   R←A ANY_FUNCTION B
-   // 2.   EXISTING_FUNCTION
-   //
-   loop(f, fun_name.size())
-      {
-        if (!Avec::is_symbol_char(fun_name[f]))
-           {
-             return open_new_function();
-           }
-      }
-
-   // all chars are symbol chars.
-   // They may or may not relate to an existing symbol.
-   //
-Symbol * fsym = Workspace::lookup_existing_symbol(fun_name);
-   if (fsym == 0)   return open_new_function();
-
-   // existing symbol: check that function is not on the SI stack.
-   //
-   if (Workspace::is_called(fun_name))
-      {
-        UCS_string & t4 = Workspace::more_error();
-        t4.clear();
-        t4.append_utf8("function ");
-        t4.append(fun_name);
-        t4.append_utf8(" could not be edited, since "
-                       "it is used on the )SI stack.");
-        return LOC;
-      }
-
-   if (fsym->get_nc() == NC_UNUSED_USER_NAME)   return open_new_function();
-
-   return open_existing_function(fsym);
-}
-//-----------------------------------------------------------------------------
-const char *
-Nabla::open_existing_function(Symbol * fsym)
+Nabla::open_new_function()
 {
    Log(LOG_nabla)
-      CERR << "opening existing function '" << fun_name << "'" << endl;
+      UERR << "creating new function '" << fun_symbol->get_name() 
+           << "' with header '" << fun_header << "'" << endl;
 
-const NameClass nc = fsym->get_nc();
-   if (nc != NC_FUNCTION && nc != NC_OPERATOR)
-      {
-        Log(LOG_nabla)
-           CERR << "symbol '" << fun_name << "' has name class " << nc
-                << " (not function or operator)" << endl;
-        return LOC;
-      }
-
-Function * function = fsym->get_function();
-   Assert(function);
-
-   if (function->get_exec_properties()[0])   NABLA_ERROR;   // locked
-
-const UCS_string ftxt = function->canonical(false);
-   Log(LOG_nabla)
-      CERR << "function is:\n" << ftxt << endl;
-
-vector<UCS_string> tlines;
-   ftxt.to_vector(tlines);
-
-   loop(t, tlines.size())   lines.push_back(FunLine(t, tlines[t]));
-
-   current_line = LineLabel(tlines.size());
+   lines.push_back(FunLine(0, fun_header));
    return 0;
 }
 //-----------------------------------------------------------------------------
 const char *
-Nabla::open_new_function()
+Nabla::open_existing_function()
 {
    Log(LOG_nabla)
-      CERR << "creating new function '" << fun_name << "'" << endl;
+      UERR << "opening existing function '" << fun_symbol->get_name()
+           << "'" << endl;
 
-int error_line = 0;
-UserFunction * ufun = UserFunction::fix(fun_name, error_line, true, LOC,
-                                        "∇ editor");
+   // this function must only be called when editing functions interactively
+   //
+   Assert(!InputFile::running_script());
 
-   if (ufun == 0)   return "Bad function header at " LOC;
-   fun_name = ufun->get_name();
+Function * function = fun_symbol->get_function();
+   Assert(function);
 
-   Log(LOG_nabla)
-      CERR << "created new function '" << fun_name << "'" << endl;
-   function_existed = false;
+   if (function->get_exec_properties()[0])
+      return "function is locked";
 
-   return open_function();
+   if (function->is_native())
+      return "function is native";
+
+   if (Workspace::is_called(fun_symbol->get_name()))
+      return "function is pendent or suspended";
+
+const UserFunction * ufun = function->get_ufun1();
+   if (ufun == 0)
+      return "function is not editable at " LOC;
+
+const UCS_string ftxt = function->canonical(false);
+   Log(LOG_nabla)   UERR << "existing function is:\n" << ftxt << endl;
+
+vector<UCS_string> tlines;
+   ftxt.to_vector(tlines);
+
+   Assert(tlines.size());
+   fun_header = tlines[0];
+   loop(t, tlines.size())
+       {
+         FunLine fl(t, tlines[t]);
+
+         // set stop and trace flags
+         //
+         loop(st, ufun->get_stop_lines().size())
+             {
+               if (t == ufun->get_stop_lines()[st])   // stop set
+                  {
+                    fl.stop_flag = true;
+                    break;
+                  }
+             }
+
+         loop(tr, ufun->get_trace_lines().size())
+             {
+               if (t == ufun->get_trace_lines()[tr])   // trace set
+                  {
+                    fl.trace_flag = true;
+                    break;
+                  }
+             }
+
+         lines.push_back(fl);
+       }
+
+   current_line = LineLabel(tlines.size());
+
+   return 0;
 }
 //-----------------------------------------------------------------------------
 const char *
 Nabla::execute_oper()
 {
-const bool have_from = edit_from.ln_major != -1;
-const bool have_to = edit_to.ln_major != -1;
-
-   if (!have_from && ecmd != ECMD_DELETE)   edit_from = lines[0].label;
-   if (!have_to)     edit_to = lines[lines.size() - 1].label;
-   if (edit_to.ln_major > lines.size() - 1)
-      { edit_to.ln_major = lines.size() - 1;   edit_to.ln_minor.clear(); };
-
    if (ecmd == ECMD_NOP)
       {
         Log(LOG_nabla)
-           CERR << "Nabla::execute_oper(NOP)" << endl;
+           UERR << "Nabla::execute_oper(NOP)" << endl;
         return 0;
       }
 
+const bool have_from = edit_from.ln_major != -1;
+const bool have_to = edit_to.ln_major != -1;
+
+   if (lines.size())
+      {
+        if (!have_from && ecmd != ECMD_DELETE)   edit_from = lines[0].label;
+        if (!have_to)     edit_to = lines[lines.size() - 1].label;
+      }
+
    if (ecmd == ECMD_SHOW)     return execute_show();
    if (ecmd == ECMD_DELETE)   return execute_delete();
    if (ecmd == ECMD_EDIT)     return execute_edit();
    if (ecmd == ECMD_ESCAPE)   return execute_escape();
 
-   CERR << "edit command " << ecmd
+   UERR << "edit command " << ecmd
         << " from " << edit_from << " to " << edit_to << endl;
    FIXME;
 
@@ -536,21 +631,48 @@ const char *
 Nabla::execute_show()
 {
    Log(LOG_nabla)
-      CERR << "Nabla::execute_oper(SHOW) from "
-           << edit_from << " to " << edit_to << " line-count " << lines.size() << endl;
+      UERR << "Nabla::execute_oper(SHOW) from " << edit_from
+           << " to " << edit_to << " line-count " << lines.size() << endl;
+
 int idx_from = find_line(edit_from);
 int idx_to   = find_line(edit_to);
 
+const LineLabel user_edit_to = edit_to;
+
    if (idx_from == -1)   edit_from = lines[idx_from = 0].label;
    if (idx_to == -1)     edit_to   = lines[idx_to = lines.size() - 1].label;
 
    Log(LOG_nabla)
-      CERR << "Nabla::execute_oper(SHOW) from "
+      UERR << "Nabla::execute_oper(SHOW) from "
            << edit_from << " to " << edit_to << endl;
 
+   if (idx_from == 0)   COUT << "    ∇" << endl;               // if header line
    for (int e = idx_from; e <= idx_to; ++e)   lines[e].print(COUT);
-   current_line = lines.back().label;
-   current_line.next();
+   if (idx_to == lines.size() - 1)   COUT << "    ∇" << endl;  // if last line
+
+   if (user_edit_to.valid())   // eg. [⎕42] or [2⎕42]
+      {
+        current_line = user_edit_to;
+        if (line_exists(current_line))   current_line.next();
+        if (line_exists(current_line))
+           {
+             // user_edit_to and its next line exist. Increase minor length
+             // That is, if both [8] and [9] exist then use [8.1]
+             //
+             current_line = user_edit_to;
+             current_line.insert();
+           }
+      }
+   else                      // eg. [42⎕] or [⎕]
+      {
+        current_line = lines.back().label;
+        current_line.next();
+      }
+
+   Log(LOG_nabla)
+      UERR << "Nabla::execute_oper(SHOW) done with current line "
+           << current_line << endl;
+
    return 0;
 }
 //-----------------------------------------------------------------------------
@@ -560,7 +682,7 @@ Nabla::execute_delete()
    // for delete we want exact numbers.
    //
 const int idx_to = find_line(LineLabel(edit_to));
-   if (idx_to == -1)   return "No/bad edit_to at " LOC;
+   if (idx_to == -1)   return "Bad line number N in [M∆N] ";
 
    if (edit_from == -1)   // delete single line
       {
@@ -571,8 +693,8 @@ const int idx_to = find_line(LineLabel(edit_to));
    // delete multiple lines
    //
 const int idx_from = find_line(LineLabel(edit_from));
-   if (idx_from == -1)       return "No/bad edit_from at " LOC;
-   if (idx_from >= idx_to)   return "idx_from >= idx_to at " LOC;
+   if (idx_from == -1)       return "Bad line number M in [M∆N] ";
+   if (idx_from >= idx_to)   return "M ≥ N in [M∆N] ";
 
    lines.erase(lines.begin() + idx_from, lines.begin() + idx_to);
    current_line = lines.back().label;
@@ -588,13 +710,77 @@ Nabla::execute_edit()
 
    current_line = edit_from;
 
+   // if the user has specified a line label without a text then we are done
+   //
+   if (current_text.size() == 0)   return 0;
+
+   // check that current_text is valid
+   //
+   if (current_line.is_header_line_number())   return edit_header_line();
+   else                                        return edit_body_line();
+}
+//-----------------------------------------------------------------------------
+const char *
+Nabla::edit_header_line()
+{
+   // parse the header and check that it is valid
+   //
+UserFunction_header header(current_text);
+   if (header.get_error() != E_NO_ERROR)
+      {
+        CERR << "BAD FUNCTION HEADER";
+        COUT << endl;
+        return 0;
+      }
+
+   // check if the function name has changed
+   //
+   Assert(fun_symbol);
+   Assert(header.FUN());
+const UCS_string & old_name = fun_symbol->get_name();
+const UCS_string & new_name = header.FUN()->get_name();
+   if (old_name != new_name)
+      {
+        // the name has changed. This is OK if the new name can be edited.
+        //
+        // the old symbol shall not be ⎕FXed when closing the editor, so we
+        // can simply forget it and continue with the new function
+        //
+        // UserFunction_header::UserFunction_header() should have triggered
+        // the creation of header.FUN() so we can assume tht it exists. We
+        // need to check that it is not a variable or an existing function.
+        //
+        if (header.FUN()->get_nc() != NC_UNUSED_USER_NAME)
+           {
+             CERR << "BAD FUNCTION HEADER";
+             COUT << endl;
+             return 0;
+           }
+      }
+
+   lines[0].text = current_text;
+   current_line.next();
+   return 0;
+}
+//-----------------------------------------------------------------------------
+const char *
+Nabla::edit_body_line()
+{
+const Parser parser(PM_FUNCTION, LOC);
+Token_string in;
+ErrorCode ec = parser.parse(current_text, in);
+   if (ec)
+      {
+        CERR << "SYNTAX ERROR";
+        COUT << endl;
+        return 0;
+      }
+
 const int idx_from = find_line(edit_from);
 
    Assert(lines.size() > 0);
    if (idx_from == -1)   // new line
       {
-        if (!current_text.size())   return 0;   // no text: done
-
         // find the largest label before edit_from (if any)
         //
         int before_idx = -1;
@@ -606,16 +792,14 @@ const int idx_from = find_line(edit_from);
 
         FunLine fl(edit_from, current_text);
         lines.insert(lines.begin() + before_idx + 1, fl);
-        current_line.next();
-        return 0;
+      }
+   else
+      {
+        // replace line
+        //
+        lines[idx_from].text = current_text;
       }
 
-
-   // replace line
-   //
-   if (!current_text.size())   return 0;   // no text: done
-
-   lines[idx_from].text = current_text;
    current_line.next();
    return 0;
 }
@@ -624,12 +808,12 @@ const char *
 Nabla::execute_escape()
 {
    lines.clear();
-   open_function();
+   lines.push_back(FunLine(0, fun_header));
    return 0;
 }
 //-----------------------------------------------------------------------------
 int
-Nabla::find_line(const LineLabel & lab)
+Nabla::find_line(const LineLabel & lab) const
 {
    if (lab.ln_major == -1)   return -1;
 
@@ -642,15 +826,27 @@ void
 Nabla::FunLine::print(ostream & out) const
 {
    label.print(out);
-   out << " " << text << endl;
+
+   // print a space unless text is a label or a comment
+   //
+   if (!text.is_comment_or_label())   out << " ";
+   out << text << endl;
 }
 //-----------------------------------------------------------------------------
 void
 LineLabel::print(ostream & out) const
 {
-   out << "[" << ln_major;
-   if (ln_minor.size())   out << "." << ln_minor;
-   out << "]";
+UCS_string ucs("[");
+   ucs.append_number(ln_major);
+   if (ln_minor.size())
+      {
+        ucs.append_utf8(".");
+        ucs.append(ln_minor);
+      }
+   ucs.append_utf8("]");
+
+   while (ucs.size() < 5)   ucs.append_utf8(" ");
+   out << ucs;
 }
 //-----------------------------------------------------------------------------
 UCS_string
@@ -686,6 +882,12 @@ const Unicode cc = ln_minor[ln_minor.size() - 1];
    else                     ln_minor.append(UNI_ASCII_1);
 }
 //-----------------------------------------------------------------------------
+void
+LineLabel::insert()
+{
+   ln_minor.append(UNI_ASCII_1);
+}
+//-----------------------------------------------------------------------------
 bool
 LineLabel::operator ==(const LineLabel & other) const
 {
