@@ -20,8 +20,12 @@
 
 
 #include <dlfcn.h>
+#include <errno.h>
+#include <string.h>
+#include <unistd.h>
 
 #include "Error.hh"
+#include "makefile.h"
 #include "NativeFunction.hh"
 #include "Symbol.hh"
 #include "Workspace.hh"
@@ -34,88 +38,39 @@ NativeFunction::NativeFunction(const UCS_string & so_name,
    : Function(ID_USER_SYMBOL, TOK_FUN2),
      handle(0),
      name(apl_name),
+     original_so_path(so_name),
      so_path(so_name),
      valid(false),
      close_fun(0)
 {
-   // open .so file...
-   //
-   {
-     UTF8_string lib_name(so_path);
-     handle = 0;
+UCS_string t4;
+   handle = open_so_file(t4, so_path);
+   if (handle == 0)
+      {
+         Workspace::more_error() = t4;
+         return;
+      }
 
-     if ((strchr(lib_name.c_str(), '/') == 0) &&
-         (strchr(lib_name.c_str(), '\\') == 0))
-        {
-          // the lib_name contains no path prefix. Try:
-          //
-          // PKGLIBDIR/lib_name
-          // PKGLIBDIR/lib_name.so
-          // PKGLIBDIR/lib_name.dylib
-          //
-          UTF8_string pkg_path = PKGLIBDIR;
-          pkg_path.append('/');
-          pkg_path.append(lib_name);
-          handle = dlopen(pkg_path.c_str(), RTLD_NOW);
-          if (handle == 0)   // no luck: try pkg_path.so
-             {
-               UTF8_string pkg_path__so(pkg_path);
-               pkg_path__so.append(UTF8_string(".so"));
-               handle = dlopen(pkg_path__so.c_str(), RTLD_NOW);
-             }
-          if (handle == 0)   // still no luck: try pkg_path.dylib
-             {
-               UTF8_string pkg_path__dylib(pkg_path);
-               pkg_path__dylib.append(UTF8_string(".dylib"));
-               handle = dlopen(pkg_path__dylib.c_str(), RTLD_NOW);
-             }
-        }
-
-     if (handle == 0)   // lib_name was not in PKGLIBDIR.
-        {
-          // Try:
-          //
-          // lib_name
-          // lib_name.so
-          // lib_name.dylib
-
-          handle = dlopen(lib_name.c_str(), RTLD_NOW);
-          if (handle == 0)   // no luck: try lib_name.so
-             {
-               UTF8_string lib_name__so(lib_name);
-               lib_name__so.append(UTF8_string(".so"));
-               handle = dlopen(lib_name__so.c_str(), RTLD_NOW);
-             }
-          if (handle == 0)   // still no luck: try lib_name.dylib
-             {
-               UTF8_string lib_name__dylib(lib_name);
-               lib_name__dylib.append(UTF8_string(".dylib"));
-               handle = dlopen(lib_name__dylib.c_str(), RTLD_NOW);
-             }
-        }
-
-     if (handle == 0)   // all dlopen() failed
-        {
-          Workspace::more_error() = UCS_string(dlerror());
-          return;
-        }
-   }
+   t4 = UCS_string("shared library ");
+   t4.append(so_name);
+   t4.append_utf8(" ");
 
    // get the function multiplexer
    //
 void * fmux = dlsym(handle, "get_function_mux");
      if (!fmux)
         {
-          CERR << "shared library is lacking the mandatory "
+          CERR << "shared library " << so_name << " is lacking the mandatory "
                   "function get_function_mux() !" << endl;
-          Workspace::more_error() = UCS_string(
-                                  "invalid .so file (no get_function_mux())");
+          t4.append_utf8("is invalid (no get_function_mux())");
+          Workspace::more_error() = t4;
           return;
         }
 
 void * (*get_function_mux)(const char *) = (void * (*)(const char *))fmux;
 
-   // get the mandatory function get_signature() which returns the function signature
+   // get the mandatory function get_signature() which returns
+   //  the function signature
    //
    {
      void * get_sig = get_function_mux("get_signature");
@@ -123,8 +78,8 @@ void * (*get_function_mux)(const char *) = (void * (*)(const char *))fmux;
         {
           CERR << "shared library is lacking the mandatory "
                   "function signature() !" << endl;
-          Workspace::more_error() = UCS_string(
-                                "invalid .so file (no get_signature())");
+          t4.append_utf8("is invalid (no get_signature())");
+          Workspace::more_error() = t4;
           return;
         }
 
@@ -136,7 +91,7 @@ void * (*get_function_mux)(const char *) = (void * (*)(const char *))fmux;
    // this function disappears
    {
      void * cfun = get_function_mux("close_fun");
-     if (cfun)   close_fun = (void (*)(Cause))cfun;
+     if (cfun)   close_fun = (bool (*)(Cause, const NativeFunction *))cfun;
      else        close_fun = 0;
    }
 
@@ -144,36 +99,40 @@ void * (*get_function_mux)(const char *) = (void * (*)(const char *))fmux;
    //
 Symbol * sym = Workspace::lookup_symbol(apl_name);
    Assert(sym);
-   if (!sym->can_be_defined())
+
+const char * why_not = sym->cant_be_defined();
+   if (why_not)
       {
-        Workspace::more_error() = UCS_string("Symbol is locked");
+        Workspace::more_error() = UCS_string(why_not);
         return;
       }
 
    /// read function pointers...
    //
+#define Th const NativeFunction * th
+
 #define ev(fun, args) f_ ## fun = (Token (*) args ) get_function_mux(#fun)
 
-   ev(eval_        , ()                              );
+   ev(eval_        , (                                Th));
 
-   ev(eval_B       , (Vr B)                          );
-   ev(eval_AB      , (Vr A, Vr B)                    );
-   ev(eval_XB      , (Vr X, Vr B)                    );
-   ev(eval_AXB     , (Vr A, Vr X, Vr B)              );
+   ev(eval_B       , (                          Vr B, Th));
+   ev(eval_AB      , (Vr A,                     Vr B, Th));
+   ev(eval_XB      , (                    Vr X, Vr B, Th));
+   ev(eval_AXB     , (Vr A,               Vr X, Vr B, Th));
 
-   ev(eval_LB      , (Fr LO, Vr B)                   );
-   ev(eval_ALB     , (Vr A, Fr LO, Vr B)             );
-   ev(eval_LXB     , (Fr LO, Vr X, Vr B)             );
-   ev(eval_ALXB    , (Vr A, Fr LO, Vr X, Vr B)       );
+   ev(eval_LB      , (      Fr LO,              Vr B, Th));
+   ev(eval_ALB     , (Vr A, Fr LO,              Vr B, Th));
+   ev(eval_LXB     , (      Fr LO,        Vr X, Vr B, Th));
+   ev(eval_ALXB    , (Vr A, Fr LO,        Vr X, Vr B, Th));
 
-   ev(eval_LRB     , (Fr LO, Fr RO, Vr B)            );
-   ev(eval_LRXB    , (Fr LO, Fr RO, Vr X, Vr B)      );
-   ev(eval_ALRB    , (Vr A, Fr LO, Fr RO, Vr B)      );
-   ev(eval_ALRXB   , (Vr A, Fr LO, Fr RO, Vr X, Vr B));
+   ev(eval_LRB     , (      Fr LO, Fr RO,       Vr B, Th));
+   ev(eval_LRXB    , (      Fr LO, Fr RO, Vr X, Vr B, Th));
+   ev(eval_ALRB    , (Vr A, Fr LO, Fr RO,       Vr B, Th));
+   ev(eval_ALRXB   , (Vr A, Fr LO, Fr RO, Vr X, Vr B, Th));
 
-   ev(eval_fill_B  , (Vr B)                          );
-   ev(eval_fill_AB , (Vr A, Vr B)                    );
-   ev(eval_ident_Bx, (Vr B, Axis x)                  );
+   ev(eval_fill_B  , (                          Vr B, Th));
+   ev(eval_fill_AB , (Vr A,                     Vr B, Th));
+   ev(eval_ident_Bx, (Vr B,             Axis x,       Th));
 #undef ev
 
    // compute function tag based on the signature
@@ -186,13 +145,15 @@ Symbol * sym = Workspace::lookup_symbol(apl_name);
    if (is_operator())   sym->set_nc(NC_OPERATOR, this);
    else                 sym->set_nc(NC_FUNCTION, this);
 
+   Workspace::more_error().clear();
    valid = true;
    valid_functions.push_back(this);
 }
 //-----------------------------------------------------------------------------
 NativeFunction::~NativeFunction()
 {
-   if (handle)   dlclose(handle);
+  Log(LOG_UserFunction__enter_leave)
+      CERR << "Native function " << get_name() << " deleted." << endl;
 
    loop(v, valid_functions.size())
       {
@@ -203,13 +164,163 @@ NativeFunction::~NativeFunction()
       }
 }
 //-----------------------------------------------------------------------------
+void *
+NativeFunction::try_one_file(const char * filename, UCS_string & t4)
+{
+const int t4_len = t4.size();
+   t4.append_utf8("    file ");
+   t4.append_utf8(filename);
+
+   if (access(filename, R_OK) != 0)
+      {
+        while (t4.size() < t4_len + 44)     t4.append_utf8(" ");
+        t4.append_utf8(" (");
+        t4.append_utf8(strerror(errno));
+        t4.append_utf8(")\n");
+        return 0;
+      }
+
+void * handle = dlopen(filename, RTLD_NOW);
+   if (handle == 0)
+      {
+        const char * err = dlerror();
+        if (strrchr(err, ':'))   err = 1 + strrchr(err, ':');
+
+        while (t4.size() < t4_len + 44)     t4.append_utf8(" ");
+        t4.append_utf8(" (");
+        t4.append_utf8(err);
+        t4.append_utf8(" )\n");
+      }
+
+   return handle;
+}
+//-----------------------------------------------------------------------------
+void *
+NativeFunction::open_so_file(UCS_string & t4, UCS_string & so_path)
+{
+   // prepare a )MORE error message containing the file names tried.
+   //
+   t4.clear();
+   t4.append_utf8("Could not find shared library '");
+   t4.append(so_path);
+   t4.append_utf8("'\n"
+                  "The following directories and file names were tried:\n");
+
+   // if the name starts with / (or \ on Windows) or .
+   // then take it as is without changes.
+   //
+   if (so_path[0] == UNI_ASCII_SLASH     ||
+       so_path[0] == UNI_ASCII_BACKSLASH ||
+       so_path[0] == UNI_ASCII_FULLSTOP)
+      {
+        UTF8_string filename(so_path);
+        void * handle = try_one_file(filename.c_str(), t4);
+
+        if (handle == 0)
+           {
+             t4.append_utf8(
+            "NOTE: Filename extensions are NOT automatically added "
+            "when a full path\n"
+            "      (i.e. a path starting with / or .) is used.");
+           }
+        return handle;
+      }
+
+   // otherwise try Makefile__pkglibdir . /usr/lib/apl and /usr/local/lib/apl,
+   // avoiding duplicates
+   //
+UTF8_string utf_so_path(so_path);
+const char * dirs[] =
+{
+  Makefile__pkglibdir,    // the normal case
+  "/usr/lib/apl",
+  "/usr/local/lib/apl",
+  ".",
+  "./native",             // if make install was not performed
+  "./emacs_mode",         // if make install was not performed
+};
+
+   // most likely Makefile__pkglibdir is /usr/lib/apl or /usr/local/lib/apl.
+   // don't try them twice.
+   //
+   if (!strcmp(Makefile__pkglibdir, dirs[1]))   dirs[1] = 0;
+   if (!strcmp(Makefile__pkglibdir, dirs[2]))   dirs[2] = 0;
+
+   loop(d, sizeof(dirs) / sizeof(*dirs))
+       {
+         if (dirs[d] == 0)   continue;
+
+         UTF8_string dir_so_path(dirs[d]);
+         dir_so_path.append(UTF8_string("/"));
+         dir_so_path.append(utf_so_path);
+
+         UTF8_string dir_only(dir_so_path);
+         dir_only[strrchr(dir_only.c_str(), '/') - dir_only.c_str()] = 0;
+         if (access(dir_only.c_str(), R_OK | X_OK))
+            {
+              t4.append_utf8("    directory ");
+              t4.append_utf8(dir_only.c_str());
+              t4.append_utf8("\n");
+              continue;   // new directory
+            }
+
+         // try filename, then filename.so, and then filename.dylib
+         // unless the filename has an extension already.
+         //
+         const char * exts[] = { ".so", ".dylib", "" };
+
+         bool has_extension = false;
+         loop(e, sizeof(exts)/sizeof(*exts))
+            {
+              if (*exts[e])   // not "no extension"
+                 {
+                   const char * ext = exts[e];
+                   const char * end = dir_so_path.c_str();
+                   end += strlen(end) - strlen(ext);
+                   if (!strcmp(exts[e],  end))
+                      {
+                        has_extension = true;
+                        break;
+                      }
+                 }
+            }
+
+         loop(e, sizeof(exts)/sizeof(*exts))
+             {
+               if (has_extension && *exts[e])   continue;
+
+               UTF8_string filename(dir_so_path);
+               if (exts[e])   filename.append(UTF8_string(exts[e]));
+
+               void * handle = try_one_file(filename.c_str(), t4);
+               if (handle)   // found library
+                  {
+                    so_path = UCS_string(filename);   // update so_path
+                    return handle;
+                  }
+             }
+       }
+
+   return 0;
+}
+//-----------------------------------------------------------------------------
 void
 NativeFunction::cleanup()
 {
+   // delete in reverse construction order
+   //
    loop(v, valid_functions.size())
       {
-        void (*close_fun)(Cause cause) = valid_functions[v]->close_fun;
-        if (close_fun)   (*close_fun)(CAUSE_SHUTDOWN);
+        NativeFunction & fun = *valid_functions.back();
+        if (fun.close_fun && fun.handle)
+           {
+             const bool do_dlclose = (*fun.close_fun)(CAUSE_SHUTDOWN, &fun);
+             if (do_dlclose)
+                {
+                  dlclose(fun.handle);
+                  fun.handle = 0;
+                }
+           }
         delete valid_functions[v];
       }
 }
@@ -217,7 +328,16 @@ NativeFunction::cleanup()
 void
 NativeFunction::destroy()
 {
-   if (close_fun)   (*close_fun)(CAUSE_ERASED);
+   if (close_fun)   (*close_fun)(CAUSE_ERASED, this);
+   if (close_fun && handle)
+      {
+        const bool do_dlclose = (*close_fun)(CAUSE_ERASED, this);
+        if (do_dlclose)
+           {
+             dlclose(handle);
+             handle = 0;
+           }
+      }
    delete this;
 }
 //-----------------------------------------------------------------------------
@@ -229,7 +349,26 @@ NativeFunction::fix(const UCS_string & so_name,
    //
    loop(v, valid_functions.size())
       {
-        if (so_name == valid_functions[v]->so_path)   return valid_functions[v];
+        NativeFunction * fun = valid_functions[v];
+        if (so_name == fun->so_path)
+           {
+             // the NativeFunction object exists, but may have been
+             // )ERASEd at APL level. If so, then re-install it.
+             //
+             Symbol * sym = Workspace::lookup_symbol(fun->get_name());
+             Assert(sym);
+
+             const char * why_not = sym->cant_be_defined();
+             if (why_not)
+                {
+                  Workspace::more_error() = UCS_string(why_not);
+                  return 0;
+                }
+
+             if (fun->is_operator())   sym->set_nc(NC_OPERATOR, fun);
+             else                      sym->set_nc(NC_FUNCTION, fun);
+             return fun;
+           }
       }
 
 NativeFunction * new_function = new NativeFunction(so_name, function_name);
@@ -275,7 +414,7 @@ UCS_string ind(indent, UNI_ASCII_SPACE);
 UCS_string
 NativeFunction::canonical(bool with_lines) const
 {
-   return so_path;
+   return original_so_path;
 }
 //-----------------------------------------------------------------------------
 ostream &
@@ -287,10 +426,45 @@ NativeFunction::print(std::ostream & out) const
    return out;
 }
 //-----------------------------------------------------------------------------
+UCS_string
+NativeFunction::load_emacs_library(const char * emacs_arg)
+{
+UCS_string so_path("libemacs");
+UCS_string t4;
+
+void * handle = open_so_file(t4, so_path);
+   if (handle == 0)   return t4;
+
+   t4 = UCS_string("found emacs library ");
+   t4.append(so_path);
+
+void * emacs_start = dlsym(handle, "emacs_start");
+   if (emacs_start == 0)
+      {
+        t4.append_utf8(", but it\n   it is lacking the mandatory "
+                       "function emacs_start()\n");
+        return t4;
+      }
+
+UTF8_string so_path_utf(so_path);
+const int error = ((int (*)(const char *, const char *))emacs_start)
+                  (emacs_arg, (const char *)so_path_utf.c_str());
+   if (error)
+      {
+        t4.append_utf8(", but emacs_start()  returned error ");
+        t4.append_number(error);
+        t4.append_utf8("\n");
+        return t4;
+      }
+
+   t4.clear();   // success
+   return t4;
+}
+//-----------------------------------------------------------------------------
 Token
 NativeFunction::eval_()
 {
-   if (f_eval_)   return (*f_eval_)();
+   if (f_eval_)   return (*f_eval_)(this);
 
    SYNTAX_ERROR;
 }
@@ -298,7 +472,7 @@ NativeFunction::eval_()
 Token
 NativeFunction::eval_B(Value_P B)
 {
-   if (f_eval_B)   return (*f_eval_B)(B);
+   if (f_eval_B)   return (*f_eval_B)(B, this);
 
    SYNTAX_ERROR;
 }
@@ -306,7 +480,7 @@ NativeFunction::eval_B(Value_P B)
 Token
 NativeFunction::eval_AB(Value_P A, Value_P B)
 {
-   if (f_eval_AB)   return (*f_eval_AB)(A, B);
+   if (f_eval_AB)   return (*f_eval_AB)(A, B, this);
 
    SYNTAX_ERROR;
 }
@@ -314,7 +488,7 @@ NativeFunction::eval_AB(Value_P A, Value_P B)
 Token
 NativeFunction::eval_LB(Token & LO, Value_P B)
 {
-   if (f_eval_LB)   return (*f_eval_LB)(*LO.get_function(), B);
+   if (f_eval_LB)   return (*f_eval_LB)(*LO.get_function(), B, this);
 
    SYNTAX_ERROR;
 }
@@ -322,7 +496,7 @@ NativeFunction::eval_LB(Token & LO, Value_P B)
 Token
 NativeFunction::eval_ALB(Value_P A, Token & LO, Value_P B)
 {
-   if (f_eval_ALB)   return (*f_eval_ALB)(A, *LO.get_function(), B);
+   if (f_eval_ALB)   return (*f_eval_ALB)(A, *LO.get_function(), B, this);
 
    SYNTAX_ERROR;
 }
@@ -331,7 +505,7 @@ Token
 NativeFunction::eval_LRB(Token & LO, Token & RO, Value_P B)
 {
    if (f_eval_LRB)   return (*f_eval_LRB)(*LO.get_function(),
-                                          *RO.get_function(), B);
+                                          *RO.get_function(), B, this);
 
    SYNTAX_ERROR;
 }
@@ -340,7 +514,7 @@ Token
 NativeFunction::eval_ALRB(Value_P A, Token & LO, Token & RO, Value_P B)
 {
    if (f_eval_ALRB)   return (*f_eval_ALRB)(A, *LO.get_function(),
-                                               *RO.get_function(), B);
+                                               *RO.get_function(), B, this);
 
    SYNTAX_ERROR;
 }
@@ -350,7 +524,7 @@ NativeFunction::eval_XB(Value_P X, Value_P B)
 {
    // call axis variant if present, or else the non-axis variant.
    //
-   if (f_eval_XB)   return (*f_eval_XB)(X, B);
+   if (f_eval_XB)   return (*f_eval_XB)(X, B, this);
    else             return eval_B(B);
 }
 //-----------------------------------------------------------------------------
@@ -359,7 +533,7 @@ NativeFunction::eval_AXB(Value_P A, Value_P X, Value_P B)
 {
    // call axis variant if present, or else the non-axis variant.
    //
-   if (f_eval_AXB)   return (*f_eval_AXB)(A, X, B);
+   if (f_eval_AXB)   return (*f_eval_AXB)(A, X, B, this);
    else              return eval_AB(A, B);
 }
 //-----------------------------------------------------------------------------
@@ -368,7 +542,7 @@ NativeFunction::eval_LXB(Token & LO, Value_P X, Value_P B)
 {
    // call axis variant if present, or else the non-axis variant.
    //
-   if (f_eval_LXB)   return (*f_eval_LXB)(*LO.get_function(), X, B);
+   if (f_eval_LXB)   return (*f_eval_LXB)(*LO.get_function(), X, B, this);
    else              return eval_LB(LO, B);
 }
 //-----------------------------------------------------------------------------
@@ -377,7 +551,7 @@ NativeFunction::eval_ALXB(Value_P A, Token & LO, Value_P X, Value_P B)
 {
    // call axis variant if present, or else the non-axis variant.
    //
-   if (f_eval_ALXB)   return (*f_eval_ALXB)(A, *LO.get_function(), X, B);
+   if (f_eval_ALXB)   return (*f_eval_ALXB)(A, *LO.get_function(), X, B, this);
    else               return eval_ALB(A, LO, B);
 }
 //-----------------------------------------------------------------------------
@@ -387,7 +561,7 @@ NativeFunction::eval_LRXB(Token & LO, Token & RO, Value_P X, Value_P B)
    // call axis variant if present, or else the non-axis variant.
    //
    if (f_eval_LRXB)   return (*f_eval_LRXB)(*LO.get_function(),
-                                            *RO.get_function(), X, B);
+                                            *RO.get_function(), X, B, this);
    else                return eval_LRB(LO, RO, B);
 }
 //-----------------------------------------------------------------------------
@@ -397,14 +571,14 @@ NativeFunction::eval_ALRXB(Value_P A, Token & LO, Token & RO, Value_P X, Value_P
    // call axis variant if present, or else the non-axis variant.
    //
    if (f_eval_ALRXB)   return (*f_eval_ALRXB)(A, *LO.get_function(),
-                                                 *RO.get_function(), X, B);
+                                                 *RO.get_function(), X, B, this);
    else                return eval_ALRB(A, LO, RO, B);
 }
 //-----------------------------------------------------------------------------
 Token
 NativeFunction::eval_fill_B(Value_P B)
 {
-   if (f_eval_fill_B)   return (*f_eval_fill_B)(B);
+   if (f_eval_fill_B)   return (*f_eval_fill_B)(B, this);
 
    SYNTAX_ERROR;
 }
@@ -412,7 +586,7 @@ NativeFunction::eval_fill_B(Value_P B)
 Token
 NativeFunction::eval_fill_AB(Value_P A, Value_P B)
 {
-   if (f_eval_fill_AB)   return (*f_eval_fill_AB)(A, B);
+   if (f_eval_fill_AB)   return (*f_eval_fill_AB)(A, B, this);
 
    SYNTAX_ERROR;
 }
@@ -420,7 +594,7 @@ NativeFunction::eval_fill_AB(Value_P A, Value_P B)
 Token
 NativeFunction::eval_identity_fun(Value_P B, Axis axis)
 {
-   if (f_eval_ident_Bx)   return (*f_eval_ident_Bx)(B, axis);
+   if (f_eval_ident_Bx)   return (*f_eval_ident_Bx)(B, axis, this);
 
    SYNTAX_ERROR;
 }
