@@ -23,25 +23,28 @@
 #include <math.h>
 
 #include "Workspace.hh"
-#include "Value.hh"
+#include "Value.icc"
 #include "Error.hh"
 #include "ComplexCell.hh"
 #include "FloatCell.hh"
 #include "IntCell.hh"
 
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 NumericCell::bif_not(Cell * Z) const
 {
+const APL_Float qct = Workspace::get_CT();
+
+   if (!is_near_bool(qct))   return E_DOMAIN_ERROR;
+
    if (get_near_bool(Workspace::get_CT()))   new (Z) IntCell(0);
    else                                      new (Z) IntCell(1);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
 void
 NumericCell::do_binomial(Cell * Z, APL_Integer a, APL_Integer b, bool negate)
 {
-double nom_f = 1.0;
-APL_Integer nom_i = 1;
 APL_Integer den_max = a;
 APL_Integer den_min = b - a;
 
@@ -54,6 +57,44 @@ APL_Integer den_min = b - a;
         den_min = a;
       }
 
+   /* special cases: a = 0, 1, 2, b-2, b-1, b
+     
+      / b \         / b \         / b \    .
+      |   | = 1,    |   | = b,    |   | = b*(b-1)/2
+     
+      \ 0 /         \ 1 /         \ 2 /
+    */
+   if (den_min <= 2)
+      {
+        if (den_min == 0)
+           {
+             if (negate)   new (Z) IntCell(-1);
+             else          new (Z) IntCell(1);
+           }
+        else if (den_min == 1)
+           {
+             if (negate)   new (Z) IntCell(-b);
+             else          new (Z) IntCell(b);
+           }
+        else    // a == 2 or b-2
+           {
+             // either b or b-1 must be even
+             if (b & 1)   // b-1 even
+                {
+                  if (negate)   new (Z) IntCell( - b * ((b-1)/2));
+                  else          new (Z) IntCell(   b * ((b-1)/2));
+                }
+             else         // b even
+                {
+                  if (negate)   new (Z) IntCell(- (b-1) * (b/2));
+                  else          new (Z) IntCell(  (b-1) * (b/2));
+                }
+           }
+        return;
+      }
+
+double nom_f = 1.0;
+APL_Integer nom_i = 1;
    // compute b! / den_max! == den_max+1 * den_max+2 * ... * b
    //
    for (APL_Integer n = b; n > den_max; --n)
@@ -62,25 +103,25 @@ APL_Integer den_min = b - a;
          nom_f *= n;
        }
 
-   if (nom_f > BIG_INT64_F)   // integer overflow
+   if (nom_f > SMALL_INT)   // integer overflow
       {
         for (APL_Integer n = den_min; n > 1; --n)
             nom_f /= n;
 
-        if (negate)   new(Z)   FloatCell(-nom_f);
-        else          new(Z)   FloatCell( nom_f);
+        if (negate)   new(Z)   FloatCell(- nom_f);
+        else          new(Z)   FloatCell(  nom_f);
       }
    else
       {
         for (APL_Integer n = den_min; n > 1; --n)
             nom_i /= n;
 
-        if (negate)   new(Z)   IntCell(-nom_i);
-        else          new(Z)   IntCell( nom_i);
+        if (negate)   new(Z)   IntCell(- nom_i);
+        else          new(Z)   IntCell(  nom_i);
       }
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 NumericCell::bif_binomial(Cell * Z, const Cell * A) const
 {
 const APL_Float qct = Workspace::get_CT();
@@ -99,7 +140,7 @@ const APL_Float qct = Workspace::get_CT();
         const APL_Complex gam_1_b__a = ComplexCell::gamma(r_1_b__a, i_b_a);
 
         new (Z) ComplexCell(gam_1_b / (gam_1_a * gam_1_b__a));
-        return;
+        return E_NO_ERROR;
       }
 
    if (!is_near_int(qct) || !A->is_near_int(qct))   // non-integer result
@@ -108,12 +149,12 @@ const APL_Float qct = Workspace::get_CT();
         const APL_Float r_1_b    = 1.0 + get_real_value();
         const APL_Float r_1_b__a = r_1_b - A->get_real_value();
 
-        if (r_1_a < 0    && is_near_int(r_1_a, qct))      DOMAIN_ERROR;
-        if (r_1_b < 0    && is_near_int(r_1_b, qct))      DOMAIN_ERROR;
-        if (r_1_b__a < 0 && is_near_int(r_1_b__a, qct))   DOMAIN_ERROR;
+        if (r_1_a < 0    && is_near_int(r_1_a, qct))      return E_DOMAIN_ERROR;
+        if (r_1_b < 0    && is_near_int(r_1_b, qct))      return E_DOMAIN_ERROR;
+        if (r_1_b__a < 0 && is_near_int(r_1_b__a, qct))   return E_DOMAIN_ERROR;
 
         new (Z) FloatCell(  tgamma(r_1_b) / (tgamma(r_1_a) * tgamma(r_1_b__a)));
-        return;
+        return E_NO_ERROR;
       }
 
 const APL_Integer a = A->get_near_int(qct);
@@ -126,52 +167,54 @@ int how = 0;
 
    switch(how)
       {
-        case 0: do_binomial(Z, a, b, false);                       return;
-        case 1: new (Z) IntCell(0);                                return;
-        case 2: Assert(0 && "Impossible case 2");                  return;
-        case 3: do_binomial(Z, a, a - (b + 1), a & 1);               return;
-        case 4: new (Z) IntCell(0);                                return;
-        case 5: Assert(0 && "Impossible case 5");                  return;
-        case 6: do_binomial(Z, -(b + 1), -(a + 1), (b - a) & 1);   return;
-        case 7: new (Z) IntCell(0);                                return;
+        case 0: do_binomial(Z, a, b, false);                       break;
+        case 1: new (Z) IntCell(0);                                break;
+        case 2: Assert(0 && "Impossible case 2");                  break;
+        case 3: do_binomial(Z, a, a - (b + 1), a & 1);             break;
+        case 4: new (Z) IntCell(0);                                break;
+        case 5: Assert(0 && "Impossible case 5");                  break;
+        case 6: do_binomial(Z, -(b + 1), -(a + 1), (b - a) & 1);   break;
+        case 7: new (Z) IntCell(0);                                break;
       }
+
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void NumericCell::bif_and(Cell * Z, const Cell * A) const
+ErrorCode
+NumericCell::bif_and(Cell * Z, const Cell * A) const
 {
 const APL_Float qct = Workspace::get_CT();
 
    // if either value is 0 then return 0
    //
-   if (A->is_near_zero(qct))
+   if (A->is_near_zero(qct) || is_near_zero(qct))
       {
         new (Z) IntCell(0);
-        return;
+        return E_NO_ERROR;
       }
 
-   if (is_near_zero(qct))
+   // if both args are 1 then return the classical A ∧ B
+   //
+   if (A->is_near_one(qct) && is_near_one(qct))
       {
-        new (Z) IntCell(0);
-        return;
+        new (Z) IntCell(1);
+        return E_NO_ERROR;
       }
 
-   if (!A->is_near_real(qct) || !is_near_real(qct))   // complex
+   // handle the complex case before the int case because
+   // is_near_int() can be true for complex numbers.
+   //
+   if (A->is_complex_cell() || is_complex_cell())   // complex
       {
         // a or b is complex; we assume (require) Gaussian integers.
         //
-        const APL_Complex gcd = cpx_gcd(A->get_complex_value(),
+        APL_Complex gcd;
+        const ErrorCode err = cpx_gcd(gcd, A->get_complex_value(),
                                            get_complex_value(), qct);
+        if (err)   return err;
 
         new (Z) ComplexCell(A->get_complex_value() * (get_complex_value()/gcd));
-        return;
-      }
-
-   // if both args are boolean then return the classical A ∧ B
-   //
-   if (A->is_near_bool(qct) && is_near_bool(qct))
-      {
-        new (Z) IntCell(1);
-        return;
+        return E_NO_ERROR;
       }
 
    // if both args are int then return the least common multiple of them
@@ -180,9 +223,11 @@ const APL_Float qct = Workspace::get_CT();
       {
         const APL_Integer a = A->get_near_int(qct);
         const APL_Integer b =    get_near_int(qct);
-        const APL_Integer gcd = int_gcd(a, b);
+        APL_Integer gcd;
+        const ErrorCode err = int_gcd(gcd, a, b);
+        if (err)   return err;
         new (Z) IntCell(a * (b / gcd));
-        return;
+        return E_NO_ERROR;
       }
 
    // if both args are real then return the (real) least common multiple of them
@@ -191,30 +236,35 @@ const APL_Float qct = Workspace::get_CT();
       {
         const APL_Float a = A->get_real_value();
         const APL_Float b =    get_real_value();
-        const APL_Float gcd = flt_gcd(a, b, qct);
+        APL_Float gcd;
+        const ErrorCode err = flt_gcd(gcd, a, b, qct);
+        if (err)   return err;
         new (Z) FloatCell(a * (b / gcd));
-        return;
+        return E_NO_ERROR;
       }
 
-   DOMAIN_ERROR;   // char ?
+   return E_DOMAIN_ERROR;   // char ?
 }
 //-----------------------------------------------------------------------------
-void NumericCell::bif_nand(Cell * Z, const Cell * A) const
+ErrorCode
+NumericCell::bif_nand(Cell * Z, const Cell * A) const
 {
 const APL_Float qct = Workspace::get_CT();
    if ( A->get_near_bool(qct) && get_near_bool(qct))   new (Z) IntCell(0);
    else                                                new (Z) IntCell(1);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 NumericCell::bif_nor(Cell * Z, const Cell * A) const
 {
 const APL_Float qct = Workspace::get_CT();
    if ( A->get_near_bool(qct) || get_near_bool(qct))   new (Z) IntCell(0);
    else                                                new (Z) IntCell(1);
+   return E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-void
+ErrorCode
 NumericCell::bif_or(Cell * Z, const Cell * A) const
 {
 const APL_Float qct = Workspace::get_CT();
@@ -225,6 +275,7 @@ const APL_Float qct = Workspace::get_CT();
       {
         if ( A->get_near_bool(qct) || get_near_bool(qct))   new (Z) IntCell(1);
         else                                                new (Z) IntCell(0);
+        return E_NO_ERROR;
       }
 
    if (!A->is_near_real(qct) || !is_near_real(qct))   // complex
@@ -232,22 +283,24 @@ const APL_Float qct = Workspace::get_CT();
         if (A->is_near_zero(qct))
            {
              new (Z) ComplexCell(get_complex_value());
-             return;
+             return E_NO_ERROR;
            }
 
         if (is_near_zero(qct))
            {
              new (Z) ComplexCell(A->get_complex_value());
-             return;
+             return E_NO_ERROR;
            }
 
         // a or b is complex; we assume (require) Gaussian integers.
         //
-        const APL_Complex gcd = cpx_gcd(A->get_complex_value(),
+        APL_Complex gcd;
+        const ErrorCode err = cpx_gcd(gcd, A->get_complex_value(),
                                            get_complex_value(), qct);
+        if (err)   return err;
 
         new (Z) ComplexCell(gcd);
-        return;
+        return E_NO_ERROR;
       }
 
    // if both args are int then return the greatest common divisor of them
@@ -256,9 +309,11 @@ const APL_Float qct = Workspace::get_CT();
       {
         const APL_Integer a = A->get_near_int(qct);
         const APL_Integer b =    get_near_int(qct);
-        const APL_Integer gcd = int_gcd(a, b);
+        APL_Integer gcd;
+        const ErrorCode err = int_gcd(gcd, a, b);
+        if (err)   return err;
         new (Z) IntCell(gcd);
-        return;
+        return E_NO_ERROR;
       }
 
    // if both args are real then return the (real) greatest common divisor
@@ -267,94 +322,14 @@ const APL_Float qct = Workspace::get_CT();
       {
         const APL_Float a = A->get_real_value();
         const APL_Float b =    get_real_value();
-        const APL_Float gcd = flt_gcd(a, b, qct);
+        APL_Float gcd;
+        const ErrorCode err = flt_gcd(gcd, a, b, qct);
+        if (err)   return err;
         new (Z) FloatCell(gcd);
-        return;
+        return E_NO_ERROR;
       }
 
-   DOMAIN_ERROR;   // char ?
-}
-//-----------------------------------------------------------------------------
-APL_Integer
-NumericCell::int_gcd(APL_Integer a, APL_Integer b)
-{
-   if (a == 0x8000000000000000ULL)   DOMAIN_ERROR;
-   if (b == 0x8000000000000000ULL)   DOMAIN_ERROR;
-
-   if (a < 0)   a = - a;
-   if (b < 0)   b = - b;
-   if (b < a)
-      {
-         const APL_Integer _b = b;
-         b = a;
-         a = _b;
-      }
-
-   // at this point 0 ≤ a ≤ b
-   //
-   for (;;)
-       {
-         if (a == 0)   return b;
-         const APL_Integer r = b%a;
-         b = a;
-         a = r;
-       }
-}
-//-----------------------------------------------------------------------------
-APL_Float
-NumericCell::flt_gcd(APL_Float a, APL_Float b, APL_Float qct)
-{
-   if (a < 0)   a = - a;
-   if (b < 0)   b = - b;
-   if (b < a)
-      {
-         const APL_Float _b = b;
-         b = a;
-         a = _b;
-      }
-
-   // at this point 0 ≤ a ≤ b
-   //
-   for (;;)
-       {
-         if (is_near_zero(a, qct))   return b;
-         const APL_Float r = fmod(b, a);
-         b = a;
-         a = r;
-       }
-}
-//-----------------------------------------------------------------------------
-APL_Complex
-NumericCell::cpx_gcd(APL_Complex a, APL_Complex b, APL_Float qct)
-{
-   if (!is_near_int(a.real(), qct))   DOMAIN_ERROR;
-   if (!is_near_int(a.imag(), qct))   DOMAIN_ERROR;
-   if (!is_near_int(b.real(), qct))   DOMAIN_ERROR;
-   if (!is_near_int(b.imag(), qct))   DOMAIN_ERROR;
-
-   // make a and b true integers
-   //
-   a = APL_Complex(round(a.real()), a.imag());
-   b = APL_Complex(round(b.real()), b.imag());
-
-   for (;;)
-       {
-         if (abs(a) > abs(b))   // make ∣b∣ > ∣a∣
-            {
-               const APL_Complex _b = b;
-               b = a;
-               a = _b;
-            }
-
-         if (abs(a) < 0.2)   return b;
-
-         const APL_Complex xy = b/a;
-         const APL_Complex q(round(xy.real()), round(xy.imag()));
-         const APL_Complex r(b - q*a);
-
-         b = a;
-         a = r;
-       }
+   return E_DOMAIN_ERROR;   // char ?
 }
 //-----------------------------------------------------------------------------
 APL_Complex
@@ -381,3 +356,86 @@ APL_Complex z;
    return z;
 }
 //-----------------------------------------------------------------------------
+ErrorCode
+NumericCell::int_gcd(APL_Integer & z, APL_Integer a, APL_Integer b)
+{
+   if (a == 0x8000000000000000ULL)   return E_DOMAIN_ERROR;
+   if (b == 0x8000000000000000ULL)   return E_DOMAIN_ERROR;
+
+   if (a < 0)   a = - a;
+   if (b < 0)   b = - b;
+   if (b < a)
+      {
+         const APL_Integer _b = b;
+         b = a;
+         a = _b;
+      }
+
+   // at this point 0 ≤ a ≤ b
+   //
+   for (;;)
+       {
+         if (a == 0)   { z = b;   return E_NO_ERROR; }
+         const APL_Integer r = b%a;
+         b = a;
+         a = r;
+       }
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+NumericCell::flt_gcd(APL_Float & z, APL_Float a, APL_Float b, APL_Float qct)
+{
+   if (a < 0)   a = - a;
+   if (b < 0)   b = - b;
+   if (b < a)
+      {
+         const APL_Float _b = b;
+         b = a;
+         a = _b;
+      }
+
+   // at this point 0 ≤ a ≤ b
+   //
+   for (;;)
+       {
+         if (is_near_zero(a, qct))   { z = b;   return E_NO_ERROR; }
+         const APL_Float r = fmod(b, a);
+         b = a;
+         a = r;
+       }
+}
+//-----------------------------------------------------------------------------
+ErrorCode
+NumericCell::cpx_gcd(APL_Complex & z, APL_Complex a, APL_Complex b,
+                     APL_Float qct)
+{
+   if (!is_near_int(a.real(), qct))   return E_DOMAIN_ERROR;
+   if (!is_near_int(a.imag(), qct))   return E_DOMAIN_ERROR;
+   if (!is_near_int(b.real(), qct))   return E_DOMAIN_ERROR;
+   if (!is_near_int(b.imag(), qct))   return E_DOMAIN_ERROR;
+
+   // make a and b true integers
+   //
+   a = APL_Complex(round(a.real()), a.imag());
+   b = APL_Complex(round(b.real()), b.imag());
+
+   for (;;)
+       {
+         if (abs(a) > abs(b))   // make ∣b∣ > ∣a∣
+            {
+               const APL_Complex _b = b;
+               b = a;
+               a = _b;
+            }
+
+         if (abs(a) < 0.2)   { z = b;   return E_NO_ERROR; }
+
+         const APL_Complex xy = b/a;
+         const APL_Complex q(round(xy.real()), round(xy.imag()));
+         const APL_Complex r(b - q*a);
+
+         b = a;
+         a = r;
+       }
+}
+//-----------------------------------------------------------------------------
