@@ -22,28 +22,41 @@
 
 #if HAVE_CURSES_H && HAVE_CURSES_H
 
-#include <curses.h>
-#include <term.h>
+# if defined(__sun) && defined(__SVR4)
+#  define NOMACROS
+# endif
 
-#else
+# include <curses.h>
+# include <term.h>
 
-#define tputs(x, y, z)
-#define setupterm(x, y, z) -1
+// curses #defines erase() and tab() on Solaris, conflicting with eg.
+// vector::erase() and others
+# ifdef erase
+#  undef erase
+# endif
+
+# ifdef tab
+#  undef tab
+# endif
+
+#else // not (HAVE_CURSES_H && HAVE_CURSES_H)
+
+# define tputs(x, y, z)
+# define setupterm(x, y, z) -1
 
 #endif
 
 #include "Command.hh"
 #include "Common.hh"
 #include "DiffOut.hh"
+#include "InputFile.hh"
 #include "main.hh"
 #include "Output.hh"
 #include "PrintOperator.hh"
 #include "Svar_DB.hh"
-#include "TestFiles.hh"
 
 bool Output::colors_enabled = false;
 bool Output::colors_changed = false;
-bool Output::print_sema_held = false;
 bool Output::use_curses = true;
 
 int Output::color_CIN_foreground = 0;
@@ -91,11 +104,16 @@ bool ErrOut::used = false;
 DiffOut dout_filebuf(false);
 DiffOut uerr_filebuf(true);
 
-ostream CIN(&cin_filebuf);
+// Android defines its own CIN, COUT, CERR, and UERR ostreams
+#ifndef WANT_ANDROID
+
+ostream CIN (&cin_filebuf);
 ostream COUT(&dout_filebuf);
 ostream CERR(cerr_filebuf.use());
 ostream UERR(&uerr_filebuf);
 
+#endif
+
 extern ostream & get_CERR();
 ostream & get_CERR()
 {
@@ -116,7 +134,7 @@ char Output::color_COUT[100] = CSI "0;30;48;2;255;255;255m";
 /// VT100 escape sequence to change to cerr color
 char Output::color_CERR[100] = CSI "0;35;48;2;255;255;255m";
 
-/// VT100 escape sequence to change to cerr color
+/// VT100 escape sequence to change to uerr color
 char Output::color_UERR[100] = CSI "0;35;48;2;255;255;255m";
 
 /// VT100 escape sequence to reset colors to their default
@@ -129,53 +147,25 @@ char Output::clear_EOL[100] = CSI "K";
 int
 CinOut::overflow(int c)
 {
-   if (do_not_echo)   return 0;
-   if (!Output::print_sema_held)
-      {
-        Svar_DB::start_print(LOC);
-        Output::print_sema_held = true;
-      }
+   if (!InputFile::echo_current_file())   return 0;
 
    Output::set_color_mode(Output::COLM_INPUT);
    cerr << (char)c;
-
-   if (c == 0x0A)
-      {
-        if (Output::print_sema_held)
-           {
-             Svar_DB::end_print(LOC);
-             Output::print_sema_held = false;
-           }
-      }
-
    return 0;
 }
 //-----------------------------------------------------------------------------
 int
 ErrOut::overflow(int c)
 {
-   if (!Output::print_sema_held)
-      {
-        Svar_DB::start_print(LOC);
-        Output::print_sema_held = true;
-      }
-
    Output::set_color_mode(Output::COLM_ERROR);
    cerr << (char)c;
-
-   if (c == '\n')
-      {
-        Svar_DB::end_print(LOC);
-        Output::print_sema_held = false;
-      }
-
    return 0;
 }
 //-----------------------------------------------------------------------------
 void
 Output::init()
 {
-int ret = setupterm(NULL, fileno(stdout), NULL);
+const int ret = setupterm(NULL, fileno(stdout), NULL);
    if (ret != 0)
       {
         use_curses = false;
@@ -183,14 +173,14 @@ int ret = setupterm(NULL, fileno(stdout), NULL);
 }
 //-----------------------------------------------------------------------------
 int
-Output::putc_stderr(int ch)
+Output::putc_stderr(TPUTS_arg3 ch)
 {
    cerr << (char)ch;
    return ch;
 }
 //-----------------------------------------------------------------------------
 int
-Output::putc_stdout(int ch)
+Output::putc_stdout(TPUTS_arg3 ch)
 {
    cout << (char)ch;
    return ch;
@@ -222,6 +212,9 @@ Output::reset_colors()
       }
 }
 //-----------------------------------------------------------------------------
+/// make Solaris happy
+#define tparm10(x, y) tparm(x, y, 0, 0, 0, 0, 0, 0, 0, 0)
+
 void
 Output::set_color_mode(Output::ColorMode mode)
 {
@@ -238,37 +231,38 @@ Output::set_color_mode(Output::ColorMode mode)
         switch(color_mode)
            {
              case COLM_INPUT:
-                  tputs(tparm(set_foreground, color_CIN_foreground),
+                  tputs(tparm10(set_foreground, color_CIN_foreground),
                               1, putc_stdout);
-                  tputs(tparm(set_background, color_CIN_background),
+                  tputs(tparm10(set_background, color_CIN_background),
                               1, putc_stdout);
                   tputs(clr_eol, 1, putc_stdout);
                   break;
 
              case COLM_OUTPUT:
-                  tputs(tparm(set_foreground, color_COUT_foreground),
+                  tputs(tparm10(set_foreground, color_COUT_foreground),
                               1, putc_stdout);
-                  tputs(tparm(set_background, color_COUT_background),
+                  tputs(tparm10(set_background, color_COUT_background),
                               1, putc_stdout);
                   tputs(clr_eol, 1, putc_stdout);
                   break;
 
              case COLM_ERROR:
-                  tputs(tparm(set_foreground, color_CERR_foreground),
+                  tputs(tparm10(set_foreground, color_CERR_foreground),
                               1, putc_stderr);
-                  tputs(tparm(set_background, color_CERR_background),
+                  tputs(tparm10(set_background, color_CERR_background),
                               1, putc_stderr);
                   tputs(clr_eol, 1, putc_stderr);
                   break;
 
              case COLM_UERROR:
-                  tputs(tparm(set_foreground, color_UERR_foreground),
+                  tputs(tparm10(set_foreground, color_UERR_foreground),
                               1, putc_stdout);
-                  tputs(tparm(set_background, color_UERR_background),
+                  tputs(tparm10(set_background, color_UERR_background),
                               1, putc_stdout);
                   tputs(clr_eol, 1, putc_stdout);
                   break;
 
+             default: break;
            }
       }
    else
@@ -279,6 +273,7 @@ Output::set_color_mode(Output::ColorMode mode)
              case COLM_OUTPUT: cout << color_COUT << clear_EOL;   break;
              case COLM_ERROR:  cerr << color_CERR << clear_EOL;   break;
              case COLM_UERROR: cout << color_UERR << clear_EOL;   break;
+             default: break;
            }
       }
 }
