@@ -32,7 +32,7 @@
 #include "Symbol.hh"
 #include "SystemLimits.hh"
 #include "SystemVariable.hh"
-#include "Value.hh"
+#include "Value.icc"
 #include "Tokenizer.hh"
 #include "Workspace.hh"
 
@@ -59,7 +59,7 @@ Parser::parse(const Token_string & input, Token_string & tos) const
    Log(LOG_parse)
       {
         CERR << "parse 1 [" << input.size() << "]: ";
-        print_token_list(CERR, input, 0);
+        input.print(CERR);
         CERR << endl;
       }
 
@@ -115,7 +115,7 @@ Parser::parse_statement(Token_string & tos) const
    Log(LOG_parse)
       {
         CERR << "parse 2 [" << tos.size() << "]: ";
-        print_token_list(CERR, tos, 0);
+        tos.print(CERR);
       }
 
    // 2. convert groups like '(' val val...')' into single APL values
@@ -131,7 +131,7 @@ Parser::parse_statement(Token_string & tos) const
    Log(LOG_parse)
       {
         CERR << "parse 3 [" << tos.size() << "]: ";
-        print_token_list(CERR, tos, 0);
+        tos.print(CERR);
       }
 
    // 3. convert vectors like 1 2 3 or '1' 2 '3' into single APL values
@@ -142,7 +142,7 @@ Parser::parse_statement(Token_string & tos) const
    Log(LOG_parse)
       {
         CERR << "parse 4 [" << tos.size() << "]: ";
-        print_token_list(CERR, tos, 0);
+        tos.print(CERR);
       }
 
    // 4. mark symbol left of ← as LSYMB
@@ -151,7 +151,7 @@ Parser::parse_statement(Token_string & tos) const
    Log(LOG_parse)
       {
         CERR << "parse 5 [" << tos.size() << "]: ";
-        print_token_list(CERR, tos, 0);
+        tos.print(CERR);
       }
 
    // 5. resolve ambiguous / ⌿ \ and ⍀ operators/functions
@@ -160,7 +160,7 @@ Parser::parse_statement(Token_string & tos) const
    Log(LOG_parse)
       {
         CERR << "parse 6 [" << tos.size() << "]: ";
-        print_token_list(CERR, tos, 0);
+        tos.print(CERR);
       }
 
    return E_NO_ERROR;
@@ -172,7 +172,7 @@ Parser::collect_constants(Token_string & tos) const
    Log(LOG_collect_constants)
       {
         CERR << "collect_constants [" << tos.size() << " token] in: ";
-        print_token_list(CERR, tos, 0);
+        tos.print(CERR);
       }
 
    // convert several items in vector notation into a single APL value
@@ -232,7 +232,7 @@ Parser::collect_constants(Token_string & tos) const
    Log(LOG_collect_constants)
       {
         CERR << "collect_constants [" << tos.size() << " token] out: ";
-        print_token_list(CERR, tos, 0);
+        tos.print(CERR);
       }
 }
 //-----------------------------------------------------------------------------
@@ -242,7 +242,7 @@ Parser::collect_groups(Token_string & tos) const
    Log(LOG_collect_constants)
       {
         CERR << "collect_groups [" << tos.size() << " token] in: ";
-        print_token_list(CERR, tos, 0);
+        tos.print(CERR);
       }
 
 int opening = -1;
@@ -282,7 +282,7 @@ int opening = -1;
                       {
                         CERR << "collect_groups [" << tos.size()
                              << " token] out: ";
-                        print_token_list(CERR, tos, 0);
+                        tos.print(CERR);
                       }
                    return true;
 
@@ -430,8 +430,8 @@ Parser::remove_nongrouping_parantheses(Token_string & tos)
 
                // case 2. We have tos[t] = ( X ) ...
                //
-               // (X) : "not grouping" if X is a skalar. 
-               // If X is non-skalar, enclose it
+               // (X) : "not grouping" if X is a scalar. 
+               // If X is non-scalar, enclose it
                //
                progress = true;
                move_1(tos[t + 2], tos[t + 1], LOC);
@@ -464,8 +464,8 @@ Parser::degrade_scan_reduce(Token_string & tos)
                  {
                    // the (back-) slash is a function, not an operator.
                    //
-                   const int funtag = tos[src].get_tag() & ~TC_MASK | TC_FUN12;
-                   tos[src].ChangeTag((TokenTag)funtag);
+                   const int ftag = (tos[src].get_tag() & ~TC_MASK) | TC_FUN12;
+                   tos[src].ChangeTag((TokenTag)ftag);
                  }
             }
        }
@@ -504,6 +504,8 @@ Parser::check_if_value(const Token_string & tos, int pos)
 
         case TC_SYMBOL:
              return (tos[pos].get_tag() == TOK_P_SYMB);   // if value
+
+        default: break;
       }
 
    return false;
@@ -531,66 +533,66 @@ Parser::create_value(Token_string & tos, int pos, int count) const
       {
         CERR << "create_value(" << __LINE__ << ") tos[" << tos.size()
              <<  "]  pos " << pos << " count " << count << " in:";
-        print_token_list(CERR, tos, 0);
+        tos.print(CERR);
       }
 
-   if (count == 1)   create_skalar_value(tos[pos]);
+   if (count == 1)   create_scalar_value(tos[pos]);
    else              create_vector_value(tos, pos, count);
 
    Log(LOG_create_value)
       {
         CERR << "create_value(" << __LINE__ << ") tos[" << tos.size()
              <<  "]  pos " << pos << " count " << count << " out:";
-        print_token_list(CERR, tos, 0);
+        tos.print(CERR);
       }
 }
 //-----------------------------------------------------------------------------
 void
-Parser::create_skalar_value(Token & output) const
+Parser::create_scalar_value(Token & output) const
 {
    switch(output.get_tag())
       {
         case TOK_CHARACTER:
              {
-               Value_P skalar(new Value(LOC));
+               Value_P scalar(new Value(LOC));
 
-               new (&skalar->get_ravel(0))  CharCell(output.get_char_val());
-               skalar->check_value(LOC);
-               Token tok(TOK_APL_VALUE3, skalar);
+               new (&scalar->get_ravel(0))  CharCell(output.get_char_val());
+               scalar->check_value(LOC);
+               Token tok(TOK_APL_VALUE3, scalar);
                move_1(output, tok, LOC);
              }
              return;
 
         case TOK_INTEGER:
              {
-               Value_P skalar(new Value(LOC));
+               Value_P scalar(new Value(LOC));
 
-               new (&skalar->get_ravel(0))   IntCell(output.get_int_val());
-               skalar->check_value(LOC);
-               Token tok(TOK_APL_VALUE3, skalar);
+               new (&scalar->get_ravel(0))   IntCell(output.get_int_val());
+               scalar->check_value(LOC);
+               Token tok(TOK_APL_VALUE3, scalar);
                move_1(output, tok, LOC);
              }
              return;
 
         case TOK_REAL:
              {
-               Value_P skalar(new Value(LOC));
+               Value_P scalar(new Value(LOC));
 
-               new (&skalar->get_ravel(0))  FloatCell(output.get_flt_val());
-               skalar->check_value(LOC);
-               Token tok(TOK_APL_VALUE3, skalar);
+               new (&scalar->get_ravel(0))  FloatCell(output.get_flt_val());
+               scalar->check_value(LOC);
+               Token tok(TOK_APL_VALUE3, scalar);
                move_1(output, tok, LOC);
              }
              return;
 
         case TOK_COMPLEX:
              {
-               Value_P skalar(new Value(LOC));
+               Value_P scalar(new Value(LOC));
 
-               new (&skalar->get_ravel(0))   ComplexCell(output.get_cpx_real(),
+               new (&scalar->get_ravel(0))   ComplexCell(output.get_cpx_real(),
                                                          output.get_cpx_imag());
-               skalar->check_value(LOC);
-               Token tok(TOK_APL_VALUE3, skalar);
+               scalar->check_value(LOC);
+               Token tok(TOK_APL_VALUE3, scalar);
                move_1(output, tok, LOC);
              }
              return;
@@ -598,6 +600,8 @@ Parser::create_skalar_value(Token & output) const
         case TOK_APL_VALUE1:
         case TOK_APL_VALUE3:
              return;
+
+        default: break;
       }
 
    CERR << "Unexpected token " << output.get_tag() << ": " << output << endl;
@@ -643,6 +647,8 @@ Value_P vector(new Value(count, LOC));
                  new (addr) PointerCell(tok.get_apl_val());
                  tok.clear(LOC);   // invalidate token
                  break;
+
+              default: Assert(0);
             }
        }
 
@@ -654,7 +660,7 @@ Token tok(TOK_APL_VALUE3, vector);
    Log(LOG_create_value)
       {
         CERR << "create_value [" << tos.size() << " token] out: ";
-        print_token_list(CERR, tos, 0);
+        tos.print(CERR);
       }
 }
 //-----------------------------------------------------------------------------
@@ -722,16 +728,10 @@ Parser::mark_lsymb(Token_string & tos)
 
                    case TC_END:     prev = 0;   // stop prev loop
                                     break;
+
+                   default: break;
                  }
             }
       }
 }
 //-----------------------------------------------------------------------------
-void
-Parser::print_token_list(ostream & out, const Token_string & tos, int from)
-{
-   loop(t, tos.size() - from)   out << "`" << tos[from + t] << "  ";
-
-   out << endl;
-}
-//-----------------------------------------------------------------------------
