@@ -18,8 +18,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-#include <vector>
-
+#include "Bif_OPER2_RANK.hh"
 #include "Common.hh"
 #include "DerivedFunction.hh"
 #include "Executable.hh"
@@ -100,6 +99,15 @@ Prefix::syntax_error(const char * loc)
           }
       }
 
+   // see if error was caused by a function not returning a value.
+   // in that case we throw a value error instead of a syntax error.
+   //
+   loop (s, size())
+      {
+        if (at(s).tok.get_Class() == TC_VOID)
+           throw_apl_error(E_VALUE_ERROR, loc);
+      }
+
    throw_apl_error(get_assign_state() == ASS_none ? E_SYNTAX_ERROR
                                                   : E_LEFT_SYNTAX_ERROR, loc);
 }
@@ -181,7 +189,7 @@ int count = 0;
         const Token & tok = at(s).tok;
         if (tok.get_ValueType() != TV_VAL)      continue;
 
-        if (tok.get_apl_val().get() == &value)
+        if (tok.get_apl_val()->is_or_contains(value))
            {
              out << prefix << " Fifo [" << s << "]" << endl;
              ++count;
@@ -209,15 +217,13 @@ Prefix::get_range_high() const
 Function_PC high = lookahead_high;
    if (high == Function_PC_invalid)   high = at(0).pc;
 
-   if (best->misc)   --high;
+   if (best && best->misc)   --high;
    return high;
 }
 //-----------------------------------------------------------------------------
 Function_PC
 Prefix::get_range_low() const
 {
-Function_PC high = lookahead_high;
-
    // if the stack is not empty then return the PC of the lowest element
    //
    if (size() > 0)     return at(size() - 1).pc;
@@ -230,6 +236,74 @@ Function_PC high = lookahead_high;
    return lookahead_high;
 }
 //-----------------------------------------------------------------------------
+bool
+Prefix::value_expected()
+{
+   // return true iff the current saved_lookahead token (which has been tested
+   // to be a TC_INDEX token) is the index of a value and false it it is
+   // a function axis
+
+   // if it contains semicolons then get_ValueType() is TV_INDEX and
+   // it MUST be a value.
+   //
+   if (saved_lookahead.tok.get_ValueType() == TV_INDEX)   return true;
+
+int brack_count = 0;
+   for (int pc = PC; pc < body.size();)
+      {
+        const Token & tok = body[pc++];
+        switch(tok.get_Class())
+           {
+               case TC_R_BRACK:  // skip to opening [
+                    //
+                    brack_count = 1;   // this [...]
+                    for (;pc < body.size(); ++pc)
+                        {
+                          if (body[pc].get_Class() == TC_R_BRACK)   // another ]
+                             {
+                               ++brack_count;
+                             }
+                          else if (body[pc].get_Class() == TC_L_BRACK)
+                             {
+                               --brack_count;
+                               if (brack_count == 0)   break;
+                               ++pc;
+                               continue;   // inner for (.,,)
+                             }
+                        }
+                    ++pc;
+                    continue;   // outer for (.,,)
+
+               case TC_END:      goto done;
+
+               case TC_FUN0:   return true;   // niladic function is a value
+               case TC_FUN12:  return false;  // function
+
+               case TC_SYMBOL:
+                    {
+                      const Symbol * sym = tok.get_sym_ptr();
+                      const NameClass nc = sym->get_nc();
+                      
+                      if (nc == NC_FUNCTION)   return false;
+                      if (nc == NC_OPERATOR)   return false;
+                      return true;   // value
+                    }
+             
+               case TC_RETURN:   goto done;
+               case TC_VALUE:   return true;
+
+               default: continue;
+           }
+
+         break;  // for (;;)
+      }
+
+   // this is a syntax error.
+   //
+done:
+   return false;
+}
+//-----------------------------------------------------------------------------
 void
 Prefix::unmark_all_values() const
 {
@@ -244,6 +318,9 @@ Prefix::unmark_all_values() const
 }
 //-----------------------------------------------------------------------------
 
+#define PH(name, idx, prio, misc, len, fun) \
+   { #name, idx, prio, misc, len, &Prefix::reduce_ ## fun, #fun },
+
 #include "Prefix.def"   // the hash table
 
 Token
@@ -296,7 +373,7 @@ grow:
           Symbol * sym = tl.tok.get_sym_ptr();
           if (tl.tok.get_tag() == TOK_LSYMB2)
              {
-               // this is the last token (C) of a vector assignment
+               // this is the last token C of a vector assignment
                // (A B ... C)←. We return C and let the caller do the rest
                //
                sym->resolve(tl.tok, true);
@@ -432,6 +509,8 @@ found_prefix:
           case TC_FUN12:
                more = best->prio < BS_OP_RO && next == TC_OPER2;
                break;
+
+          default: break;
         }
 
      if (more)
@@ -475,9 +554,9 @@ found_prefix:
         goto again;
       }
 
-   if (action == RA_NEXT_STAT)
+   if (action == RA_PUSH_NEXT)
       {
-        Log(LOG_prefix_parser)   CERR << "RA_NEXT_STAT" << endl;
+        Log(LOG_prefix_parser)   CERR << "RA_PUSH_NEXT" << endl;
         goto grow;
       }
 
@@ -521,6 +600,62 @@ Prefix::replace_AB(Value_P old_value, Value_P new_value)
    return false;
 }
 //-----------------------------------------------------------------------------
+Token * Prefix::locate_L()
+{
+   // expect at least A f B (so we have at0(), at1() and at2()
+
+   if (prefix_len < 3)   return 0;
+
+   if (at1().get_Class() != TC_FUN12 &&
+       at1().get_Class() != TC_OPER1 &&
+       at1().get_Class() != TC_OPER2)   return 0;
+
+   if (at0().get_Class() == TC_VALUE)   return &at0();
+   return 0;
+}
+//-----------------------------------------------------------------------------
+Token *
+Prefix::locate_X()
+{
+   // expect at least B X (so we have at0() and at1() and at2()
+
+   if (prefix_len < 2)   return 0;
+
+   // either at0() (for monadic f X B) or at1() (for dyadic A f X B) must
+   // be a function or operator
+   //
+   for (int x = put - 2; x >= put - prefix_len; --x)
+       {
+         if (content[x].tok.get_Class() == TC_INDEX)   // found X
+            {
+              return &content[x].tok;
+            }
+       }
+
+   return 0;
+}
+//-----------------------------------------------------------------------------
+Token * Prefix::locate_R()
+{
+   // expect at least f B (so we have at0(), at1() and at2()
+
+   if (prefix_len < 2)   return 0;
+
+   // either at0() (for monadic f B) or at1() (for dyadic A f B) must
+   // be a function or operator
+   //
+   if (at0().get_Class() != TC_FUN12 &&
+       at0().get_Class() != TC_OPER1 &&
+       at0().get_Class() != TC_OPER2 &&
+       at1().get_Class() != TC_FUN12 &&
+       at1().get_Class() != TC_OPER1 &&
+       at1().get_Class() != TC_OPER2)   return 0;
+
+Token * ret = &content[put - prefix_len].tok;
+   if (ret->get_Class() == TC_VALUE)   return ret;
+   return 0;
+}
+//-----------------------------------------------------------------------------
 void
 Prefix::print(ostream & out, int indent) const
 {
@@ -565,10 +700,18 @@ Prefix::reduce_N___()
 {
    Assert1(prefix_len == 1);
 
-Token result = si.eval_(at0());
-   move_1(at0(), result, LOC);
-   at0().SET_temp();
-   set_action(at0());
+Token result = at0().get_function()->eval_();
+   if (result.get_tag() == TOK_ERROR)
+      {
+        Token_loc tl(result, get_range_low());
+        push(tl);
+        action = RA_RETURN;
+        return;
+      }
+
+   pop_args_push_result(result);
+   result.SET_temp();
+   set_action(result);
 }
 //-----------------------------------------------------------------------------
 void
@@ -576,9 +719,31 @@ Prefix::reduce_MISC_F_B_()
 {
    Assert1(prefix_len == 2);
 
-   pop_args_push_result(si.eval_B(at0(), at1()));
-   at0().SET_temp();
-   set_action(at0());
+   if (saved_lookahead.tok.get_Class() == TC_INDEX)
+      {
+        if (value_expected())
+           {
+             // push [...] and read one more token
+             //
+             push(saved_lookahead);
+             saved_lookahead.tok.clear(LOC);
+             action = RA_PUSH_NEXT;
+             return;
+           }
+      }
+
+Token result = at0().get_function()->eval_B(at1().get_apl_val());
+   if (result.get_tag() == TOK_ERROR)
+      {
+        Token_loc tl(result, get_range_low());
+        push(tl);
+        action = RA_RETURN;
+        return;
+      }
+
+   pop_args_push_result(result);
+   result.SET_temp();
+   set_action(result);
 }
 //-----------------------------------------------------------------------------
 void
@@ -586,12 +751,35 @@ Prefix::reduce_MISC_F_C_B()
 {
    Assert1(prefix_len == 3);
 
+   if (saved_lookahead.tok.get_Class() == TC_INDEX)
+      {
+        if (value_expected())
+           {
+             // push [...] and read one more token
+             //
+             push(saved_lookahead);
+             saved_lookahead.tok.clear(LOC);
+             action = RA_PUSH_NEXT;
+             return;
+           }
+      }
+
    if (at1().get_ValueType() != TV_VAL)   SYNTAX_ERROR;
    if (!at1().get_apl_val())              SYNTAX_ERROR;
 
-   pop_args_push_result(si.eval_XB(at0(), at1(), at2()));
-   at0().SET_temp();
-   set_action(at0());
+Token result = at0().get_function()->eval_XB(at1().get_apl_val(),
+                                             at2().get_apl_val());
+   if (result.get_tag() == TOK_ERROR)
+      {
+        Token_loc tl(result, get_range_low());
+        push(tl);
+        action = RA_RETURN;
+        return;
+      }
+
+   pop_args_push_result(result);
+   result.SET_temp();
+   set_action(result);
 }
 //-----------------------------------------------------------------------------
 void
@@ -599,9 +787,19 @@ Prefix::reduce_A_F_B_()
 {
    Assert1(prefix_len == 3);
 
-   pop_args_push_result(si.eval_AB(at0(), at1(), at2()));
-   at0().SET_temp();
-   set_action(at0());
+Token result = at1().get_function()->eval_AB(at0().get_apl_val(),
+                                             at2().get_apl_val());
+   if (result.get_tag() == TOK_ERROR)
+      {
+        Token_loc tl(result, get_range_low());
+        push(tl);
+        action = RA_RETURN;
+        return;
+      }
+
+   pop_args_push_result(result);
+   result.SET_temp();
+   set_action(result);
 }
 //-----------------------------------------------------------------------------
 void
@@ -620,9 +818,20 @@ Prefix::reduce_A_F_C_B()
    if (at2().get_ValueType() != TV_VAL)   SYNTAX_ERROR;
    if (!at2().get_apl_val())              SYNTAX_ERROR;
 
-   pop_args_push_result(si.eval_AXB(at0(), at1(), at2(), at3()));
-   at0().SET_temp();
-   set_action(at0());
+Token result = at1().get_function()->eval_AXB(at0().get_apl_val(),
+                                              at2().get_apl_val(),
+                                              at3().get_apl_val());
+   if (result.get_tag() == TOK_ERROR)
+      {
+        Token_loc tl(result, get_range_low());
+        push(tl);
+        action = RA_RETURN;
+        return;
+      }
+
+   pop_args_push_result(result);
+   result.SET_temp();
+   set_action(result);
 }
 //-----------------------------------------------------------------------------
 void
@@ -632,12 +841,6 @@ Prefix::reduce_A_M_C_B()
 }
 //-----------------------------------------------------------------------------
 void
-Prefix::reduce_N_M__()
-{
-   reduce_F_M__();
-}
-//-----------------------------------------------------------------------------
-void
 Prefix::reduce_F_M__()
 {
 DerivedFunction * derived =
@@ -673,21 +876,52 @@ Prefix::reduce_B_D_G_()
 }
 //-----------------------------------------------------------------------------
 void
-Prefix::reduce_B_D_N_()
-{
-   reduce_F_D_G_();   // same as F2
-}
-//-----------------------------------------------------------------------------
-void
 Prefix::reduce_F_D_B_()
 {
-   reduce_F_D_G_();   // same as G2
-}
-//-----------------------------------------------------------------------------
-void
-Prefix::reduce_N_D_B_()
-{
-   reduce_F_D_G_();   // same as G2
+   // same as G2, except for ⍤
+   //
+   if (at1().get_function()->get_Id() != ID_OPER2_RANK)
+      {
+         reduce_F_D_G_();
+         return;
+      }
+
+   // we have f ⍤ y_B with y_B glued beforehand. Unglue it.
+   //
+Value_P y123;
+Value_P B;
+   Bif_OPER2_RANK::split_y123_B(at2().get_apl_val(), y123, B);
+Token new_y123(TOK_APL_VALUE1, y123);
+
+DerivedFunction * derived = Workspace::SI_top()->fun_oper_cache.get(LOC);
+   new (derived) DerivedFunction(at0(), at1().get_function(), new_y123, LOC);
+
+Token result = Token(TOK_FUN2, derived);
+
+   if (!B)   // only y123, no B (e.g. (f ⍤[X] 1 2 3)
+      {
+        pop_args_push_result(result);
+      }
+   else      // a new B split off from the original B
+      {
+        // save locations of ⍤ and B
+        //
+        Function_PC pc_D = at(1).pc;
+        Function_PC pc_B = at(3).pc;
+
+        pop_and_discard();   // pop F
+        pop_and_discard();   // pop C
+        pop_and_discard();   // pop B (old)
+
+        Token new_B(TOK_APL_VALUE1, B);
+        Token_loc tl_B(new_B, pc_B);
+        Token_loc tl_derived(result, pc_D);
+        push(tl_B);
+        push(tl_derived);
+      }
+
+   action = RA_CONTINUE;
+
 }
 //-----------------------------------------------------------------------------
 void
@@ -702,21 +936,46 @@ DerivedFunction * derived =
 }
 //-----------------------------------------------------------------------------
 void
-Prefix::reduce_F_D_N_()
+Prefix::reduce_F_D_C_B()
 {
-   reduce_F_D_G_();   // same as G2
-}
-//-----------------------------------------------------------------------------
-void
-Prefix::reduce_N_D_G_()
-{
-   reduce_F_D_G_();   // same as G2
-}
-//-----------------------------------------------------------------------------
-void
-Prefix::reduce_N_D_N_()
-{
-   reduce_F_D_G_();   // same as G2
+   // we have f ⍤ [X] y_B with y_B glued beforehand. Unglue it.
+   //
+Value_P y123;
+Value_P B;
+   Bif_OPER2_RANK::split_y123_B(at3().get_apl_val(), y123, B);
+Token new_y123(TOK_APL_VALUE1, y123);
+
+Value_P v_idx = at2().get_axes();
+DerivedFunction * derived = Workspace::SI_top()->fun_oper_cache.get(LOC);
+   new (derived) DerivedFunction(at0(), at1().get_function(),
+                                 v_idx, new_y123, LOC);
+
+Token result = Token(TOK_FUN2, derived);
+
+   if (!B)   // only y123, no B (e.g. (f ⍤[X] 1 2 3)
+      {
+        pop_args_push_result(result);
+      }
+   else      // a new B split off from the original B
+      {
+        // save locations of ⍤ and B
+        //
+        Function_PC pc_D = at(1).pc;
+        Function_PC pc_B = at(3).pc;
+
+        pop_and_discard();   // pop F
+        pop_and_discard();   // pop D
+        pop_and_discard();   // pop C
+        pop_and_discard();   // pop B (old)
+
+        Token new_B(TOK_APL_VALUE1, B);
+        Token_loc tl_B(new_B, pc_B);
+        Token_loc tl_derived(result, pc_D);
+        push(tl_B);   
+        push(tl_derived);   
+      }
+
+   action = RA_CONTINUE;
 }
 //-----------------------------------------------------------------------------
 void
@@ -830,6 +1089,49 @@ Token result = Token(TOK_APL_VALUE2, B);
 }
 //-----------------------------------------------------------------------------
 void
+Prefix::reduce_V_ASS_F_()
+{
+   // named lambda: V ← { ... }
+   //
+Function * F = at2().get_function();
+   if (!F->is_lambda())   SYNTAX_ERROR;
+
+Symbol * V = at0().get_sym_ptr();
+   V->assign_named_lambda(F, LOC);
+
+Value_P Z(new Value(V->get_name(), LOC));
+Token result = Token(TOK_APL_VALUE2, Z);
+   pop_args_push_result(result);
+
+   set_assign_state(ASS_none);
+   action = RA_CONTINUE;
+}
+//-----------------------------------------------------------------------------
+void
+Prefix::reduce_V_ASS_N_()
+{
+   // named lambda: V ← { ... } niladic (same as reduce_V_ASS_F_)
+   //
+   reduce_V_ASS_F_();
+}
+//-----------------------------------------------------------------------------
+void
+Prefix::reduce_V_ASS_M_()
+{
+   // named lambda: V ← { ... } monadic operator (same as reduce_V_ASS_F_)
+   //
+   reduce_V_ASS_F_();
+}
+//-----------------------------------------------------------------------------
+void
+Prefix::reduce_V_ASS_D_()
+{
+   // named lambda: V ← { ... } dyadic operator (same as reduce_V_ASS_F_)
+   //
+   reduce_V_ASS_F_();
+}
+//-----------------------------------------------------------------------------
+void
 Prefix::reduce_RBRA___()
 {
    Assert1(prefix_len == 1);
@@ -960,52 +1262,23 @@ const int count = vector_ass_count();
         return;
       }
 
-   // B should be a vector with count elements.
+   // vector assignment.
    //
+DynArray(Symbol *, symbols, count + 1);
+   symbols[0] = at0().get_sym_ptr();
+   loop(c, count)
+      {
+        Token_loc tl = lookahead();
+        Assert1(tl.tok.get_tag() == TOK_LSYMB2);   // by vector_ass_count()
+        Symbol * V = tl.tok.get_sym_ptr();
+        Assert(V);
+        symbols[c + 1] = V;
+      }
+
 Value_P B = at3().get_apl_val();
-   if (B->is_skalar())
-      {
-        if (B->get_ravel(0).is_pointer_cell())   // disclose B if enclosed
-           B = B->get_ravel(0).get_pointer_value();
+   Symbol::vector_assignment(symbols, count + 1, B);
 
-        Symbol * V = at0().get_sym_ptr();
-        V->assign(B, LOC);
-        loop(c, count)
-           {
-             Token_loc tl = lookahead();
-             Assert1(tl.tok.get_tag() == TOK_LSYMB2);   // by vector_ass_count()
-             V = tl.tok.get_sym_ptr();
-             V->assign(B, LOC);
-           }
-      }
-   else
-      {
-        if (B->get_rank() != 1)   RANK_ERROR;
-        if (B->element_count() != (count + 1))   LENGTH_ERROR;
-
-        // assign last element of B to V
-        //
-        {
-          Symbol * V = at0().get_sym_ptr();
-          Cell & cell = B->get_ravel(count);
-          Value_P B_last(new Value(cell, LOC));
-          V->assign(B_last, LOC);
-        }
-
-        // assign other elements of B to other variables...
-        //
-        loop(c, count)
-           {
-             Token_loc tl = lookahead();
-             Assert1(tl.tok.get_tag() == TOK_LSYMB2);   // by vector_ass_count()
-             Symbol * V = tl.tok.get_sym_ptr();
-             Cell & cell = B->get_ravel(count - c - 1);
-             Value_P B_cell(new Value(cell, LOC));
-             V->assign(B_cell, LOC);
-           }
-      }
-
-    set_assign_state(ASS_none);
+   set_assign_state(ASS_none);
 
    // clean up stack
    //
@@ -1025,13 +1298,14 @@ Prefix::reduce_END_VOID__()
    if (size() != 2)   syntax_error(LOC);
 
 const bool end_of_line = at0().get_tag() == TOK_ENDL;
+const bool trace = (at0().get_int_val() & 1) != 0;
 
    pop_and_discard();   // pop END
    pop_and_discard();   // pop VOID
 
 Token Void(TOK_VOID);
-   si.statement_result(Void);
-   action = RA_NEXT_STAT;
+   si.statement_result(Void, trace);
+   action = RA_PUSH_NEXT;
    if (attention_raised && end_of_line)
       {
         attention_raised = false;
@@ -1048,13 +1322,13 @@ Prefix::reduce_END_B__()
    if (size() != 2)   syntax_error(LOC);
 
 const bool end_of_line = at0().get_tag() == TOK_ENDL;
+const bool trace = (at0().get_int_val() & 1) != 0;
 
    pop_and_discard();   // pop END
 Token B = pop().tok;    // pop B
+   si.statement_result(B, trace);
 
-   si.statement_result(B);
-
-   action = RA_NEXT_STAT;
+   action = RA_PUSH_NEXT;
    if (attention_raised && end_of_line)
       {
         attention_raised = false;
@@ -1073,7 +1347,17 @@ Prefix::reduce_END_GOTO_B_()
    // at0() is either TOK_END or TOK_ENDL.
    //
 const bool end_of_line = at0().get_tag() == TOK_ENDL;
+const bool trace = at0().get_Class() == TC_END &&
+                  (at0().get_int_val() & 1) != 0;
+
 Value_P line = at2().get_apl_val();
+   if (trace && line->element_count() > 0)
+      {
+        const int64_t line_num = line->get_line_number(Workspace::get_CT());
+        Token bra(TOK_BRANCH, line_num);
+        si.statement_result(bra, true);
+      }
+
 
 const Token result = si.jump(line);
 
@@ -1097,7 +1381,7 @@ const Token result = si.jump(line);
 
    if (result.get_tag() == TOK_VOID)   // branch not taken, e.g. →''
       {
-        action = RA_NEXT_STAT;
+        action = RA_PUSH_NEXT;
         if (attention_raised && end_of_line)
            {
              attention_raised = false;
@@ -1118,7 +1402,7 @@ const Function_PC new_pc = si.get_PC();
       }
 
    Assert1(size() == 0);
-   action = RA_NEXT_STAT;
+   action = RA_PUSH_NEXT;
    if (attention_raised && end_of_line)
       {
         attention_raised = false;
@@ -1134,7 +1418,49 @@ Prefix::reduce_END_GOTO__()
 
    if (size() != 2)   syntax_error(LOC);
 
-Token result = Token(TOK_ESCAPE);
+const bool trace = at0().get_Class() == TC_END &&
+                  (at0().get_int_val() & 1) != 0;
+   if (trace)
+      {
+        Token bra(TOK_ESCAPE);
+        si.statement_result(bra, true);
+      }
+
+   // the statement is → which could mean TOK_ESCAPE (normal →) or
+   //  TOK_STOP_LINE from S∆←line
+   //
+   if (at1().get_tag() == TOK_STOP_LINE)   // S∆ line
+      {
+        const UserFunction * ufun = si.get_executable()->get_ufun();
+        if (ufun && ufun->get_exec_properties()[2])
+           {
+              // the function ignores attention (aka. weak interrupt)
+              //
+              pop_and_discard();   // pop() END
+              pop_and_discard();   // pop() GOTO
+              action = RA_CONTINUE;
+              return;
+           }
+
+        COUT << si.function_name() << "[" << si.get_line() << "]" << endl;
+        Token result(TOK_ERROR, E_STOP_LINE);
+        pop_args_push_result(result);
+        action = RA_RETURN;
+      }
+   else
+      {
+        Token result(TOK_ESCAPE);
+        pop_args_push_result(result);
+        action = RA_RETURN;
+      }
+}
+//-----------------------------------------------------------------------------
+void
+Prefix::reduce_RETC_VOID__()
+{
+   Assert1(prefix_len == 2);
+
+Token result = Token(TOK_VOID);
    pop_args_push_result(result);
    action = RA_RETURN;
 }
@@ -1178,7 +1504,6 @@ Prefix::reduce_RETC___()
              {
                const UserFunction * ufun = si.get_executable()->get_ufun();
                Assert1(ufun);
-               ufun->pop_local_vars();
                at0().clear(LOC);
              }
              action = RA_RETURN;
@@ -1188,7 +1513,9 @@ Prefix::reduce_RETC___()
              {
                const UserFunction * ufun = si.get_executable()->get_ufun();
                Assert1(ufun);
-               Value_P Z = ufun->pop_local_vars();
+               Symbol * ufun_Z = ufun->get_sym_Z();
+               Value_P Z;
+               if (ufun_Z)   Z = ufun_Z->get_value();
                if (!!Z)
                   {
                     Log(LOG_prefix_parser)
@@ -1206,6 +1533,8 @@ Prefix::reduce_RETC___()
              }
              action = RA_RETURN;
              return;
+
+        default: break;
       }
 
    // not reached
@@ -1244,6 +1573,19 @@ Prefix::reduce_RETC_GOTO_B_()
    if (size() != 3)   syntax_error(LOC);
 
    reduce_END_GOTO_B_();
+
+   if (action == RA_PUSH_NEXT)
+      {
+        // reduce_END_GOTO_B_() has detected a non-taken branch (e.g. →'')
+        // and wants to continue with the next statement.
+        // We are in ⍎ mode, so there is no next statement and we
+        // RA_RETURN a TOK_VOID instead of RA_PUSH_NEXT.
+        //
+        Token tok(TOK_VOID);
+        Token_loc tl(tok, Function_PC_0);
+        push(tl);
+      }
+
    action = RA_RETURN;
 }
 //-----------------------------------------------------------------------------
