@@ -22,31 +22,44 @@
 #include <string.h>
 
 #include "Function.hh"
+#include "main.hh"
 #include "Output.hh"
 #include "PointerCell.hh"
 #include "PrintBuffer.hh"
 #include "PrintOperator.hh"
-#include "Value.hh"
+#include "Value.icc"
 
 //-----------------------------------------------------------------------------
 PrintBuffer::PrintBuffer()
+   : complete(true)
 {
 }
 //-----------------------------------------------------------------------------
 PrintBuffer::PrintBuffer(const UCS_string & ucs, const ColInfo & ci)
-   : col_info(ci)
+   : col_info(ci),
+     complete(true)
 {
    buffer.push_back(ucs);
 }
 //-----------------------------------------------------------------------------
-PrintBuffer::PrintBuffer(const Value & value, const PrintContext & _pctx)
+PrintBuffer::PrintBuffer(const Value & value, const PrintContext & _pctx,
+                         ostream * out)
+   : complete(false)
 {
+   // Note: if ostream is non-0 then this value may be inccomplete
+   // (as indicated by member complete if it is huge. This is to speed
+   // up printing if the value is discarded after having been printed
+
 const PrintStyle outer_style = _pctx.get_style();
 const bool framed = outer_style & (PST_CS_MASK | PST_CS_OUTER);
 PrintContext pctx(_pctx);
    pctx.set_style(PrintStyle(outer_style &~ PST_CS_OUTER));
 
-   if (value.is_skalar())
+const ShapeItem ec = value.element_count();
+const uint64_t ii_count = interrupt_count;
+const bool huge = out && ec > 10000;
+
+   if (value.is_scalar())
       {
         PrintContext pctx1(pctx);
         if (value.get_ravel(0).need_scaling(pctx))   pctx1.set_scaled();
@@ -56,292 +69,360 @@ PrintContext pctx(_pctx);
         // pad the value unless it is framed
         if (value.compute_depth() > 1 && !framed)
            {
-                pad_l(UNI_iPAD_U4, 1);
-                pad_r(UNI_iPAD_U5, 1);
+             pad_l(UNI_iPAD_U4, 1);
+             pad_r(UNI_iPAD_U5, 1);
            }
 
         add_outer_frame(outer_style);
-        return;
+        goto maybe_print_it;
       }
 
-const ShapeItem ec = value.element_count();
-   if (ec == 0)   // empty value of any dimension
-      {
-         const Shape sh = value.get_shape().without_axis(value.get_rank() - 1);
-         if (sh.get_volume() <= 1)   // empty vector
-            {
-              if (pctx.get_style() == PR_APL_FUN)
-                 {
-                   if (value.get_ravel(0).is_character_cell())   // ''
-                      {
-                        UCS_string ucs;
-                        ucs.append(UNI_SINGLE_QUOTE);
-                        ucs.append(UNI_SINGLE_QUOTE);
-
-                        ColInfo ci;
-                        *this = PrintBuffer(ucs, ci);
-                        add_outer_frame(outer_style);
-                        return;
-                      }
-                 }
-              add_outer_frame(outer_style);
-              return;   // 0 rows
-            }
-
-         // value has > 0 rows. Compute how many lines we need.
-         //
-         ShapeItem lines = sh.get_volume();
-         loop(s, sh.get_rank())
-             lines += s * (sh.get_shape_item(sh.get_rank() - s - 1));
-
-         buffer.resize(lines);
-         add_outer_frame(outer_style);
-         return;
-      }
-
-const ShapeItem cols = value.get_last_shape_item();
-const ShapeItem rows = ec/cols;
-
-   if (pctx.get_style() == PR_APL_FUN)
-      {
-        UCS_string ucs;
-
-        if (value.is_char_vector())
-           {
-             ucs.append(UNI_SINGLE_QUOTE);
-             loop(e, ec)
-                {
-                  const Unicode uni = value.get_ravel(e).get_char_value();
-                  ucs.append(uni);
-                  if (uni == UNI_SINGLE_QUOTE)   ucs.append(uni);   // ' -> ''
-                }
-             ucs.append(UNI_SINGLE_QUOTE);
-           }
-        else
-           {
-             loop(e, ec)
-                {
-                  PrintBuffer pb = value.get_ravel(e)
-                                        .character_representation(pctx);
-                  if (e)   ucs.append(UNI_ASCII_SPACE);
-                  ucs.append(UCS_string(pb, 0, pctx.get_PW()));
-                }
-           }
-
-        ColInfo ci;
-        *this = PrintBuffer(ucs, ci);
-        add_outer_frame(outer_style);
-        return;
-      }
-   else if (pctx.get_style() & PST_HEXDUMP)
-      {
-        UCS_string ucs;
-        const char * char_fmt;
-        const char * int_fmt;
-        if (pctx.get_style() & PST_HEXlower)
-           {
-             if (pctx.get_style() & PST_PACKED)   char_fmt = "%04x";
-             else                                 char_fmt = "%02x";
-             if (sizeof(APL_Integer) == 4)        int_fmt  = "%08x";
-             else                                 int_fmt  = "%016x";
-           }
-        else
-           {
-             if (pctx.get_style() & PST_PACKED)   char_fmt = "%04X";
-             else                                 char_fmt = "%02X";
-             if (sizeof(APL_Integer) == 4)        int_fmt  = "%08X";
-             else                                 int_fmt  = "%016X";
-           }
-
-        loop(e, ec)
-           {
-             const Cell & cell = value.get_ravel(e);
-             char cc[200];
-             if (cell.is_character_cell())
-                {
-                  const Unicode uni = cell.get_char_value();
-                  snprintf(cc, sizeof(cc) - 1, char_fmt, uni);
-                }
-              else if (cell.is_integer_cell())
-                {
-                  const APL_Integer val = cell.get_int_value();
-                  snprintf(cc, sizeof(cc) - 1, int_fmt, val);
-                  
-                }
-              else
-                {
-                  CERR << "??celltype= " << cell.get_classname()
-                       << " ??" << endl;;
-                  DOMAIN_ERROR;
-                }
-
-              loop(c, sizeof(cc))
-                 {
-                   if (cc[c])   ucs.append(Unicode(cc[c]));
-                   else         break;
-                 }
-           }
-
-        ColInfo ci;
-        *this = PrintBuffer(ucs, ci);
-        add_outer_frame(outer_style);
-        return;
-      }
-
-   // 1. create a vector with a bool per column that tells if the
-   // column needs scaling (i.e. exponential format) or not
-   //
-vector<bool> scaling;
    {
-     scaling.reserve(cols);
-     loop(x, cols)
+     if (ec == 0)   // empty value of any dimension
         {
-          bool need_scaling = false;
-          loop(y, rows)
-             {
-               if (value.get_ravel(x + y*cols).need_scaling(pctx))
-                  {
-                    need_scaling = true;
-                    break;
-                  }
-             }
-          scaling.push_back(need_scaling);
-        }
-   }
-
-   // 2. create a matrix of items.
-   //    An item of the matrix is a top-level value (possibly nested),
-   //    therefore we have (⍴,value) items. Items are rectangular.
-   //
-typedef vector<PrintBuffer> PB_row;
-vector<PB_row> item_matrix;
-
-   item_matrix.reserve(rows);
-   loop(y, rows)
-      {
-        vector<PrintBuffer> item_row;
-        item_row.reserve(cols);
-        ShapeItem row_height = 0;
-
-        loop(x, cols)
-            {
-              PrintContext pctx1 = pctx;
-              if (scaling[x])   pctx1.set_scaled();
-              const Cell & cell = value.get_ravel(x + y*cols);
-              PrintBuffer item = cell.character_representation(pctx1);
-
-              if (row_height < item.get_height())
-                 row_height = item.get_height();
-
-              Assert(item.is_rectangular());
-              item_row.push_back(item);
-            }
-
-        // pad all items to the same height
-        loop(x, cols)
-           {
-             item_row[x].pad_height(UNI_iPAD_U6, row_height);
-             Assert(item_row[x].is_rectangular());
-           }
-        item_matrix.push_back(item_row);
-      }
-
-
-   // 3. align all columns (which pads them to the same width).
-   //
-   loop(x, cols)
-      {
-        ColInfo col_info_x;
-        loop(y, rows)   col_info_x.consider(item_matrix[y][x].get_info());
-
-        loop(y, rows)   item_matrix[y][x].align(col_info_x);
-
-      }
-
-   // 4. combine colums, then rows.
-   //
-int32_t last_spacing = 0;
-bool last_notchar = false;
-
-   loop(x, cols)
-      {
-        PrintBuffer pcol;
-
-        loop(y, rows)
-         {
-           // separator row(s)
-           if (y)
+           if (value.get_rank() == 1)   // vector: 1 line
               {
-                ShapeItem prod = 1;
-                loop(r, value.get_rank() - 2)
+                if (pctx.get_style() == PR_APL_FUN)
                    {
-                     prod *= value.get_shape_item(value.get_rank() - r - 2);
-                     if (y % prod == 0)
-                        pcol.append_ucs(UCS_string(pcol.get_width(0),
-                                                      UNI_iPAD_L0));
-                     else
-                        break;
+                     if (value.get_ravel(0).is_character_cell())   // ''
+                        {
+                          UCS_string ucs("''");
+                          ColInfo ci;
+                          *this = PrintBuffer(ucs, ci);
+                          add_outer_frame(outer_style);
+                          goto maybe_print_it;
+                        }
+
+                     if (value.get_ravel(0).is_numeric())   // ⍬
+                        {
+                          UCS_string ucs("⍬");
+                          ColInfo ci;
+                          *this = PrintBuffer(ucs, ci);
+                          add_outer_frame(outer_style);
+                          goto maybe_print_it;
+                        }
                    }
+
+                UCS_string ucs;   // empty
+                append_ucs(ucs);
+                add_outer_frame(outer_style);
+                goto maybe_print_it;   // 1 rows
               }
 
-             const PB_row & row = item_matrix[y];
-             const PrintBuffer & item = row[x];
-             if (y)   pcol.add_row(item);
-             else     pcol = item;
+           const Shape sh = value.get_shape().without_axis(value.get_rank()-1);
+           if (sh.get_volume() <= 1)   // empty vector
+              {
+                add_outer_frame(outer_style);
+                goto maybe_print_it;   // 0 rows
+              }
+
+           // value has > 0 rows. Compute how many lines we need.
+           //
+           ShapeItem lines = sh.get_volume();
+           loop(s, sh.get_rank())
+               lines += s * (sh.get_shape_item(sh.get_rank() - s - 1));
+
+           buffer.resize(lines);
+           add_outer_frame(outer_style);
+           goto maybe_print_it;
+        }
+
+     const ShapeItem cols = value.get_last_shape_item();
+     const ShapeItem rows = ec/cols;
+
+     if (pctx.get_style() == PR_APL_FUN)
+        {
+          UCS_string ucs;
+
+          if (value.is_char_vector())
+             {
+               ucs.append(UNI_SINGLE_QUOTE);
+               loop(e, ec)
+                  {
+                    const Unicode uni = value.get_ravel(e).get_char_value();
+                    ucs.append(uni);
+                    if (uni == UNI_SINGLE_QUOTE)   ucs.append(uni);   // ' -> ''
+                  }
+               ucs.append(UNI_SINGLE_QUOTE);
+             }
+          else
+             {
+               loop(e, ec)
+                  {
+                    PrintBuffer pb = value.get_ravel(e)
+                                          .character_representation(pctx);
+                    if (e)   ucs.append(UNI_ASCII_SPACE);
+                    ucs.append(UCS_string(pb, 0, pctx.get_PW()));
+                  }
+             }
+
+          ColInfo ci;
+          *this = PrintBuffer(ucs, ci);
+          add_outer_frame(outer_style);
+          goto maybe_print_it;
+        }
+
+     // 1. create a vector with a bool per column that tells if the
+     // column needs scaling (i.e. exponential format) or not
+     //
+     vector<bool> scaling;
+     {
+       scaling.reserve(cols);
+       loop(x, cols)
+          {
+            bool need_scaling = false;
+            loop(y, rows)
+               {
+                 if (huge && (ii_count != interrupt_count))   goto interrupted;
+                 if (value.get_ravel(x + y*cols).need_scaling(pctx))
+                    {
+                      need_scaling = true;
+                      break;
+                    }
+               }
+            scaling.push_back(need_scaling);
+          }
+     }
+
+     // 2. create a matrix of items.
+     //    An item of the matrix is a top-level cell (possibly nested),
+     //    therefore we have (⍴,value) items. Items are rectangular.
+     //
+     typedef vector<PrintBuffer> PB_row;
+     vector<PB_row> item_matrix;
+
+     item_matrix.reserve(rows);
+     loop(y, rows)
+        {
+          vector<PrintBuffer> item_row;
+          item_row.reserve(cols);
+          ShapeItem row_height = 0;
+
+          loop(x, cols)
+              {
+                if (huge && (ii_count != interrupt_count))   goto interrupted;
+
+                PrintContext pctx1 = pctx;
+                if (scaling[x])   pctx1.set_scaled();
+                const Cell & cell = value.get_ravel(x + y*cols);
+                PrintBuffer item = cell.character_representation(pctx1);
+
+                if (row_height < item.get_height())
+                   row_height = item.get_height();
+
+                Assert(item.is_rectangular());
+                item_row.push_back(item);
+              }
+
+          // pad all items to the same height
+          loop(x, cols)
+             {
+                if (huge && (ii_count != interrupt_count))   goto interrupted;
+
+               item_row[x].pad_height(UNI_iPAD_U6, row_height);
+               Assert(item_row[x].is_rectangular());
+             }
+          item_matrix.push_back(item_row);
+        }
+
+
+     // 3. align all columns (which pads them to the same width).
+     //
+     loop(x, cols)
+        {
+          ColInfo col_info_x;
+          loop(y, rows)
+              {
+                if (huge && (ii_count != interrupt_count))   goto interrupted;
+                col_info_x.consider(item_matrix[y][x].get_info());
+              }
+
+          loop(y, rows)
+              {
+                if (huge && (ii_count != interrupt_count))   goto interrupted;
+                item_matrix[y][x].align(col_info_x);
+              }
+        }
+
+     // 4. combine colums, then rows.
+     //
+     int last_spacing = 0;
+     bool last_notchar = false;
+
+     vector<PrintBuffer> pcols;
+     pcols.reserve(cols);
+
+     loop(x, cols)
+        {
+          PrintBuffer pcol;
+
+          loop(y, rows)
+           {
+             if (huge && (ii_count != interrupt_count))   goto interrupted;
+
+             // separator row(s)
+             //
+             if (y)
+                {
+                  const UCS_string sepa_row(pcol.get_width(0), UNI_iPAD_L0);
+                  ShapeItem prod = 1;
+                  loop(r, value.get_rank() - 2)
+                     {
+                       prod *= value.get_shape_item(value.get_rank() - r - 2);
+                       if (y % prod == 0)   pcol.append_ucs(sepa_row);
+                       else                 break;
+                     }
+                }
+
+               const PB_row & row = item_matrix[y];
+               const PrintBuffer & item = row[x];
+               if (y)   pcol.add_row(item);
+               else     pcol = item;
+           }
+
+          bool not_char = false;   // determined by get_col_spacing()
+	  const int col_spacing = value.get_col_spacing(not_char, x, framed);
+
+          const int32_t max_spacing = (col_spacing > last_spacing) 
+                                    ?  col_spacing : last_spacing;
+          int32_t no_char_spacing = 0;
+
+          if (huge && (ii_count != interrupt_count))   goto interrupted;
+
+          if (x)   // not first column
+             {
+                if (last_notchar)
+                   {
+                     pcols.back().pad_r(UNI_iPAD_U2, 1);
+                     ++no_char_spacing;
+                   }
+                else if (not_char)
+                   {
+                     pcol.pad_l(UNI_iPAD_U3, 1);
+                     ++no_char_spacing;
+                   }
+
+                // we want a total spacing of 'max_spacing'
+                // but we deduct the 'no_char_spacing' chars ² and ³
+                // that were appended above.
+                //
+                const int u7_pad_len = max_spacing - no_char_spacing;
+                if (u7_pad_len)
+                   {
+                     pcols.back().pad_r(UNI_iPAD_U7, u7_pad_len);
+                   }
+             }
+
+          pcols.push_back(pcol);
+
+          if (huge && (ii_count != interrupt_count))   goto interrupted;
+
+          last_spacing = col_spacing;
+          last_notchar = not_char;
+        }
+
+     // combine columns
+     //
+     for (int dx = 1; dx < cols; dx += dx)
+     for (int xx = dx; xx < cols; xx += dx)
+         {
+           pcols[xx - dx].append_col(pcols[xx]);
          }
+     *this = pcols[0];
 
-        bool not_char = false;   // determined by get_col_spacing()
-	const int32_t col_spacing = value.get_col_spacing(not_char, x, framed);
+     if (value.compute_depth() > 1 && !framed)
+        {
+          pad_l(UNI_iPAD_U8, 1);
+          pad_r(UNI_iPAD_U9, 1);
+        }
 
-        const int32_t max_spacing = (col_spacing > last_spacing) 
-                                  ?  col_spacing : last_spacing;
-        int32_t no_char_spacing = 0;
+     if (!is_rectangular())
+        {
+          Q1(get_height())
+          loop(h, get_height())   CERR << "w=" << get_width(h) << "*" << endl;
+          loop(h, get_height())   CERR << "*" << get_line(h) << "*" << endl;
+        }
 
-        if (x == 0)
-           {
-              *this = pcol;
-           }
-        else
-           {
-              if (last_notchar)
+     Assert(is_rectangular());
+     add_outer_frame(outer_style);
+   }
+
+maybe_print_it:
+   if (huge)   // ergo: out
+      {
+        print_interruptible(*out, value.get_rank(), _pctx.get_PW());
+      }
+   else if (out)
+      {
+        UCS_string ucs(*this, value.get_rank(), _pctx.get_PW());
+        if (ucs.size() == 0)   return;
+        *out << ucs << endl;
+      }
+   complete = true;
+   return;
+
+interrupted:
+   attention_raised = false;
+   interrupt_raised = false;
+   *out << endl << "INTERRUPT" << endl;
+   return;
+}
+//-----------------------------------------------------------------------------
+void
+PrintBuffer::print_interruptible(ostream & out, Rank rank, int quad_PW)
+{
+   if (get_height() == 0)   return;      // empty PrintBuffer
+
+const int total_width = get_width(0);
+
+   // breakpoints can get rather large, so we reserve enough space
+vector<int> breakpoints;
+   breakpoints.reserve(2*total_width/quad_PW);
+
+   // print rows, breaking at breakpoints
+   //
+   loop(row, get_height())
+       {
+         if (row)   out << endl;   // end previous row
+         int col = 0;
+         int b = 0;
+
+         while (col < total_width)
+            {
+              int chunk_len;
+              if (row == 0)   // first row: set up breakpoints
                  {
-                   pad_r(UNI_iPAD_U2, 1);
-                   ++no_char_spacing;
+                   chunk_len = get_line(0).compute_chunk_length(quad_PW, col);
+                   breakpoints.push_back(chunk_len);
                  }
-              else if (not_char)
+              else
                  {
-                   pcol.pad_l(UNI_iPAD_U3, 1);
-                   ++no_char_spacing;
+                   chunk_len = breakpoints[b++];
                  }
 
-              // append the column. we want a total spacing of 'max_spacing'
-              // but we deduct the 'no_char_spacing' chars ² and ³
-              // that were appended already.
-              //
-              add_column(UNI_iPAD_U7, max_spacing - no_char_spacing, pcol);
-           }
+              if (col)   out << endl << "      ";
+              UCS_string trow(get_line(row), col, chunk_len);
+              trow.remove_trailing_padchars();
 
-        last_spacing = col_spacing;
-        last_notchar = not_char;
-      }
+              loop(t, trow.size())
+                  {
+                     const Unicode uni = trow[t];
+                     if (is_iPAD_char(uni))   out << " ";
+                     else                     out << uni;
+                  }
+              col += chunk_len;
 
-   if (value.compute_depth() > 1 && !framed)
-      {
-        pad_l(UNI_iPAD_U8, 1);
-        pad_r(UNI_iPAD_U9, 1);
-      }
+              if (interrupt_raised)
+                 {
+                   out << endl << "INTERRUPT" << endl;
+                   attention_raised = false;
+                   interrupt_raised = false;
+                   return;
+                 }
+            }
+       }
 
-   if (!is_rectangular())
-      {
-        Q1(get_height())
-        loop(h, get_height())   CERR << "w=" << get_width(h) << "*" << endl;
-        loop(h, get_height())   CERR << "*" << get_line(h) << "*" << endl;
-      }
-
-   Assert(is_rectangular());
-   add_outer_frame(outer_style);
-
-// debug("PrintBuffer::PrintBuffer(Value_P, ...)");
+   out << endl;
 }
 //-----------------------------------------------------------------------------
 UCS_string
@@ -845,14 +926,15 @@ PrintBuffer::align_dot(ColInfo & COL_INFO)
    else
       {
         // right-align char items
-        size_t LEN = COL_INFO.int_len + COL_INFO.fract_len;
-        size_t len = col_info.int_len + col_info.fract_len;
+        size_t LEN = COL_INFO.total_len();
+        size_t len = col_info.total_len();
         if (LEN > len)
            {
              const size_t diff = LEN - len;
              pad_l(UNI_iPAD_L5, diff);
              col_info.int_len   = COL_INFO.int_len;
              col_info.fract_len = COL_INFO.fract_len;
+             col_info.real_len = COL_INFO.real_len;
            }
       }
 
