@@ -28,57 +28,84 @@
 #include "ProcessorID.hh"
 #include "Quad_SVx.hh"
 #include "Svar_signals.hh"
+#include "UserPreferences.hh"
+
+extern const char * prog_name();
 
 AP_num3 ProcessorID::id(NO_AP, AP_NULL, AP_NULL);
 
 Network_Profile ProcessorID::network_profile;
 
-uint16_t ProcessorID::APnnn_port = 0;
-UdpSocket ProcessorID::APnnn_socket(true, LOC);
-bool ProcessorID::doing_SV = true;
-
 //-----------------------------------------------------------------------------
 bool
-ProcessorID::init(bool do_sv, int proc_id, int par_id)
+ProcessorID::init(bool log_startup)
 {
-   doing_SV = do_sv;
-   id.proc = (AP_num)proc_id;
-   id.parent = par_id ? (AP_num)par_id : AP_NULL;
+   if (log_startup)
+      {
+        CERR << "uprefs.user_do_svars:   " << uprefs.user_do_svars   << endl
+             << "uprefs.system_do_svars: " << uprefs.system_do_svars << endl
+             << "uprefs.requested_id:    " << uprefs.requested_id    << endl
+             << "uprefs.requested_par:   " << uprefs.requested_par   << endl;
+      }
+
+   id.proc = (AP_num)uprefs.requested_id;
+   id.parent = uprefs.requested_par ? (AP_num)uprefs.requested_par : AP_NULL;
    id.grand = AP_NULL;
 
-   if (proc_id == 0)
+   if (!uprefs.system_do_svars)
       {
-        // no --id option in argv: use first free user ID
+        // shared variables are disabled, so Svar_DB is unavailable,
+        // we use id.proc of 1000 if no ID is provided and otherwise
+        // trust the provided ID.
         //
-        id.proc = Svar_DB::get_unused_id();
+        if (id.proc == 0)   id.proc = (AP_num)1000;
+        if (log_startup)   CERR << "id.proc: " << id.proc << " at " LOC << endl;
+        return false;
+      }
+
+   if (id.proc == 0)
+      {
+        // no --id option in argv: use first free user ID. We poll IDs starting
+        // at 1000 and stop at the first ID that is not registered.
+        //
+        for (AP_num proc = AP_FIRST_USER; ; proc = AP_num(proc + 1))
+            {
+              id.proc = proc;
+              if (!Svar_DB::is_registered_id(AP_num(id.proc)))   break;
+            }
+
+        if (log_startup)   CERR << "id.proc: " << id.proc << " at " LOC << endl;
       }
    else
       {
         // --id option provided in argv: check that it is not in use
         //
-        if (!Svar_DB::is_unused_id(AP_num(proc_id)))
+        if (Svar_DB::is_registered_id(AP_num(id.proc)))
            {
-             CERR << _("*** Another APL interpreter with --id ")
-                  << proc_id <<  _(" is already running") << endl;
+             CERR << "*** Another APL interpreter with --id "
+                  << id.proc <<  " is already running" << endl;
  
              return true;
            }
       }
 
-   if (!do_sv)   return false;
-
-   Quad_SVx::start_AP(id.proc, true);
-   APnnn_port = Svar_DB::get_udp_port(id.proc, id.parent);
-   if (APnnn_port == 0)
+   if (log_startup)
       {
-        CERR << "*** Failed to start APnnn: processor " << id.proc
-             << " will not accept incoming shared"
-                " variable offers. Expect surprises." << endl;
-        doing_SV = false;
-        return false;   // no error in order to continue.
+        CERR << "Processor ID was completely initialized: "
+             << id.proc << ":" << id.parent << ":" << id.grand << endl
+             << "system_do_svars is: " << uprefs.system_do_svars << endl;
+
       }
 
-   new (&APnnn_socket) UdpClientSocket(LOC, APnnn_port, 0);
+   // if we have an APserver then let it know our IDs
+   //
+const TCP_socket sock = Svar_DB::get_DB_tcp();
+   if (sock != NO_TCP_SOCKET)
+      {
+        string progname(prog_name());
+        REGISTER_PROCESSOR_c request(sock, id.proc, id.parent, id.grand,
+                             0, progname);
+      }
 
    return false;   // no error
 }
@@ -319,10 +346,6 @@ const char * loc = 0;
 void
 ProcessorID::disconnect()
 {
-   if (APnnn_socket.is_valid())
-      {
-        DISCONNECT_c signal(APnnn_socket);
-        APnnn_socket.udp_close();
-      }
+   Svar_DB::DB_tcp_error(0, 0, 0);
 }
 //-----------------------------------------------------------------------------
