@@ -21,26 +21,29 @@
 #include <stdlib.h>
 
 #include "Quad_RL.hh"
-#include "Value.hh"
+#include "Value.icc"
 #include "IntCell.hh"
 
+uint64_t Quad_RL::state = 0;
+
 //=============================================================================
 Quad_RL::Quad_RL()
    : SystemVariable(ID_Quad_RL)
 {
 Value_P value(new Value(LOC));
 
-   new (&value->get_ravel(0)) IntCell(1);
+const unsigned int seed = reset_seed();
+
+   new (value->next_ravel()) IntCell(seed);
    value->check_value(LOC);
 
-   srandom(1);
    Symbol::assign(value, LOC);
 }
 //-----------------------------------------------------------------------------
 void
 Quad_RL::assign(Value_P value, const char * loc)
 {
-   if (!value->is_skalar_or_len1_vector())
+   if (!value->is_scalar_or_len1_vector())
       {
         if (value->get_rank() > 1)   RANK_ERROR;
         else                         LENGTH_ERROR;
@@ -53,15 +56,31 @@ const APL_Integer val = cell.get_int_value();
    Symbol::assign(value, LOC);
 }
 //-----------------------------------------------------------------------------
-APL_Integer
+uint64_t
 Quad_RL::get_random()
 {
-const APL_Integer r1 = random();
-const APL_Integer seed = random();
+   state *= Knuth_a;
+   state += Knuth_c;
+
    Assert(value_stack.size());
    if (value_stack.back().name_class != NC_VARIABLE)   VALUE_ERROR;
 
-   new (&value_stack.back().apl_val->get_ravel(0))   IntCell(seed);
-   return seed << 32 | r1;
+   new (&value_stack.back().apl_val->get_ravel(0))   IntCell(state);
+   return state;
+}
+//-----------------------------------------------------------------------------
+void
+Quad_RL::push()
+{
+   // clone the current state
+   //
+   Symbol::push_value(IntScalar(state, LOC));
+}
+//-----------------------------------------------------------------------------
+void
+Quad_RL::pop()
+{
+   Symbol::pop();
+   state = value_stack.back().apl_val->get_ravel(0).get_int_value();
 }
 //=============================================================================
