@@ -34,6 +34,7 @@
 #include "Quad_SVx.hh"
 #include "Svar_DB.hh"
 #include "Svar_signals.hh"
+#include "UserPreferences.hh"
 #include "Workspace.hh"
 
 extern char **environ;
@@ -50,6 +51,10 @@ Quad_SVS Quad_SVS::fun;
 
 APL_time_us Quad_SVE::timer_end = 0;
 
+TCP_socket get_TCP_for_key(SV_key key)
+{
+   return Svar_DB::get_DB_tcp();
+}
 //=============================================================================
 /**
     return true iff \b filename is an executable file
@@ -65,56 +70,71 @@ const char * end = strchr(file_and_args, ' ');
 }
 //-----------------------------------------------------------------------------
 void
-Quad_SVx::start_AP(AP_num proc, bool startup)
+Quad_SVx::start_AP(AP_num ap)
 {
+   if (!uprefs.system_do_svars)   // something went wrong
+      {
+        if (uprefs.user_do_svars)   // user wanted APs
+           {
+             // user wanted APs, but something went wrong
+             //
+             CERR << "*** Not starting AP << ap because the connecton to "
+                     " APserver has failed earlier." << endl;
+           }
+        else                        // user gave --noSV
+           {
+             CERR << "*** Not starting AP " << ap
+                  << " because --noSV (or equivalent) was given." << endl
+                  << " That conflicts with the use of ⎕SVxxx functions "
+                     "and variables." << endl;
+           }
+        ATTENTION;
+      }
+
 const char * dirs[] = { "", "/APs" };
    enum { dircount = sizeof(dirs) / sizeof(*dirs) };
 
 const AP_num own_ID = ProcessorID::get_own_ID();
 const AP_num par_ID = ProcessorID::get_parent_ID();
 
-   // start_AP() is called from two places:
-   // on start-up of the interpreter (and then startup == true), or
-   // after making an offer to an AP that is not yet running
-   //
+const char * verbose = "";
+   Log(LOG_shared_variables)   verbose = " -v";
+
+char filename[PATH_MAX + 1];
+
+bool found_executable = false;
    for (int d = 0; d < dircount; ++d)
        {
-         char filename[PATH_MAX + 1];
-         const char * verbose = "";
-         Log(LOG_shared_variables)   verbose = " -v";
-         if (startup)
-            snprintf(filename, PATH_MAX,
-                     "%s%s/APnnn --id %u --par %u%s --ppid %u",
-                     LibPaths::get_APL_bin_path(), dirs[d],
-                     own_ID, par_ID, verbose, getpid());
-         else
-            snprintf(filename, PATH_MAX,
-                     "%s%s/AP%u --id %u --par %u --gra %u --auto%s --ppid %u",
-                     LibPaths::get_APL_bin_path(), dirs[d], proc, proc,
-                     own_ID, par_ID, verbose, getpid());
+         snprintf(filename, PATH_MAX,
+                  "%s%s/AP%u --id %u --par %u --gra %u --auto%s",
+                  LibPaths::get_APL_bin_path(), dirs[d], ap, ap,
+                  own_ID, par_ID, verbose);
 
-         if (!is_executable(filename))   continue;
-
-         Log(LOG_shared_variables)
-            CERR << "found executable: " << filename << endl;
-
-         FILE * fp = popen(filename, "r");
-         if (fp == 0)
-            {
-              CERR << "popen(" << filename << " failed: " << strerror(errno)
-                   << endl;
-              continue;
-            }
-
-         for (int cc; (cc = getc(fp)) != EOF;)   CERR << (char)cc;
-         CERR << endl;
-
-         pclose(fp);
-         return;
+         found_executable = is_executable(filename);
+         if (found_executable)   break;
        }
 
-   CERR << "No binary found for AP " << proc << " (interpreter path = "
-        << LibPaths::get_APL_bin_path() << ")" << endl;
+   if (!found_executable)
+      {
+        CERR << "No binary found for AP " << ap << " (interpreter path = "
+             << LibPaths::get_APL_bin_path() << ")" << endl;
+        return;
+      }
+
+FILE * fp = popen(filename, "r");
+   if (fp == 0)
+      {
+        CERR << "popen(" << filename << " failed: " << strerror(errno)
+             << endl;
+        return;
+      }
+
+   for (int cc; (cc = getc(fp)) != EOF;)   CERR << (char)cc;
+   CERR << endl;
+
+   pclose(fp);
+
+   usleep(100000);   // give new AP time to register with APserver
 }
 //=============================================================================
 Token
@@ -140,7 +160,7 @@ Value_P Z(var_count > 1 ? new Value(shZ, LOC) : new Value(4, LOC));
         // construct control value
         //
         int ctl = 0;
-        if (A->is_skalar())
+        if (A->is_scalar())
            {
              const bool val = A->get_ravel(0).get_near_bool(qct);
               ctl = val ? ALL_SVAR_CONTROLS : NO_SVAR_CONTROL;
@@ -166,7 +186,8 @@ Value_P Z(var_count > 1 ? new Value(shZ, LOC) : new Value(4, LOC));
         if (sym == 0)   throw_symbol_error(vars[z], LOC);
 
         const SV_key key = sym->get_SV_key();
-        ctl = Svar_DB::set_control(key, Svar_Control(ctl));
+        Svar_DB::set_control(key, Svar_Control(ctl));
+        ctl = Svar_DB::get_control(key);
 
         new (Z->next_ravel())   IntCell(ctl & SET_BY_1 ? 1 : 0);
         new (Z->next_ravel())   IntCell(ctl & SET_BY_2 ? 1 : 0);
@@ -211,7 +232,7 @@ Value_P Z(var_count > 1 ? new Value(shZ, LOC) : new Value(4, LOC));
 Quad_SVE::Quad_SVE()
    : NL_SystemVariable(ID_Quad_SVE)
 {
-   Symbol::assign(Value::Zero_P, LOC);
+   Symbol::assign(IntScalar(0, LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 void
@@ -220,7 +241,7 @@ Quad_SVE::assign(Value_P value, const char * loc)
    // ⎕SVE←X merely remembers the timer expiration time. The timer is only
    // started if ⎕SVE is referenced (in Quad_SVE::get_apl_value())
    //
-   if (!value->is_skalar())   RANK_ERROR;
+   if (!value->is_scalar())   RANK_ERROR;
 
 const APL_time_us duration = 1000000 * value->get_ravel(0).get_real_value();
    if (duration < 0)   DOMAIN_ERROR;
@@ -242,65 +263,31 @@ const APL_time_us duration = 1000000 * value->get_ravel(0).get_real_value();
 Value_P
 Quad_SVE::get_apl_value() const
 {
-const APL_time_us current_time = now();
-const APL_time_us wait = timer_end - current_time;
+   for (;;)
+       {
+         Svar_event events = SVE_NO_EVENTS;
+         SV_key key = Svar_DB::get_events(events, ProcessorID::get_id());
 
-   // if timer has expired, return 0
-   //
-   if (wait <= 0)
-      {
-        Svar_DB::clear_all_events();
-        return Value::Zero_P;
-      }
+          if (key || (events != SVE_NO_EVENTS))   break; // something happend
 
-   // At this point the timer is still running.
-   // we use a new UDP port for event reporting and communicate it to APnnn
-   //
-bool got_event = false;
-   {
-     UdpClientSocket event_sock(LOC, ProcessorID::get_APnnn_port());
-     Log(LOG_shared_variables)   event_sock.set_debug(CERR);
+         if (timer_end < now())   // timer expired: return 0
+            {
+              Svar_DB::clear_all_events(ProcessorID::get_id());
+              return IntScalar(0, LOC);
+            }
 
-     {
-       START_EVENT_REPORTING_c start(event_sock, event_sock.get_local_port());
-     }
-
-     uint8_t buf[Signal_base::get_class_size()];
-     const Signal_base * resp = Signal_base::recv(event_sock, buf, wait / 1000);
-
-     if (resp)   // got event
-        {
-          got_event = true;
-          Log(LOG_shared_variables)
-             {
-               const uint64_t key = resp->get__GOT_EVENT__key();
-               const offered_SVAR * svar = Svar_DB::find_var(SV_key(key));
-               CERR << "⎕SVE got event '"
-                    << event_name(Svar_event(resp->get__GOT_EVENT__event()))
-                    << "' from shared variable ";
-               if (svar)   svar->print_name(CERR);
-               else        CERR << "unknown key " << key;
-               CERR << endl;
-             }
-        }
-
-     STOP_EVENT_REPORTING_c stop(event_sock);
-     event_sock.udp_close();
-   }
+          usleep(50000);  // wait 50 ms
+       }
 
    // referencing ⎕SVE always clears all events
    //
-   Svar_DB::clear_all_events();
-
-   // if no event has occurred (timeout) then return 0
-   //
-  if (!got_event)   return Value::Zero_P;
+   Svar_DB::clear_all_events(ProcessorID::get_id());
 
    // we have got an event; return remaining time.
    //
 const APL_Float remaining = timer_end - now();
 
-   if (remaining < 0)   return Value::Zero_P;
+   if (remaining < 0)   return IntScalar(0, LOC);
 
 Value_P Z(new Value(LOC));
    new (&Z->get_ravel(0))   FloatCell(0.000001 * remaining);
@@ -329,7 +316,7 @@ Value_P Z(var_count > 1 ? new Value(var_count, LOC) : new Value(LOC));
 
    loop(z, var_count)
       {
-        ShapeItem a = A->is_skalar() ? 0 : z;
+        ShapeItem a = A->is_scalar() ? 0 : z;
         uint32_t vname[MAX_SVAR_NAMELEN + 1];
         const int vlen = surrogates[z].size();
         if (vlen > MAX_SVAR_NAMELEN)   DOMAIN_ERROR;
@@ -404,24 +391,21 @@ bool auto_started = false;
 
    if (to_ap && to_ap < AP_FIRST_USER)   // to_proc is an AP
       {
-        if (!Svar_DB::is_registered(to_proc))
+        if (!Svar_DB::is_registered_id(to_proc))
            {
-             start_AP(to_ap, false);
+             start_AP(to_ap);
              auto_started = true;
            }
       }
    else                         // to_proc is an independent processor
       {
-        new (&to_proc)   AP_num3(to_ap, AP_NULL, AP_NULL);
+        new (&to_proc)   AP_num3(to_ap);
       }
 
-Svar_partner from;
-   from.clear();
-   from.id = ProcessorID::get_id();
-   from.pid = getpid();
-   from.port = ProcessorID::get_APnnn_port();
+Svar_partner from(ProcessorID::get_id(), NO_TCP_SOCKET);
 
-offered_SVAR * svar = Svar_DB::match_or_make(vname, to_proc, from, coupling);
+const SV_key key = Svar_DB::match_or_make(vname, to_proc, from);
+   coupling = Svar_DB::get_coupling(key);
 
    if (coupling == SV_COUPLED)
       {
@@ -432,7 +416,7 @@ offered_SVAR * svar = Svar_DB::match_or_make(vname, to_proc, from, coupling);
         //
         if (auto_started)   coupling = SV_OFFERED;
         
-        return svar->key;
+        return key;
       }
 
    // if something went wrong, complain.
@@ -446,17 +430,7 @@ offered_SVAR * svar = Svar_DB::match_or_make(vname, to_proc, from, coupling);
 
    Assert(coupling == SV_OFFERED);
 
-   // if the coupling partner was registered, then a signal exchange between
-   // APnnn and the partner is taking place. Wait a little until the signal
-   // exchange is finished, so that the next ⎕SVO will return SV_COUPLED.
-   //
-   for (int t = 0; t < 20; ++t)   // wait at most 20 milliseconds
-       {
-         if (!(svar->offering.flags & OSV_OFFER_SENT))   break;
-         usleep(1000);   // wait 1 ms
-       }
-
-   return svar->key;
+   return key;
 }
 //-----------------------------------------------------------------------------
 /**
@@ -534,7 +508,7 @@ Quad_SVQ::get_processors()
    //
    // 2. running processors that have offered variables in Svar_DB.
    //
-vector<int32_t> processors;
+vector<AP_num> processors;
 
    // case 1...
    //
@@ -548,10 +522,18 @@ const char * dirs[] = { "", "/APs" };
                   dirs[d]);
          dirname[PATH_MAX] = 0;
          DIR * dir = opendir(dirname);
+
+         // the APs directory only exists below the src directory (i.e when
+         // apl is started locally without installing it. We do not complain
+         // if the APs directory does not exist.
+         //
          if (dir == 0)
             {
-              CERR << "Could not open " << dirname << " : "
-                   << strerror(errno) << endl;
+              if (*dirs[d] == 0)
+                 {
+                   CERR << "Could not open directory " << dirname << " : "
+                        << strerror(errno) << " at " << LOC << endl;
+                 }
               continue;
             }
 
@@ -560,7 +542,9 @@ const char * dirs[] = { "", "/APs" };
                 dirent * entry = readdir(dir);
                 if (entry == 0)   break;   // directory done
 
+#ifdef _DIRENT_HAVE_D_TYPE
                 if (entry->d_type != DT_REG)   continue; // not a regular file
+#endif
 
                 char filename[PATH_MAX + 1];
                 snprintf(filename, PATH_MAX, "%s/%s", dirname, entry->d_name);
@@ -575,7 +559,7 @@ const char * dirs[] = { "", "/APs" };
                 snprintf(expected, sizeof(expected), "AP%u", apnum);
                 if (strcmp(entry->d_name, expected))   continue;
 
-                processors.push_back(apnum);
+                processors.push_back((AP_num)apnum);
              }
 
          closedir(dir);
@@ -583,7 +567,7 @@ const char * dirs[] = { "", "/APs" };
 
    // case 2...
    //
-   Svar_DB::get_processors(ProcessorID::get_own_ID(), processors);
+   Svar_DB::get_offering_processors(ProcessorID::get_own_ID(), processors);
 
    // sort and remove duplicates
    //
@@ -624,32 +608,47 @@ Value_P Z(new Value(sorted.size(), LOC));
 Value_P
 Quad_SVQ::get_variables(AP_num proc)
 {
-vector<const uint32_t *> vars;
-   Svar_DB::get_variables(ProcessorID::get_own_ID(), proc, vars);
+vector<uint32_t> varnames;
+   Svar_DB::get_offered_variables(ProcessorID::get_own_ID(), proc, varnames);
 
-const int count = vars.size();
-vector<int> varlen;
+   // varnames is a sequence of 0-terminated Unicodes
+   //
+vector<int> var_lengths;   // including terminating 0
+int last_zero = -1;
+   loop(v, varnames.size())
+      {
+        if (varnames[v] == 0)
+           {
+             var_lengths.push_back(v + 1 - last_zero);
+             last_zero = v;
+           }
+        
+      }
+
+const int count = var_lengths.size();
 ShapeItem max_len = 0;
    loop(z, count)
-      {
-        int len = 0;
-        const uint32_t * vz = vars[z];
-        for (; *vz; ++vz)   ++len;
-        varlen.push_back(len);
-         if (max_len < len)   max_len = len;
-      }
+       if (max_len < (var_lengths[z] - 1))   max_len = var_lengths[z] - 1;
 
 const Shape shZ(count, max_len);
 Value_P Z(new Value(shZ, LOC));
+int v = 0;
 
    loop(z, count)
    loop(c, max_len)
       {
-        if (c < varlen[z])   new (Z->next_ravel()) CharCell(Unicode(vars[z][c]));
-        else                 new (Z->next_ravel()) CharCell(UNI_ASCII_SPACE);
+        if (c < var_lengths[z])
+           {
+             if (varnames[v] == 0)   ++v;
+             new (Z->next_ravel()) CharCell(Unicode(varnames[v++]));
+           }
+        else
+           {
+             new (Z->next_ravel()) CharCell(UNI_ASCII_SPACE);
+           }
       }
 
-   Z->set_default(*Value::Spc_P);   // prototype: character
+   Z->set_default_Spc();   // prototype: character
    return Z;
 }
 //=============================================================================
