@@ -36,7 +36,7 @@ Quad_TF  Quad_TF::fun;
 Token
 Quad_TF::eval_AB(Value_P A, Value_P B)
 {
-   // A should be an integer skalar or 1-element_vector with value 1 or 2
+   // A should be an integer scalar or 1-element_vector with value 1 or 2
    //
    if (A->get_rank() > 0)         RANK_ERROR;
    if (A->element_count() != 1)   LENGTH_ERROR;
@@ -93,7 +93,7 @@ Value_P
 Quad_TF::tf1(const UCS_string & name)
 {
 NamedObject * obj = Workspace::lookup_existing_name(name);
-   if (obj == 0)   return Value::Str0_P;
+   if (obj == 0)   return Str0(LOC);
 
 const Function * function = obj->get_function();
 
@@ -102,7 +102,7 @@ const Function * function = obj->get_function();
         if (obj->is_user_defined())   return tf1(name, *function);
 
         // quad function or primitive: return ''
-        return Value::Str0_P;
+        return Str0(LOC);
       }
 
 Symbol * symbol = obj->get_symbol();
@@ -114,7 +114,7 @@ Symbol * symbol = obj->get_symbol();
         /* not a variable: fall through */
       }
 
-   return Value::Str0_P;
+   return Str0(LOC);
 }
 //-----------------------------------------------------------------------------
 Value_P
@@ -144,7 +144,7 @@ const ShapeItem ec = val->element_count();
              const Cell & cell = val->get_ravel(e);
              if (!cell.is_character_cell())
                 {
-                  return  Value::Str0_P;
+                  return  Str0(LOC);
                 }
 
              ucs.append(cell.get_char_value());
@@ -152,7 +152,6 @@ const ShapeItem ec = val->element_count();
       }
    else   // number
       {
-        const APL_Float qct = Workspace::get_CT();
         loop(e, ec)
            {
              ucs.append(UNI_ASCII_SPACE);
@@ -164,7 +163,7 @@ const ShapeItem ec = val->element_count();
                 }
              else // if (cell.is_near_real(qct))
                 {
-                  PrintContext pctx(PR_APL_MIN, MAX_Quad_PP, 0.0, MAX_Quad_PW);
+                  PrintContext pctx(PR_APL_MIN, MAX_Quad_PP, MAX_Quad_PW);
                   const APL_Float value = cell.get_real_value();
                   bool scaled = true;
                   UCS_string ucs1(value, scaled, pctx);
@@ -341,7 +340,7 @@ Token
 Quad_TF::tf2(const UCS_string & name)
 {
 NamedObject * obj = Workspace::lookup_existing_name(name);
-   if (obj == 0)   return Token(TOK_APL_VALUE1, Value::Str0_P);
+   if (obj == 0)   return Token(TOK_APL_VALUE1, Str0(LOC));
 
 const Function * function = obj->get_function();
 
@@ -356,7 +355,7 @@ const Function * function = obj->get_function();
            }
 
         // quad function or primitive: return ''
-        return Token(TOK_APL_VALUE1, Value::Str0_P);
+        return Token(TOK_APL_VALUE1, Str0(LOC));
       }
 
 Symbol * symbol = obj->get_symbol();
@@ -368,7 +367,7 @@ Symbol * symbol = obj->get_symbol();
         /* not a variable: fall through */
       }
 
-   return Token(TOK_APL_VALUE1, Value::Str0_P);
+   return Token(TOK_APL_VALUE1, Str0(LOC));
 }
 //-----------------------------------------------------------------------------
 Token
@@ -377,11 +376,11 @@ Quad_TF::tf2_var(const UCS_string & var_name, Value_P value)
 UCS_string ucs(var_name);
    ucs.append(UNI_LEFT_ARROW);
 
-const bool error = tf2_ravel(0, ucs, value);
+const bool error = tf2_value(0, ucs, value);
 
    // return '' for invalid values.
    //
-   if (error)   return Token(TOK_APL_VALUE1, Value::Str0_P);
+   if (error)   return Token(TOK_APL_VALUE1, Str0(LOC));
 
 Value_P Z(new Value(ucs, LOC));
    Z->check_value(LOC);
@@ -440,6 +439,25 @@ Token_string tos;
 
    // Try ⎕FX fun-text
    //
+   if (tos.size() == 3)   // native function
+      {
+        if (tos[0].get_Class() != TC_VALUE)    return UCS_string();
+        if (tos[1].get_tag() != TOK_Quad_FX)   return UCS_string();
+        if (tos[2].get_Class() != TC_VALUE)    return UCS_string();
+
+        Value_P fname   =  tos[0].get_apl_val();
+        Value_P so_path =  tos[2].get_apl_val();
+
+        const Token tok = Quad_FX::fun.eval_AB(so_path, fname);
+        if (tok.get_Class() == TC_VALUE)   // ⎕FX successful
+           {
+             Value_P val = tok.get_apl_val();
+             return UCS_string(*val.get());
+           }
+
+        return UCS_string();
+      }
+
    if (tos.size() != 2)                   return UCS_string();
    if (tos[0].get_tag() != TOK_Quad_FX)   return UCS_string();
    if (tos[1].get_Class() != TC_VALUE)    return UCS_string();
@@ -458,12 +476,14 @@ Token_string tos;
    return UCS_string();
 }
 //-----------------------------------------------------------------------------
-void
+bool
 Quad_TF::tf2_shape(UCS_string & ucs, const Shape & shape)
 {
-   // dont print anything if shape is a skalar or non-empty vector
-   if (shape.get_volume() != 0 && shape.get_rank() <=1)   return;
+   // dont print anything if shape is a scalar or non-empty vector
+   //
+   if (shape.get_volume() != 0 && shape.get_rank() <= 1)   return false;
 
+   ucs.append(UNI_ASCII_L_PARENT);
    loop(r, shape.get_rank())
       {
         if (r)   ucs.append(UNI_ASCII_SPACE);
@@ -471,10 +491,11 @@ Quad_TF::tf2_shape(UCS_string & ucs, const Shape & shape)
       }
 
    ucs.append(UNI_RHO);
+   return true;
 }
 //-----------------------------------------------------------------------------
 bool
-Quad_TF::tf2_ravel(int level, UCS_string & ucs, Value_P value)
+Quad_TF::tf2_value(int level, UCS_string & ucs, Value_P value)
 {
 const ShapeItem ec = value->element_count();
 
@@ -500,12 +521,12 @@ const ShapeItem ec = value->element_count();
            }
 
         // other empty values: print ( shape ⍴ prototype )
-        ucs.append(UNI_ASCII_L_PARENT);
 
-        tf2_shape(ucs, value->get_shape());
+        const bool rho_used = tf2_shape(ucs, value->get_shape());
+        Assert(rho_used);
 
         Value_P proto = value->prototype(LOC);
-        tf2_ravel(level + 1, ucs, proto);
+        tf2_value(level + 1, ucs, proto);
         ucs.append(UNI_ASCII_R_PARENT);
         return  false;
       }
@@ -514,7 +535,7 @@ const ShapeItem ec = value->element_count();
    //
    if (!value->NOTCHAR())
       {
-        tf2_shape(ucs, value->get_shape());
+        const bool rho_used = tf2_shape(ucs, value->get_shape());
 
         // check if ⎕UCS is needed. 
         //
@@ -556,15 +577,23 @@ const ShapeItem ec = value->element_count();
                 }
              ucs.append(UNI_SINGLE_QUOTE);
            }
+
+        if (rho_used) ucs.append(UNI_ASCII_R_PARENT);
         return  false;
       }
 
 const Depth depth = value->compute_depth();
-   if (depth && level)   ucs.append(UNI_ASCII_L_PARENT);
+bool need_parenth = depth && level;
 
    // maybe print shape followed by ⍴
    //
-   tf2_shape(ucs, value->get_shape());
+bool have_parenth = tf2_shape(ucs, value->get_shape());
+
+   if (need_parenth && !have_parenth)
+      {
+        have_parenth = true;
+        ucs.append(UNI_ASCII_L_PARENT);
+      }
 
    loop(e, ec)
       {
@@ -573,7 +602,7 @@ const Depth depth = value->compute_depth();
 
         if (cell.is_pointer_cell())
            {
-             if (tf2_ravel(level + 1, ucs, cell.get_pointer_value()))
+             if (tf2_value(level + 1, ucs, cell.get_pointer_value()))
                 return true;
            }
         else if (cell.is_lval_cell())
@@ -582,17 +611,17 @@ const Depth depth = value->compute_depth();
            }
         else if (cell.is_complex_cell())
            {
-             PrintContext pctx(PR_APL_MIN, MAX_Quad_PP, 0.0, MAX_Quad_PW);
+             PrintContext pctx(PR_APL_MIN, MAX_Quad_PP, MAX_Quad_PW);
              bool scaled = true;
              UCS_string ucs1(cell.get_real_value(), scaled, pctx);
              UCS_string ucs2(cell.get_imag_value(), scaled, pctx);
              ucs.append(ucs1);
              ucs.append(UNI_ASCII_J);
-             ucs.append(ucs1);
+             ucs.append(ucs2);
            }
         else if (cell.is_float_cell())
            {
-             PrintContext pctx(PR_APL_MIN, MAX_Quad_PP, 0.0, MAX_Quad_PW);
+             PrintContext pctx(PR_APL_MIN, MAX_Quad_PP, MAX_Quad_PW);
              bool scaled = true;
              UCS_string ucs1(cell.get_real_value(), scaled, pctx);
              ucs.append(ucs1);
@@ -606,7 +635,7 @@ const Depth depth = value->compute_depth();
            }
       }
 
-   if (depth && level)   ucs.append(UNI_ASCII_R_PARENT);
+   if (have_parenth)   ucs.append(UNI_ASCII_R_PARENT);
 
    return false;
 }
@@ -774,6 +803,22 @@ Quad_TF::tf2_fun_ucs(UCS_string & ucs, const UCS_string & fun_name,
                     const Function & fun)
 {
 const UCS_string text = fun.canonical(false);
+
+   if (fun.is_native())
+      {
+        CERR << "Warning: the workspace contains a native function '"
+             << fun_name << "', making the" << endl
+             << "   .atf output file incompatible with other APL interpreters."
+             << endl;
+
+        ucs.append_utf8("'");
+        ucs.append(fun_name);
+        ucs.append_utf8("' ⎕FX '");
+        ucs.append(text);
+        ucs.append_utf8("'");
+        return;
+      }
+
 vector<UCS_string> lines;
    text.to_vector(lines);
 
@@ -861,20 +906,20 @@ Value_P
 Quad_TF::tf3(const UCS_string & name)
 {
 NamedObject * obj = Workspace::lookup_existing_name(name);
-   if (obj == 0)   return Value::Str0_P;
+   if (obj == 0)   return Str0(LOC);
 
    if (obj->get_function())
       {
         // 3⎕TF is not defined for functions.
         //
-        return Value::Str0_P;
+        return Str0(LOC);
       }
 
 Symbol * symbol = obj->get_symbol();
    if (symbol)
       {
-        Value_P value = symbol->get_apl_value();
-        if (!!value)
+        const Value & value = *symbol->get_apl_value();
+        if (&value)
            {
              CDR_string cdr;
              CDR::to_CDR(cdr, value);
@@ -885,6 +930,6 @@ Symbol * symbol = obj->get_symbol();
         /* not a variable: fall through */
       }
 
-   return Value::Str0_P;
+   return Str0(LOC);
 }
 //-----------------------------------------------------------------------------
