@@ -27,11 +27,12 @@
 #include "Output.hh"
 #include "Prefix.hh"
 #include "StateIndicator.hh"
+#include "SystemLimits.hh"
 #include "UserFunction.hh"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
-StateIndicator::StateIndicator(const Executable * exec, StateIndicator * _par)
+StateIndicator::StateIndicator(Executable * exec, StateIndicator * _par)
    : executable(exec),
      safe_execution(false),
      eoc_handler(0),
@@ -39,7 +40,6 @@ StateIndicator::StateIndicator(const Executable * exec, StateIndicator * _par)
      level(_par ? 1 + _par->get_level() : 0),
      error(E_NO_ERROR, LOC),
      current_stack(*this, exec->get_body()),
-     eval_arg_F(0),
      parent(_par)
 {
 }
@@ -51,13 +51,15 @@ StateIndicator::~StateIndicator()
    //
    current_stack.cleanup();
 
-   // some args could come from the body, so we erase them BEFORE the body.
+   // if executable is a user defined function then pop its local vars.
+   // otherwise delete the body token
    //
-   clear_args(LOC);
-
-   // delete the body token unless executable is a user defined function
-   //
-   if ((executable->get_parse_mode() != PM_FUNCTION))
+   if (executable->get_parse_mode() == PM_FUNCTION)
+      {
+        const UserFunction * ufun = get_executable()->get_ufun();
+        if (ufun)   ufun->pop_local_vars();
+      }
+   else
       {
          Assert1(executable);
          delete executable;
@@ -66,62 +68,6 @@ StateIndicator::~StateIndicator()
 }
 //-----------------------------------------------------------------------------
 void
-StateIndicator::clear_args(const char * loc)
-{
-   // discard duplicate args
-   //
-   if (eval_arg_B == eval_arg_A)   ptr_clear(eval_arg_A, loc);
-   if (eval_arg_B == eval_arg_X)   ptr_clear(eval_arg_X, loc);
-   if (eval_arg_A == eval_arg_X)   ptr_clear(eval_arg_X, loc);
-
-   if (!!eval_arg_A)
-      {
-        ptr_clear(eval_arg_A, loc);
-      }
-
-   if (!!eval_arg_B)
-      {
-        ptr_clear(eval_arg_B, loc);
-      }
-
-   if (!!eval_arg_X)
-      {
-        ptr_clear(eval_arg_X, loc);
-      }
-
-   eval_arg_F = 0;
-}
-//-----------------------------------------------------------------------------
-void
-StateIndicator::set_args(Token * A, Token & F, Token * X, Token * B)
-{
-   if (A)
-      {
-        eval_arg_A = A->get_apl_val();
-      }
-
-   if (X)
-      {
-        // X can be a single value (which is OK) or an IndexExpr (which is not)
-        //
-        if (X->get_ValueType() != TV_VAL)
-           {
-             IndexExpr & idx = X->get_index_val();
-             idx.extract_all();   // clear all values in X
-             SYNTAX_ERROR;
-           }
-        eval_arg_X = X->get_axes();
-      }
-
-   if (B)
-      {
-        eval_arg_B = B->get_apl_val();
-      }
-
-   eval_arg_F = F.get_function();
-}
-//-----------------------------------------------------------------------------
-void
 StateIndicator::goon(Function_Line new_line, const char * loc)
 {
 const Function_PC pc = get_executable()->get_ufun()->pc_for_line(new_line);
@@ -130,8 +76,19 @@ const Function_PC pc = get_executable()->get_ufun()->pc_for_line(new_line);
       CERR << "Continue SI[" << level << "] at line " << new_line
            << " pc=" << pc << " at " << loc << endl;
 
-   set_PC(pc);
-   Log(LOG_prefix_parser)   CERR << "GOTO " << get_PC() << endl;
+   if (get_executable()->get_body()[pc].get_tag() == TOK_STOP_LINE)   // S∆
+      {
+        // pc points to a S∆ token. We are jumping back from immediate
+        // execution, so we don't want to stop again.
+        //
+        set_PC(pc + 2);
+      }
+   else
+      {
+        set_PC(pc);
+      }
+
+   Log(LOG_prefix_parser)   CERR << "GOTO [" << get_line() << "]" << endl;
 
    current_stack.reset(LOC);
 }
@@ -139,64 +96,8 @@ const Function_PC pc = get_executable()->get_ufun()->pc_for_line(new_line);
 void
 StateIndicator::retry(const char * loc)
 {
-   CERR << endl << "RETRY" << endl << endl;
-
-   if (eval_arg_F == 0)
-      {
-        CERR << endl << "Can't RETRY: no function" << endl << endl;
-        return;
-      }
-
-   if (!eval_arg_B)
-      {
-        CERR << endl << "Can't RETRY: no right argument" << endl << endl;
-        return;
-      }
-
-Token Z;
-Token B(TOK_APL_VALUE1, eval_arg_B);
-int phrase_len = 1;
-
-   if (!!eval_arg_A)   // dyadic
-      {
-        Token A(TOK_APL_VALUE1, eval_arg_A);
-        Token F2(TOK_FUN2, eval_arg_F);
-
-        if (!!eval_arg_X)   // dyadic with axis
-           {
-             Token X(TOK_APL_VALUE1, eval_arg_X);
-             move_2(Z, eval_AXB(A, F2, X, B), LOC);
-             phrase_len = 4;
-           }
-        else              // dyadic without axis
-           {
-             move_2(Z, eval_AB(A, F2, B), LOC);
-             phrase_len = 3;
-           }
-      }
-   else              // monadic
-      {
-        Token F1(TOK_FUN1, eval_arg_F);
-
-        if (!!eval_arg_X)   // monadic with axis
-           {
-             Token X(TOK_APL_VALUE1, eval_arg_X);
-             move_2(Z, eval_XB(F1, X, B), LOC);
-             phrase_len = 3;
-           }
-        else              // monadic without axis
-           {
-             move_2(Z, eval_B(F1, B), LOC);
-             phrase_len = 2;
-           }
-      }
-
-   if (Z.get_Class() == TC_VALUE)   // retry successful
-      {
-        Token_loc tl_Z(Z, Function_PC(get_PC() - 1));
-        clear_args(LOC);
-        current_stack.push(tl_Z);
-      }
+   Log(LOG_StateIndicator__push_pop || LOG_prefix_parser)
+      CERR << endl << "RETRY " << loc << ")" << endl;
 }
 //-----------------------------------------------------------------------------
 UCS_string
@@ -232,6 +133,7 @@ StateIndicator::print(ostream & out) const
 {
    out << "Depth:    " << level << endl;
    out << "Exec:     " << executable << endl;
+   out << "Safe ex:  " << (safe_execution ? "yes" : "no") << endl;
 
    Assert(executable);
 
@@ -270,16 +172,6 @@ StateIndicator::print(ostream & out) const
 }
 //-----------------------------------------------------------------------------
 void
-StateIndicator::clear(ostream & out)
-{
-const UserFunction * ufun = get_executable()->get_ufun();
-   if (ufun)
-      {
-         Value_P Z = ufun->pop_local_vars();
-      }
-}
-//-----------------------------------------------------------------------------
-void
 StateIndicator::list(ostream & out, SI_mode mode) const
 {
    Assert(this);
@@ -406,16 +298,23 @@ const UserFunction * ufun = executable->get_ufun();
         const Function_PC from = current_stack.get_range_low();
         Function_PC to   = current_stack.get_range_high();
         const Function_PC2 error_range(from, to);
-        // current_stack.print_range(CERR);
-        // CERR  << endl;
         executable->set_error_info(err, error_range);
       }
 
    // print error, unless we are in safe execution mode.
    //
-   if (!safe_execution)   err.print_em(UERR, LOC);
-
-   Workspace::update_EM_ET(err);   // update ⎕EM and ⎕ET
+   {
+     bool print_error = true;
+     for (const StateIndicator * si = this; si; si = si->get_parent())
+         {
+           if (si->get_safe_execution())
+              {
+                print_error = false;
+                break;
+              }
+         }
+     if (print_error)   err.print_em(UERR, LOC);
+   }
 
    error = err;
 }
@@ -452,6 +351,7 @@ StateIndicator::jump(Value_P value)
    if (value->element_count() == 0)     // →''
       {
         // →'' in immediate execution means resume (retry) suspended function
+        //
         if (get_executable()->get_parse_mode() == PM_STATEMENT_LIST)
            return Token(TOK_BRANCH, int64_t(Function_Retry));
 
@@ -476,10 +376,10 @@ const UserFunction * ufun = get_executable()->get_ufun();
 void
 StateIndicator::escape()  
 {
-const UserFunction * ufun = get_executable()->get_ufun();
-   if (ufun)
+   if (eoc_handler)
       {
-        Value_P Z = ufun->pop_local_vars();
+        Token tok(TOK_ESCAPE);
+        eoc_handler(tok, eoc_arg);
       }
 }
 //-----------------------------------------------------------------------------
@@ -497,10 +397,6 @@ Token result = current_stack.reduce_statements();
 void
 StateIndicator::unmark_all_values() const
 {
-   if (!!eval_arg_A)   eval_arg_A->unmark();
-   if (!!eval_arg_X)   eval_arg_X->unmark();
-   if (!!eval_arg_B)   eval_arg_B->unmark();
-
    Assert(executable);
    executable->unmark_all_values();
 
@@ -512,15 +408,6 @@ StateIndicator::show_owners(ostream & out, const Value & value) const
 {
 int count = 0;
 
-   if (&value == eval_arg_A.get())
-      { out << "    SI[" << level << "] eval_arg_A" << endl;   ++count; }
-
-   if (&value == eval_arg_X.get())
-      { out << "    SI[" << level << "] eval_arg_X" << endl;   ++count; }
-
-   if (&value == eval_arg_B.get())
-      { out << "    SI[" << level << "] eval_arg_B" << endl;   ++count; }
-
    Assert(executable);
 char cc[100];
    snprintf(cc, sizeof(cc), "    SI[%d] ", level);
@@ -542,349 +429,57 @@ StateIndicator::info(ostream & out, const char * loc) const
 }
 //-----------------------------------------------------------------------------
 Value_P
-StateIndicator::get_L() const
+StateIndicator::get_L()
 {
-   if (eval_arg_F == 0)                 return Value_P();   // no function
-   if (eval_arg_F->is_user_defined())   return Value_P();   // user defined function
-   return eval_arg_A;
+Token * tok_L = current_stack.locate_L();
+   if (tok_L)   return tok_L->get_apl_val();
+   return Value_P();
+}
+//-----------------------------------------------------------------------------
+Value_P
+StateIndicator::get_R()
+{
+Token * tok_R = current_stack.locate_R();
+   if (tok_R)   return tok_R->get_apl_val();
+   return Value_P();
+}
+//-----------------------------------------------------------------------------
+Value_P
+StateIndicator::get_X()
+{
+Token * tok_X = current_stack.locate_X();
+   if (tok_X)   return tok_X->get_apl_val();
+   return Value_P();
 }
 //-----------------------------------------------------------------------------
 void
 StateIndicator::set_L(Value_P new_value)
 {
-Value_P old_value = eval_arg_A;
+Token * tok_L = current_stack.locate_L();
+   if (tok_L == 0)   return;
 
-   eval_arg_A = new_value;
-}
-//-----------------------------------------------------------------------------
-Value_P
-StateIndicator::get_X() const
-{
-   if (eval_arg_F == 0)                 return Value_P();   // no function
-   if (eval_arg_F->is_user_defined())   return Value_P();   // user defined function
-   return eval_arg_X;
-}
-//-----------------------------------------------------------------------------
-void
-StateIndicator::set_X(Value_P new_value)
-{
-Value_P old_value = eval_arg_X;
-
-   eval_arg_X = new_value;
-}
-//-----------------------------------------------------------------------------
-Value_P
-StateIndicator::get_R() const
-{
-   if (eval_arg_F == 0)                 return Value_P();   // no function
-   if (eval_arg_F->is_user_defined())   return Value_P();   // user defined function
-   return eval_arg_B;
+Value_P old_value = tok_L->get_apl_val();   // so that 
+   move_2(*tok_L, Token(tok_L->get_tag(), new_value), LOC);
 }
 //-----------------------------------------------------------------------------
 void
 StateIndicator::set_R(Value_P new_value)
 {
-Value_P old_value = eval_arg_B;
+Token * tok_R = current_stack.locate_R();
+   if (tok_R == 0)   return;
 
-   eval_arg_B = new_value;
-}
-//-----------------------------------------------------------------------------
-bool
-StateIndicator::replace_arg(Value_P old_value, Value_P new_value)
-{
-   current_stack.replace_AB(old_value, new_value);
-
-bool success = false;
-   if (old_value == eval_arg_A)
-      {
-        eval_arg_A = new_value;
-        success = true;
-      }
-
-   if (old_value == eval_arg_B)
-      {
-        eval_arg_B = new_value;
-        success = true;
-      }
-
-   if (old_value == eval_arg_X)
-      {
-        eval_arg_X = new_value;
-        success = true;
-      }
-
-   return success;
+Value_P old_value = tok_R->get_apl_val();   // so that 
+   move_2(*tok_R, Token(tok_R->get_tag(), new_value), LOC);
 }
 //-----------------------------------------------------------------------------
 void
-StateIndicator::recover_from_error(const char * loc)
+StateIndicator::set_X(Value_P new_value)
 {
-   Value::finish_incomplete(loc);
-}
-//-----------------------------------------------------------------------------
-/// a macro for the common part of all StateIndicator::eval_XXX() calls
-/// catch exceptions and check for errors,
-#define FINISH_EVAL \
-   catch (Error e)                              \
-      {                                         \
-       recover_from_error(LOC);                 \
-       return Token(TOK_ERROR, e.error_code);   \
-      }                                         \
-   catch (...)                                  \
-      {                                         \
-        FIXME;                                  \
-      }                                         \
-                                                \
-   return Token(TOK_ERROR, error.error_code);
+Token * tok_X = current_stack.locate_X();
+   if (tok_X == 0)   return;
 
-//-----------------------------------------------------------------------------
-Token
-StateIndicator::eval_(Token & fun)
-{
-   set_args(0, fun, 0, 0);
-
-   try
-      {
-        const Token result = eval_arg_F->eval_();
-
-        if (result.get_tag() != TOK_ERROR)
-           {
-             clear_args(LOC);
-             return result;
-           }
-      }
-
-   FINISH_EVAL
-}
-//-----------------------------------------------------------------------------
-Token
-StateIndicator::eval_B(Token & fun, Token & B)
-{
-   set_args(0, fun, 0, &B);
-
-   try
-      {
-        const Token result = eval_arg_F->eval_B(eval_arg_B);
-
-        if (result.get_tag() != TOK_ERROR)
-           {
-             clear_args(LOC);
-             return result;
-           }
-      }
-
-   FINISH_EVAL
-}
-//-----------------------------------------------------------------------------
-Token
-StateIndicator::eval_AB(Token & A, Token & fun, Token & B)
-{
-   set_args(&A, fun, 0, &B);
-
-   try
-      {
-        const Token result = eval_arg_F->eval_AB(eval_arg_A, eval_arg_B);
-
-        if (result.get_tag() != TOK_ERROR)
-           {
-             clear_args(LOC);
-             return result;
-           }
-      }
-
-   FINISH_EVAL
-}
-//-----------------------------------------------------------------------------
-Token
-StateIndicator::eval_XB(Token & fun, Token & X, Token & B)
-{
-   set_args(0, fun, &X, &B);
-
-   try
-      {
-        const Token result = eval_arg_F->eval_XB(eval_arg_X,
-                                                 eval_arg_B);
-        if (result.get_tag() != TOK_ERROR)
-           {
-             clear_args(LOC);
-             return result;
-           }
-      }
-
-   FINISH_EVAL
-}
-//-----------------------------------------------------------------------------
-Token
-StateIndicator::eval_AXB(Token & A, Token & fun, Token & X, Token & B)
-{
-   set_args(&A, fun, &X, &B);
-
-   try
-      {
-        const Token result = eval_arg_F->eval_AXB(eval_arg_A,
-                                                  eval_arg_X,
-                                                  eval_arg_B);
-        if (result.get_tag() != TOK_ERROR)
-           {
-             clear_args(LOC);
-             return result;
-           }
-      }
-
-   FINISH_EVAL
-}
-//-----------------------------------------------------------------------------
-Token
-StateIndicator::eval_LB(Token & LO, Token & oper, Token & B)
-{
-   set_args(0, oper, 0, &B);
-
-   try
-      {
-        const Token result = eval_arg_F->eval_LB(LO, eval_arg_B);
-        if (result.get_tag() != TOK_ERROR)
-           {
-             clear_args(LOC);
-             return result;
-           }
-      }
-
-   FINISH_EVAL
-}
-//-----------------------------------------------------------------------------
-Token
-StateIndicator::eval_ALB(Token & A, Token & LO, Token & oper, Token & B)
-{
-   set_args(&A, oper, 0, &B);
-
-   try
-      {
-        const Token result = eval_arg_F->eval_ALB(eval_arg_A, LO, eval_arg_B);
-        if (result.get_tag() != TOK_ERROR)
-           {
-             clear_args(LOC);
-             return result;
-           }
-      }
-
-   FINISH_EVAL
-}
-//-----------------------------------------------------------------------------
-Token
-StateIndicator::eval_LXB(Token & LO, Token & oper, Token & X, Token & B)
-{
-   set_args(0, oper, &X, &B);
-
-   try
-      {
-        const Token result = eval_arg_F->eval_LXB(LO, eval_arg_X, eval_arg_B);
-        if (result.get_tag() != TOK_ERROR)
-           {
-             clear_args(LOC);
-             return result;
-           }
-      }
-
-   FINISH_EVAL
-}
-//-----------------------------------------------------------------------------
-Token
-StateIndicator::eval_LRB(Token & LO, Token & oper, Token & RO, Token & B)
-{
-   set_args(0, oper, 0, &B);
-
-   try
-      {
-        const Token result = eval_arg_F->eval_LRB(LO, RO, eval_arg_B);
-        if (result.get_tag() != TOK_ERROR)
-           {
-             clear_args(LOC);
-             return result;
-           }
-      }
-
-   FINISH_EVAL
-}
-//-----------------------------------------------------------------------------
-Token
-StateIndicator::eval_ALXB(Token & A, Token & LO, Token & oper,
-                          Token & X, Token & B)
-{
-   set_args(&A, oper, &X, &B);
-
-   try
-      {
-        const Token result = eval_arg_F->eval_ALXB(eval_arg_A, LO, eval_arg_X,
-                                                   eval_arg_B);
-        if (result.get_tag() != TOK_ERROR)
-           {
-             clear_args(LOC);
-             return result;
-           }
-      }
-
-   FINISH_EVAL
-}
-//-----------------------------------------------------------------------------
-Token
-StateIndicator::eval_ALRB(Token & A, Token & LO, Token & oper,
-                          Token & RO, Token & B)
-{
-   set_args(&A, oper, 0, &B);
-
-   try
-      {
-        const Token result = eval_arg_F->eval_ALRB(eval_arg_A, LO, RO,
-                                                   eval_arg_B);
-        if (result.get_tag() != TOK_ERROR)
-           {
-             clear_args(LOC);
-             return result;
-           }
-      }
-
-   FINISH_EVAL
-}
-//-----------------------------------------------------------------------------
-Token
-StateIndicator::eval_LRXB(Token & LO, Token & oper,
-                          Token & X, Token & RO, Token & B)
-{
-   set_args(0, oper, &X, &B);
-
-   try
-      {
-        const Token result = eval_arg_F->eval_LRXB(LO, RO, eval_arg_X,
-                                                   eval_arg_B);
-        if (result.get_tag() != TOK_ERROR)
-           {
-             clear_args(LOC);
-             return result;
-           }
-      }
-
-   FINISH_EVAL
-}
-//-----------------------------------------------------------------------------
-Token
-StateIndicator::eval_ALRXB(Token & A, Token & LO, Token & oper,
-                           Token & X, Token & RO, Token & B)
-{
-   set_args(&A, oper, &X, &B);
-
-   try
-      {
-        const Token result = eval_arg_F->eval_ALRXB(eval_arg_A, LO, RO,
-                                                    eval_arg_X,
-                                                    eval_arg_B);
-        if (result.get_tag() != TOK_ERROR)
-           {
-             clear_args(LOC);
-             return result;
-           }
-      }
-
-   FINISH_EVAL
+Value_P old_value = tok_X->get_apl_val();   // so that 
+   move_2(*tok_X, Token(tok_X->get_tag(), new_value), LOC);
 }
 //-----------------------------------------------------------------------------
 Function_Line
@@ -894,12 +489,23 @@ StateIndicator::get_line() const
 }
 //-----------------------------------------------------------------------------
 void
-StateIndicator::statement_result(Token & result)
+StateIndicator::statement_result(Token & result, bool trace)
 {
    Log(LOG_StateIndicator__enter_leave)
       CERR << "StateIndicator::statement_result(pmode="
            << get_parse_mode_name() << ", result=" << result << endl;
 
+   if (trace)
+      {
+        const UserFunction * ufun = executable->get_ufun();
+        if (ufun && (ufun->get_exec_properties()[0] == 0))
+           {
+             const Function_Line line =
+                            executable->get_line((Function_PC)(get_PC() - 1));
+             result.show_trace(COUT, ufun->get_name(), line);
+           }
+      }
+
    fun_oper_cache.reset();
 
 // if (get_executable()->get_parse_mode() == PM_EXECUTE)   return;
@@ -915,20 +521,51 @@ Value_P B(result.get_apl_val());
 
    // print TOK_APL_VALUE and TOK_APL_VALUE1, but not TOK_APL_VALUE2
    //
-   if (tag == TOK_APL_VALUE1 || tag == TOK_APL_VALUE3)
+   if (tag != TOK_APL_VALUE1 && tag != TOK_APL_VALUE3)   return;
+
+   Quad_QUOTE::done(false, LOC);
+
+const int boxing_format = Command::get_boxing_format();
+   if (boxing_format == -1)
       {
-        Quad_QUOTE::done(false, LOC);
-        const int boxing_format = Command::get_boxing_format();
-        if (boxing_format == -1)
+        if (Quad_SYL::print_length_limit &&
+            B->element_count() >= Quad_SYL::print_length_limit)
            {
-             B->print(COUT);
+             // the value exceeds print_length_limit.
+             // We cut the longest dimension in half until we are below the
+             // limit
+             //
+             Shape sh(B->get_shape());
+             while (sh.get_volume() >= Quad_SYL::print_length_limit)
+                {
+                  Rank longest = 0;
+                  loop(r, sh.get_rank())
+                     {
+                       if (sh.get_shape_item(r) > sh.get_shape_item(longest))
+                          longest = r;
+                     }
+
+                  sh.set_shape_item(longest, sh.get_shape_item(longest) / 2);
+                }
+
+             Value_P B1 = Bif_F12_TAKE::do_take(sh, B).get_apl_val();
+             B1->print(COUT);
+
+             CERR << "      *** display of value was truncated (limit "
+                     "⎕SYL[⎕IO + " << Quad_SYL::SYL_PRINT_LIMIT
+                  << "] reached)  ***" << endl;
            }
         else
            {
-             Value_P B1 = Quad_CR::do_CR(boxing_format, *B.get());
-             B1->print(COUT);
+             B->print(COUT);
            }
       }
+   else
+      {
+        const PrintContext pctx = Workspace::get_PrintContext(PST_NONE);
+        Value_P B1 = Quad_CR::do_CR(boxing_format, *B.get(), pctx);
+        B1->print(COUT);
+      }
 }
 //-----------------------------------------------------------------------------
 Unicode
