@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2014  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -35,9 +35,8 @@
 #include "Symbol.hh"
 #include "Svar_signals.hh"
 #include "SystemVariable.hh"
-#include "UdpSocket.hh"
 #include "UserFunction.hh"
-#include "Value.hh"
+#include "Value.icc"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
@@ -61,8 +60,7 @@ ostream &
 Symbol::print_verbose(ostream & out) const
 {
    out << "Symbol ";
-   print(out);
-   out << endl;
+   print(out) << " " << (const void *)this << endl;
 
    loop(v, value_stack.size())
        {
@@ -96,12 +94,19 @@ Symbol::print_verbose(ostream & out) const
                    {
                      Function * fun = item.sym_val.function;
                      Assert(fun);
-                     fun->print_properties(out, 8);
-                      out << endl;
-                   }
-                   continue;
-            }
 
+                     fun->print_properties(out, 4);
+                     out << "    ⎕NC:            " << item.name_class
+                         << endl
+                         << "    addr:           " << (const void *)fun
+                         << endl;
+
+                     out << endl;
+                   }
+                   break;
+
+              default: break;
+            }
        }
 
    return out;
@@ -138,103 +143,17 @@ ValueStackItem & vs = value_stack.back();
 
              new_value = new_value->clone(loc);
 
-             // un-assign and erase old value
-             //
-             Workspace::replace_arg(vs.apl_val, new_value);
-
              vs.apl_val = new_value;
              if (monitor_callback)   monitor_callback(*this, SEV_ASSIGNED);
              return;
 
         case NC_SHARED_VAR:
-             {
-               CDR_string cdr;
-               CDR::to_CDR(cdr, new_value);
-               if (cdr.size() > MAX_SVAR_SIZE)   LIMIT_ERROR_SVAR;
-               string data((const char *)cdr.get_items(), cdr.size());
-               const SV_key key = get_SV_key();
-
-               // wait for shared variable to be ready
-               //
-               offered_SVAR * svar =  Svar_DB::find_var(get_SV_key());
-               if (svar == 0)
-                  {
-                    Log(LOG_shared_variables)
-                       CERR << " Svar_DB::find_var() failed." << endl;
-
-                    Workspace::more_error() =
-                       UCS_string("Svar_DB corrupt or erased?");
-                    VALUE_ERROR;
-                  }
-
-               for (int w = 0; ; ++w)
-                   {
-                     if (svar->may_set(w))   // ready for writing
-                        {
-                          if (w)
-                             {
-                               Log(LOG_shared_variables)
-                                  CERR << " - OK." << endl;
-                             }
-                          break;
-                         }
-
-                     if (w == 0)
-                        {
-                          Log(LOG_shared_variables)
-                             {
-                               CERR << "Shared variable ";
-                               svar->print_name(CERR);
-                               CERR << " is blocked on set. Waiting ...";
-                             }
-                        }
-                     else if (w%25 == 0)
-                        {
-                          Log(LOG_shared_variables)   CERR << ".";
-                        }
-
-                     usleep(10000);   // wait 10 ms
-                   }
-
-               // update shared var state (BEFORE sending request to peer)
-               svar->set_state(false, loc);
-               {
-                 UdpClientSocket sock(loc, svar->data_owner_port());
-                 ASSIGN_VALUE_c request(sock, get_SV_key(), data);
-
-                 uint8_t buffer[Signal_base::get_class_size()];
-                 const Signal_base * response =
-                                     Signal_base::recv(sock, buffer, 10000);
-
-                 if (response == 0)
-                    {
-                      cerr << "TIMEOUT on signal ASSIGN_VALUE_c" << endl;
-                      VALUE_ERROR;
-                    }
-
-                 const ErrorCode ec =
-                                 ErrorCode(response->get__ASSIGNED__error());
-                 if (ec)
-                    {
-                      Log(LOG_shared_variables)
-                         {
-                           Error e(ec,
-                                  response->get__ASSIGNED__error_loc().c_str());
-                           cerr << Error::error_name(ec) << " assigning "
-                                << get_name() << ", detected at "
-                                << response->get__ASSIGNED__error_loc()
-                                << endl;
-                         }
-
-                      throw_apl_error(ec, loc);
-                    }
-               }
-             }
+             assign_shared_variable(new_value, loc);
              if (monitor_callback)   monitor_callback(*this, SEV_ASSIGNED);
              return;
+
+        default: SYNTAX_ERROR;
       }
-             
-   SYNTAX_ERROR;
 }
 //-----------------------------------------------------------------------------
 void
@@ -265,7 +184,7 @@ const ShapeItem ec_B = B->element_count();
 const ShapeItem ec_X = X->element_count();
 const APL_Integer qio = Workspace::get_IO();
 const APL_Float qct = Workspace::get_CT();
-const int incr_B = (ec_B == 1) ? 0 : 1;   // maybe skalar extend B
+const int incr_B = (ec_B == 1) ? 0 : 1;   // maybe scalar extend B
 const Cell * cX = &X->get_ravel(0);
 const Cell * cB = &B->get_ravel(0);
 
@@ -300,16 +219,16 @@ Symbol::assign_indexed(IndexExpr & IX, Value_P B)   // A[IX;...] ← B
 Value_P A = get_apl_value();
    if (A->get_rank() != IX.value_count())   INDEX_ERROR;
 
-   // B must either B a skalar (and is then skalar extended to the size
+   // B must either be a scalar (and is then scalar extended to the size
    // of the updated area, or else have the shape of the concatenated index
    // items for example:
    //
    //  X:   X1    ; X2    ; X3
    //  ⍴B:  b1 b2   b3 b4   b5 b6
    //
-   if (1 && !B->is_skalar())
+   if (1 && !B->is_scalar())
       {
-        // remove dimensions with len 1 from X's and B's shapes...
+        // remove dimensions with len 1 from the shapes of X and B...
         // if we see an empty Xn then we return.
         //
         Shape B1;
@@ -320,7 +239,7 @@ Value_P A = get_apl_value();
            }
 
         Shape IX1;
-        loop(ix, IX.value_count())
+        for (ShapeItem ix = IX.value_count() - 1; ix >= 0; --ix)
             {
               const Value * ix_val = IX.values[ix].get();
               if (ix_val)   // normal index
@@ -357,7 +276,8 @@ const int incr_B = (ec_B == 1) ? 0 : 1;
 
    while (!mult.done())
       {
-        Cell & dest = A->get_ravel(mult.next());
+        const ShapeItem offset_A = mult.next();
+        Cell & dest = A->get_ravel(offset_A);
         dest.release(LOC);   // free sub-values etc (if any)
         dest.init(*cB);
         cB += incr_B;
@@ -366,8 +286,28 @@ const int incr_B = (ec_B == 1) ? 0 : 1;
    if (monitor_callback)   monitor_callback(*this, SEV_ASSIGNED);
 }
 //-----------------------------------------------------------------------------
-Value_P
-Symbol::pop(bool erase)
+void
+Symbol::assign_named_lambda(Function * lambda, const char * loc)
+{
+ValueStackItem & vs = value_stack.back();
+
+   switch(vs.name_class)
+      {
+        case NC_UNUSED_USER_NAME:
+        case NC_FUNCTION:
+        case NC_OPERATOR:
+
+             vs.name_class = NC_FUNCTION;
+             vs.sym_val.function = lambda;
+             if (monitor_callback)   monitor_callback(*this, SEV_ASSIGNED);
+             return;
+
+        default: SYNTAX_ERROR;
+      }
+}
+//-----------------------------------------------------------------------------
+void
+Symbol::pop()
 {
    Assert(value_stack.size());
 const ValueStackItem & vs = value_stack.back();
@@ -377,17 +317,15 @@ const ValueStackItem & vs = value_stack.back();
         Value_P ret = vs.apl_val;
         Log(LOG_SYMBOL_push_pop)
            {
-             CERR << "-pop-value " << symbol;
-             if (erase)   CERR  << " (and erase) ";
-             CERR << " flags " << ret->get_flags() << " ";
+             CERR << "-pop-value " << symbol
+                  << " flags " << ret->get_flags() << " ";
              if (value_stack.size() == 0)   CERR << " (last)";
              CERR << " addr " << (const void *)ret.get() << endl;
            }
-        if (erase)   ptr_clear(ret, LOC);
+        ptr_clear(ret, LOC);
 
         value_stack.pop_back();
         if (monitor_callback)   monitor_callback(*this, SEV_POPED);
-        return ret;
       }
    else
       {
@@ -400,7 +338,6 @@ const ValueStackItem & vs = value_stack.back();
            }
         value_stack.pop_back();
         if (monitor_callback)   monitor_callback(*this, SEV_POPED);
-        return Value_P();
       }
 }
 //-----------------------------------------------------------------------------
@@ -498,16 +435,16 @@ const ValueStackItem & vs = value_stack.back();
    return Value_P();
 }
 //-----------------------------------------------------------------------------
-bool
-Symbol::can_be_defined() const
+const char *
+Symbol::cant_be_defined() const
 {
-   if (value_stack.size() > 1)         return false;   // localized
-   if (Workspace::is_called(symbol))   return false;  // on stack
+// if (value_stack.size() > 1)         return "symbol was localized";
+   if (Workspace::is_called(symbol))   return "function is called";
 
-   if (value_stack.back().name_class == NC_UNUSED_USER_NAME)   return true;
-   if (value_stack.back().name_class == NC_FUNCTION)           return true;
-   if (value_stack.back().name_class == NC_OPERATOR)           return true;
-   return false;
+   if (value_stack.back().name_class == NC_UNUSED_USER_NAME)   return 0;   // OK
+   if (value_stack.back().name_class == NC_FUNCTION)           return 0;   // OK
+   if (value_stack.back().name_class == NC_OPERATOR)           return 0;   // OK
+   return "bad name class";
 }
 //-----------------------------------------------------------------------------
 Value_P
@@ -527,11 +464,9 @@ Symbol::can_be_assigned() const
       {
         case NC_UNUSED_USER_NAME:
         case NC_VARIABLE:
-        case NC_SHARED_VAR:
-             return true;
+        case NC_SHARED_VAR: return true;
+        default:            return false;
       }
-
-   return false;
 }
 //-----------------------------------------------------------------------------
 SV_key
@@ -559,9 +494,8 @@ Symbol::get_function() const
       {
         case NC_FUNCTION:
         case NC_OPERATOR: return value_stack.back().sym_val.function;
+        default:          return 0;
       }
-
-   return 0;
 }
 //-----------------------------------------------------------------------------
 Function *
@@ -579,7 +513,6 @@ void
 Symbol::get_attributes(int mode, Cell * dest) const
 {
 const ValueStackItem & vs = value_stack.back();
-bool is_var = false;
 bool has_result = false;
 int fun_valence = 0;
 int oper_valence = 0;
@@ -591,7 +524,6 @@ const int * exec_properties = exec_prop;
       {
         case NC_LABEL:
         case NC_VARIABLE:
-             is_var = 1;
              has_result = true;
              break;
 
@@ -608,6 +540,8 @@ const int * exec_properties = exec_prop;
              created = vs.sym_val.function->get_creation_time();
              exec_properties = vs.sym_val.function->get_exec_properties();
              break;
+
+        default: break;
       }
 
    switch(mode)
@@ -654,9 +588,9 @@ const int * exec_properties = exec_prop;
 
         case 4: {
                   Value_P val = get_apl_value();
-                  const int CDR_type = val->get_CDR_type();
-                  const int brutto = val->total_size_brutto(CDR_type);
-                  const int data = val->data_size(CDR_type);
+                  const CDR_type cdr_type = val->get_CDR_type();
+                  const int brutto = val->total_size_brutto(cdr_type);
+                  const int data = val->data_size(cdr_type);
 
                   new (dest + 0) IntCell(brutto);
                   new (dest + 1) IntCell(data);
@@ -708,107 +642,8 @@ Symbol::resolve(Token & tok, bool left_sym)
              return;
 
         case NC_SHARED_VAR:
-             {
-               if (left_sym)   return;   // leave symbol as is
-
-               const SV_key key = get_SV_key();
-
-               // wait for shared variable to be ready
-               //
-               offered_SVAR * svar =  Svar_DB::find_var(key);
-               if (svar == 0)
-                  {
-                    Log(LOG_shared_variables)
-                       CERR << " Svar_DB::find_var() failed." << endl;
-
-                    Workspace::more_error() =
-                       UCS_string("Svar_DB corrupt or erased?");
-                    VALUE_ERROR;
-                  }
-
-
-               for (int w = 0; ; ++w)
-                   {
-                     if (svar->may_use(w))   // ready for reading
-                        {
-                          if (w)
-                             {
-                               Log(LOG_shared_variables)
-                                  cerr << " - OK." << endl;
-                             }
-                          break;
-                        }
-
-                     if (w == 0)
-                        {
-                          Log(LOG_shared_variables)
-                             {
-                               cerr << "apl" << ProcessorID::get_id().proc
-                                    << ": Shared variable ";
-                               svar->print_name(cerr);
-                               cerr << " is blocked on use. Waiting ...";
-                             }
-                        }
-                     else if (w%25 == 0)
-                        {
-                          Log(LOG_shared_variables)   cerr << ".";
-                        }
-
-                     usleep(10000);   // wait 10 ms
-                   }
-
-               UdpClientSocket sock(LOC, svar->data_owner_port());
-
-               GET_VALUE_c request(sock, get_SV_key());
-               uint8_t buffer[Signal_base::get_class_size()];
-               const Signal_base * response =
-                                   Signal_base::recv(sock, buffer, 10000);
-
-               if (response == 0)
-                  {
-                    cerr << "TIMEOUT on signal GET_VALUE" << endl;
-                    VALUE_ERROR;
-                  }
-
-               const ErrorCode err(ErrorCode(response->get__VALUE_IS__error()));
-               if (err)
-                  {
-                    Log(LOG_shared_variables)
-                       {
-                         cerr << Error::error_name(err) << " referencing "
-                              << get_name() << ", detected at "
-                              << response->get__VALUE_IS__error_loc() << endl;
-                       }
-
-                    throw_apl_error(err, 
-                                  response->get__VALUE_IS__error_loc().c_str());
-                  }
-
-               CDR_string cdr;
-               const string & data = response->get__VALUE_IS__value();
-               loop(d, data.size())   cdr.append(data[d]);
-               Value_P value = CDR::from_CDR(cdr, LOC);
-
-               if (!value)     VALUE_ERROR;
-
-               // update shared var state (AFTER sending request to peer)
-               {
-                 offered_SVAR * svar =  Svar_DB::find_var(get_SV_key());
-                 if (svar == 0)
-                    {
-                      Log(LOG_shared_variables)
-                         CERR << " Svar_DB::find_var() failed." << endl;
-
-                      Workspace::more_error() =
-                         UCS_string("Svar_DB corrupt or erased?");
-                      VALUE_ERROR;
-                    }
-
-                 svar->set_state(true, LOC);
-               }
-               value->check_value(LOC);
-               new (&tok) Token(TOK_APL_VALUE1, value);
-             }
+             if (left_sym)   return;   // leave symbol as is
+             resolve_shared_variable(tok);
              return;
 
         default:
@@ -867,9 +702,8 @@ Symbol::resolve_class(bool left)
                return (valence == 2) ? TC_OPER2 : TC_OPER1;
              }
 
+        default: return TC_SYMBOL;
       }
-
-   return TC_SYMBOL;
 }
 //-----------------------------------------------------------------------------
 int
@@ -966,12 +800,14 @@ ValueStackItem & vs = value_stack.back();
    if (vs.name_class != NC_SHARED_VAR)   return NO_COUPLING;
 
 const SV_key key = get_SV_key();
-const SV_Coupling coupling = Svar_DB::retract_var(key);
+const SV_Coupling old_coupling = Svar_DB::get_coupling(key);
+
+   Svar_DB::retract_var(key);
 
    set_SV_key(0);
    vs.name_class = NC_UNUSED_USER_NAME;
 
-   return coupling;
+   return old_coupling;
 }
 //-----------------------------------------------------------------------------
 void
@@ -1028,7 +864,7 @@ UCS_string data;
                data.append(UNI_ASCII_A);
                data.append(get_name());
                data.append(UNI_LEFT_ARROW);
-               Quad_TF::tf2_ravel(0, data, value_stack[0].apl_val);
+               Quad_TF::tf2_value(0, data, value_stack[0].apl_val);
              }
              break;
 
@@ -1046,7 +882,7 @@ UCS_string data;
 
                for (char * cp = buffer + strlen(buffer);
                     cp < (buffer + 72); )   *cp++ = ' ';
-                sprintf(buffer + 72, "%8.8lld\r\n", seq++);
+                sprintf(buffer + 72, "%8.8lld\r\n", (long long)(seq++));
                 fwrite(buffer, 1, 82, out);
 
                // write function record(s)
@@ -1069,11 +905,11 @@ UCS_string data;
         loop(uu, 71)
            {
              unsigned char cc = ' ';
-             if (u < data.size()) cc = Command::unicode_to_cp(data[u++]);
+             if (u < data.size()) cc = Avec::unicode_to_cp(data[u++]);
              buffer[1 + uu] = cc;
            }
 
-        sprintf(buffer + 72, "%8.8lld\r\n", seq++);
+        sprintf(buffer + 72, "%8.8lld\r\n", (long long)(seq++));
         fwrite(buffer, 1, 82, out);
       }
 }
@@ -1096,7 +932,11 @@ Symbol::unmark_all_values() const
                    {
                      const UserFunction * ufun =
                                           item.sym_val.function->get_ufun1();
-                     Assert(ufun);
+
+                     // ufun can be 0 for example if F is a function argument
+                     // of a defined operator and F is a primitive function
+                     //
+                     if (ufun == 0)   break;
 
                      const Token_string & body = ufun->get_body();
                      loop(b, body.size())
@@ -1111,6 +951,8 @@ Symbol::unmark_all_values() const
                         }
                    }
                    break;
+
+              default: break;
             }
        }
 }
@@ -1126,7 +968,7 @@ int count = 0;
          switch(item.name_class)
             {
               case NC_VARIABLE:
-                   if (item.apl_val.get() == &value)
+                   if (item.apl_val->is_or_contains(value))
                       {
                          out << "    Variable[vs=" << v << "] "
                             << get_name() << endl;
@@ -1137,15 +979,20 @@ int count = 0;
               case NC_FUNCTION:
               case NC_OPERATOR:
                    {
-                     const Executable * ufun =
-                                          item.sym_val.function->get_ufun1();
-                     Assert(ufun);
-                     char cc[100];
-                     snprintf(cc, sizeof(cc), "    VS[%lld] ", v);
-                     count += ufun->show_owners(cc, out, value);
-
+                     const Function * fun = item.sym_val.function;
+                     const Executable * ufun = fun->get_ufun1();
+                     Assert(ufun || fun->is_native());
+                     if (ufun)
+                        {
+                          char cc[100];
+                          snprintf(cc, sizeof(cc), "    VS[%lld] ",
+                                   (long long)v);
+                          count += ufun->show_owners(cc, out, value);
+                        }
                    }
                    break;
+
+              default: break;
             }
        }
 
@@ -1153,14 +1000,13 @@ int count = 0;
 }
 //-----------------------------------------------------------------------------
 void
-Symbol::vector_assignment(vector<Symbol *> & symbols, Value_P values)
+Symbol::vector_assignment(Symbol * * symbols, int sym_count, Value_P values)
 {
-const int sym_count = symbols.size();
    if (values->get_rank() > 1)   RANK_ERROR;
-   if (!values->is_skalar() &&
+   if (!values->is_scalar() &&
        values->element_count() != sym_count)   LENGTH_ERROR;
 
-const int incr = values->is_skalar() ? 0 : 1;
+const int incr = values->is_scalar() ? 0 : 1;
 const Cell * cV = &values->get_ravel(0);
    loop(s, sym_count)
       {
@@ -1176,7 +1022,7 @@ const Cell * cV = &values->get_ravel(0);
              sym->assign(val, LOC);
            }
 
-        cV += incr;   // skalar extend values
+        cV += incr;   // scalar extend values
       }
 }
 //-----------------------------------------------------------------------------
@@ -1223,7 +1069,9 @@ const ValueStackItem & vs = (*this)[0];
               out << text[u];
               if (text[u] == '\n' && u < text.size() - 1)   out << " ";
            }
-        out << "∇" << endl << endl;
+        if (ufun->get_exec_properties()[0])   out << "⍫";
+        else                                  out << "∇";
+        out << endl << endl;
       }
 }
 //-----------------------------------------------------------------------------
@@ -1232,7 +1080,7 @@ Symbol::clear_vs()
 {
    while (value_stack.size() > 1)
       {
-        Value_P deleted = pop(true);
+        pop();
       }
 
 ValueStackItem & tos = value_stack[0];
@@ -1253,17 +1101,10 @@ ValueStackItem & tos = value_stack[0];
 
         case NC_FUNCTION:
         case NC_OPERATOR:
-             if (tos.sym_val.function->is_native())   break;
-
-             {
-               const UserFunction * ufun = tos.sym_val.function->get_ufun1();
-               Assert(ufun);
-               const Executable * exec = ufun;
-               Assert(!Workspace::oldest_exec(exec));
-             }
-
-             delete tos.sym_val.function;
+             tos.sym_val.function->destroy();
              break;
+
+        default: break;
       }
 }
 //-----------------------------------------------------------------------------
@@ -1273,3 +1114,222 @@ operator <<(ostream & out, const Symbol & sym)
    return sym.print(out);
 }
 //-----------------------------------------------------------------------------
+void
+Symbol::assign_shared_variable(Value_P new_value, const char * loc)
+{
+   // put new_value into a CDR string
+   //
+CDR_string cdr;
+   CDR::to_CDR(cdr, *new_value);
+   if (cdr.size() > MAX_SVAR_SIZE)   LIMIT_ERROR_SVAR;
+
+string data((const char *)cdr.get_items(), cdr.size());
+
+   // wait for shared variable to be ready
+   //
+const bool ws_to_ws = Svar_DB::is_ws_to_ws(get_SV_key());
+
+   for (int w = 0; ; ++w)
+       {
+         if (Svar_DB::may_set(get_SV_key(), w))   // ready for writing
+            {
+              if (w)
+                 {
+                   Log(LOG_shared_variables)
+                      CERR << " - OK." << endl;
+                 }
+              break;
+            }
+
+         if (w == 0)
+            {
+              Log(LOG_shared_variables)
+                 {
+                   CERR << "Shared variable ";
+                   for (const uint32_t * varname =
+                                         Svar_DB::get_varname(get_SV_key());
+                        varname && *varname; ++varname)
+                       CERR << (Unicode)(*varname++);
+                   CERR << " is blocked on set. Waiting ...";
+                 }
+            }
+         else if (w%25 == 0)
+            {
+              Log(LOG_shared_variables)   CERR << ".";
+            }
+
+         usleep(10000);   // wait 10 ms
+       }
+
+const TCP_socket tcp = Svar_DB::get_DB_tcp();
+
+   if (ws_to_ws)
+      {
+        // variable shared between workspaces. It is stored on APserver
+        //
+
+        // update shared var state (BEFORE sending request to peer)
+        //
+        ASSIGN_WSWS_VAR_c(tcp, get_SV_key(), data);
+        return;
+      }
+
+   // update shared var state (BEFORE sending request to peer)
+   //
+   Svar_DB::set_state(get_SV_key(), false, loc);
+
+   ASSIGN_VALUE_c request(tcp, get_SV_key(), data);
+
+char * del = 0;
+char buffer[2*MAX_SIGNAL_CLASS_SIZE];
+const Signal_base * response = Signal_base::recv_TCP(tcp, buffer,
+                                                     sizeof(buffer), del, 0);
+
+   if (response == 0)
+      {
+        cerr << "TIMEOUT on signal ASSIGN_VALUE" << endl;
+        if (del)   delete del;
+        VALUE_ERROR;
+      }
+
+const ErrorCode ec = ErrorCode(response->get__SVAR_ASSIGNED__error());
+   if (ec)
+      {
+        Log(LOG_shared_variables)
+           {
+             Error e(ec, response->get__SVAR_ASSIGNED__error_loc().c_str());
+             cerr << Error::error_name(ec) << " assigning "
+                  << get_name() << ", detected at "
+                  << response->get__SVAR_ASSIGNED__error_loc()
+                  << endl;
+           }
+
+        if (del)   delete del;
+        throw_apl_error(ec, loc);
+      }
+
+   if (del)   delete del;
+}
+//-----------------------------------------------------------------------------
+void
+Symbol::resolve_shared_variable(Token & tok)
+{
+   // wait for shared variable to be ready
+   //
+const bool ws_to_ws = Svar_DB::is_ws_to_ws(get_SV_key());
+
+   for (int w = 0; ; ++w)
+       {
+         if (Svar_DB::may_use(get_SV_key(), w))   // ready for reading
+            {
+              if (w)
+                 {
+                   Log(LOG_shared_variables)   cerr << " - OK." << endl;
+                 }
+              break;
+            }
+
+         if (w == 0)
+            {
+              Log(LOG_shared_variables)
+                 {
+                   CERR << "apl" << ProcessorID::get_id().proc
+                                    << ": Shared variable ";
+                   for (const uint32_t * varname =
+                                         Svar_DB::get_varname(get_SV_key());
+                        varname && *varname; ++varname)
+                       CERR << (Unicode)(*varname++);
+                   CERR << " is blocked on use. Waiting ...";
+                 }
+            }
+         else if (w%25 == 0)
+            {
+              Log(LOG_shared_variables)   cerr << ".";
+            }
+
+         usleep(10000);   // wait 10 ms
+       }
+
+const TCP_socket tcp = Svar_DB::get_DB_tcp();
+
+   if (ws_to_ws)
+      {
+        // variable shared between workspaces. It is stored on APserver
+        //
+        READ_WSWS_VAR_c(tcp, get_SV_key());
+
+        char * del = 0;
+        char buffer[MAX_SIGNAL_CLASS_SIZE + 40000];
+        const Signal_base * response =
+              Signal_base::recv_TCP(tcp, buffer, sizeof(buffer), del, 0);
+
+        if (response == 0)
+           {
+             if (del)   delete del;
+             CERR << "no response to signal READ_WSWS_VAR" << endl;
+             VALUE_ERROR;
+           }
+
+        const string & data = response->get__WSWS_VALUE_IS__cdr_value();
+        if (data.size() == 0)
+           {
+             if (del)   delete del;
+             CERR << "no data in signal WSWS_VALUE_IS" << endl;
+             VALUE_ERROR;
+           }
+
+        CDR_string cdr;
+        loop(d, data.size())   cdr.append(data[d]);
+        if (del)   delete del;
+
+        Value_P value = CDR::from_CDR(cdr, LOC);
+        if (!value)     VALUE_ERROR;
+
+        value->check_value(LOC);
+        new (&tok) Token(TOK_APL_VALUE1, value);
+        return;
+      }
+
+GET_VALUE_c request(tcp, get_SV_key());
+
+char * del = 0;
+char buffer[MAX_SIGNAL_CLASS_SIZE + 40000];
+const Signal_base * response = Signal_base::recv_TCP(tcp, buffer,
+                                                     sizeof(buffer), del, 0);
+
+   if (response == 0)
+      {
+        CERR << "TIMEOUT on signal GET_VALUE" << endl;
+        VALUE_ERROR;
+      }
+
+const ErrorCode err(ErrorCode(response->get__VALUE_IS__error()));
+   if (err)
+      {
+        Log(LOG_shared_variables)
+           {
+             cerr << Error::error_name(err) << " referencing "
+                  << get_name() << ", detected at "
+                  << response->get__VALUE_IS__error_loc() << endl;
+           }
+
+        if (del)   delete del;
+        throw_apl_error(err, response->get__VALUE_IS__error_loc().c_str());
+      }
+
+const string & data = response->get__VALUE_IS__cdr_value();
+CDR_string cdr;
+   loop(d, data.size())   cdr.append(data[d]);
+   if (del)   delete del;
+
+Value_P value = CDR::from_CDR(cdr, LOC);
+   if (!value)     VALUE_ERROR;
+
+   // update shared var state (AFTER sending request to peer)
+   //
+   Svar_DB::set_state(get_SV_key(), true, LOC);
+
+   value->check_value(LOC);
+   new (&tok) Token(TOK_APL_VALUE1, value);
+}
+//-----------------------------------------------------------------------------
