@@ -2,7 +2,7 @@
     This file is part of GNU APL, a free implementation of the
     ISO/IEC Standard 13751, "Programming Language APL, Extended"
 
-    Copyright (C) 2008-2013  Dr. Jürgen Sauermann
+    Copyright (C) 2008-2014  Dr. Jürgen Sauermann
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,7 +18,6 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-#include <langinfo.h>
 #include <stdint.h>
 #include <stdlib.h>
 #include <sys/resource.h>   // for ⎕WA
@@ -29,6 +28,7 @@
 # include <utmpx.h>           // for ⎕UL
 #endif
 
+#include "Bif_F12_FORMAT.hh"
 #include "CDR.hh"
 #include "CharCell.hh"
 #include "Common.hh"
@@ -42,7 +42,8 @@
 #include "StateIndicator.hh"
 #include "SystemVariable.hh"
 #include "UserFunction.hh"
-#include "Value.hh"
+#include "UserPreferences.hh"
+#include "Value.icc"
 #include "Workspace.hh"
 
 UCS_string Quad_QUOTE::prompt;
@@ -50,9 +51,9 @@ UCS_string Quad_QUOTE::prompt;
 ShapeItem Quad_SYL::si_depth_limit = 0;
 ShapeItem Quad_SYL::value_count_limit = 0;
 ShapeItem Quad_SYL::ravel_count_limit = 0;
+ShapeItem Quad_SYL::print_length_limit = 0;
 
-int Quad_ARG::argc = 0;
-const char ** Quad_ARG::argv = 0;
+Unicode Quad_AV::qav[MAX_AV];
 
 //=============================================================================
 void
@@ -107,9 +108,9 @@ SystemVariable::get_attributes(int mode, Cell * dest) const
 
         case 4: {
                   Value_P val = get_apl_value();
-                  const int CDR_type = val->get_CDR_type();
-                  const int brutto = val->total_size_brutto(CDR_type);
-                  const int data = val->data_size(CDR_type);
+                  const CDR_type cdr_type = val->get_CDR_type();
+                  const int brutto = val->total_size_brutto(cdr_type);
+                  const int data = val->data_size(cdr_type);
 
                   new (dest + 0) IntCell(brutto);
                   new (dest + 1) IntCell(data);
@@ -124,13 +125,24 @@ SystemVariable::get_attributes(int mode, Cell * dest) const
 Quad_AV::Quad_AV()
    : RO_SystemVariable(ID_Quad_AV)
 {
-   Symbol::assign(Value::AV_P, LOC);
+Value_P AV(new Value(MAX_AV, LOC));
+   loop(cti, MAX_AV)
+       {
+         const int av_pos = Avec::get_av_pos(CHT_Index(cti));
+         const Unicode uni = Avec::unicode(CHT_Index(cti));
+
+         qav[av_pos] = uni;   // remember unicode for indexed_at()
+         new (&AV->get_ravel(av_pos))  CharCell(uni);
+       }
+   AV->check_value(LOC);
+
+   Symbol::assign(AV, LOC);
 }
 //-----------------------------------------------------------------------------
 Unicode
 Quad_AV::indexed_at(uint32_t pos)
 {
-   if (pos < MAX_AV)   return Value::AV_P->get_ravel(pos).get_char_value();
+   if (pos < MAX_AV)   return qav[pos];
    return UNI_AV_MAX;
 }
 //=============================================================================
@@ -166,15 +178,17 @@ Quad_ARG::Quad_ARG()
 Value_P
 Quad_ARG::get_apl_value() const
 {
+const int argc = uprefs.expanded_argv.size();
+
 Value_P Z(new Value(argc, LOC));
 Cell * C = &Z->get_ravel(0);
 
-   loop(a, Quad_ARG::argc)
+   loop(a, argc)
       {
-        const char * arg = Quad_ARG::argv[a];
+        const char * arg = uprefs.expanded_argv[a];
         const int len = strlen(arg);
         Value_P val(new Value(len, LOC));
-        loop(l, len)   new (&val->get_ravel(l))   CharCell(Unicode(arg[l]));
+        loop(l, len)   new (val->next_ravel())   CharCell(Unicode(arg[l]));
 
         new (C++)   PointerCell(val);
       }
@@ -184,21 +198,15 @@ Cell * C = &Z->get_ravel(0);
 }
 //=============================================================================
 Quad_CT::Quad_CT()
-   : SystemVariable(ID_Quad_CT),
-     current_ct(DEFAULT_Quad_CT)
+   : SystemVariable(ID_Quad_CT)
 {
-Value_P value(new Value(LOC));
-   new (&value->get_ravel(0)) FloatCell(current_ct);
-
-   value->check_value(LOC);
-
-   Symbol::assign(value, LOC);
+   Symbol::assign(FloatScalar(DEFAULT_Quad_CT, LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 void
 Quad_CT::assign(Value_P value, const char * loc)
 {
-   if (!value->is_skalar_or_len1_vector())
+   if (!value->is_scalar_or_len1_vector())
       {
         if (value->get_rank() > 1)   RANK_ERROR;
         else                         LENGTH_ERROR;
@@ -214,28 +222,45 @@ APL_Float val = cell.get_real_value();
    // APL2 "discourages" the use of ⎕CT > 1E¯9.
    // We round down to MAX_Quad_CT (= 1E¯9) instead.
    //
-   if (val > Value::Max_CT_P->get_ravel(0).get_real_value())
+   if (val > MAX_Quad_CT)
       {
-        value = Value::Max_CT_P;
+        Symbol::assign(FloatScalar(MAX_Quad_CT, LOC), LOC);
+      }
+   else
+      {
+        Symbol::assign(value, LOC);
       }
-
-   Symbol::assign(value, LOC);
-   current_ct = value->get_ravel(0).get_real_value();
 }
 //=============================================================================
 Value_P
 Quad_EM::get_apl_value() const
 {
-   if (error.error_code == E_NO_ERROR)
+const Error * err = 0;
+
+   for (StateIndicator * si = Workspace::SI_top(); si; si = si->get_parent())
+       {
+         if (si->get_error().error_code != E_NO_ERROR)
+            {
+              err = &si->get_error();
+              break;
+            }
+         if (si->get_executable()->get_parse_mode() == PM_FUNCTION)   break;
+       }
+
+   if (err == 0)   // no SI entry with non-zero error code
       {
         // return 3 0⍴' '
         //
-        return Value::Str3_0_P;
+        Shape sh((ShapeItem)3, (ShapeItem)0);
+        Value_P Z(new Value(sh, LOC));
+        new (&Z->get_ravel(0))   CharCell(UNI_ASCII_SPACE);
+        Z->check_value(LOC);
+        return Z;
       }
 
-const UCS_string msg_1 = error.get_error_line_1();
-const UCS_string msg_2 = error.get_error_line_2();
-const UCS_string msg_3 = error.get_error_line_3();
+const UCS_string msg_1 = err->get_error_line_1();
+const UCS_string msg_2 = err->get_error_line_2();
+const UCS_string msg_3 = err->get_error_line_3();
 
    // compute max line length of the 3 error lines,
    //
@@ -272,10 +297,24 @@ Value_P Z(new Value(sh, LOC));
 Value_P
 Quad_ET::get_apl_value() const
 {
+StateIndicator * si = Workspace::SI_top();
+ErrorCode ec = E_NO_ERROR;
+
+   for (; si; si = si->get_parent())
+       {
+         ec = si->get_error().error_code;
+         if (ec != E_NO_ERROR)
+            {
+              break;
+            }
+
+         if (si->get_executable()->get_parse_mode() == PM_FUNCTION)   break;
+       }
+
 Value_P Z(new Value(2, LOC));
 
-   new (&Z->get_ravel(0)) IntCell(Error::error_major(error.error_code));
-   new (&Z->get_ravel(1)) IntCell(Error::error_minor(error.error_code));
+   new (&Z->get_ravel(0)) IntCell(Error::error_major(ec));
+   new (&Z->get_ravel(1)) IntCell(Error::error_minor(ec));
 
    Z->check_value(LOC);
    return Z;
@@ -283,32 +322,39 @@ Value_P Z(new Value(2, LOC));
 //=============================================================================
 Quad_FC::Quad_FC() : SystemVariable(ID_Quad_FC)
 {
-   set_default();
+Value_P QFC(new Value(6, LOC));
+   new (QFC->next_ravel()) CharCell(UNI_ASCII_FULLSTOP);
+   new (QFC->next_ravel()) CharCell(UNI_ASCII_COMMA);
+   new (QFC->next_ravel()) CharCell(UNI_STAR_OPERATOR);
+   new (QFC->next_ravel()) CharCell(UNI_ASCII_0);
+   new (QFC->next_ravel()) CharCell(UNI_ASCII_UNDERSCORE);
+   new (QFC->next_ravel()) CharCell(UNI_OVERBAR);
+   QFC->check_value(LOC);
 
-Value_P value(new Value(6, LOC));
-
-   loop(c, 6)   new (&value->get_ravel(c))  CharCell(current_fc[c]);
-
-   value->check_value(LOC);
-
-   Symbol::assign(value, LOC);
+   Symbol::assign(QFC, LOC);
 }
 //-----------------------------------------------------------------------------
 void
-Quad_FC::set_default()
+Quad_FC::push()
 {
-   current_fc[0] = UNI_ASCII_FULLSTOP;
-   current_fc[1] = UNI_ASCII_COMMA;
-   current_fc[2] = UNI_STAR_OPERATOR;
-   current_fc[3] = UNI_ASCII_0;
-   current_fc[4] = UNI_ASCII_UNDERSCORE;
-   current_fc[5] = UNI_OVERBAR;
+   Symbol::push();
+
+Value_P QFC(new Value(6, LOC));
+   new (QFC->next_ravel()) CharCell(UNI_ASCII_FULLSTOP);
+   new (QFC->next_ravel()) CharCell(UNI_ASCII_COMMA);
+   new (QFC->next_ravel()) CharCell(UNI_STAR_OPERATOR);
+   new (QFC->next_ravel()) CharCell(UNI_ASCII_0);
+   new (QFC->next_ravel()) CharCell(UNI_ASCII_UNDERSCORE);
+   new (QFC->next_ravel()) CharCell(UNI_OVERBAR);
+   QFC->check_value(LOC);
+
+   Symbol::assign(QFC, LOC);
 }
 //-----------------------------------------------------------------------------
 void
 Quad_FC::assign(Value_P value, const char * loc)
 {
-   if (!value->is_skalar_or_vector())   RANK_ERROR;
+   if (!value->is_scalar_or_vector())   RANK_ERROR;
 
 ShapeItem value_len = value->element_count();
    if (value_len > 6)   value_len = 6;
@@ -316,19 +362,22 @@ ShapeItem value_len = value->element_count();
    loop(c, value_len)
        if (!value->get_ravel(c).is_character_cell())   DOMAIN_ERROR;
 
-   // value is OK.
+   // new value is correct. 
    //
-   if (value_len < 6)   set_default();
-
-   loop(c, value_len)
-      {
-        current_fc[c] = value->get_ravel(c).get_char_value();
-      }
-
+Unicode fc[6] = { UNI_ASCII_FULLSTOP, UNI_ASCII_COMMA,      UNI_STAR_OPERATOR,
+                  UNI_ASCII_0,        UNI_ASCII_UNDERSCORE, UNI_OVERBAR };
+                  
+   loop(c, 6)   if (c < value_len)
+         fc[c] = value->get_ravel(c).get_char_value();
+                  
    // 0123456789,. are forbidden for ⎕FC[4 + ⎕IO]
    //
-   if (Bif_F12_FORMAT::is_control_char(current_fc[4]))
-      current_fc[4] = UNI_ASCII_SPACE;
+   if (Bif_F12_FORMAT::is_control_char(fc[4]))
+      fc[4] = UNI_ASCII_SPACE;
+
+UCS_string ucs(fc, 6);
+Value_P new_val(new Value(ucs, LOC));
+   Symbol::assign(new_val, LOC);
 }
 //-----------------------------------------------------------------------------
 void
@@ -347,7 +396,7 @@ Quad_FC::assign_indexed(IndexExpr & IX, Value_P value)
 void
 Quad_FC::assign_indexed(Value_P X, Value_P value)
 {
-   // we don't do skalar extension but require indices to match the value.
+   // we don't do scalar extension but require indices to match the value.
    //
 ShapeItem ec = X->element_count();
    if (ec != value->element_count())   INDEX_ERROR;
@@ -357,68 +406,56 @@ ShapeItem ec = X->element_count();
    if (ec > 6)   ec = 6;
 
 const APL_Integer qio = Workspace::get_IO();
+Unicode fc[6];
+   {
+     Value_P old = get_apl_value();
+     loop(e, 6)   fc[e] = old->get_ravel(e).get_char_value();
+   }
+
    loop(e, ec)
       {
         const APL_Integer idx = X->get_ravel(e).get_int_value() - qio;
         if (idx < 0)   continue;
         if (idx > 5)   continue;
 
-        current_fc[idx] = value->get_ravel(e).get_char_value();
+        fc[idx] = value->get_ravel(e).get_char_value();
       }
 
    // 0123456789,. are forbidden for ⎕FC[4 + ⎕IO]
    //
-   if (Bif_F12_FORMAT::is_control_char(current_fc[4]))
-      current_fc[4] = UNI_ASCII_SPACE;
-}
-//-----------------------------------------------------------------------------
-Value_P Quad_FC::get_apl_value() const
-{
-Value_P v1(new Value(6, LOC));
-   loop(c, 6)   new (&v1->get_ravel(c))   CharCell(current_fc[c]);
-   v1->check_value(LOC);
-   return v1;
+   if (Bif_F12_FORMAT::is_control_char(fc[4]))
+      fc[4] = UNI_ASCII_SPACE;
+
+UCS_string ucs(fc, 6);
+Value_P new_val(new Value(ucs, LOC));
+   Symbol::assign(new_val, LOC);
 }
 //=============================================================================
 Quad_IO::Quad_IO()
-   : SystemVariable(ID_Quad_IO),
-     current_io(1)
+   : SystemVariable(ID_Quad_IO)
 {
-Value_P value(new Value(LOC));
-
-   new (&value->get_ravel(0)) IntCell(current_io);
-
-   value->check_value(LOC);
-
-   Symbol::assign(value, LOC);
+   Symbol::assign(IntScalar(1, LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 void
 Quad_IO::assign(Value_P value, const char * loc)
 {
-   if (!value->is_skalar_or_len1_vector())
+   if (!value->is_scalar_or_len1_vector())
       {
         if (value->get_rank() > 1)   RANK_ERROR;
         else                         LENGTH_ERROR;
       }
 
-const Cell & cell = value->get_ravel(0);
-   current_io = value->get_ravel(0).get_near_bool(0.1);
-   Symbol::assign(current_io ? Value::One_P : Value::Zero_P, LOC);
-}
-//-----------------------------------------------------------------------------
-int
-Quad_IO::expunge()
-{
-   current_io = -1;
-   Symbol::expunge();
-   return 1;
+   if (value->get_ravel(0).get_near_bool(0.1))
+      Symbol::assign(IntScalar(1, LOC), LOC);
+   else
+      Symbol::assign(IntScalar(0, LOC), LOC);
 }
 //=============================================================================
 Quad_L::Quad_L()
  : NL_SystemVariable(ID_Quad_L)
 {
-   Symbol::assign(Value::Zero_P, LOC);
+   Symbol::assign(IntScalar(0, LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 void
@@ -437,7 +474,7 @@ StateIndicator * si = Workspace::SI_top_fun();
 Value_P
 Quad_L::get_apl_value() const
 {
-const StateIndicator * si = Workspace::SI_top_error();
+StateIndicator * si = Workspace::SI_top_error();
    if (si)
       {
         Value_P ret = si->get_L();
@@ -450,7 +487,7 @@ const StateIndicator * si = Workspace::SI_top_error();
 Quad_LC::Quad_LC()
    : RO_SystemVariable(ID_Quad_LC)
 {
-   Symbol::assign(Value::Zero_P, LOC);
+   Symbol::assign(IntScalar(0, LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 Value_P
@@ -475,7 +512,7 @@ Value_P Z(new Value(len, LOC));
             new (Z->next_ravel())   IntCell(si->get_line());
        }
 
-   Z->set_default(*Value::Zero_P);
+   Z->set_default_Zero();
    Z->check_value(LOC);
    return Z;
 }
@@ -483,7 +520,7 @@ Value_P Z(new Value(len, LOC));
 Quad_LX::Quad_LX()
    : NL_SystemVariable(ID_Quad_LX)
 {
-   Symbol::assign(Value::Str0_P, LOC);
+   Symbol::assign(Str0(LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 void
@@ -495,90 +532,12 @@ Quad_LX::assign(Value_P value, const char * loc)
    Symbol::assign(value, LOC);
 }
 //=============================================================================
-Quad_NLT::Quad_NLT()
- : NL_SystemVariable(ID_Quad_NLT)
-{
-   // get current locale
-   //
-const char * locale = setlocale(LC_CTYPE, 0);   // e,g, "en_US.utf8"
-   if (locale == 0)
-      {
-        Symbol::assign(Value::Str0_P, LOC);
-          return;
-      }
-
-UCS_string ulocale(locale);
-
-Value_P Z(new Value(ulocale, LOC));
-   Z->set_default(*Value::Spc_P);
-   Z->check_value(LOC);
-   Symbol::assign(Z, LOC);
-}
-//-----------------------------------------------------------------------------
-void
-Quad_NLT::assign(Value_P new_value, const char * loc)
-{
-   if (new_value->get_rank() > 1)      RANK_ERROR;
-   if (!new_value->is_char_string())   DOMAIN_ERROR;
-
-   // expect language [_territory] [.code‐set] [@modifier]
-   // e.g. en_US.utf8
-   //
-   // we only check that it is 7-bit ASCII and leave it to setlocale()
-   // to check if new_value is OK.
-   //
-const ShapeItem len = new_value->element_count();
-   if (len > 100)   LENGTH_ERROR;
-
-   // setlocale() can mess up readline, so we want to use it only if new_value
-   // is different from the current value
-   //
-Value_P old_value = get_apl_value();
-   if (len == old_value->element_count())
-      {
-        bool same = true;
-        loop(l, len)
-           {
-             if (new_value->get_ravel(l).get_char_value() !=
-                 old_value->get_ravel(l).get_char_value())
-                {
-                  same = false;
-                  break;
-                }
-           }
-
-        if (same)   return;
-      }
-
-char new_locale[len + 1];
-   loop(l, len)
-      {
-         new_locale[l] = new_value->get_ravel(l).get_char_value();
-         if (new_locale[l] < ' ')    DOMAIN_ERROR;
-         if (new_locale[l] > 0x7E)   DOMAIN_ERROR;
-      }
-
-   new_locale[len] = 0;
-
-const char * locale = setlocale(LC_ALL, new_locale);
-   if (locale == 0)   DOMAIN_ERROR;
-
-   // setlocale() seems to mess up readline, so we need to re-init it
-   //
-   Input::init(false);
-
-   // setlocale() has not complained, so we can set the ⎕NLT;
-   //
-   Symbol::assign(new_value, LOC);
-}
-//=============================================================================
 Quad_PP::Quad_PP()
-   : SystemVariable(ID_Quad_PP),
-     current_pp(DEFAULT_Quad_PP)
+   : SystemVariable(ID_Quad_PP)
 {
 Value_P value(new Value(LOC));
 
-   new (&value->get_ravel(0)) IntCell(current_pp);
+   new (&value->get_ravel(0)) IntCell(DEFAULT_Quad_PP);
 
    value->check_value(LOC);
    Symbol::assign(value, LOC);
@@ -587,37 +546,25 @@ Value_P value(new Value(LOC));
 void
 Quad_PP::assign(Value_P value, const char * loc)
 {
-   if (!value->is_skalar_or_len1_vector())
+   if (!value->is_scalar_or_len1_vector())
       {
         if (value->get_rank() > 1)   RANK_ERROR;
         else                         LENGTH_ERROR;
       }
 
 const Cell & cell = value->get_ravel(0);
-const APL_Integer val = cell.get_near_int(0.1);
+APL_Integer val = cell.get_near_int(0.1);
    if (val < MIN_Quad_PP)   DOMAIN_ERROR;
 
-   if (val > MAX_Quad_PP)
-      {
-        current_pp = MAX_Quad_PP;
-        value = Value::Max_PP_P;
-      }
-   else
-      {
-        current_pp = val;
-      }
+   if (val > MAX_Quad_PP)   val = MAX_Quad_PP;
 
-   Symbol::assign(value, LOC);
+   Symbol::assign(IntScalar(val, LOC), LOC);
 }
 //=============================================================================
 Quad_PR::Quad_PR()
-   : SystemVariable(ID_Quad_PR),
-     current_pr(UNI_ASCII_SPACE)
+   : SystemVariable(ID_Quad_PR)
 {
-Value_P value(new Value(current_pr, LOC));
-
-   value->check_value(LOC);
-   Symbol::assign(value, LOC);
+   Symbol::assign(Spc(LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 void
@@ -627,22 +574,19 @@ UCS_string ucs = value->get_UCS_ravel();
 
    if (ucs.size() > 1)   LENGTH_ERROR;
 
-   current_pr = ucs;
-
    Symbol::assign(value, LOC);
 }
 //=============================================================================
 Quad_PS::Quad_PS()
-   : SystemVariable(ID_Quad_PS),
-     current_ps(PR_APL)
+   : SystemVariable(ID_Quad_PS)
 {
-   Symbol::assign(Value::Zero_P, LOC);
+   Symbol::assign(IntScalar(0, LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 void
 Quad_PS::assign(Value_P value, const char * loc)
 {
-   if (!value->is_skalar_or_len1_vector())
+   if (!value->is_scalar_or_len1_vector())
       {
         if (value->get_rank() > 1)   RANK_ERROR;
         else                         LENGTH_ERROR;
@@ -650,20 +594,20 @@ Quad_PS::assign(Value_P value, const char * loc)
 
 const Cell & cell = value->get_ravel(0);
 const APL_Integer val = cell.get_near_int(0.1);
-   if      (val == 0)   current_ps = PR_APL;
-   else if (val == 1)   current_ps = PR_APL_FUN;
-   else if (val == 2)   current_ps = PR_BOXED_CHAR;
-   else if (val == 3)   current_ps = PR_BOXED_GRAPHIC;
-   else                 DOMAIN_ERROR;
-
-   new (&value->get_ravel(0))  IntCell(val);
-   Symbol::assign(value, LOC);
+   switch(val)
+      {
+        case 0:
+        case 1:
+        case 2:
+        case 3:  Symbol::assign(IntScalar(val, LOC), LOC);   return;
+        default: DOMAIN_ERROR;
+      }
 }
 //=============================================================================
 Quad_PT::Quad_PT()
    : RO_SystemVariable(ID_Quad_PT)
 {
-   Symbol::assign(Value::Zero_P, LOC);
+   Symbol::assign(IntScalar(0, LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 Value_P
@@ -677,21 +621,15 @@ Value_P Z(new Value(LOC));
 }
 //=============================================================================
 Quad_PW::Quad_PW()
-   : SystemVariable(ID_Quad_PW),
-     current_pw(80)
+   : SystemVariable(ID_Quad_PW)
 {
-Value_P value(new Value(LOC));
-
-   new (&value->get_ravel(0)) IntCell(current_pw);
-
-   value->check_value(LOC);
-   Symbol::assign(value, LOC);
+   Symbol::assign(IntScalar(DEFAULT_Quad_PW, LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 void
 Quad_PW::assign(Value_P value, const char * loc)
 {
-   if (!value->is_skalar_or_len1_vector())
+   if (!value->is_scalar_or_len1_vector())
       {
         if (value->get_rank() > 1)   RANK_ERROR;
         else                         LENGTH_ERROR;
@@ -706,7 +644,6 @@ const APL_Integer val = cell.get_near_int(0.1);
    // max val is system specific. Ignore larger values.
    if (val > MAX_Quad_PW)   return;
 
-   current_pw = val;
    Symbol::assign(value, LOC);
 }
 //=============================================================================
@@ -747,7 +684,6 @@ Quad_QUOTE::Quad_QUOTE()
 Value_P dummy(new Value(UCS_string(LOC), LOC));
    dummy->set_complete();
    Symbol::assign(dummy, LOC);
-   dummy->set_forever();
 }
 //-----------------------------------------------------------------------------
 void
@@ -772,7 +708,7 @@ Quad_QUOTE::assign(Value_P value, const char * loc)
            << prompt << "]" << endl;
 
 PrintContext pctx(PR_QUOTE_Quad);
-PrintBuffer pb(*value, pctx);
+PrintBuffer pb(*value, pctx, 0);
    if (pb.get_height() > 1)   // multi line output: flush and restart corking
       {
         loop(y, pb.get_height())
@@ -782,11 +718,15 @@ PrintBuffer pb(*value, pctx);
              COUT << prompt;
            }
       }
-   else
+   else if (pb.get_height() > 0)
       {
         COUT << pb.l1().no_pad();
         prompt.append(pb.l1());
       }
+   else   // empty output
+      {
+        // nothing to do
+      }
 
    Log(LOG_cork)
       CERR << "Quad_QUOTE::assign() done, buffer = ["
@@ -800,11 +740,13 @@ Quad_QUOTE::get_apl_value() const
       CERR << "Quad_QUOTE::get_apl_value() called, buffer = ["
            << prompt << "]" << endl;
 
-   // get_quad_cr_line() calls done(), so we save the current prompt.
+   // get_quad_cr_line() may call done(), so we save the current prompt.
    //
-UCS_string old_prompt(prompt);
+const UCS_string old_prompt(prompt);
 
 UCS_string in = Input::get_quad_cr_line(old_prompt);
+   done(false, LOC);   // if get_quad_cr_line() has not called it
+
 const UCS_string qpr = Workspace::get_PR();
 
    if (qpr.size() > 0)   // valid prompt replacement char
@@ -825,7 +767,7 @@ Value_P Z(new Value(in, LOC));
 Quad_R::Quad_R()
  : NL_SystemVariable(ID_Quad_R)
 {
-   Symbol::assign(Value::Zero_P, LOC);
+   Symbol::assign(IntScalar(0, LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 void
@@ -844,7 +786,7 @@ StateIndicator * si = Workspace::SI_top_fun();
 Value_P
 Quad_R::get_apl_value() const
 {
-const StateIndicator * si = Workspace::SI_top_error();
+StateIndicator * si = Workspace::SI_top_error();
    if (si)
       {
         Value_P ret = si->get_R();
@@ -865,7 +807,7 @@ Quad_SYL::assign(Value_P value, const char * loc)
    // this assign is called from the constructor in order to trigger the
    // creation of a symbol for Quad_SYL.
    //
-   Symbol::assign(Value::Zero_P, LOC);
+   Symbol::assign(IntScalar(0, LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 void
@@ -897,8 +839,8 @@ Quad_SYL::assign_indexed(Value_P X, Value_P B)
 {
 const APL_Float qct = Workspace::get_CT();
 
-   if (!(X->is_int_skalar(qct) || X->is_int_vector(qct)))   INDEX_ERROR;
-   if (!(B->is_int_skalar(qct) || B->is_int_vector(qct)))   DOMAIN_ERROR;
+   if (!(X->is_int_scalar(qct) || X->is_int_vector(qct)))   INDEX_ERROR;
+   if (!(B->is_int_scalar(qct) || B->is_int_vector(qct)))   DOMAIN_ERROR;
    if (X->element_count() != B->element_count())            LENGTH_ERROR;
 
 const ShapeItem ec = X->element_count();
@@ -929,6 +871,17 @@ const APL_Integer qio = Workspace::get_IO();
 
              ravel_count_limit = cells;
            }
+#if CORE_COUNT_WANTED == -3
+        else if (x == SYL_CURRENT_CORES)   // number of cores
+           {
+             setup_cores((CoreCount)b);
+           }
+#endif
+        else if (x == SYL_PRINT_LIMIT)   // print length limit
+           {
+             if (b < 1000)   DOMAIN_ERROR;
+             print_length_limit = b;
+           }
         else
            {
              INDEX_ERROR;
@@ -943,6 +896,7 @@ const Shape sh(SYL_MAX, 2);
 Value_P Z( new Value(sh, LOC));
 
 #define syl2(n, e, v) syl1(n, e, v)
+#define syl3(n, e, v) syl1(n, e, v)
 #define syl1(n, _e, v) \
   new (Z->next_ravel()) PointerCell(Value_P(new Value(UCS_string(UTF8_string(n)), LOC))); \
   new (Z->next_ravel()) IntCell(v);
@@ -955,19 +909,19 @@ Value_P Z( new Value(sh, LOC));
 Quad_TC::Quad_TC()
    : RO_SystemVariable(ID_Quad_TC)
 {
-   Symbol::assign(Value::Quad_TC_P, LOC);
-}
-//-----------------------------------------------------------------------------
-Value_P
-Quad_TC::get_apl_value() const
-{
-   return Value::Quad_TC_P;
+Value_P QCT(new Value(3, LOC));
+   new (QCT->next_ravel()) CharCell(UNI_ASCII_BS);
+   new (QCT->next_ravel()) CharCell(UNI_ASCII_CR);
+   new (QCT->next_ravel()) CharCell(UNI_ASCII_LF);
+   QCT->check_value(LOC);
+
+   Symbol::assign(QCT, LOC);
 }
 //=============================================================================
 Quad_TS::Quad_TS()
    : RO_SystemVariable(ID_Quad_TS)
 {
-   Symbol::assign(Value::Zero_P, LOC);
+   Symbol::assign(IntScalar(0, LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 Value_P
@@ -991,6 +945,19 @@ Value_P Z(new Value(7, LOC));
 //=============================================================================
 Quad_TZ::Quad_TZ()
    : SystemVariable(ID_Quad_TZ)
+{
+   offset_seconds = compute_offset();
+
+   // ⎕TZ is the offset in hours between GMT and local time
+   //
+   if (offset_seconds % 3600 == 0)   // full hour
+      Symbol::assign(IntScalar(offset_seconds/3600, LOC), LOC);
+   else
+      Symbol::assign(FloatScalar(offset_seconds/3600, LOC), LOC);
+}
+//-----------------------------------------------------------------------------
+int
+Quad_TZ::compute_offset()
 {
    // the GNU/Linux timezone variable conflicts with function timezone() on
    // other systems. timezone.tz_minuteswest in gettimeofday() does not
@@ -1017,25 +984,13 @@ int diff_minutes = local_minutes - gm_minutes;
    if (local_days < gm_days)        diff_minutes -= 24*60;   // local < gmean
    else if (local_days > gm_days)   diff_minutes += 24*60;   // local > gmean
 
-   offset_seconds = 60*diff_minutes;
-
-   // ⎕TZ is the offset in hours between GMT and local time
-   //
-Value_P value(new Value(LOC));
-   if (offset_seconds % 3600 == 0)   // full hour
-      new (&value->get_ravel(0))   IntCell(offset_seconds/3600);
-   else
-      new (&value->get_ravel(0))   FloatCell(offset_seconds/3600.0);
-
-   value->check_value(LOC);
-
-   Symbol::assign(value, LOC);
+   return 60*diff_minutes;
 }
 //-----------------------------------------------------------------------------
 void
 Quad_TZ::assign(Value_P value, const char * loc)
 {
-   if (!value->is_skalar())   RANK_ERROR;
+   if (!value->is_scalar())   RANK_ERROR;
 
    // ignore values outside [-12 ... 14], DOMAIN ERROR for bad types.
 
@@ -1099,7 +1054,7 @@ Value_P Z(new Value(LOC));
 Quad_WA::Quad_WA()
    : RO_SystemVariable(ID_Quad_WA)
 {
-   Symbol::assign(Value::Zero_P, LOC);
+   Symbol::assign(IntScalar(0, LOC), LOC);
 }
 //-----------------------------------------------------------------------------
 Value_P
@@ -1166,3 +1121,35 @@ uint64_t proc_mem = 0;            // memory as reported proc/mem_info
    return Z;
 }
 //=============================================================================
+Quad_X::Quad_X()
+ : NL_SystemVariable(ID_Quad_X)
+{
+   Symbol::assign(IntScalar(0, LOC), LOC);
+}
+//-----------------------------------------------------------------------------
+void
+Quad_X::assign(Value_P value, const char * loc)
+{
+StateIndicator * si = Workspace::SI_top_fun();
+   if (si == 0)   return;
+
+   // ignore assignments if error was SYNTAX ERROR or VALUE ERROR
+   //
+   if (si->get_error().is_syntax_or_value_error())   return;
+
+   si->set_X(value);
+}
+//-----------------------------------------------------------------------------
+Value_P
+Quad_X::get_apl_value() const
+{
+StateIndicator * si = Workspace::SI_top_error();
+   if (si)
+      {
+        Value_P ret = si->get_X();
+        if (!!ret)   return ret;
+      }
+
+   VALUE_ERROR;
+}
+//=============================================================================
