@@ -113,6 +113,8 @@ operator << (ostream & out, TokenClass tc)
         tcc(TC_QUOTE)
         tcc(TC_OFF)
         tcc(TC_SI_LEAVE)
+
+        default: break;
       }
 #undef tcc
 
@@ -145,7 +147,7 @@ operator << (ostream & out, const Token & token)
         if (value->get_rank())   out << value->get_shape();
 
         const PrintContext pctx(PR_APL);
-        PrintBuffer pb(*value, pctx);
+        PrintBuffer pb(*value, pctx, 0);
         bool more = pb.get_height() > 1;
         if (pb.get_height() > 0)
            {
@@ -220,6 +222,8 @@ operator << (ostream & out, const Token & token)
         case TC_LINE:     return out << "\n[" << token.get_fun_line() << "] ";
         case TC_NUMERIC:  return out << "Numeric";
         case TC_NEWLINE:  return out << "{LFeed}";
+
+        default: break;
       }
 
    return out <<  "{-unknown Token " << token.get_tag() << "-}";
@@ -330,6 +334,8 @@ Token::print_function(ostream & out) const
         case TOK_FUN2:      return get_function()->print(out << "USER-F2 ");
         case TOK_OPER1:     return get_function()->print(out << "USER-OP1 ");
         case TOK_OPER2:     return get_function()->print(out << "USER-OP2 ");
+
+        default: break;
       }
 
    // unknown tag.
@@ -376,7 +382,7 @@ Token::print_value(ostream & out) const
                         break;
                       }
 
-                   const PrintContext pctx(PR_APL, 2, 0.02, 80);
+                   const PrintContext pctx(PR_APL, 2, 80);
                    out << UCS_string(v->get_ravel(e)
                                        .character_representation(pctx), 0, 80)
                        << " ";
@@ -396,7 +402,6 @@ Token::print_value(ostream & out) const
         case TOK_Quad_LC:
         case TOK_Quad_LX:
         case TOK_Quad_L:
-        case TOK_Quad_NLT:
         case TOK_Quad_PP:
         case TOK_Quad_PR:
         case TOK_Quad_PS:
@@ -406,6 +411,8 @@ Token::print_value(ostream & out) const
         case TOK_Quad_SVE:
         case TOK_Quad_TZ:
         case TOK_Quad_WA:   return print_quad(out);
+
+        default: break;
       }
 
    out << "{ unknown value Token " << tag;
@@ -418,6 +425,76 @@ Token::print_value(ostream & out) const
    return out <<  ", }";
 }
 //-----------------------------------------------------------------------------
+void
+Token::show_trace(ostream & out, const UCS_string & fun_name, 
+                  Function_Line line) const
+{
+UCS_string fn = fun_name;
+   fn.append_utf8("[");
+   fn.append_number(line);
+   fn.append_utf8("] ");
+
+   out << fn;
+
+   switch(get_tag())
+      {
+        case TOK_APL_VALUE1:
+        case TOK_APL_VALUE2:
+        case TOK_APL_VALUE3:
+             break;   // continue below
+
+        case TOK_BRANCH:
+             out << "→" << get_int_val() << endl;
+             return;
+
+        case TOK_ESCAPE:
+             out << "→" << endl;
+             return;
+
+        case TOK_VOID:
+             out << endl;
+             return;
+
+        default: Q1(*this)
+                 FIXME;
+      }
+
+   // print a value
+
+PrintContext pctx = Workspace::get_PrintContext();
+const Value & val = *get_apl_val();
+   if (val.get_rank() == 0)   // scalar
+      {
+        pctx.set_style(PR_APL_MIN);
+      }
+   else if (val.get_rank() == 1)   // vector
+      {
+        if (val.element_count() == 0 &&   // empty vector
+            (val.get_ravel(0).is_character_cell() ||
+             val.get_ravel(0).is_numeric()))
+           {
+             out << endl;
+             return;
+           }
+            
+        pctx.set_style(PR_APL_MIN);
+      }
+   else                  // matrix or higher
+      {
+        pctx.set_style((PrintStyle)(pctx.get_style() | PST_NO_FRACT_0));
+      }
+
+PrintBuffer pb(val, pctx, 0);
+const UCS_string indent(fn.size(), UNI_ASCII_SPACE);
+   loop(l, pb.get_height())
+      {
+        if (l)   out << indent;
+        out << pb.get_line(l).no_pad() << endl;
+      }
+
+   if (pb.get_height() == 0)   out << endl;
+}
+//-----------------------------------------------------------------------------
 ostream &
 Token::print_quad(ostream & out) const
 {
@@ -463,8 +540,8 @@ UCS_string ucs;
 
         case TC_VALUE:
              {
-               PrintContext pctx(style, 16, 1.0E-18, 80);
-               PrintBuffer pbuf(*get_apl_val(), pctx);
+               PrintContext pctx(style, 16, 80);
+               PrintBuffer pbuf(*get_apl_val(), pctx, 0);
                if (pbuf.get_height() == 0)   return ucs;
                return pbuf.l1();
              }
@@ -536,33 +613,39 @@ Token::short_class_name(TokenClass cls)
 {
    switch(cls)
       {
-        case TC_INVALID:   return "INV";
-        case TC_END:      return "END";
-        case TC_RETURN:    return "RET";
-        case TC_VALUE:     return "VAL";
-        case TC_INDEX:     return "IDX";
-        case TC_PINDEX:    return "PIDX";
-        case TC_SYMBOL:    return "SYM";
+        case TC_ASSIGN:    return "←";
+        case TC_R_ARROW:   return "→";
+        case TC_L_BRACK:   return "[";
+        case TC_R_BRACK:   return "]";
+        case TC_END:       return "END";
         case TC_FUN0:      return "F0";
         case TC_FUN12:     return "F12";
+        case TC_INDEX:     return "IDX";
         case TC_OPER1:     return "OP1";
         case TC_OPER2:     return "OP2";
-        case TC_R_PARENT:  return ")";
-        case TC_R_BRACK:   return "]";
         case TC_L_PARENT:  return "(";
-        case TC_L_BRACK:   return "[/;";
-        case TC_R_ARROW:   return "→";
-        case TC_ASSIGN:    return "←";
-        case TC_LINE:      return "LINE";
+        case TC_R_PARENT:  return ")";
+        case TC_RETURN:    return "RET";
+        case TC_SYMBOL:    return "SYM";
+        case TC_VALUE:     return "VAL";
+
+        case TC_PINDEX:    return "PIDX";
         case TC_VOID:      return "VOID";
+
+        case TC_OFF:       return "OFF";
+        case TC_SI_LEAVE:  return "LEAVE";
+        case TC_LINE:      return "LINE";
         case TC_DIAMOND:   return "◊";
         case TC_NUMERIC:   return "NUMB";
         case TC_SPACE:     return "SPACE";
         case TC_NEWLINE:   return "LF";
         case TC_COLON:     return ":";
         case TC_QUOTE:     return "QUOTE";
-        case TC_OFF:       return "OFF";
-        case TC_SI_LEAVE:  return "LEAVE";
+        case TC_L_CURLY:   return "{";
+        case TC_R_CURLY:   return "}";
+
+        case TC_INVALID:   return "INV";
+        default:           break;
       }
 
    return "???";
@@ -592,33 +675,39 @@ Token::class_name(TokenClass tc)
 
    switch(tc)
       {
-        tcn( TC_INVALID  )
+        tcn( TC_ASSIGN   )
+        tcn( TC_R_ARROW  )
+        tcn( TC_L_BRACK  )
+        tcn( TC_R_BRACK  )
         tcn( TC_END      )
-        tcn( TC_RETURN   )
-        tcn( TC_VALUE    )
-        tcn( TC_INDEX    )
-        tcn( TC_PINDEX   )
-        tcn( TC_SYMBOL   )
         tcn( TC_FUN0     )
         tcn( TC_FUN12    )
+        tcn( TC_INDEX    )
         tcn( TC_OPER1    )
         tcn( TC_OPER2    )
-        tcn( TC_R_PARENT )
-        tcn( TC_R_BRACK  )
         tcn( TC_L_PARENT )
-        tcn( TC_L_BRACK  )
-        tcn( TC_R_ARROW  )
-        tcn( TC_ASSIGN   )
-        tcn( TC_LINE     )
+        tcn( TC_R_PARENT )
+        tcn( TC_RETURN   )
+        tcn( TC_SYMBOL   )
+        tcn( TC_VALUE    )
+
+        tcn( TC_PINDEX   )
         tcn( TC_VOID     )
+
+        tcn( TC_OFF      )
+        tcn( TC_SI_LEAVE )
+        tcn( TC_LINE     )
         tcn( TC_DIAMOND  )
         tcn( TC_NUMERIC  )
         tcn( TC_SPACE    )
         tcn( TC_NEWLINE  )
         tcn( TC_COLON    )
         tcn( TC_QUOTE    )
-        tcn( TC_OFF      )
-        tcn( TC_SI_LEAVE )
+        tcn( TC_L_CURLY  )
+        tcn( TC_R_CURLY  )
+
+        tcn( TC_INVALID  )
+        default: break;
       }
 
    return "*** Obscure token class ***";
@@ -694,3 +783,15 @@ move_2(Token & dst, const Token & src, const char * loc)
       }
 }
 //-----------------------------------------------------------------------------
+void
+Token_string::print(ostream & out, int from) const
+{
+   loop(t, size() - from)
+       {
+         const Token & tok = (*this)[from + t];
+         out << "`" << tok << "  ";
+       }
+
+   out << endl;
+}
+//-----------------------------------------------------------------------------
