@@ -20,10 +20,11 @@
 
 #include <string.h>
 
+#include "Bif_F12_FORMAT.hh"
 #include "Bif_OPER1_COMMUTE.hh"
 #include "Bif_OPER1_EACH.hh"
 #include "Bif_OPER2_PRODUCT.hh"
-#include "Bif_OPER1_RANK.hh"
+#include "Bif_OPER2_RANK.hh"
 #include "Bif_OPER1_REDUCE.hh"
 #include "Bif_OPER1_SCAN.hh"
 #include "CharCell.hh"
@@ -38,7 +39,7 @@
 #include "SystemLimits.hh"
 #include "SystemVariable.hh"
 #include "Tokenizer.hh"
-#include "Value.hh"
+#include "Value.icc"
 #include "Workspace.hh"
 
 //-----------------------------------------------------------------------------
@@ -305,6 +306,7 @@ Token tok = Avec::uni_to_token(uni, LOC);
         sys(F2_LEFT,       F2_LEFT)
         sys(F12_LOGA,      F12_LOGA)
         sys(F12_MINUS,     F12_MINUS)
+        sys(F12_NEQUIV,    F12_NEQUIV)
         sys(F12_PARTITION, F12_PARTITION)
         sys(F12_PICK,      F12_PICK)
         sys(F12_PLUS,      F12_PLUS)
@@ -329,7 +331,7 @@ Token tok = Avec::uni_to_token(uni, LOC);
 
         sys(OPER1_EACH,    OPER1_EACH)
         sys(OPER1_COMMUTE, OPER1_COMMUTE)
-        sys(OPER1_RANK,    OPER1_RANK)
+        sys(OPER2_RANK,    OPER2_RANK)
         sys(OPER1_REDUCE,  OPER1_REDUCE)
         sys(OPER1_REDUCE1, OPER1_REDUCE1)
         sys(OPER1_SCAN,    OPER1_SCAN)
@@ -355,15 +357,11 @@ Tokenizer::tokenize_quad(Source<Unicode> & src, Token_string & tos)
 UCS_string ucs(UNI_Quad_Quad);
    Assert(ucs[0]);
 
-   ucs.append((src.rest() > 0) ? src[0] : Invalid_Unicode);
-   ucs.append((src.rest() > 1) ? src[1] : Invalid_Unicode);
-   ucs.append((src.rest() > 2) ? src[2] : Invalid_Unicode);
-
-   // allow lowercase ⎕xx
-   //
-   if (ucs[1] >= 'a' && ucs[1] <= 'z')   ucs[1] = (Unicode)(ucs[1] - 32);
-   if (ucs[2] >= 'a' && ucs[2] <= 'z')   ucs[2] = (Unicode)(ucs[2] - 32);
-   if (ucs[3] >= 'a' && ucs[3] <= 'z')   ucs[3] = (Unicode)(ucs[3] - 32);
+   if (src.rest() > 0)   ucs.append(src[0]);
+   if (src.rest() > 1)   ucs.append(src[1]);
+   if (src.rest() > 2)   ucs.append(src[2]);
+   if (src.rest() > 3)   ucs.append(src[3]);
+   if (src.rest() > 4)   ucs.append(src[4]);
 
 int len = 0;
 const Token t = Workspace::get_quad(ucs, len);
@@ -384,17 +382,22 @@ const Unicode uni = src.get();
    Assert(uni == UNI_SINGLE_QUOTE);
 
 UCS_string string_value;
+bool got_end = false;
 
    while (src.rest())
        {
          const Unicode uni = src.get();
 
-         // a single ' is the end of the string, while a double ' (i.e. '')
-         // a single '. 
-         //
          if (uni == UNI_SINGLE_QUOTE)
             {
-              if ((src.rest() == 0) || (*src != UNI_SINGLE_QUOTE))   break;
+              // a single ' is the end of the string, while a double '
+              // (i.e. '') is a single '. 
+              //
+              if ((src.rest() == 0) || (*src != UNI_SINGLE_QUOTE))
+                 {
+                   got_end = true;
+                   break;
+                 }
 
               string_value.append(UNI_SINGLE_QUOTE);
               ++src;      // skip the second '
@@ -414,14 +417,12 @@ UCS_string string_value;
             }
        }
 
-   if (string_value.size() == 1)   // skalar
+   if (!got_end)   throw_parse_error(E_NO_STRING_END, LOC, loc);
+
+   if (string_value.size() == 1)   // scalar
       {
         tos.append(Token(TOK_CHARACTER, string_value[0]));
       }
-   else if (string_value.size() == 0)
-      {
-        tos.append(Token(TOK_APL_VALUE1, Value::Str0_P), LOC);
-      }
    else
       {
         tos.append(Token(TOK_APL_VALUE1,
@@ -445,6 +446,7 @@ Tokenizer::tokenize_string2(Source<Unicode> & src, Token_string & tos)
    }
 
 UCS_string string_value;
+bool got_end = false;
 
    while (src.rest())
        {
@@ -452,6 +454,7 @@ UCS_string string_value;
 
          if (uni == UNI_ASCII_DOUBLE_QUOTE)     // terminating "
             {
+              got_end = true;
               break;
             }
          else if (uni == UNI_ASCII_CR)          // ignore CR
@@ -489,10 +492,8 @@ UCS_string string_value;
             }
        }
 
-   if (string_value.size() == 0)
-      {
-        tos.append(Token(TOK_APL_VALUE1, Value::Str0_P), LOC);
-      }
+   if (!got_end)   throw_parse_error(E_NO_STRING_END, LOC, loc);
+
    else
       {
         tos.append(Token(TOK_APL_VALUE1,
@@ -594,7 +595,6 @@ bool
 Tokenizer::tokenize_real(Source<Unicode> &src, APL_Float & flt_val,
                          APL_Integer & int_val)
 {
-const int32_t src_size = src.rest();
 bool dot_seen = false;
 int E_pos = 0;   // position of 'e' in utf
 int minus_pos = 0;   // position of (exponent-) ¯ in utf
@@ -654,7 +654,9 @@ UTF8_string utf;
    if (utf.size() == minus_pos)   return false;
    {
      const int real_cnt = sscanf(utf.c_str(), "%lg", &flt_val);
-     const int int_cnt  = sscanf(utf.c_str(), "%lld", &int_val);
+     long long int_val1;
+                          sscanf(utf.c_str(), "%lld", &int_val1);
+     int_val = int_val1;
      if (digit_seen && (real_cnt > 0))   return true;   // valid real number
    }
 
@@ -758,6 +760,38 @@ UCS_string symbol;
          ++src;
        }
 
+   if (symbol.size() > 2 && symbol[1] == UNI_DELTA  &&
+       (symbol[0] == UNI_ASCII_S || symbol[0] == UNI_ASCII_T))
+      {
+        // S∆ or T∆
+
+        while (src.rest() && *src <= UNI_ASCII_SPACE)   src.get();   // spaces
+        UCS_string symbol1(symbol, 2, symbol.size() - 2);   // without S∆/T∆
+        Value_P AB(new Value(symbol1, LOC));
+        Function * ST = 0;
+        if (symbol[0] == UNI_ASCII_S) ST = &Quad_STOP::fun;
+        else                          ST = &Quad_TRACE::fun;
+
+        const bool assigned = (src.rest() && *src == UNI_LEFT_ARROW);
+        if (assigned)   // dyadic: AB ∆fun
+           {
+             src.get();                                // skip ←
+             Log(LOG_tokenize)
+                CERR << "Stop/Trace assigned: " << symbol1 << endl;
+             tos.append(Token(TOK_APL_VALUE1, AB));   // left argument of ST
+             tos.append(Token(TOK_FUN2, ST));
+           }
+        else
+           {
+             Log(LOG_tokenize)
+                CERR << "Stop/Trace referenved: " << symbol1 << endl;
+             tos.append(Token(TOK_FUN2, ST));
+             tos.append(Token(TOK_APL_VALUE1, AB));   // right argument of ST
+           }
+
+        return;
+      }
+
 Symbol * sym = Workspace::lookup_symbol(symbol);
    tos.append(Token(TOK_SYMBOL, sym));
 }
