@@ -28,7 +28,7 @@
 #include "PrintOperator.hh"
 #include "UCS_string.hh"
 #include "UTF8_string.hh"
-#include "Value.hh"
+#include "Value.icc"
 
 //-----------------------------------------------------------------------------
 UCS_string::UCS_string(const char * cstring)
@@ -64,8 +64,7 @@ UCS_string::UCS_string(const UTF8_string & utf)
              utf.dump_hex(CERR << "Bad UTF8 string: ", 40)
                                << " at " << LOC <<  endl;
              Backtrace::show(__FILE__, __LINE__);
-             Assert(0 && "Error in UCS_string::UCS_string()");
-             break;
+             return;
            }
 
         uint32_t uni = 0;
@@ -74,7 +73,13 @@ UCS_string::UCS_string(const UTF8_string & utf)
               if (i >= utf.size())
                  {
                    utf.dump_hex(CERR << "Truncated UTF8 string: ", 40)
-                      << " at " << LOC <<  endl;
+                      << " len " << utf.size() << " at " << LOC <<  endl;
+                   if (utf.size() >= 40)
+                      {
+                         UTF8_string end(utf.get_items() + utf.size() - 10, 10);
+                         end.dump_hex(CERR << endl << "(ending with : ", 20)
+                                           << ")" << endl;
+                      }
                    return;
                  }
 
@@ -82,9 +87,15 @@ UCS_string::UCS_string(const UTF8_string & utf)
               if ((subc & 0xC0) != 0x80)
                  {
                    utf.dump_hex(CERR << "Bad UTF8 string: ", 40)
-                      << " at " << LOC <<  endl;
+                      << " len " << utf.size() << " at " << LOC <<  endl;
+                   if (utf.size() >= 40)
+                      {
+                         UTF8_string end(utf.get_items() + utf.size() - 10, 10);
+                         end.dump_hex(CERR << endl << "(ending with : ", 20)
+                                           << ")" << endl;
+                      }
                    Backtrace::show(__FILE__, __LINE__);
-                   Assert(0 && "Error in UCS_string::UCS_string()");
+                   return;
                  }
 
               bx  <<= 6;
@@ -276,39 +287,10 @@ UCS_string::UCS_string(const PrintBuffer & pb, Rank rank, int quad_PW)
 {
    if (pb.get_height() == 0)   return;      // empty PrintBuffer
 
-const int pb_width = pb.get_width(0);
+const int total_width = pb.get_width(0);
+
 vector<int> breakpoints;
-
-   // compute breakpoints
-   //
-   for (int col = 0; col < pb_width;)
-       {
-         const int max_chunk_width = col ? quad_PW - 6 : quad_PW;
-
-         // find breakpoint
-         //
-         int chunk_len = max_chunk_width;
-         const int rest = pb_width - col;
-         if (chunk_len < rest)   // rest too large
-            {
-              // search backwards for break char
-              //
-              int pos = col + chunk_len;
-              if (pos > pb_width)   pos = pb_width;
-              while (--pos > col)
-                  {
-                    const Unicode uni = pb.get_line(0)[pos];
-                    if (uni == UNI_iPAD_U2 || uni == UNI_iPAD_U3)
-                       {
-                          chunk_len = pos - col + 1;
-                          break;
-                       }
-                  }
-            }
-
-         breakpoints.push_back(chunk_len);
-         col += chunk_len;
-       }
+   breakpoints.reserve(2*total_width/quad_PW);
 
    // print rows, breaking at breakpoints
    //
@@ -316,20 +298,59 @@ vector<int> breakpoints;
        {
          if (row)   append(UNI_ASCII_LF);   // end previous row
          int col = 0;
-         loop(b, breakpoints.size())
+         int b = 0;
+
+         while (col < total_width)
             {
-              const int chunk_len = breakpoints[b];
+              int chunk_len;
+              if (row == 0)   // first row: set up breakpoints
+                 {
+                   chunk_len = pb.get_line(0).compute_chunk_length(quad_PW,col);
+                   breakpoints.push_back(chunk_len);
+                 }
+              else
+                 {
+                   chunk_len = breakpoints[b++];
+                 }
+
               if (col)   append_utf8("\n      ");
               UCS_string trow(pb.get_line(row), col, chunk_len);
               trow.remove_trailing_padchars();
               append(trow);
+
               col += chunk_len;
             }
        }
 
-   // replacing pad chars with blanks.
+   // replace pad chars with blanks.
    //
-   loop(u, size())   if (is_iPAD_char((*this)[u]))   (*this)[u] = UNI_ASCII_SPACE;
+   loop(u, size())
+       {
+         if (is_iPAD_char((*this)[u]))   (*this)[u] = UNI_ASCII_SPACE;
+       }
+}
+//-----------------------------------------------------------------------------
+int
+UCS_string::compute_chunk_length(int quad_PW, int col) const
+{
+int chunk_len = quad_PW;
+
+   if (col)   chunk_len -= 6;   // subsequent line inden
+
+int pos = col + chunk_len;
+   if (pos >= size())   return size() - col;
+
+   while (--pos > col)
+      {
+         const Unicode uni = (*this)[pos];
+         if (uni == UNI_iPAD_U2 || uni == UNI_iPAD_U3)
+            {
+               chunk_len = pos - col + 1;
+               break;
+            }
+      }
+
+   return chunk_len;
 }
 //-----------------------------------------------------------------------------
 void
@@ -370,6 +391,25 @@ UCS_string::remove_trailing_padchars()
       }
 }
 //-----------------------------------------------------------------------------
+void
+UCS_string::remove_trailing_whitespaces()
+{
+   while (size() && (last() <= UNI_ASCII_SPACE))   pop();
+}
+//-----------------------------------------------------------------------------
+void
+UCS_string::remove_leading_whitespaces()
+{
+int count = 0;
+   loop(s, size())
+      {
+        if ((*this)[s] <= UNI_ASCII_SPACE)   ++count;
+        else                                 break;
+      }
+
+   drop_leading(count);
+}
+//-----------------------------------------------------------------------------
 /// constructor
 UCS_string::UCS_string(const Value & value)
    : Simple_string<Unicode>(0, 0)
@@ -483,6 +523,23 @@ UCS_string::starts_iwith(const char * prefix) const
    return *prefix == 0;   
 }
 //-----------------------------------------------------------------------------
+bool 
+UCS_string::starts_iwith(const UCS_string & prefix) const
+{
+   if (prefix.size() > size())   return false;
+
+   loop(p, prefix.size())
+      {
+        int c1 = (*this)[p];
+        int c2 = prefix[p];
+        if (c1 >= 'a' && c1 <= 'z')   c1 -= 'a' - 'A';
+        if (c2 >= 'a' && c2 <= 'z')   c2 -= 'a' - 'A';
+        if (c1 != c2)   return false;
+      }
+
+   return true;
+}
+//-----------------------------------------------------------------------------
 UCS_string
 UCS_string::no_pad() const
 {
@@ -539,6 +596,21 @@ UCS_string ret;
    return ret;
 }
 //-----------------------------------------------------------------------------
+bool
+UCS_string::is_comment_or_label() const
+{
+   if (size() == 0)                               return false;
+   if ((*this)[0] == UNI_ASCII_NUMBER_SIGN)       return true;   // comment
+   if ((*this)[0] == UNI_COMMENT)                 return true;   // comment
+   loop(t, size())
+       {
+         if ((*this)[t] == UNI_ASCII_COLON)       return true;   // label
+         if (!Avec::is_symbol_char((*this)[t]))   return false;
+       }
+
+   return false;
+}
+//-----------------------------------------------------------------------------
 void
 UCS_string::append_utf8(const UTF8 * str)
 {
@@ -550,6 +622,19 @@ const UCS_string ucs(utf);
 }
 //-----------------------------------------------------------------------------
 void
+UCS_string::append_quoted(const UCS_string & other)
+{
+   append(UNI_ASCII_DOUBLE_QUOTE);
+   loop(s, other.size())
+       {
+          const Unicode uni = other[s];
+          if (uni == UNI_ASCII_DOUBLE_QUOTE)   append(UNI_ASCII_BACKSLASH);
+          append(uni);
+       }
+   append(UNI_ASCII_DOUBLE_QUOTE);
+}
+//-----------------------------------------------------------------------------
+void
 UCS_string::append_ascii(const char * str)
 {
    while (*str)   append(Unicode(*str++));
@@ -559,7 +644,7 @@ void
 UCS_string::append_number(ShapeItem num)
 {
 char cc[40];
-   snprintf(cc, sizeof(cc) - 1, "%lld", num);
+   snprintf(cc, sizeof(cc) - 1, "%lld", (long long)num);
    loop(c, sizeof(cc))
       {
         if (cc[c])   append(Unicode(cc[c]));
@@ -613,12 +698,6 @@ const size_t len = result.back().size();
    return max_len;
 }
 //-----------------------------------------------------------------------------
-Unicode
-UCS_string::upper(Unicode uni)
-{
-   return (uni < 'a' || uni > 'z') ? uni : Unicode(uni - ' ');
-}
-//-----------------------------------------------------------------------------
 int
 UCS_string::atoi() const
 {
@@ -674,10 +753,37 @@ operator << (ostream & os, Unicode uni)
 ostream &
 operator << (ostream & os, const UCS_string & ucs)
 {
-   loop(u, ucs.size())   os << ucs[u];
+const int fill_len = os.width() - ucs.size();;
+
+   if (fill_len > 0)
+      {
+        os.width(0);
+        loop(u, ucs.size())   os << ucs[u];
+        loop(f, fill_len)     os << os.fill();
+      }
+   else
+      {
+        loop(u, ucs.size())   os << ucs[u];
+      }
+
    return os;
 }
 //-----------------------------------------------------------------------------
+bool
+UCS_string::lexical_before(const UCS_string other) const
+{
+   loop(u, size())
+      {
+        if (u >= other.size())   return false;   // other is a prefix of this
+        if ((*this)[u] < other[u])   return true;
+        if ((*this)[u] > other[u])   return false;
+      }
+
+   // at this point the common part of this and other is equal, If other
+   // is longer then this is a prefix of other (and this comes before other)
+   return other.size() > size();
+}
+//-----------------------------------------------------------------------------
 UCS_string
 UCS_string::from_int(int64_t value)
 {
@@ -871,7 +977,7 @@ UCS_string::round_last_digit()
                  {
                    items[q] = Unicode(cc + 1);
                    if (cc != UNI_ASCII_9)   break;    // 0-8 rounded up: stop
-                   items[q] = UNI_ASCII_0;   // 9 rounded up: say 0 and continue
+                   items[q] = UNI_ASCII_0;    // 9 rounded up: say 0 and repeat
                  }
             }
       }
@@ -879,4 +985,122 @@ UCS_string::round_last_digit()
    shrink(size() - 1);
    if (items[size() - 1] == UNI_ASCII_FULLSTOP)   shrink(size() - 1);
 }
-//-----------------------------------------------------------------------------
+//----------------------------------------------------------------------------
+bool
+UCS_string::contains(Unicode uni)
+{
+   loop(u, size())   if ((*this)[u] == uni)   return true;
+   return false;
+}
+//----------------------------------------------------------------------------
+void
+UCS_string::sort_names(const UCS_string ** names, int count)
+{
+   if (count < 2)   return;
+
+   for (int h = 0; h < (count - 1); ++h)
+       {
+         // find smallest above (including) h
+         //
+         int smallest = h;
+         for (int j = h + 1; j < count; ++j)
+             {
+               if (names[smallest]->compare(*names[j]) == COMP_GT)
+                  {
+                    const UCS_string * tmp = names[smallest];
+                    names[smallest] = names[j];
+                    names[j] = tmp;
+                  }
+             }
+       }
+}
+//----------------------------------------------------------------------------
+void
+UCS_string::compute_column_width(vector<int> & result,
+                                 const UCS_string ** names, int name_count,
+                                 int tab_size, int quad_PW)
+{
+   if (name_count < 2)
+      {
+        result.clear();
+        if (name_count)   result.push_back(names[0]->size());
+        else              result.push_back(quad_PW);
+        return;
+      }
+
+   // compute block counts (one block having tab_size characters)
+   //
+const int max_blocks = (quad_PW + 1) / tab_size;
+DynArray(int, name_blocks, name_count);
+   loop(n, name_count)
+       {
+         name_blocks[n] = 1 + (1 + names[n]->size()) / tab_size;
+       }
+
+   // compute max number of column blocks based on first line blocks
+   //
+int max_col = -1;
+   {
+     int blocks = 0;
+     loop(n, name_count)
+         {
+           if ((blocks + name_blocks[n]) < max_blocks)   // name_blocks[n] fits
+              blocks += name_blocks[n];
+           else                                          // max_blocks exceeded
+             {
+               max_col = n - 1;
+               break;
+             }
+         }
+
+     if (max_col == -1)   // all blocks fit
+        {
+          result.clear();
+          loop(n, name_count)
+              {
+                result.push_back(name_blocks[n]);
+              }
+          return;
+        }
+   }
+
+   // decrease max_col until all names fit...
+   //
+   for (;max_col > 1; --max_col)
+      {
+        result.clear();
+        int free_blocks = max_blocks;
+        loop(n, name_count)   // try to fit blocks into result
+            {
+              const int col_n = n % max_col;
+              const int bn = name_blocks[n];
+              if (n < max_col)   // first row: append column
+                 {
+                   free_blocks -= bn;
+                   if (free_blocks < 0)   break;
+                   result.push_back(bn);
+                 }
+              else if (bn > result[col_n])   
+                 {
+                   free_blocks -= bn - result[col_n];
+                   if (free_blocks < 0)   break;
+                   result[col_n] = bn;
+                 }
+            }
+
+        if (free_blocks >= 0)   return;   // success
+      }
+
+   // single colums
+   //
+   result.clear();
+int max_nb = 0;
+   loop(n, name_count)
+      {
+        if (max_nb < name_blocks[n])   max_nb = name_blocks[n];
+      }
+   result.push_back(max_nb);
+}
+//----------------------------------------------------------------------------
+
+//----------------------------------------------------------------------------
