@@ -19,7 +19,6 @@
 */
 
 #include <stdio.h>
-#include <vector>
 #include <string.h>
 #include <fcntl.h>
 #include <unistd.h>
@@ -32,9 +31,8 @@
 #include "Parser.hh"
 #include "StateIndicator.hh"
 #include "Symbol.hh"
-#include "TestFiles.hh"
 #include "UserFunction.hh"
-#include "Value.hh"
+#include "Value.icc"
 #include "Workspace.hh"
 
 //=============================================================================
@@ -59,6 +57,9 @@ const struct _header_pattern
 /// left lambda argument
 #define __a  TOK_ALPHA
 
+/// left lambda operator argument
+#define __au  TOK_ALPHA_U
+
 /// a niladic function
 #define __F0 TOK_SYMBOL
 
@@ -71,9 +72,15 @@ const struct _header_pattern
 /// a monadic operator
 #define __OP1 TOK_L_PARENT, TOK_SYMBOL, TOK_SYMBOL, TOK_R_PARENT
 
+/// a monadic operator in lambda
+#define __op1 TOK_L_PARENT, TOK_ALPHA_U, TOK_SYMBOL, TOK_R_PARENT
+
 /// a dyadic operator
 #define __OP2 TOK_L_PARENT, TOK_SYMBOL, TOK_SYMBOL, TOK_SYMBOL, TOK_R_PARENT
 
+/// a dyadic operator in lambda
+#define __op2 TOK_L_PARENT, TOK_ALPHA_U, TOK_SYMBOL, TOK_OMEGA_U, TOK_R_PARENT
+
 /// an axis
 #define __x   TOK_L_BRACK,  TOK_CHI, TOK_R_BRACK
 
@@ -86,6 +93,9 @@ const struct _header_pattern
 /// right lambda argument
 #define __b  TOK_OMEGA
 
+/// left lambda operator argument
+#define __ou  TOK_OMEGA_U
+
    // niladic
    //
  { SIG_F0             , 1,  1, {           __F0,            } },
@@ -108,9 +118,9 @@ const struct _header_pattern
 
  { SIG_Z_F1_B         , 3,  4, { __z,      __F1,       __b, } },
  { SIG_Z_F1_X_B       , 4,  7, { __z,      __F1,  __x, __b, } },
- { SIG_Z_LO_OP1_B     , 4,  7, { __z,      __OP1,      __b, } },
- { SIG_Z_LO_OP1_X_B   , 5, 10, { __z,      __OP1, __x, __b, } },
- { SIG_Z_LO_OP2_RO_B  , 5,  8, { __z,      __OP2,      __b, } },
+ { SIG_Z_LO_OP1_B     , 4,  7, { __z,      __op1,      __b, } },
+ { SIG_Z_LO_OP1_X_B   , 5, 10, { __z,      __op1, __x, __b, } },
+ { SIG_Z_LO_OP2_RO_B  , 5,  8, { __z,      __op2,      __b, } },
 
    // dyadic
    //
@@ -128,9 +138,9 @@ const struct _header_pattern
 
  { SIG_Z_A_F2_B       , 4,  5, { __z, __a, __F2,       __b } },
  { SIG_Z_A_F2_X_B     , 5,  8, { __z, __a, __F2,  __x, __b } },
- { SIG_Z_A_LO_OP1_B   , 5,  8, { __z, __a, __OP1,      __b } },
- { SIG_Z_A_LO_OP1_X_B , 6, 11, { __z, __a, __OP1, __x, __b } },
- { SIG_Z_A_LO_OP2_RO_B, 6,  9, { __z, __a, __OP2,      __b } },
+ { SIG_Z_A_LO_OP1_B   , 5,  8, { __z, __a, __op1,      __b } },
+ { SIG_Z_A_LO_OP1_X_B , 6, 11, { __z, __a, __op1, __x, __b } },
+ { SIG_Z_A_LO_OP2_RO_B, 6,  9, { __z, __a, __op2,      __b } },
 };
 
 /// the number of signatures
@@ -138,16 +148,16 @@ enum { PATTERN_COUNT = sizeof(header_patterns) / sizeof(*header_patterns) };
 
 //-----------------------------------------------------------------------------
 UserFunction_header::UserFunction_header(const UCS_string text)
-  : error(E_SYNTAX_ERROR),
-    error_loc(0),
+  : from_lambda(false),
+    error(E_SYNTAX_ERROR),
+    error_cause(0),
     sym_Z(0),
     sym_A(0),
     sym_LO(0),
     sym_FUN(0),
     sym_RO(0),
     sym_X(0),
-    sym_B(0),
-    from_lambda(false)
+    sym_B(0)
 {
 UCS_string header_line;
 
@@ -159,7 +169,7 @@ UCS_string header_line;
          else                            header_line.append(uni);
        }
 
-   if (header_line.size() == 0)   { error_loc = LOC;   return; }
+   if (header_line.size() == 0)   { error_cause = LOC;   return; }
 
    Log(LOG_UserFunction__set_line)
       {
@@ -177,7 +187,7 @@ Token_string tos;
      const Parser parser(PM_FUNCTION, LOC);
      const ErrorCode err = parser.parse(header_line, tos);
 
-     if (err != E_NO_ERROR)   { error = err;   error_loc = LOC;  return; }
+     if (err != E_NO_ERROR)   { error = err;   error_cause = LOC;  return; }
    }
 
    // count symbols before first semicolon, allow one symbol too much.
@@ -217,7 +227,7 @@ Fun_signature signature = SIG_NONE;
            }
       }
 
-   if (signature == SIG_NONE)   { error_loc = LOC;   return; }
+   if (signature == SIG_NONE)   { error_cause = LOC;   return; }
 
    // note: constructor has set all symbol pointers to 0!
    // store symbol pointers according to signature.
@@ -241,15 +251,20 @@ Fun_signature signature = SIG_NONE;
       {
         if (tos[tos_idx++].get_tag() != TOK_SEMICOL)
            {
-             error_loc = LOC;
+             error_cause = LOC;
              return;
            }
 
         if  (tos_idx >= tos.size())   break;   // local vars done
 
         const TokenTag tag = tos[tos_idx].get_tag();
-        if (tag != TOK_SYMBOL && tag != TOK_Quad_IO
-                              && tag != TOK_Quad_CT)
+        if (tag != TOK_SYMBOL && tag != TOK_Quad_CT
+                              && tag != TOK_Quad_FC
+                              && tag != TOK_Quad_IO
+                              && tag != TOK_Quad_PP
+                              && tag != TOK_Quad_PR
+                              && tag != TOK_Quad_PW
+                              && tag != TOK_Quad_RL)
            {
              CERR << "Offending token at " LOC " is: " << tos[tos_idx] << endl;
              return;
@@ -263,7 +278,8 @@ Fun_signature signature = SIG_NONE;
 //-----------------------------------------------------------------------------
 UserFunction_header::UserFunction_header(Fun_signature sig,
                                          const UCS_string fname)
-  : error(E_SYNTAX_ERROR),
+  : from_lambda(true),
+    error(E_SYNTAX_ERROR),
     function_name(fname),
     sym_Z(0),
     sym_A(0),
@@ -271,15 +287,33 @@ UserFunction_header::UserFunction_header(Fun_signature sig,
     sym_FUN(0),
     sym_RO(0),
     sym_X(0),
-    sym_B(0),
-    from_lambda(true)
+    sym_B(0)
 {
-                       sym_Z = &Workspace::get_v_LAMBDA();
-   if (sig & SIG_A)    sym_A = &Workspace::get_v_ALPHA();
-   if (sig & SIG_LO)   sym_X = &Workspace::get_v_ALPHA_U();
-   if (sig & SIG_RO)   sym_X = &Workspace::get_v_OMEGA_U();
-   if (sig & SIG_B)    sym_B = &Workspace::get_v_OMEGA();
-   if (sig & SIG_X)    sym_X = &Workspace::get_v_CHI();
+   // make sure that sig is valid
+   //
+Fun_signature sig1 = (Fun_signature)(sig | SIG_FUN);
+bool valid_signature = false;
+   loop(p, PATTERN_COUNT)
+      {
+        if (header_patterns[p].signature == sig1)
+          {
+            valid_signature = true;
+            break;
+          }
+      }
+
+   if (!valid_signature)   
+      {
+         error = E_SYNTAX_ERROR;
+         return;
+      }
+
+                       sym_Z  = &Workspace::get_v_LAMBDA();
+   if (sig & SIG_A)    sym_A  = &Workspace::get_v_ALPHA();
+   if (sig & SIG_LO)   sym_LO = &Workspace::get_v_ALPHA_U();
+   if (sig & SIG_RO)   sym_RO = &Workspace::get_v_OMEGA_U();
+   if (sig & SIG_B)    sym_B  = &Workspace::get_v_OMEGA();
+   if (sig & SIG_X)    sym_X  = &Workspace::get_v_CHI();
 
    if (sig & SIG_FUN)   // named lambda
       {
@@ -289,22 +323,19 @@ UserFunction_header::UserFunction_header(Fun_signature sig,
    error = E_NO_ERROR;
 }
 //-----------------------------------------------------------------------------
-Value_P
+void
 UserFunction_header::pop_local_vars() const
 {
-   loop(l, label_values.size())   label_values[l].sym->pop(true);
+   loop(l, label_values.size())   label_values[l].sym->pop();
 
-   loop(l, local_vars.size())   local_vars[l]->pop(true);
+   loop(l, local_vars.size())   local_vars[l]->pop();
 
-   if (sym_B)    sym_B->pop(true);
-   if (sym_X)    sym_X->pop(true);
-   if (sym_A)    sym_A->pop(true);
-   if (sym_LO)   sym_LO->pop(true);
-   if (sym_RO)   sym_RO->pop(true);
-
-   if (!sym_Z)   return Value_P();
-
-   return sym_Z->pop(false);
+   if (sym_B)    sym_B ->pop();
+   if (sym_X)    sym_X ->pop();
+   if (sym_A)    sym_A ->pop();
+   if (sym_LO)   sym_LO->pop();
+   if (sym_RO)   sym_RO->pop();
+   if (sym_Z)    sym_Z ->pop();
 }
 //-----------------------------------------------------------------------------
 void
@@ -327,6 +358,7 @@ UserFunction_header::check_duplicate_symbols()
    check_duplicate_symbol(sym_LO);
    check_duplicate_symbol(sym_FUN);
    check_duplicate_symbol(sym_RO);
+   check_duplicate_symbol(sym_X);
    check_duplicate_symbol(sym_B);
 
    loop(l, local_vars.size())
@@ -347,6 +379,7 @@ int count = 0;
    if (sym == sym_LO)    ++count;
    if (sym == sym_FUN)   ++count;
    if (sym == sym_RO)    ++count;
+   if (sym == sym_X)     ++count;
    if (sym == sym_B)     ++count;
 
    loop(l, local_vars.size())   if (sym == local_vars[l])   ++count;
@@ -406,10 +439,13 @@ UCS_string ind(indent, UNI_ASCII_SPACE);
 
    if (label_values.size())
       {
-        out << ind << "Labels:         ";
+        out << ind << "Labels:        ";
         loop(l, label_values.size())
-                   out << " " << *label_values[l].sym
-                       << " = " << label_values[l].line;
+           {
+             if (l)   out << ",";
+             out << " " << *label_values[l].sym
+                 << "=" << label_values[l].line;
+           }
         out << endl;
       }
 }
@@ -432,11 +468,11 @@ UserFunction_header::eval_common()
 }
 //=============================================================================
 UserFunction::UserFunction(const UCS_string txt,
-                           int & error_line, const char * & error_loc,
+                           int & error_line, const char * & error_cause,
                            bool keep_existing, const char * loc,
                            const UTF8_string & _creator)
-  : Executable(txt, loc),
-    Function(ID_USER_SYMBOL, TOK_FUN2),
+  : Function(ID_USER_SYMBOL, TOK_FUN2),
+    Executable(txt, loc),
     header(txt),
     creator(_creator)
 {
@@ -452,7 +488,7 @@ UserFunction::UserFunction(const UCS_string txt,
    if (header.get_error() != E_NO_ERROR)   // bad header
       {
         error_line = 0;
-        error_loc = header.get_error_loc();
+        error_cause = header.get_error_cause();
         DEFN_ERROR;
       }
 
@@ -467,56 +503,24 @@ UserFunction::UserFunction(const UCS_string txt,
 
    // check that the function can be defined.
    //
-   if (!header.FUN()->can_be_defined())   DEFN_ERROR;
+const char * why_not = header.FUN()->cant_be_defined();
+   if (why_not)
+      {
+         error_cause = why_not;
+         DEFN_ERROR;
+      }
      
 Function * old_function = header.FUN()->get_function();
-   if (old_function && keep_existing)    DEFN_ERROR;
-
-   for (int l = 1; l < get_text_size(); ++l)
+   if (old_function && keep_existing)
       {
-        error_line = l;
-        line_starts.push_back(Function_PC(body.size()));
-
-        const UCS_string & text = get_text(l);
-        parse_body_line(Function_Line(l), text, loc);
-        setup_lambdas();
+        error_cause = "function exists";
+        DEFN_ERROR;
       }
 
-#if 0
-   // resolve named lambdas
-   //
-   loop(l, named_lambdas.size())
-      {
-        UserFunction * lambda = named_lambdas[l];
-        const UCS_string & lambda_name = lambda->get_name();
-        loop(b, body.size())
-           {
-             if (body[b].get_ValueType() != TV_SYM)   continue;
-
-             Symbol * sym = body[b].get_sym_ptr();
-             if (!sym)                                continue;
-
-             const UCS_string & sym_name = sym->get_name();
-             if (lambda_name != sym_name)             continue;
-
-             move_2(body[b], lambda->get_token(), LOC);
-           }
-      }
-#endif
-
-   Log(LOG_UserFunction__fix)
-      {
-        CERR << "body.size() is " << body.size() << endl
-             << "line_starts.size() is " << line_starts.size() <<endl; 
-      }
-
-   // let [0] be the end of the function.
-   line_starts[0] = Function_PC(body.size());
-
-   if (header.Z())   body.append(Token(TOK_RETURN_SYMBOL, header.Z()), LOC);
-   else              body.append(Token(TOK_RETURN_VOID), LOC);
+   parse_body(error_line, loc);
 
    header.check_duplicate_symbols();
+   if (error_line > 0)   error_cause = "Error in function body";
 
    if (header.LO())   header.FUN()->set_nc(NC_OPERATOR, this);
    else               header.FUN()->set_nc(NC_FUNCTION, this);
@@ -533,30 +537,37 @@ Function * old_function = header.FUN()->get_function();
 //-----------------------------------------------------------------------------
 UserFunction::UserFunction(Fun_signature sig, const UCS_string & fname,
                            const UCS_string & text, const Token_string & bdy)
-  : Executable(sig, fname, text, LOC),
-    Function(ID_USER_SYMBOL, TOK_FUN0),
+  : Function(ID_USER_SYMBOL, TOK_FUN0),
+    Executable(sig, fname, text, LOC),
     header(sig, fname),
     creator(UNI_LAMBDA)
 {
+   set_creation_time(now());
+
    exec_properties[0] = 0;
    exec_properties[1] = 0;
    exec_properties[2] = 0;
    exec_properties[3] = 0;
 
+   if (header.get_error() != E_NO_ERROR)   // bad header
+      {
+        DEFN_ERROR;
+      }
+
    if      (header.RO())   tag = TOK_OPER2;
    else if (header.LO())   tag = TOK_OPER1;
    else if (header.A())    tag = TOK_FUN2;
    else if (header.B())    tag = TOK_FUN1;
    else                    tag = TOK_FUN0;
 
-   parse_body_line(Function_Line_0, bdy, LOC);
+   parse_body_line(Function_Line_0, bdy, false, LOC);
    setup_lambdas();
    line_starts.push_back(Function_PC(bdy.size() - 1));
 
    if (header.FUN())
       {
         if (header.LO())   header.FUN()->set_nc(NC_OPERATOR, this);
-   else                    header.FUN()->set_nc(NC_FUNCTION, this);
+        else               header.FUN()->set_nc(NC_FUNCTION, this);
       }
 }
 //-----------------------------------------------------------------------------
@@ -898,6 +909,20 @@ UserFunction::eval_ALRXB(Value_P A, Token & LO, Token & RO,
    return Token(TOK_SI_PUSHED);
 }
 //-----------------------------------------------------------------------------
+Token
+UserFunction::eval_fill_B(Value_P B)
+{
+Value_P Z = B->clone(LOC);
+   return Token(TOK_APL_VALUE1, Z);
+}
+//-----------------------------------------------------------------------------
+Token
+UserFunction::eval_fill_AB(Value_P A, Value_P B)
+{
+Value_P Z = B->clone(LOC);
+   return Token(TOK_APL_VALUE1, Z);
+}
+//-----------------------------------------------------------------------------
 void
 UserFunction::set_locked_error_info(Error & error) const
 {
@@ -909,7 +934,7 @@ UCS_string & message_2 = error.error_message_2;
         if (!!val_A)
            {
              PrintContext pctx(PR_APL_FUN);
-             PrintBuffer pb(*val_A, pctx);
+             PrintBuffer pb(*val_A, pctx, 0);
              message_2.append(UCS_string(pb, 1, DEFAULT_Quad_PW));
              message_2.append(UNI_ASCII_SPACE);
            }
@@ -924,7 +949,7 @@ UCS_string & message_2 = error.error_message_2;
            {
              message_2.append(UNI_ASCII_SPACE);
              PrintContext pctx(PR_APL_FUN);
-             PrintBuffer pb(*val_B, pctx);
+             PrintBuffer pb(*val_B, pctx, 0);
              message_2.append(UCS_string(pb, 1, DEFAULT_Quad_PW));
            }
       }
@@ -932,6 +957,98 @@ UCS_string & message_2 = error.error_message_2;
    error.right_caret = error.left_caret + message_2.size() - 7;
 }
 //-----------------------------------------------------------------------------
+void
+UserFunction::set_trace_stop(Function_Line * lines, int line_count, bool stop)
+{
+   // Sort lines
+   //
+DynArray(bool, ts_lines, line_starts.size());
+   loop(ts, line_starts.size())   ts_lines[ts] = false;
+   loop(ll, line_count)
+      {
+        Function_Line line = lines[ll];
+        if (line >= 1 && line < line_starts.size())   ts_lines[line] = true;
+      }
+
+   if (stop)
+      {
+        stop_lines.clear();
+        loop(ts, line_starts.size())
+           {
+             if (ts_lines[ts])   stop_lines.push_back((Function_Line)ts);
+           }
+      }
+   else
+      {
+        trace_lines.clear();
+        loop(ts, line_starts.size())
+           {
+             if (ts_lines[ts])   trace_lines.push_back((Function_Line)ts);
+           }
+      }
+
+int error_line = -1;
+   parse_body(error_line, LOC);
+}
+//-----------------------------------------------------------------------------
+void
+UserFunction::parse_body(int & error_line, const char * loc)
+{
+   line_starts.clear();
+   line_starts.push_back(Function_PC_0);   // will be set later.
+
+   clear_body();
+
+   for (int l = 1; l < get_text_size(); ++l)
+      {
+        bool stop_line = false;
+        loop(s, stop_lines.size())
+           {
+             if (stop_lines[s] == l)
+                {
+                  stop_line = true;
+                  break;
+                }
+           }
+
+        bool trace_line = false;
+        loop(t, trace_lines.size())
+           {
+             if (trace_lines[t] == l)
+                {
+                  trace_line = true;
+                  break;
+                }
+           }
+
+        error_line = l;
+        line_starts.push_back(Function_PC(body.size()));
+
+        if (stop_line)
+           {
+             body.append(Token(TOK_STOP_LINE));
+             const int64_t tr = 0;
+             body.append(Token(TOK_END, tr));
+           }
+
+        const UCS_string & text = get_text(l);
+        parse_body_line(Function_Line(l), text, trace_line, loc);
+      }
+      setup_lambdas();
+
+   Log(LOG_UserFunction__fix)
+      {
+        CERR << "body.size() is " << body.size() << endl
+             << "line_starts.size() is " << line_starts.size() <<endl; 
+      }
+
+   // let [0] be the end of the function.
+   line_starts[0] = Function_PC(body.size());
+
+   if (header.Z())   body.append(Token(TOK_RETURN_SYMBOL, header.Z()), LOC);
+   else              body.append(Token(TOK_RETURN_VOID), LOC);
+}
+//-----------------------------------------------------------------------------
 UserFunction *
 UserFunction::load(const char * workspace, const char * function)
 {
@@ -1007,7 +1124,7 @@ UTF8_string utf((const UTF8 *)start, len);
    utf.erase(len);
 
 UCS_string ucs(utf);
-   munmap(start, st.st_size);
+   munmap((char *)start, st.st_size);
    close(in);
 
 int error_line = -1;
@@ -1025,21 +1142,21 @@ UserFunction::pc_for_line(int l) const
 UserFunction *
 UserFunction::fix(const UCS_string & text, int & error_line,
                   bool keep_existing, const char * loc,
-                  const UTF8_string &  creator)
+                  const UTF8_string & creator)
 {
    Log(LOG_UserFunction__fix)
       {
         CERR << "fix pmode=user function:" << endl << text << endl
-             <<  "------------------- fix --" << endl;
+             <<  "------------------- UserFunction::fix() --" << endl;
       }
 
    if (Workspace::SI_top())   Workspace::SI_top()->set_safe_execution(true);
 
 UserFunction * fun = 0;
-const char * error_loc = 0;
+const char * error_cause = 0;
    try
       {
-        fun = new UserFunction(text, error_line, error_loc,
+        fun = new UserFunction(text, error_line, error_cause,
                                keep_existing, loc, creator);
       }
    catch (Error err)
@@ -1047,7 +1164,25 @@ const char * error_loc = 0;
         err.print(CERR);
         if (Workspace::SI_top())
            Workspace::SI_top()->set_safe_execution(false);
-        if (error_loc) Workspace::more_error() = UCS_string(error_loc);
+
+        if (error_cause)
+           {
+             Log(LOG_UserFunction__fix)
+                CERR << "Error: " << error_cause << endl;
+             Workspace::more_error() = UCS_string(error_cause);
+           }
+
+         if (error_line == 0)
+           {
+             Workspace::more_error().append_utf8(" (function header)");
+           }
+         else if (error_line > 0)
+           {
+             Workspace::more_error().append_utf8(" (function line [");
+             Workspace::more_error().append_number(error_line);
+             Workspace::more_error().append_utf8("])");
+           }
+                
         return 0;
       }
    catch (...)
@@ -1073,6 +1208,8 @@ const char * error_loc = 0;
    //
    if (error_line != -1)
       {
+        Log(LOG_UserFunction__fix)
+           CERR << "Line: " << error_line << endl;
         delete fun;
         return 0;
       }
@@ -1108,7 +1245,9 @@ UserFunction::print_properties(ostream & out, int indent) const
 {
    header.print_properties(out, indent);
 UCS_string ind(indent, UNI_ASCII_SPACE);
-   out << ind << "Body Lines:     " << line_starts.size() << endl;
+   out << ind << "Body Lines:      " << line_starts.size() << endl
+       << ind << "Creator:         " << get_creator()      << endl
+       << ind << "Body: " << body << endl;
 }
 //-----------------------------------------------------------------------------
 const UCS_string &
@@ -1138,6 +1277,9 @@ Function_Line
 UserFunction::get_line(Function_PC pc) const
 {
    Assert(pc >= 0);
+
+   // search line_starts backwards until a line with non-greater pc is found.
+   //
    for (int l = line_starts.size() - 1; l > 0; --l)
        {
          if (line_starts[l] <= pc)   return Function_Line(l);
@@ -1176,7 +1318,7 @@ UserFunction::line_prefix(Function_Line l)
 {
 UCS_string ucs;
    ucs.append(UNI_ASCII_L_BRACK);
-   if (l < 10)   ucs.append(UNI_ASCII_SPACE);
+   if (l < Function_Line_10)   ucs.append(UNI_ASCII_SPACE);
 
 char cc[40];
    snprintf(cc, sizeof(cc), "%d", l);
