@@ -32,9 +32,8 @@
 #include "PointerCell.hh"
 #include "PrintOperator.hh"
 #include "SystemVariable.hh"
-#include "TestFiles.hh"
 #include "UCS_string.hh"
-#include "Value.hh"
+#include "Value.icc"
 #include "ValueHistory.hh"
 #include "Workspace.hh"
 
@@ -43,6 +42,9 @@ ShapeItem Value::total_ravel_count = 0;
 
 // the static Value instances are defined in StaticObjects.cc
 
+void * Value::deleted_values = 0;
+int Value::deleted_values_count = 0;
+
 //-----------------------------------------------------------------------------
 void
 Value::init_ravel()
@@ -131,118 +133,6 @@ Value::Value(const char * loc)
    init_ravel();
 }
 //-----------------------------------------------------------------------------
-Value::Value(const char * loc, Value_how how)
-   : DynamicObject(loc),
-     flags(VF_forever),
-     valid_ravel_items(0)
-{
-   // default shape is skalar
-   //
-   ADD_EVENT(this, VHE_Create, 0, loc);
-   switch(how)
-      {
-        case Value_how_Zero ... Value_how_Nine:
-             init_ravel();
-             new (&get_ravel(0)) IntCell(how);
-             break;
-
-        case Value_how_Minus_One:
-             init_ravel();
-             new (&get_ravel(0)) IntCell(-1);
-             break;
-
-        case Value_how_Spc:
-             init_ravel();
-             new (&get_ravel(0)) CharCell(UNI_ASCII_SPACE);
-             break;
-
-        case Value_how_Idx0:
-             new (&shape) Shape(0);
-             init_ravel();
-             new (&get_ravel(0)) IntCell(0);
-             break;
-
-        case Value_how_Str0:
-             new (&shape) Shape(0);
-             init_ravel();
-             new (&get_ravel(0)) CharCell(UNI_ASCII_SPACE);
-             break;
-
-        case Value_how_zStr0:
-             init_ravel();
-             {
-               Value_P z(new Value(loc, Value_how_Str0));   // ''
-               new (&get_ravel(0)) PointerCell(z);               // ⊂''
-             }
-             break;
-
-        case Value_how_Str0_0:
-             new (&shape) Shape(ShapeItem(0), ShapeItem(0));
-             init_ravel();
-             new (&get_ravel(0)) CharCell(UNI_ASCII_SPACE);
-             break;
-
-        case Value_how_Str3_0:
-             new (&shape) Shape(ShapeItem(3), ShapeItem(0));
-             init_ravel();
-             new (&get_ravel(0)) CharCell(UNI_ASCII_SPACE);
-             break;
-
-        case Value_how_Max:
-             init_ravel();
-             new (&get_ravel(0)) FloatCell(BIG_FLOAT);
-             break;
-
-        case Value_how_Min:
-             init_ravel();
-             new (&get_ravel(0)) FloatCell(-BIG_FLOAT);
-             break;
-
-        case Value_how_AV:
-             new (&shape) Shape(MAX_AV);
-             init_ravel();
-             loop(cti, MAX_AV)
-                 {
-                   const int pos = Avec::get_av_pos(CHT_Index(cti));
-                   const Unicode uni = Avec::unicode(CHT_Index(cti));
-
-                   new (&get_ravel(pos))  CharCell(uni);
-                 }
-             break;
-
-        case Value_how_Max_CT:
-             init_ravel();
-             new (&get_ravel(0)) FloatCell(MAX_Quad_CT);
-             break;
-
-        case Value_how_Max_PP:
-             init_ravel();
-             new (&get_ravel(0)) IntCell(MAX_Quad_PP);
-             break;
-
-        case Value_how_Quad_NLT:
-             {
-                const char * cp = "En_US.nlt";
-                const ShapeItem len = strlen(cp);
-                new (&shape) Shape(len);
-                init_ravel();
-                loop(l, len)   new (&get_ravel(l)) CharCell(Unicode(cp[l]));
-             }
-             break;
-
-        case Value_how_Quad_TC:
-             new (&shape) Shape(3);
-             init_ravel();
-             new (&get_ravel(0)) CharCell(UNI_ASCII_BS);
-             new (&get_ravel(1)) CharCell(UNI_ASCII_CR);
-             new (&get_ravel(2)) CharCell(UNI_ASCII_LF);
-             break;
-      }
-
-   check_value(LOC);
-   owner_count = 1;
-}
-//-----------------------------------------------------------------------------
 Value::Value(ShapeItem sh, const char * loc)
    : DynamicObject(loc, &all_values),
      shape(sh),
@@ -264,7 +154,29 @@ Value::Value(const UCS_string & ucs, const char * loc)
 
    if (ucs.size())
       {
-        loop(l, ucs.size())   new (&get_ravel(l)) CharCell(ucs[l]);
+        loop(l, ucs.size())   new (next_ravel()) CharCell(ucs[l]);
+      }
+   else   // empty string: set prototype
+      {
+        new (&get_ravel(0)) CharCell(UNI_ASCII_SPACE);
+      }
+
+   set_complete();
+}
+//-----------------------------------------------------------------------------
+Value::Value(const UTF8_string & utf, const char * loc)
+   : DynamicObject(loc, &all_values),
+     shape(utf.size()),
+     flags(VF_NONE),
+     valid_ravel_items(0)
+{
+   ADD_EVENT(this, VHE_Create, 0, loc);
+   init_ravel();
+
+   if (utf.size())
+      {
+        loop(l, utf.size())
+            new (next_ravel()) CharCell((Unicode)(utf[l] & 0xFF));
       }
    else   // empty string: set prototype
       {
@@ -386,15 +298,15 @@ const ShapeItem value_count = new_value->nz_element_count();
 const Cell * src = &new_value->get_ravel(0);
 Cell * C = &get_ravel(0);
 
-const int src_incr = new_value->is_skalar() ? 0 : 1;
+const int src_incr = new_value->is_scalar() ? 0 : 1;
 
-   if (is_skalar())
+   if (is_scalar())
       {
         if (!C->is_lval_cell())   LEFT_SYNTAX_ERROR;
         Cell * dest = C->get_lval_value();   // can be 0!
         if (dest)   dest->release(LOC);   // free sub-values etc (if any)
 
-        if (new_value->is_skalar())
+        if (new_value->is_scalar())
            {
              dest->init(new_value->get_ravel(0));
            }
@@ -405,7 +317,7 @@ const int src_incr = new_value->is_skalar() ? 0 : 1;
         return;
       }
 
-   if (value_count != dest_count)   LENGTH_ERROR;
+   if (!new_value->is_scalar() && value_count != dest_count)   LENGTH_ERROR;
 
    loop(d, dest_count)
       {
@@ -439,6 +351,26 @@ const ShapeItem ec = nz_element_count();
    return false;
 }
 //-----------------------------------------------------------------------------
+bool
+Value::is_or_contains(const Value & value) const
+{
+   if (this == 0)        return false;   // not a valie value
+
+   if (this == &value)   return true;
+
+const Cell * C = &get_ravel(0);
+   loop(e, nz_element_count())
+      {
+        if (C->is_pointer_cell())
+           {
+             if (C->get_pointer_value()->is_or_contains(value))   return true;;
+           }
+        ++C;
+      }
+
+   return false;
+}
+//-----------------------------------------------------------------------------
 void
 Value::flag_info(const char * loc, ValueFlags flag, const char * flag_name,
                  bool set) const
@@ -469,7 +401,7 @@ Value::mark_all_dynamic_values()
         dob != &all_values; dob = dob->get_prev())
        {
          Value * val = (Value *)dob;
-         if (!val->is_forever())   val->set_marked();
+         val->set_marked();
        }
 }
 //-----------------------------------------------------------------------------
@@ -640,39 +572,25 @@ Value::list_one(ostream & out, bool show_owners)
       {
         out << "   Flags =";
         char sep = ' ';
-        if (is_forever())    { out << sep << "FOREVER";    sep = '+'; }
         if (is_complete())   { out << sep << "COMPLETE";   sep = '+'; }
         if (is_marked())     { out << sep << "MARKED";     sep = '+'; }
         if (is_temp())       { out << sep << "TEMP";       sep = '+'; }
-        out << ",";
       }
    else
       {
-        out << "   Flags = NONE,";
+        out << "   Flags = NONE";
       }
 
-   out << " Rank = " << get_rank();
-   if (get_rank())
-      {
-        out << ", Shape =";
-        loop(r, get_rank())   out << " " << get_shape_item(r);
-      }
-
-   out << ":" << endl;
+   out << ", ⍴" << get_shape() << " ≡" << compute_depth() << ":" << endl;
    print(out);
    out << endl;
+
    if (!show_owners)   return out;
 
    // print owners...
    //
    out << "Owners of " << (const void *)(intptr_t)this << ":" << endl;
 
-   // static variables
-   //
-#define stv_def(x) \
-   if (this == x ## _P.get())   out << "(static) Value::_" #x << endl;
-#include "StaticValues.def"
-
    Workspace::show_owners(out, *this);
 
    out << "---------------------------" << endl << endl;
@@ -780,13 +698,14 @@ Value::is_char_vector() const
 {
    if (get_rank() != 1)   return false;
 
-   loop(c, get_shape_item(0))
-      if (!get_ravel(c).is_character_cell())   return false;
+const Cell * C = &get_ravel(0);
+   loop(c, nz_element_count())   // also check prototype
+      if (!C++->is_character_cell())   return false;   // not char
    return true;
 }
 //-----------------------------------------------------------------------------
 bool
-Value::is_char_skalar() const
+Value::is_char_scalar() const
 {
    if (get_rank() != 0)   return false;
 
@@ -851,7 +770,7 @@ Value::is_int_vector(APL_Float qct) const
 }
 //-----------------------------------------------------------------------------
 bool
-Value::is_int_skalar(APL_Float qct) const
+Value::is_int_scalar(APL_Float qct) const
 {
    if (get_rank() != 0)   return false;
 
@@ -879,7 +798,6 @@ Value::is_simple() const
 const ShapeItem count = element_count();
 const Cell * C = &get_ravel(0);
 
-
    loop(c, count)
        {
          if (C->is_pointer_cell())   return false;
@@ -890,10 +808,35 @@ const Cell * C = &get_ravel(0);
    return true;
 }
 //-----------------------------------------------------------------------------
+bool
+Value::is_one_dimensional() const
+{
+   // lrm would not return false if the value itself is, e.g. a matrix.
+   // That is wrong, however
+   //
+   if (get_rank() > 1)   return false;
+
+const ShapeItem count = nz_element_count();
+const Cell * C = &get_ravel(0);
+
+   loop(c, count)
+       {
+         if (C->is_pointer_cell())
+            {
+             Value_P sub_val = C->get_pointer_value();
+             if (sub_val->get_rank() > 1)                return false;
+             if (!sub_val->is_one_dimensional())         return false;
+            }
+         ++C;
+       }
+
+   return true;   // all items are scalars or vectors
+}
+//-----------------------------------------------------------------------------
 Depth
 Value::compute_depth() const
 {
-   if (is_skalar())
+   if (is_scalar())
       {
         if (get_ravel(0).is_pointer_cell())
            {
@@ -921,17 +864,56 @@ Depth sub_depth = 0;
 }
 //-----------------------------------------------------------------------------
 CellType
-Value::compute_cell_types() const
+Value::flat_cell_types() const
 {
-int32_t ctypes = CT_NONE;
+int32_t ctypes = 0;
 
-const int64_t count = element_count();
+const ShapeItem count = nz_element_count();
+   loop(c, count)   ctypes |= get_ravel(c).get_cell_type();
+
+   return CellType(ctypes);
+}
+//-----------------------------------------------------------------------------
+CellType
+Value::flat_cell_subtypes() const
+{
+int32_t ctypes = 0;
+
+const ShapeItem count = nz_element_count();
+   loop(c, count)   ctypes |= get_ravel(c).get_cell_subtype();
+
+   return CellType(ctypes);
+}
+//-----------------------------------------------------------------------------
+CellType
+Value::deep_cell_types() const
+{
+int32_t ctypes = 0;
+
+const ShapeItem count = nz_element_count();
    loop(c, count)
       {
        const Cell & cell = get_ravel(c);
         ctypes |= cell.get_cell_type();
         if (cell.is_pointer_cell())
-           ctypes |= cell.get_pointer_value()->compute_cell_types();
+            ctypes |= cell.get_pointer_value()->deep_cell_types();
+      }
+
+   return CellType(ctypes);
+}
+//-----------------------------------------------------------------------------
+CellType
+Value::deep_cell_subtypes() const
+{
+int32_t ctypes = 0;
+
+const ShapeItem count = nz_element_count();
+   loop(c, count)
+      {
+       const Cell & cell = get_ravel(c);
+        ctypes |= cell.get_cell_subtype();
+        if (cell.is_pointer_cell())
+           ctypes |= cell.get_pointer_value()->deep_cell_subtypes();
       }
 
    return CellType(ctypes);
@@ -1018,7 +1000,6 @@ Value::index(Value_P X) const
 
 const Shape shape_Z(X->get_shape());
 Value_P Z(new Value(shape_Z, LOC));
-const ShapeItem ec = shape_Z.get_volume();
 const ShapeItem max_idx = element_count();
 const APL_Integer qio = Workspace::get_IO();
 const APL_Float qct = Workspace::get_CT();
@@ -1062,7 +1043,7 @@ Value::get_single_axis(Rank max_axis) const
 const APL_Float   qct = Workspace::get_CT();
 const APL_Integer qio = Workspace::get_IO();
 
-   if (!is_skalar_or_len1_vector())     AXIS_ERROR;
+   if (!is_scalar_or_len1_vector())     AXIS_ERROR;
 
    if (!get_ravel(0).is_near_int(qct))   AXIS_ERROR;
 
@@ -1129,8 +1110,8 @@ const ShapeItem len_B = B->element_count();
              << "with shape " << B->get_shape() << endl;
       }
 
-   Assert(A->is_skalar_or_vector());
-   Assert(B->is_skalar_or_vector());
+   Assert(A->is_scalar_or_vector());
+   Assert(B->is_scalar_or_vector());
 
 Value_P Z(new Value(len_A + len_B, LOC));
 
@@ -1153,14 +1134,14 @@ Value::glue_strand_closed(Token & result, Value_P A, Value_P B,
              << " to non-strand " << endl << *B << endl;
       }
 
-   Assert(A->is_skalar_or_vector());
+   Assert(A->is_scalar_or_vector());
 
 const ShapeItem len_A = A->element_count();
 Value_P Z(new Value(len_A + 1, LOC));
 
    loop(a, len_A)   Z->next_ravel()->init(A->get_ravel(a));
 
-   if (B->is_simple_skalar())
+   if (B->is_simple_scalar())
       {
         Z->next_ravel()->init(B->get_ravel(0));
       }
@@ -1185,12 +1166,12 @@ Value::glue_closed_strand(Token & result, Value_P A, Value_P B,
              << " to strand " << endl << *B << endl;
       }
 
-   Assert(B->is_skalar_or_vector());
+   Assert(B->is_scalar_or_vector());
 
 const ShapeItem len_B = B->element_count();
 Value_P Z(new Value(len_B + 1, LOC));
 
-   if (A->is_simple_skalar())
+   if (A->is_simple_scalar())
       {
         Z->next_ravel()->init(A->get_ravel(0));
       }
@@ -1218,7 +1199,7 @@ Value::glue_closed_closed(Token & result, Value_P A, Value_P B,
       }
 
 Value_P Z(new Value(2, LOC));
-   if (A->is_simple_skalar())
+   if (A->is_simple_scalar())
       {
         Z->next_ravel()->init(A->get_ravel(0));
       }
@@ -1227,7 +1208,7 @@ Value_P Z(new Value(2, LOC));
         new (Z->next_ravel()) PointerCell(A);
       }
 
-   if (B->is_simple_skalar())
+   if (B->is_simple_scalar())
       {
         Z->next_ravel()->init(B->get_ravel(0));
       }
@@ -1248,30 +1229,18 @@ Value::check_value(const char * loc)
    // if value was initialized by means of the next_ravel() mechanism,
    // then all cells are supposed to be OK.
    //
-   if (valid_ravel_items && valid_ravel_items >= element_count())   return;
-
-uint32_t error_count = 0;
-const CellType ctype = get_ravel(0).get_cell_type();
-   switch(ctype)
+   if (valid_ravel_items && valid_ravel_items >= element_count())
       {
-        case CT_CHAR:
-        case CT_POINTER:
-        case CT_CELLREF:
-        case CT_INT:
-        case CT_FLOAT:
-        case CT_COMPLEX:   break;   // OK
-
-        default:
-                 CERR << endl
-                      << "*** check_value(" << loc << ") detects:" << endl
-                      << "   bad ravel[0] (CellType " << ctype << ")" << endl;
-                 ++error_count;
+        set_complete();
+        return;
       }
 
-    loop(c, element_count())
+uint32_t error_count = 0;
+const Cell * C = &get_ravel(0);
+
+    loop(c, nz_element_count())
        {
-         const Cell * cell = &get_ravel(c);
-         const CellType ctype = cell->get_cell_type();
+         const CellType ctype = C->get_cell_type();
          switch(ctype)
             {
               case CT_CHAR:
@@ -1283,10 +1252,10 @@ const CellType ctype = get_ravel(0).get_cell_type();
 
               default:
                  CERR << endl
-                      << "*** check_value(" << loc << ") detects:"
-                      << endl
+                      << "*** check_value(" << loc << ") detects:" << endl
                       << "   bad ravel[" << c << "] (CellType "
                       << ctype << ")" << endl;
+
                  ++error_count;
             }
 
@@ -1295,6 +1264,8 @@ const CellType ctype = get_ravel(0).get_cell_type();
               CERR << endl << "..." << endl;
               break;
             }
+
+         ++C;
        }
 
    if (error_count)
@@ -1311,10 +1282,10 @@ const CellType ctype = get_ravel(0).get_cell_type();
 }
 //-----------------------------------------------------------------------------
 int
-Value::total_size_netto(int CDR_type) const
+Value::total_size_netto(CDR_type cdr_type) const
 {
-   if (CDR_type != 7)   // not nested
-      return 16 + 4*get_rank() + data_size(CDR_type);
+   if (cdr_type != 7)   // not nested
+      return 16 + 4*get_rank() + data_size(cdr_type);
 
    // nested: header + offset-array + sub-values.
    //
@@ -1335,7 +1306,7 @@ int size = 16 + 4*get_rank() + 4*ec;   // top_level size
         else if (cell.is_pointer_cell())
            {
              Value_P sub_val = cell.get_pointer_value();
-             const int sub_type = sub_val->get_CDR_type();
+             const CDR_type sub_type = sub_val->get_CDR_type();
              size += sub_val->total_size_brutto(sub_type);
            }
          else
@@ -1346,11 +1317,11 @@ int size = 16 + 4*get_rank() + 4*ec;   // top_level size
 }
 //-----------------------------------------------------------------------------
 int
-Value::data_size(int CDR_type) const
+Value::data_size(CDR_type cdr_type) const
 {
 const ShapeItem ec = nz_element_count();
    
-   switch (CDR_type) 
+   switch (cdr_type) 
       {
         case 0: return (ec + 7) / 8;   // 1/8 byte bit, rounded up
         case 1: return 4*ec;           //   4 byte integer
@@ -1378,7 +1349,7 @@ int size = 0;
         else if (cell.is_pointer_cell())
            {
              Value_P sub_val = cell.get_pointer_value();
-             const int sub_type = sub_val->get_CDR_type();
+             const CDR_type sub_type = sub_val->get_CDR_type();
              size += sub_val->data_size(sub_type);
            }
          else
@@ -1388,7 +1359,7 @@ int size = 0;
    return size;
 }
 //-----------------------------------------------------------------------------
-int
+CDR_type
 Value::get_CDR_type() const
 {
 const ShapeItem ec = nz_element_count();
@@ -1404,18 +1375,17 @@ const Cell & cell_0 = get_ravel(0);
         loop(e, ec)
            {
              const Cell & cell = get_ravel(e);
-             if (!cell.is_character_cell())   return 7;
+             if (!cell.is_character_cell())   return CDR_NEST32;
              const Unicode uni = cell.get_char_value();
              if (uni < 0)      has_big = true;
              if (uni >= 256)   has_big = true;
            }
 
-        return has_big ? 5 : 4;   // 8-bit or 32-bit char
+        return has_big ? CDR_CHAR32 : CDR_CHAR8;   // 8-bit or 32-bit char
       }
 
    if (cell_0.is_numeric())
       {
-        bool has_bool    = false;
         bool has_int     = false;
         bool has_float   = false;
         bool has_complex = false;
@@ -1426,8 +1396,8 @@ const Cell & cell_0 = get_ravel(0);
              if (cell.is_integer_cell())
                 {
                   const APL_Integer i = cell.get_int_value();
-                  if (i == 0)                   has_bool  = true;
-                  else if (i == 1)              has_bool  = true;
+                  if (i == 0)                   ;
+                  else if (i == 1)              ;
                   else if (i >  0x7FFFFFFFLL)   has_float = true;
                   else if (i < -0x80000000LL)   has_float = true;
                   else                          has_int   = true;
@@ -1440,43 +1410,62 @@ const Cell & cell_0 = get_ravel(0);
                 {
                   has_complex  = true;
                 }
-             else return 7;   // mixed: return 7
+             else return CDR_NEST32;   // mixed: return 7
            }
 
-        if (has_complex)   return 3;
-        if (has_float)     return 2;
-        if (has_int)       return 1;
-        return 0;
+        if (has_complex)   return CDR_CPLX128;
+        if (has_float)     return CDR_FLT64;
+        if (has_int)       return CDR_INT32;
+        return CDR_BOOL1;
       }
 
-   return 7;
+   return CDR_NEST32;
 }
 //-----------------------------------------------------------------------------
 ostream &
 Value::print(ostream & out) const
 {
-PrintStyle style;
-   if (get_rank() < 2)   // skalar or vector
+PrintContext pctx = Workspace::get_PrintContext();
+   if (get_rank() == 0)   // scalar
+      {
+        pctx.set_style(PR_APL_MIN);
+      }
+   else if (get_rank() == 1)   // vector
+      {
+        if (element_count() == 0 &&   // empty vector
+            (get_ravel(0).is_character_cell() || get_ravel(0).is_numeric()))
+           {
+             return out << endl;
+           }
+            
+        pctx.set_style(PR_APL_MIN);
+      }
+   else                  // matrix or higher
+      {
+        pctx.set_style((PrintStyle)(pctx.get_style() | PST_NO_FRACT_0));
+      }
+
+PrintBuffer pb(*this, pctx, &out);   // constructor prints it
+   return out;
+}
+//-----------------------------------------------------------------------------
+ostream &
+Value::print1(ostream & out, PrintContext pctx) const
+{
+int style = pctx.get_style();
+   if (get_rank() < 2)   // scalar or vector
       {
         style = PR_APL_MIN;
       }
    else                  // matrix or higher
       {
-        style = Workspace::get_PS();
-        style = PrintStyle(style | PST_NO_FRACT_0);
+        style |= PST_NO_FRACT_0;
       }
 
-PrintContext pctx(style, Workspace::get_PP(), 
-                         Workspace::get_CT(), Workspace::get_PW());
-PrintBuffer pb(*this, pctx);
+   pctx.set_style((PrintStyle)style);
 
-//   pb.debug(CERR, "Value::print()");
-
-UCS_string ucs(pb, get_rank(), pctx.get_PW());
-
-   if (ucs.size() == 0)   return out;
-
-   return out << ucs << endl;
+PrintBuffer pb(*this, pctx, &out);
+   return out;
 }
 //-----------------------------------------------------------------------------
 ostream &
@@ -1487,7 +1476,6 @@ UCS_string ind(indent, UNI_ASCII_SPACE);
        << ind << "Rank:    " << get_rank()  << endl
        << ind << "Shape:   " << get_shape() << endl
        << ind << "Flags:   " << get_flags();
-   if (is_forever())    out << " VF_forever";
    if (is_complete())   out << " VF_complete";
    if (is_marked())     out << " VF_marked";
    if (is_temp())       out << " VF_temp";
@@ -1502,8 +1490,8 @@ UCS_string ind(indent, UNI_ASCII_SPACE);
 void
 Value::debug(const char * info)
 {
-const PrintContext pctx;
-PrintBuffer pb(*this, pctx);
+const PrintContext pctx = Workspace::get_PrintContext();
+PrintBuffer pb(*this, pctx, 0);
    pb.debug(CERR, info);
 }
 //-----------------------------------------------------------------------------
@@ -1512,7 +1500,9 @@ Value::print_boxed(ostream & out, const char * info)
 {
    if (info)   out << info << endl;
 
-Value_P Z = Quad_CR::do_CR(4, *this);
+const PrintContext pctx(PST_NONE);
+
+Value_P Z = Quad_CR::do_CR(4, *this, pctx);
    out << *Z << endl;
    return out;
 }
@@ -1577,8 +1567,9 @@ Cell * c = &get_ravel(0);
 
    loop(e, ec)
       {
-        if (c->is_pointer_cell())   c->get_pointer_value()->to_proto();
-        else                        new (c) IntCell(0);
+        if (c->is_pointer_cell())          c->get_pointer_value()->to_proto();
+        else if (c->is_character_cell())   new (c) CharCell(UNI_ASCII_SPACE);
+        else                               new (c) IntCell(0);
         ++c;
       }
 }
@@ -1785,7 +1776,6 @@ int count = 0;
         dob != &all_values; dob = dob->get_prev())
        {
          Value * val = (Value *)dob;
-         if (val->is_forever())   continue;
 
          // don't print values found in the previous round.
          //
@@ -1821,7 +1811,8 @@ Value::print_stale_info(ostream & out, const DynamicObject * dob)
    try 
       {
         print_structure(out, 0, 0);
-        Value_P Z = Quad_CR::do_CR(7, *this);
+        const PrintContext pctx(PST_NONE);
+        Value_P Z = Quad_CR::do_CR(7, *this, pctx);
         Z->print(out);
         out << endl;
       }
