@@ -21,29 +21,28 @@
 #include <string.h>
 
 #include <iomanip>
-#include <vector>
 
 #include "Common.hh"
 #include "Error.hh"
+#include "InputFile.hh"
 #include "PrintOperator.hh"
-#include "TestFiles.hh"
 #include "UCS_string.hh"
-#include "Value.hh"
+#include "Value.icc"
 #include "ValueHistory.hh"
 
-VH_entry VH_entry::history[VALUEHISTORY_SIZE];
+VH_entry VH_entry::history[VALUEHISTORY_SIZE + 1];
 int VH_entry::idx = 0;
 
 //----------------------------------------------------------------------------
 VH_entry::VH_entry(const Value * _val, VH_event _ev, int _iarg,
                    const char * _loc)
-  : event(_ev),
-    val(_val),
+  : val(_val),
+    event(_ev),
     iarg(_iarg),
     loc(_loc)
 {
-   testcase_file = TestFiles::get_testfile_name();
-   testcase_line = TestFiles::get_current_lineno();
+   testcase_file = InputFile::current_filename();
+   testcase_line = InputFile::current_line_no();
 }
 //----------------------------------------------------------------------------
 void
@@ -195,7 +194,7 @@ const ValueFlags flags_before = (ValueFlags)flags;
              break;
 
         case VHE_Error:
-             out << "  " << setw(36)
+             out << "  " << setw(38)
                  << Error::error_name((ErrorCode)iarg) << " ";
              break;
 
