@@ -27,20 +27,22 @@ using namespace std;
 #include "Archive.hh"
 #include "Command.hh"
 #include "Input.hh"
+#include "InputFile.hh"
+#include "IO_Files.hh"
 #include "LibPaths.hh"
 #include "main.hh"
 #include "Output.hh"
 #include "Quad_FX.hh"
 #include "Quad_TF.hh"
-#include "TestFiles.hh"
 #include "UserFunction.hh"
+#include "UserPreferences.hh"
 #include "Workspace.hh"
 
 // Workspace::the_workspace is defined in StaticObjects.cc
 
 //-----------------------------------------------------------------------------
 void
-Workspace::push_SI(const Executable * fun, const char * loc)
+Workspace::push_SI(Executable * fun, const char * loc)
 {
    Assert1(fun);
 
@@ -77,7 +79,6 @@ Workspace::pop_SI(const char * loc)
 const Executable * exec = SI_top()->get_executable();
    Assert1(exec);
 
-
    Log(LOG_StateIndicator__push_pop)
       {
         CERR << "Pop  SI[" <<  SI_top()->get_level() << "] "
@@ -96,6 +97,18 @@ StateIndicator * del = SI_top();
    delete del;
 }
 //-----------------------------------------------------------------------------
+void
+Workspace::clear_error(const char * loc)
+{
+   // clear errors up to (including) next user defined function (see ⎕ET)
+   //
+   for (StateIndicator * si = the_workspace.SI_top(); si; si = si->get_parent())
+       {
+         si->get_error().init(E_NO_ERROR, LOC);
+         if (si->get_executable()->get_parse_mode() == PM_FUNCTION)   break;
+       }
+}
+//-----------------------------------------------------------------------------
 StateIndicator *
 Workspace::SI_top_fun()
 {
@@ -107,10 +120,10 @@ Workspace::SI_top_fun()
    return 0;   // no context wirh parse mode PM_FUNCTION
 }
 //-----------------------------------------------------------------------------
-const StateIndicator *
+StateIndicator *
 Workspace::SI_top_error()
 {
-   for (const StateIndicator * si = SI_top(); si; si = si->get_parent())
+   for (StateIndicator * si = SI_top(); si; si = si->get_parent())
        {
          if (si->get_error().error_code != E_NO_ERROR)   return si;
        }
@@ -138,7 +151,7 @@ Workspace::immediate_execution(bool exit_on_error)
                              << HEX(err.error_code) << " ("
                              << err.error_name(err.error_code) << ")" << endl;
 
-                        TestFiles::apl_error(LOC);
+                        IO_Files::apl_error(LOC);
                       }
                  }
               if (exit_on_error)   return Token(TOK_OFF);
@@ -147,7 +160,7 @@ Workspace::immediate_execution(bool exit_on_error)
             {
               CERR << "*** " << __FUNCTION__
                    << "() caught other exception ***" << endl;
-              TestFiles::apl_error(LOC);
+              IO_Files::apl_error(LOC);
               if (exit_on_error)   return Token(TOK_OFF);
             }
       }
@@ -158,6 +171,8 @@ Workspace::immediate_execution(bool exit_on_error)
 NamedObject *
 Workspace::lookup_existing_name(const UCS_string & name)
 {
+   if (name.size() == 0)   return 0;
+
    if (Avec::is_quad(name[0]))   // distinguished name
       {
         int len;
@@ -204,9 +219,19 @@ Workspace::get_quad(UCS_string ucs, int & len)
        return Token();
      }
 
-const Unicode av_0 = (ucs.size() > 1) ? ucs[1] : Invalid_Unicode;
-const Unicode av_1 = (ucs.size() > 2) ? ucs[2] : Invalid_Unicode;
-const Unicode av_2 = (ucs.size() > 3) ? ucs[3] : Invalid_Unicode;
+Unicode av_0 = (ucs.size() > 1) ? ucs[1] : Invalid_Unicode;
+Unicode av_1 = (ucs.size() > 2) ? ucs[2] : Invalid_Unicode;
+Unicode av_2 = (ucs.size() > 3) ? ucs[3] : Invalid_Unicode;
+Unicode av_3 = (ucs.size() > 4) ? ucs[4] : Invalid_Unicode;
+Unicode av_4 = (ucs.size() > 5) ? ucs[5] : Invalid_Unicode;
+
+   // lowercase to uppercase
+   //
+   if (av_0 >= 'a' && av_0 <= 'z')   av_0 = Unicode(av_0 - 0x20);
+   if (av_1 >= 'a' && av_1 <= 'z')   av_1 = Unicode(av_1 - 0x20);
+   if (av_2 >= 'a' && av_2 <= 'z')   av_2 = Unicode(av_2 - 0x20);
+   if (av_3 >= 'a' && av_3 <= 'z')   av_3 = Unicode(av_3 - 0x20);
+   if (av_4 >= 'a' && av_4 <= 'z')   av_4 = Unicode(av_4 - 0x20);
 
 #define var(t, l) { len = l + 1; \
    return Token(TOK_Quad_ ## t, &the_workspace.v_Quad_ ## t); }
@@ -264,6 +289,7 @@ const Unicode av_2 = (ucs.size() > 3) ? ucs[3] : Invalid_Unicode;
              if (av_1 == UNI_ASCII_N)
                 {
                   if (av_2 == UNI_ASCII_P)   f2(INP, 3)
+                  break;
                 }
 
              if (av_1 == UNI_ASCII_O)        var(IO, 2)
@@ -278,11 +304,7 @@ const Unicode av_2 = (ucs.size() > 3) ? ucs[3] : Invalid_Unicode;
         case UNI_ASCII_N:
              if (av_1 == UNI_ASCII_A)        f2 (NA, 2);
              if (av_1 == UNI_ASCII_C)        f2 (NC, 2);
-             if (av_1 == UNI_ASCII_L)
-                {
-                  if (av_2 == UNI_ASCII_T)   var(NLT, 3)
-                                             f1 (NL, 2)
-                }
+             if (av_1 == UNI_ASCII_L)        f1 (NL, 2)
              break;
 
         case UNI_ASCII_P:
@@ -300,6 +322,12 @@ const Unicode av_2 = (ucs.size() > 3) ? ucs[3] : Invalid_Unicode;
 
         case UNI_ASCII_S:
              if (av_1 == UNI_ASCII_I)      f1 (SI, 2)
+             if (av_1 == UNI_ASCII_T)
+                {
+                  if (av_2 == UNI_ASCII_O &&
+                      av_3 == UNI_ASCII_P)   f2 (STOP, 4)
+                  break;
+                }
              if (av_1 == UNI_ASCII_V)
                 {
                   if (av_2 == UNI_ASCII_C)   f2 (SVC, 3)
@@ -308,16 +336,25 @@ const Unicode av_2 = (ucs.size() > 3) ? ucs[3] : Invalid_Unicode;
                   if (av_2 == UNI_ASCII_Q)   f2 (SVQ, 3)
                   if (av_2 == UNI_ASCII_R)   f1 (SVR, 3)
                   if (av_2 == UNI_ASCII_S)   f1 (SVS, 3)
+                  break;
                 }
              if (av_1 == UNI_ASCII_Y)
                 {
                   if (av_2 == UNI_ASCII_L)   var(SYL, 3)
+                  break;
                 }
              break;
 
         case UNI_ASCII_T:
              if (av_1 == UNI_ASCII_C)        var(TC, 2)
              if (av_1 == UNI_ASCII_F)        f2 (TF, 2)
+             if (av_1 == UNI_ASCII_R)
+                {
+                  if (av_2 == UNI_ASCII_A &&
+                      av_3 == UNI_ASCII_C &&
+                      av_4 == UNI_ASCII_E)   f2 (TRACE, 5)
+                  break;
+                }
              if (av_1 == UNI_ASCII_S)        var(TS, 2)
              if (av_1 == UNI_ASCII_Z)        var(TZ, 2)
              break;
@@ -331,6 +368,12 @@ const Unicode av_2 = (ucs.size() > 3) ? ucs[3] : Invalid_Unicode;
         case UNI_ASCII_W:
              if (av_1 == UNI_ASCII_A)        var(WA, 2)
              break;
+
+        case UNI_ASCII_X:
+                                             var(X, 1)
+             break;
+
+        default: break;
       }
 
    var(Quad, 0);
@@ -355,10 +398,27 @@ StateIndicator * ret = 0;
 bool
 Workspace::is_called(const UCS_string & funname)
 {
+   // return true if the current-referent of funname is pendant
+   //
+Symbol * current_referent = lookup_existing_symbol(funname);
+   if (current_referent == 0)   return false;   // no such symbol
+
+   Assert(current_referent->get_Id() == ID_USER_SYMBOL);
+
+const NameClass nc = current_referent->get_nc();
+   if (nc != NC_FUNCTION && nc != NC_OPERATOR)   return false;
+
+const Function * fun = current_referent->get_function();
+   Assert(fun);
+
+const UserFunction * ufun = fun->get_ufun1();
+   if (!ufun)   return false;         // not a defined function
+
+const Executable * uexec = ufun;
+
    for (const StateIndicator * si = SI_top(); si; si = si->get_parent())
       {
-        UCS_string si_fun = si->function_name();
-       if (funname == si_fun)   return true;
+       if (uexec == si->get_executable())   return true;
       }
 
    return false;
@@ -467,8 +527,9 @@ int count = 0;
    loop(f, the_workspace.expunged_functions.size())
        {
          char cc[100];
-         snprintf(cc, sizeof(cc), "    ⎕EX[%lld] ", f);
-         count += the_workspace.expunged_functions[f]->show_owners(cc, out, value);
+         snprintf(cc, sizeof(cc), "    ⎕EX[%lld] ", (long long)f);
+         count += the_workspace.expunged_functions[f]
+                               ->show_owners(cc, out, value);
        }
 
    return count;
@@ -501,15 +562,6 @@ const int ret = the_workspace.expunged_functions.size();
 }
 //-----------------------------------------------------------------------------
 void
-Workspace::replace_arg(Value_P old_value, Value_P new_value)
-{
-   for (StateIndicator * si = SI_top(); si; si = si->get_parent())
-       {
-        if (si->replace_arg(old_value, new_value))   break;
-       }
-}
-//-----------------------------------------------------------------------------
-void
 Workspace::clear_WS(ostream & out, bool silent)
 {
    // clear the SI (pops all localized symbols)
@@ -524,7 +576,7 @@ Workspace::clear_WS(ostream & out, bool silent)
    //
    more_error().clear();
 
-   // clear the read/write system variables...
+   // clear the value stacks of read/write system variables...
    //
 #define rw_sv_def(x) get_v_ ## x().clear_vs();
 #define ro_sv_def(x)
@@ -539,6 +591,8 @@ Workspace::clear_WS(ostream & out, bool silent)
 #define ro_sv_def(x)
 #include "SystemVariable.def"
 
+   get_v_Quad_RL().reset_seed();
+
    set_WS_name(UCS_string("CLEAR WS"));
    if (!silent)   out << "CLEAR WS" << endl;
 }
@@ -548,7 +602,7 @@ Workspace::clear_SI(ostream & out)
 {
    while (SI_top())
       {
-        SI_top()->clear(out);
+        SI_top()->escape();
         pop_SI(LOC);
       }
 }
@@ -569,9 +623,11 @@ Workspace::save_WS(ostream & out, vector<UCS_string> & lib_ws)
    // )SAVE wsname
    // )SAVE libnum wsname
 
+bool name_from_WSID = false;
    if (lib_ws.size() == 0)   // no argument: use )WSID value
       {
-         lib_ws.push_back(the_workspace.WS_name);
+        name_from_WSID = true;
+        lib_ws.push_back(the_workspace.WS_name);
       }
    else if (lib_ws.size() > 2)   // too many arguments
       {
@@ -585,11 +641,12 @@ Workspace::save_WS(ostream & out, vector<UCS_string> & lib_ws)
 LibRef libref = LIB_NONE;
 UCS_string wname = lib_ws.back();
    if (lib_ws.size() == 2)   libref = (LibRef)(lib_ws.front().atoi());
-UTF8_string filename = LibPaths::get_lib_filename(libref, wname, false, "xml");
+UTF8_string filename = LibPaths::get_lib_filename(libref, wname, false,
+                                                  ".xml", 0);
 
    if (wname.compare(UCS_string("CLEAR WS")) == 0)   // don't save CLEAR WS
       {
-        COUT << "NOT SAVED: THIS WS IS " << wname << endl;
+        COUT << "NOT SAVED: THIS WS IS CLEAR WS" << endl;
         more_error() = UCS_string(
         "the workspace was not saved because 'CLEAR WS' is a special \n"
         "workspace name that cannot be saved. Use )WSID <name> first.");
@@ -598,29 +655,32 @@ UTF8_string filename = LibPaths::get_lib_filename(libref, wname, false, "xml");
 
    // dont save if workspace names differ and file exists
    //
-   {
-     if (access(filename.c_str(), F_OK) == 0)   // file exists
-        {
-          if (wname.compare(the_workspace.WS_name) != 0)   // names differ
-             {
-               COUT << "NOT SAVED: THIS WS IS " << wname << endl;
-               UCS_string & t4 = more_error();
-               t4.clear();
-               t4.append_utf8("the workspace was not saved because:\n"
-                      "   the workspace name '");
-               t4.append(the_workspace.WS_name);
-               t4.append_utf8("' of )WSID\n   does not match the name '");
-               t4.append(wname);
-               t4.append_utf8("' used in the )SAVE command\n"
-                      "   and the workspace file\n   ");
-               t4.append_utf8(filename.c_str());
-               t4.append_utf8("\n   already exists. Use )WSID ");
-               t4.append(wname);
-               t4.append_utf8(" first."); 
-               return;
-             }
-        }
-   }
+const bool file_exists = access(filename.c_str(), W_OK) == 0;
+   if (file_exists)
+      {
+        if (wname.compare(the_workspace.WS_name) != 0)   // names differ
+           {
+             COUT << "NOT SAVED: THIS WS IS "
+                  << the_workspace.WS_name << endl;
+
+             UTF8_ostream more;
+             more << "the workspace was not saved because:\n"
+                  << "   the workspace name '" << the_workspace.WS_name
+                  << "' of )WSID\n   does not match the name '" << wname
+                  << "' used in the )SAVE command\n"
+                  << "   and the workspace file\n   " << filename
+                  << "\n   already exists. Use )WSID " << wname << " first."; 
+             more_error() = UCS_string(more.get_data());
+             return;
+           }
+      }
+
+   if (uprefs.backup_before_save && backup_existing_file(filename.c_str()))
+      {
+        COUT << "NOT SAVED: COULD NOT CREATE BACKUP FILE "
+             << filename << endl;
+        return;
+      }
 
    // at this point it is OK to rename and save the workspace
    //
@@ -639,21 +699,97 @@ XML_Saving_Archive ar(outf);
 
    // print time and date to COUT
    {
-     const int offset = get_v_Quad_TZ().get_offset();
-     const YMDhmsu time(now());
-const char * tz_sign = (offset < 0) ? "" : "+";
+     const APL_time_us offset = get_v_Quad_TZ().get_offset();
+     const YMDhmsu time(now() + 1000000*offset);
+     const char * tz_sign = (offset < 0) ? "" : "+";
 
-     COUT << setfill('0') << time.year  << "-"
-          << setw(2)      << time.month << "-"
-          << setw(2)      << time.day   << "  " 
-          << setw(2)      << time.hour  << ":"
-          << setw(2)      << time.minute << ":"
-          << setw(2)      << time.second << " (GMT"
-          << tz_sign      << offset/3600 << ")"
-          << setfill(' ') << endl;
+     out << setfill('0') << time.year  << "-"
+         << setw(2)      << time.month << "-"
+         << setw(2)      << time.day   << "  " 
+         << setw(2)      << time.hour  << ":"
+         << setw(2)      << time.minute << ":"
+         << setw(2)      << time.second << " (GMT"
+         << tz_sign      << offset/3600 << ")"
+         << setfill(' ');
+
+     if (name_from_WSID)   out << " " << the_workspace.WS_name;
+     out << endl;
    }
 }
 //-----------------------------------------------------------------------------
+bool
+Workspace::backup_existing_file(const char * filename)
+{
+   // 1. if file 'filename' does not exist then no backup is needed.
+   //
+   if (access(filename, F_OK) != 0)   return false;   // OK
+
+UTF8_string backup_filename = filename;
+   backup_filename.append_str(".bak");
+
+   // 2. if backup file exists then remove it...
+   //
+   if (access(backup_filename.c_str(), F_OK) == 0)   // backup exists
+      {
+        const int err = unlink(backup_filename.c_str());
+        if (err)
+           {
+             CERR << "Could not remove backup file " << backup_filename
+                  << ": " << strerror(errno) << endl;
+             return true;   // error
+           }
+
+         if (access(backup_filename.c_str(), F_OK) == 0)   // still exists
+            {
+             CERR << "Could not remove backup file " << backup_filename << endl;
+             return true;   // error
+            }
+      }
+
+   // 3. rename file to file.bak
+   //
+const int err = rename(filename, backup_filename.c_str());
+   if (err)   // rename failed
+      {
+        CERR << "Could not rename file " << filename
+             << " to " << backup_filename
+             << ": " << strerror(errno) << endl;
+        return true;   // error
+      }
+
+   if (access(filename, F_OK) == 0)   // file still exists
+      {
+        CERR << "Could not rename file " << filename
+             << " to " << backup_filename << endl;
+        return true;   // error
+      }
+
+   return false; // OK
+}
+//-----------------------------------------------------------------------------
+void
+Workspace::load_DUMP(ostream & out, const UTF8_string & filename, int fd,
+                     bool with_LX)
+{
+   out << "loading )DUMP file " << filename << "..." << endl;
+FILE * file = fdopen(fd, "r");
+
+   // make sure that filename is not already open (which would indicate
+   // )COPY recursion
+   //
+   loop(f, InputFile::files_todo.size())
+      {
+        if (filename == InputFile::files_todo[f].filename)   // same filename
+           {
+             CERR << ")COPY " << filename << " causes recursion" << endl;
+             return;
+           }
+      }
+
+InputFile fam(filename, file, false, false, true, with_LX);
+   InputFile::files_todo.insert(InputFile::files_todo.begin(), fam);
+}
+//-----------------------------------------------------------------------------
 void
 Workspace::dump_WS(ostream & out, vector<UCS_string> & lib_ws)
 {
@@ -677,17 +813,25 @@ Workspace::dump_WS(ostream & out, vector<UCS_string> & lib_ws)
 LibRef libref = LIB_NONE;
 UCS_string wname = lib_ws.back();
    if (lib_ws.size() == 2)   libref = (LibRef)(lib_ws.front().atoi());
-UTF8_string filename = LibPaths::get_lib_filename(libref, wname, false, "apl");
+UTF8_string filename = LibPaths::get_lib_filename(libref, wname, false,
+                                                  ".apl", 0);
 
    if (wname.compare(UCS_string("CLEAR WS")) == 0)   // don't save CLEAR WS
       {
         COUT << "NOT DUMPED: THIS WS IS " << wname << endl;
         more_error() = UCS_string(
-        "the workspace was not dumped because 'CLEAR WS' is a special \n"
+        "the workspace was not dumped because 'CLEAR WS' is a special\n"
         "workspace name that cannot be dumped. Use )WSID <name> first.");
         return;
       }
 
+   if (uprefs.backup_before_save && backup_existing_file(filename.c_str()))
+      {
+        COUT << "NOT DUMPED: COULD NOT CREATE BACKUP FILE "
+             << filename << endl;
+        return;
+      }
+
 ofstream outf(filename.c_str(), ofstream::out);
    if (!outf.is_open())   // open failed
       {
@@ -699,13 +843,13 @@ ofstream outf(filename.c_str(), ofstream::out);
    // print header line, workspace name, time, and date to outf
    //
    {
-     const int offset = get_v_Quad_TZ().get_offset();
-     const YMDhmsu time(now());
+     const APL_time_us offset = get_v_Quad_TZ().get_offset();
+     const YMDhmsu time(now() + 1000000*offset);
      const char * tz_sign = (offset < 0) ? "" : "+";
 
      outf << "#!" << LibPaths::get_APL_bin_path()
           << "/" << LibPaths::get_APL_bin_name()
-          << " --script --" << endl
+          << " --script" << endl
           << endl
           << " ⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝⍝" << endl
           << "⍝" << endl
@@ -739,107 +883,123 @@ int variable_count = 0;
         << " VARIABLES)" << endl;
 }
 //-----------------------------------------------------------------------------
-// )LOAD WS, return ⎕LX of loaded WS on success
-UCS_string 
-Workspace::load_WS(ostream & out, const vector<UCS_string> & lib_ws)
+// )LOAD WS, set ⎕LX of loaded WS on success
+void
+Workspace::load_WS(ostream & out, const vector<UCS_string> & lib_ws,
+                   UCS_string & quad_lx)
 {
    if (lib_ws.size() < 1 || lib_ws.size() > 2)   // no or too many argument(s)
       {
         out << "BAD COMMAND" << endl;
         more_error() = UCS_string("too many parameters in command )LOAD");
-        return UCS_string();
+        return;
       }
 
 LibRef libref = LIB_NONE;
    if (lib_ws.size() == 2)   libref = (LibRef)(lib_ws.front().atoi());
 UCS_string wname = lib_ws.back();
-UTF8_string filename = LibPaths::get_lib_filename(libref, wname, true, "xml");
+UTF8_string filename = LibPaths::get_lib_filename(libref, wname, true,
+                                                  ".xml", ".apl");
 
-XML_Loading_Archive in(filename.c_str());
+int dump_fd = -1;
+XML_Loading_Archive in(filename.c_str(), dump_fd);
 
-   if (!in.is_open())   // open failed: try filename.xml unless already .xml
+   if (dump_fd != -1)
       {
-        if (filename.size() < 5                  ||
-            filename[filename.size() - 4] != '.' ||
-            filename[filename.size() - 3] != 'x' ||
-            filename[filename.size() - 2] != 'm' ||
-            filename[filename.size() - 1] != 'l' )
-           {
-             // filename does not end with .xml, so we try filename.xml
-             //
-             filename.append(UTF8_string(".xml"));
-             new (&in) XML_Loading_Archive(filename.c_str());
-           }
+        the_workspace.clear_WS(CERR, true);
 
-        if (!in.is_open())   // open failed again: give up
+        Log(LOG_command_IN)   CERR << "LOADING " << wname << " from file '"
+                                   << filename << "' ..." << endl;
+
+        load_DUMP(out, filename, dump_fd, true);   // closes dump_fd
+
+        // )DUMP files have no )WSID so create on from the filename
+        //
+        const char * wsid_start = strrchr(filename.c_str(), '/');
+        if (wsid_start == 0)   wsid_start = filename.c_str();
+        else                   ++wsid_start;   // skip /
+        const char * wsid_end = filename.c_str() + filename.size() - 4;
+        const UTF8_string wsid_utf8((const UTF8 *)wsid_start,
+                                    wsid_end - wsid_start);
+        const UCS_string wsid_ucs(wsid_utf8);
+        wsid(out, wsid_ucs);
+
+        // we cant set ⎕LX because it was not executed yet.
+        return;
+      }
+   else
+      {
+        if (!in.is_open())   // open failed
            {
              CERR << ")LOAD " << wname << " (file " << filename
                   << ") failed: " << strerror(errno) << endl;
-             return UCS_string();
+             return;
            }
+
+        Log(LOG_command_IN)   CERR << "LOADING " << wname << " from file '"
+                                   << filename << "' ..." << endl;
+
+        // got open file. We assume that from here on everything will be fine.
+        // clear current WS and load it from file
+        //
+        the_workspace.clear_WS(CERR, true);
+        in.read_Workspace();
       }
 
-   Log(LOG_command_IN)   CERR << "LOADING " << wname << " from file '"
-                              << filename << "' ..." << endl;
-
-   // got open file. We assume that from here on everything will be fine.
-   // clear current WS and load it from file
-   //
-   the_workspace.clear_WS(CERR, true);
-   in.read_Workspace();
-   return Workspace::get_LX();
+   if (Workspace::get_LX().size())  quad_lx = Workspace::get_LX();
 }
 //-----------------------------------------------------------------------------
 void
 Workspace::copy_WS(ostream & out, vector<UCS_string> & lib_ws_objects,
                    bool protection)
 {
+   // )COPY wsname                              wsname /absolute or relative
+   // )COPY libnum wsname                       wsname relative
+   // )COPY wsname objects...                   wsname /absolute or relative
+   // )COPY libnum wsname objects...            wsname relative
+
    if (lib_ws_objects.size() < 1)   // at least workspace name is required
       {
         out << "BAD COMMAND" << endl;
+        more_error() = UCS_string(
+                       "missing parameter(s) in command )COPY or )PCOPY");
         return;
       }
 
-   // move the first one or two items in lib_ws_objects to lib_ws
-   //
-vector<UCS_string> lib_ws;
-const bool with_lib = Command::is_lib_ref(lib_ws_objects.front());
-
-   lib_ws.push_back(lib_ws_objects[0]);
-   lib_ws_objects.erase(lib_ws_objects.begin());
-   if (with_lib)   // lib wname
+LibRef libref = LIB_NONE;
+   if (Avec::is_digit(lib_ws_objects.front()[0]))
       {
-        lib_ws.push_back(lib_ws_objects[0]);
+        libref = (LibRef)(lib_ws_objects.front().atoi());
         lib_ws_objects.erase(lib_ws_objects.begin());
       }
 
-LibRef libref = LIB_NONE;
-   if (lib_ws.size() == 2)   libref = (LibRef)(lib_ws.front().atoi());
-UCS_string wname = lib_ws.back();
-UTF8_string filename = LibPaths::get_lib_filename(libref, wname, true, "xml");
+   if (lib_ws_objects.size() < 1)   // at least workspace name is required
+      {
+        out << "BAD COMMAND" << endl;
+        more_error() = UCS_string(
+                       "missing parameter(s) in command )COPY or )PCOPY");
+        return;
+      }
 
-XML_Loading_Archive in(filename.c_str());
+UCS_string wname = lib_ws_objects.front();
+   lib_ws_objects.erase(lib_ws_objects.begin());
+
+UTF8_string filename = LibPaths::get_lib_filename(libref, wname, true,
+                                                  ".xml", ".apl");
+
+int dump_fd = -1;
+XML_Loading_Archive in(filename.c_str(), dump_fd);
+   if (dump_fd != -1)
+      {
+        load_DUMP(out, filename, dump_fd, false);   // closes dump_fd
+        return;
+      }
 
    if (!in.is_open())   // open failed: try filename.xml unless already .xml
       {
-        if (filename.size() < 5                  ||
-            filename[filename.size() - 4] != '.' ||
-            filename[filename.size() - 3] != 'x' ||
-            filename[filename.size() - 2] != 'm' ||
-            filename[filename.size() - 1] != 'l' )
-           {
-             // filename does not end with .xml, so we try filename.xml
-             //
-             filename.append(UTF8_string(".xml"));
-             new (&in) XML_Loading_Archive(filename.c_str());
-           }
-
-        if (!in.is_open())   // open failed again: give up
-           {
-             CERR << ")COPY " << wname << " (file " << filename
-                  << ") failed: " << strerror(errno) << endl;
-             return;
-           }
+        CERR << ")COPY " << wname << " (file " << filename
+             << ") failed: " << strerror(errno) << endl;
+        return;
       }
 
    Log(LOG_command_IN)   CERR << "LOADING " << wname << " from file '"
@@ -874,12 +1034,5 @@ Workspace::wsid(ostream & out, UCS_string arg)
    out << "WAS " << the_workspace.WS_name << endl;
    the_workspace.WS_name = arg;
 }
-//=============================================================================
-PrintContext::PrintContext()
-: style(Workspace::get_PS()),
-  precision(Workspace::get_PP()),
-  width(Workspace::get_PW())
-{
-}
-//=============================================================================
+//-----------------------------------------------------------------------------
 
