@@ -35,6 +35,7 @@
 #include "FollowCommand.hh"
 #include "SystemFnCommand.hh"
 #include "SystemVariableCommand.hh"
+#include "SendCommand.hh"
 
 #include <iostream>
 #include <sstream>
@@ -68,6 +69,7 @@ NetworkConnection::NetworkConnection( int socket_in )
     add_command( commands, new FollowCommand( "trace" ) );
     add_command( commands, new SystemFnCommand( "systemcommands" ) );
     add_command( commands, new SystemVariableCommand( "systemvariables" ) );
+    add_command( commands, new SendCommand( "sendcontent" ) );
 }
 
 NetworkConnection::~NetworkConnection()
@@ -145,10 +147,52 @@ std::vector<std::string> NetworkConnection::load_block( void )
     return result;
 }
 
+static char get_char_and_check( const std::string &command, std::string::const_iterator &i )
+{
+    if( i == command.end() ) {
+        throw ConnectionError( "Truncated escaped string" );
+    }
+    return *i;
+}
+
+static std::string unescape( const std::string &command )
+{
+    stringstream out;
+    for( std::string::const_iterator i = command.begin() ; i != command.end() ; i++ ) {
+        if( *i == '&' ) {
+            char buf[3];
+            buf[0] = get_char_and_check( command, ++i );
+            buf[1] = get_char_and_check( command, ++i );
+            buf[2] = 0;
+            if( get_char_and_check( command, ++i ) != ';' ) {
+                throw ConnectionError( "Illegal escape sequence" );
+            }
+
+            char *endptr;
+            long v = strtol( buf, &endptr, 16 );
+            if( *endptr != 0 ) {
+                throw ConnectionError( "Illegal character code" );
+            }
+
+            out << static_cast<unsigned char>( v );
+        }
+        else {
+            out << *i;
+        }
+    }
+
+    return out.str();
+}
+
 int NetworkConnection::process_command( const std::string &command )
 {
     ActiveWrapper lock;
-    std::vector<std::string> elements = split( command, ':' );
+    std::vector<std::string> parts = split( command, ':' );
+    std::vector<std::string> elements;
+    for( std::vector<std::string>::iterator i = parts.begin() ; i != parts.end() ; i++ ) {
+        elements.push_back( unescape( *i ) );
+    }
+
     if( elements.size() > 0 ) {
         std::string operation = elements[0];
 
