@@ -20,6 +20,7 @@
 
 #include "emacs.hh"
 #include "network.hh"
+#include "../NativeFunction.hh"
 
 #include <pthread.h>
 
@@ -32,6 +33,7 @@ extern void (*end_input)();
 
 extern "C" {
     void *get_function_mux( const char *function_name );
+    int emacs_start( const char *emacs_arg, const char *lib_path );
 }
 
 void set_active( bool v )
@@ -43,7 +45,7 @@ void set_active( bool v )
     }
     if( v ) {
         while( apl_active ) {
-            pthread_cond_wait( &apl_main_cond, &apl_main_lock );
+             pthread_cond_wait( &apl_main_cond, &apl_main_lock );
         }
     }
     apl_active = v;
@@ -72,7 +74,7 @@ Fun_signature get_signature()
 static Token list_functions( ostream &out )
 {
     out << "Information about the functions" << endl;
-    return Token(TOK_APL_VALUE1, Value::Str0_P);
+    return Token(TOK_APL_VALUE1, Str0(LOC));
 }
 
 Token
@@ -104,7 +106,15 @@ Token eval_XB(Value_P X, Value_P B)
         else {
             port = B->get_ravel( 0 ).get_near_int( qct );
         }
-        return start_listener( port );
+
+        try {
+            start_listener( port );
+        }
+        catch( InitProtocolError &error ) {
+            Workspace::more_error() = error.get_message().c_str();
+            DOMAIN_ERROR;
+        }
+        return Token(TOK_APL_VALUE1, Str0(LOC));
     }
 
     default:
@@ -112,18 +122,24 @@ Token eval_XB(Value_P X, Value_P B)
         DOMAIN_ERROR;
     }
 
-    return Token(TOK_APL_VALUE1, Value::Str0_P);
+    return Token(TOK_APL_VALUE1, Str0(LOC));
 }
 
 Token eval_AXB(const Value_P A, const Value_P X, const Value_P B)
 {
     COUT << "eval_AXB" << endl;
-    return Token(TOK_APL_VALUE1, Value::Str0_P);
+    return Token(TOK_APL_VALUE1, Str0(LOC));
 }
 
-void close_fun( Cause cause )
+bool close_fun( Cause cause, const NativeFunction *caller )
 {
-    close_listeners();
+    if( cause == CAUSE_ERASED ) {
+        return false;
+    }
+    else {
+        close_listeners();
+        return true;
+    }
 }
 
 void *get_function_mux( const char *function_name )
@@ -137,6 +153,13 @@ void *get_function_mux( const char *function_name )
     return 0;
 }
 
+int emacs_start( const char *emacs_arg, const char *lib_path )
+{
+    int port = atoi( emacs_arg );
+    start_listener( port );
+    return 0;
+}
+
 const UCS_string ucs_string_from_string( const std::string &string )
 {
     size_t length = string.size();
